// ASCIIWriter.cpp: implementation of the ASCIIWriter class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "DigitalMaterial.h"  // because of the STL warnings under Windows

#include "ASCIIWriter.h"

#include "DMLastHeader.h"
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

using namespace std;

ASCIIWriter::ASCIIWriter(const std::string &filename) : DMWriter(filename)
{
	// Create a new empty file;
	out.open(filename.c_str(), std::ios::out);
	out.close();
}

ASCIIWriter::~ASCIIWriter()
{

}

void ASCIIWriter::Open()
{
	out.open(filename.c_str(), std::ios::app);
}

void ASCIIWriter::Close()
{
	out.close();
}

void ASCIIWriter::Begin(const std::string &type, const std::string &name)
{
	if(offset == "") Open();
	out << offset << type << '\t' << name <<" {"<< endl;
	offset += "    ";
}

void ASCIIWriter::End()
{
	offset.resize(offset.length()-4);
	out << offset << "}" << endl;
	if(offset.length() == 0) Close();
}

void ASCIIWriter::Put(const Serializable *serializable, const string &name)
{
	Begin(serializable->GetType(), name);
	serializable->Save(this);
	End();
}

void ASCIIWriter::PutBool( bool b, const string& name )
{
	int iBool = (b) ? 1 : 0;
	PutInt( iBool, name );
}

void ASCIIWriter::PutInt(int n, const string &name)
{
	Begin("int", name);
	out << offset << n << endl;
	End();
}

void ASCIIWriter::PutDouble(double x, const string &name)
{
	Begin("double", name);
	out << offset << x << endl;
	End();
}

void ASCIIWriter::PutString(const string &str, const string &name)
{
	Begin("string", name);
	out << offset << str << endl;
	End();
}

void ASCIIWriter::PutIntArray(const int *data, const vector<int> &dims, const string &name)
{
	int tot = 1;
	for(unsigned int ii=1; ii<dims.size(); ++ii) tot *= dims[ii];
	Begin("intArray", name);
	if(dims.size() > 0) out << offset << dims[0];
	for(unsigned int k=1; k<dims.size(); ++k) out << "\t" << dims[k];
	out << endl;
	if(dims.size() != 0)
	{
		for(int i=0; i<dims[0]; ++i)
		{
			out << offset ;
			for(int j=0; j<tot; ++j) out << *data++ << " " ;
			out << endl;
		}
	}
	End();
}

void ASCIIWriter::PutDoubleArray(const double *data, const vector<int> &dims, const string &name)
{
	int tot = 1;
	for(unsigned int ii=1; ii<dims.size(); ++ii) tot *= dims[ii];
	Begin("doubleArray", name);
	if(dims.size() > 0) out << offset << dims[0];
	for(unsigned int k=1; k<dims.size(); ++k) out << "\t" << dims[k];
	out << endl;
	if(dims.size() != 0)
	{
		for(int i=0; i<dims[0]; ++i)
		{
			out << offset ;
			for(int j=0; j<tot; ++j) out << *data++ << " " ;
			out << endl;
		}
	}
	End();
}
