// DM_ComponentList.h: interface for the DM_ComponentList class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef DM_COMPONENTLIST
#define DM_COMPONENTLIST

#ifndef SWIG
#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000
#endif // SWIG

// Put includes here so we know what this file uses
// even though they are currently already included by
// previous headers.
#include <vector>
#include <string>
#include <algorithm> // for find
#include <exception>
class Introspective;

// The ComponentList really is a vector, so now it derives from vector.
// If you want to make the vector more encapsulated, then make the vector
// of components a member and make ComponentList a standard container.
// Otherwise, this is efficient and direct.

class DM_ComponentList  : public std::vector<Introspective *>
{
public:
	DM_ComponentList()
	{
		this->name = "Default ComponentList Name";
		current = -1;
	}
	DM_ComponentList(const std::string& name)
	{
		this->name = name;
		current = -1;	// No members on list to begin with!
	}
	virtual ~DM_ComponentList() {}

	virtual bool SelectMany() { return false; }

	virtual int GetCurrentNumber() const 
	{
		assert(current >= 0 && current < this->size());
		return current;
	}
	virtual bool IsCurrentNumber(int iPosition) const
	{
		return ( iPosition==current );
	}

	virtual void SetCurrentNumber(int current) 
	{
		if (current < 0 || current > this->size())
			throw std::runtime_error("No such component in DM_ComponentList");
		this->current = current;
	}

	virtual void UnsetCurrentNumber(int current) {
	}

	virtual void SetCurrent(const char * componentName)
	{
		int newNumber = 0; 
		while (componentName != this->at(newNumber)->GetName())
		{
			newNumber++;
			if (newNumber >= this->size()) 
				throw std::runtime_error("No such component in DM_ComponentList");
		}
		SetCurrentNumber(newNumber);
	}

	virtual Introspective * GetCurrent() const
	{
		return this->at(current);
	}
	virtual const std::string & GetName() const {return name;}

	virtual const std::vector<Introspective *>* GetCurrentSet() const { return 0; }
private:
	int current;
	std::string name;

};

// The ComponentSet implements a Select Many set of components.
// It is a vector, as seen in the componentList, but you can get
// to the chosen members by asking for the GetCurrentSet().  It
// returns a vector of current components.  It takes a little internal
// work to keep that sub-vector correct, but I think it's worth it outside
// so a program can just ask for the vector of current choices and go through it with
// a for loop.

class ComponentSet  : public DM_ComponentList
{
private:
	std::string name;
	typedef std::vector<Introspective *> LISTTYPE;
	LISTTYPE current;
	bool bSelectMany;
public:
	ComponentSet()
	{
		this->name = "Default ComponentList Name";
		bSelectMany = false;
	}

	ComponentSet(const std::string& name, bool bSelectMany = false)
	{
		this->name = name;
		this->bSelectMany = bSelectMany;
	}
	virtual ~ComponentSet() {}

	virtual bool SelectMany() { return true; }

	virtual int GetCurrentNumber() const 
	{
		if (bSelectMany) assert(0);
		if (current.empty()) return -1;

		Introspective * pCurrent = current.front();
		int i = 0;
		while (pCurrent != this->at(i)) i++;
		return i;
	}

	virtual bool IsCurrentNumber(int iPosition) const 
	{
		if (std::find(current.begin(),current.end(),this->at(iPosition))==current.end())
			return false;
		return true;
	}

	virtual void SetCurrentNumber(int iCurrent) 
	{
		if (iCurrent < 0 || iCurrent > this->size())
			throw std::runtime_error("No such component in DM_ComponentList");
		if (bSelectMany==false)
			current.erase(current.begin(),current.end());
		current.push_back(this->at(iCurrent));
	}

	virtual void UnsetCurrentNumber(int iCurrent) 
	{
		if (iCurrent < 0 || iCurrent > this->size())
			throw std::runtime_error("No such component in DM_ComponentList");
		LISTTYPE::iterator pCurrent = std::find(current.begin(),current.end(),this->at(iCurrent));
		current.erase(pCurrent);
	}

	virtual void SetCurrent(const char * componentName)
	{
		int newNumber = 0; 
		while (componentName != this->at(newNumber)->GetName())
		{
			newNumber++;
			if (newNumber >= this->size()) 
				throw std::runtime_error("No such component in DM_ComponentList");
		}
		SetCurrentNumber(newNumber);
	}

	virtual Introspective * GetCurrent() const
	{
		if (bSelectMany || current.empty()) return 0;
		return current.front();
	}
	virtual const std::string & GetName() const {return name;}

	virtual const std::vector<Introspective *>* GetCurrentSet() const { return &current; }

};

#endif // !defined(DM_COMPONENTLIST)
