// NetCDFReader.cpp: implementation of the NetCDFReader class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "DMFirstHeader.h"

#include "DigitalMaterial.h"  // because of the STL warnings under Windows

#include "NetCDFReader.h"
#include "netcdf.h"
#include "stdlib.h"
#include <cstdio>
#include "NetCDFError.h"
#include "NotEnoughMemory.h"
#include "SprintString.h"
#include "SerialFactory.h"
#include "DMLastHeader.h"

using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

NetCDFReader::NetCDFReader( const std::string& filename ) : DMReader(filename)
{
	TRACE1( "NetCDFReader::NetCDFReader Creating reader for %s.\n",
		filename.c_str() );
	bFileOpen = false;
}


NetCDFReader::~NetCDFReader()
{

}


//////////////////////////////////////////////////////////////////////
// File operations
//////////////////////////////////////////////////////////////////////

void
NetCDFReader::Open()
{
	NetCDFStructure::Open( filename, NetCDFStructure::READONLY, ncID );
	bFileOpen = true;
}


void NetCDFReader::Close()
{
	NetCDFStructure::Close( ncID );
	bFileOpen = false;
}


//////////////////////////////////////////////////////////////////////
// Operations on Serializable objects.
//////////////////////////////////////////////////////////////////////

std::string
NetCDFReader::GetType( const std::string& name )
{
	if ( !bFileOpen ) Open();
	NetCDFStructure::CheckItemName( name );
	NetCDFStructure::DownDir( path, name );

	string type = GetString( "type" );
	
	NetCDFStructure::UpDir( path );
	return type;
}


/*! This is the only method set to return a bool, and I don't know
	why.  The others all throw exceptions.
	*/
bool
NetCDFReader::Fill( Serializable *toFill, const std::string& name )
{
	ASSERT( 0 != toFill );
	if ( 0 == toFill ) throw NetCDFError("NetCDFReader","Fill",
		"Passed a pointer to a null serializable.");

	if ( !bFileOpen ) Open();
	NetCDFStructure::CheckItemName( name );
	NetCDFStructure::DownDir( path, name );

	string type = GetString( "type" );
	ASSERT( type == toFill->GetType() );
	if ( type != toFill->GetType() ) throw NetCDFError("NetCDFReader",
		"Fill", SprintString("The type on file is %s and requested is %s.",
		type.c_str(), toFill->GetType().c_str() ) );

	toFill->Load(this);

	NetCDFStructure::UpDir( path );
	if ( NetCDFStructure::IsAtRoot( path ) ) Close();
	
	return true;
}


Serializable*
NetCDFReader::Get( const std::string& name )
{
	if ( !bFileOpen ) Open();
	NetCDFStructure::CheckItemName( name );
	NetCDFStructure::DownDir( path, name );

	string type = GetString( "type" );
	Serializable *serializable = SerialFactory::Instance()->New(type);
	SUPERASSERT( 0 != serializable );
	if ( 0 == serializable ) throw NetCDFError("NetCDFReader","Get",
		"The serial factory could not return the requested type." );

	serializable->Load(this);

	NetCDFStructure::UpDir( path );
	if ( NetCDFStructure::IsAtRoot( path ) ) Close();

	return serializable;
}


//////////////////////////////////////////////////////////////////////
// Operations on atomic types
//////////////////////////////////////////////////////////////////////

bool
NetCDFReader::GetBool( const std::string& name )
{
	int iBool = GetInt( name );
	bool bIsBoolean = ( (iBool == 0) || (iBool == 1 ) );
	ASSERT( bIsBoolean );
	if ( !bIsBoolean ) throw NetCDFError( "NetCDFReader", "GetBool",
		"The boolean object did not retrieve correctly." );

	return ( iBool == 1 );
}


int
NetCDFReader::GetInt( const std::string& name )
{
	if ( !bFileOpen ) Open();
	string sItemName = NetCDFStructure::AbsolutePath( path, name );
	
	int ncIntID = GetVariableID( ncID, sItemName );

	int iReturn;
	int status = nc_get_var_int(ncID, ncIntID, &iReturn);
	ASSERT( status == NC_NOERR );
	if (status != NC_NOERR) {
		throw NetCDFError("nc_get_var_int",sItemName,nc_strerror(status));
	}

	return iReturn;
}


double
NetCDFReader::GetDouble( const std::string& name )
{
	if ( !bFileOpen ) Open();
	string sItemName = NetCDFStructure::AbsolutePath( path, name );
	
	int ncDoubleID = GetVariableID( ncID, sItemName );

	double dReturn;
	int status = nc_get_var_double(ncID, ncDoubleID, &dReturn);
	ASSERT( status == NC_NOERR );
	if (status != NC_NOERR) {
		throw NetCDFError("nc_get_var_double",sItemName,nc_strerror(status));
	}

	return dReturn;
}

/*! For the interface, strings behave like atomic types, but they
	are stored as arrays in the NetCDF file.
	*/
std::string
NetCDFReader::GetString( const std::string& name )
{
	if ( !bFileOpen ) Open();
	vector<int> dims;

	try {
		GetDimensions( dims, name );
		ASSERT( dims.size() == 1 );
		if ( dims.size() != 1 ) throw NetCDFError("NetCDFReader",
			"GetString", "String has more than one dimension." );
	} catch ( NetCDFError ) {
		// The string may be saved in an old format.
		return GetOldString( NetCDFStructure::AbsolutePath( path, name ) );
	}
	
	string sAbsoluteItem = NetCDFStructure::AbsolutePath( path, name );

	string sReturn;
	if ( dims.size() > 0 && dims[0] > 0) {
		char *data;
		try {
			// We do not store the terminating null.
			data = new char[ dims[0] + 1 ];
		} catch (std::bad_alloc) {
			data = 0;
		}
		ASSERT( data != 0 );
		if ( 0 == data ) throw NotEnoughMemory("NetCDFReader","GetIntArray",
			"Could not allocate memory during serialization." );
		
		try {
			int ncArrayID = GetVariableID( ncID,
				NetCDFStructure::DataMangle(sAbsoluteItem));
			
			int status = nc_get_var_text( ncID, ncArrayID, data );
			ASSERT( status == NC_NOERR );
			if (status != NC_NOERR) {
				throw NetCDFError( "nc_get_var_text", sAbsoluteItem,
					nc_strerror(status) );
			}
		} catch ( ... ) {
			delete[] data;
			throw;
		}
		
		// Add the terminating null
		data[ dims[0] ] = 0;
		sReturn = data;
		delete [] data;
	} // else don't add anything to the return string.

	return sReturn;
}


//////////////////////////////////////////////////////////////////////
// Operations on Arrays
//////////////////////////////////////////////////////////////////////

void
NetCDFReader::GetDimensions( std::vector<int>& dims, const std::string& name )
{
	if ( !bFileOpen ) Open();
	string sAbsoluteItem = NetCDFStructure::AbsolutePath( path, name );

	// Look for dimensions stored as NetCDF Dimension variables.
	try {
		GetRegularDimensions( dims, sAbsoluteItem );

	} catch ( NetCDFError ) {
		GetZeroDimensions( dims, sAbsoluteItem );

	}
}


void
NetCDFReader::FillIntArray( int* data, const std::string& name )
{
	if ( !bFileOpen ) Open();
	
	string sAbsoluteItem = NetCDFStructure::AbsolutePath( path, name );

	FillAbsoluteIntArray( data, sAbsoluteItem );
}


void
NetCDFReader::FillDoubleArray( double* data, const std::string& name )
{
	if ( !bFileOpen ) Open();

	string sAbsoluteItem = NetCDFStructure::AbsolutePath( path, name );

	FillAbsoluteDoubleArray( data, sAbsoluteItem );
}


int*
NetCDFReader::GetIntArray( std::vector<int>& dims, const string& name )
{
	if ( !bFileOpen ) Open();
	GetDimensions( dims, name );
	
	size_t memAlloc = DimsToSize( dims );

	int *data;
	if ( memAlloc > 0 ) {
		try {
			data = new int[ memAlloc ];
		} catch (std::bad_alloc) {
			data = 0;
		}
		ASSERT( 0 != data );
		if ( 0 == data ) throw NotEnoughMemory("NetCDFReader","GetIntArray",
			"Could not allocate memory during serialization." );
		
		FillIntArray( data, name );

	} else {
		data = 0;
	}

	return data;
}


double*
NetCDFReader::GetDoubleArray( std::vector<int> &dims, const std::string& name )
{
	if ( !bFileOpen ) Open();
	GetDimensions( dims, name );
	
	size_t memAlloc = DimsToSize( dims );

	double *data;
	if ( memAlloc > 0 ) {
		try {
			data = new double[ memAlloc ];
		} catch (std::bad_alloc) {
			data = 0;
		}
		ASSERT( 0 != data );
		if ( 0 == data ) throw NotEnoughMemory("NetCDFReader","GetDoubleArray",
			"Could not allocate memory during serialization." );

		FillDoubleArray( data, name );

	} else {
		data = 0;
	}

	return data;
}


//////////////////// PRIVATE FUNCTIONS ////////////////////////////


void
NetCDFReader::GetRegularDimensions( std::vector<int>& dims,
										const std::string& sAbsoluteItem )
{
	vector<int> listID;
	
	int dimIdx = 0;
	for (;;) {
		int ncDimID;
		string nameDim = NetCDFStructure::DimMangle( sAbsoluteItem, dimIdx );
		int dimStatus = nc_inq_dimid(ncID, nameDim.c_str(), &ncDimID);
		if ( dimStatus != NC_NOERR ) break;
		listID.push_back( ncDimID );
		dimIdx++;
	}
	// Don't assert here because it should fall through.
	if ( listID.size() <= 0 ) throw NetCDFError( "NetCDFReader","GetDimensions",
		SprintString("Did not find the dimensions for item \"%s\".",
		sAbsoluteItem.c_str() ) );
	
	dims.clear();
	for(unsigned int i=0; i<listID.size(); ++i){
		int iDim = GetDimensionLength( ncID, listID[i] );
		dims.push_back(iDim);
	}
}


void
NetCDFReader::GetZeroDimensions( std::vector<int>& dims,
								const std::string& sAbsoluteItem )
{
	// Look for dimensions stored as an array of integers.
	vector<int> fakeDims;

	GetRegularDimensions( fakeDims, NetCDFStructure::FakeDimMangle(sAbsoluteItem) );
	
	// fake dimensions are an integer array of dimensions
	ASSERT( fakeDims.size() == 1 );
	if ( fakeDims.size() != 1 ) throw NetCDFError( "NetCDFReader","GetZeroDimensions",
			SprintString("Could not find dimensions for integer array \"%s\".",
			sAbsoluteItem.c_str()) );
	size_t memAlloc = fakeDims[0];

	int *piDims;
	if ( memAlloc > 0 ) {
		try {
			piDims = new int[ memAlloc ];
		} catch (std::bad_alloc) {
			piDims = 0;
		}
		ASSERT( 0 != piDims );
		if ( 0 == piDims ) throw NotEnoughMemory("NetCDFReader","GetIntArray",
			"Could not allocate memory during serialization." );
		
		FillAbsoluteIntArray( piDims, NetCDFStructure::FakeDimMangle(sAbsoluteItem) );

	} else {
		throw NetCDFError( "NetCDFReader","GetZeroDimensions",
			SprintString("Could not find dimensions for integer array \"%s\".",
			sAbsoluteItem.c_str()) );
	}
	
	dims.clear();
	std::copy( piDims, piDims + fakeDims[0], std::back_inserter( dims ) );
	delete[] piDims;
}


void
NetCDFReader::FillAbsoluteIntArray( int* data,
								   const std::string& sAbsoluteItem )
{
	int ncArrayID = GetVariableID( ncID,
		NetCDFStructure::DataMangle(sAbsoluteItem) );
	
	int status = nc_get_var_int(ncID, ncArrayID, data);
	ASSERT( status == NC_NOERR );
	if (status != NC_NOERR) {
		throw NetCDFError("nc_get_var_int",sAbsoluteItem,nc_strerror(status));
	}
}


void
NetCDFReader::FillAbsoluteDoubleArray( double* data,
									   const std::string& sAbsoluteItem )
{
	int ncArrayID = GetVariableID( ncID,
		NetCDFStructure::DataMangle(sAbsoluteItem) );

	int status = nc_get_var_double(ncID, ncArrayID, data);
	ASSERT( status == NC_NOERR );
	if (status != NC_NOERR) {
		throw NetCDFError("nc_get_var_double",sAbsoluteItem,nc_strerror(status));
	}
}


int
NetCDFReader::GetVariableID( int ncID, const std::string& sAbsoluteItem )
{
	int ncArrayID;
	int status = nc_inq_varid(ncID, sAbsoluteItem.c_str(), &ncArrayID);
	ASSERT( status == NC_NOERR );
	if (status != NC_NOERR) {
		throw NetCDFError("nc_inq_varid",sAbsoluteItem,nc_strerror(status));}

	return ncArrayID;
}


size_t
NetCDFReader::GetDimensionLength( int ncID, int dimID )
{
	size_t iLength;
	int status = nc_inq_dimlen( ncID, dimID, &iLength );

	ASSERT( status == NC_NOERR );
	if ( status != NC_NOERR ) throw NetCDFError("NetCDFReader","GetDimensionLength",
		nc_strerror( status ) );

	ASSERT( iLength > 0 );
	if ( iLength == 0 ) throw NetCDFError("NetCDFReader","GetDimensionLength",
		"A dimension is equal to zero." );

	return iLength;
}


std::string
NetCDFReader::GetOldString( const std::string& sItemName )
{
	int ncLengthID;
	int status;

	string namesType = sItemName + "_nameLength";
	status = nc_inq_dimid(ncID, namesType.c_str(), &ncLengthID);
	ASSERT( NC_NOERR == status );
	if (status != NC_NOERR) {
		throw NetCDFError("nc_inq_dimid",namesType,nc_strerror(status));}

	size_t length;
	status = nc_inq_dimlen(ncID, ncLengthID, &length);
	ASSERT( NC_NOERR == status );
	if (status != NC_NOERR) {
		throw NetCDFError("nc_inq_dimlen",namesType,nc_strerror(status));}

	int ncTypeID;
	namesType = sItemName + "_name";
	status = nc_inq_varid(ncID, namesType.c_str(), &ncTypeID);
	ASSERT( NC_NOERR == status );
	if (status != NC_NOERR) {
		throw NetCDFError("nc_inq_varid",namesType,nc_strerror(status));}

	char *chartest = new char[length+1];
	status = nc_get_var_text(ncID, ncTypeID, chartest);
	ASSERT( NC_NOERR == status );
	if (status != NC_NOERR) {
		throw NetCDFError("nc_get_var_text",namesType,nc_strerror(status));}
	chartest[length] = 0;
	string type = chartest;
	delete [] chartest;
	return type;
}

