// NetCDFWriter.cpp: implementation of the NetCDFWriter class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/


#include "DMFirstHeader.h"
#include <memory>
#include "netcdf.h"
#include "DigitalMaterial.h"
#include <iostream>

#include "NetCDFWriter.h"
#include "NetCDFError.h"
#include "SprintString.h"
#include "DMLastHeader.h"

using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

NetCDFWriter::NetCDFWriter(const std::string &filename) : DMWriter(filename)
{
	// TRACE1("NetCDFWriter::NetCDFWriter Opening file %s\n", filename.c_str());

	int fileID;
	NetCDFStructure::CreateFile( filename, fileID );
	ncID = fileID;
	Close();
	bFileOpen = false;
}

NetCDFWriter::~NetCDFWriter() {}

/*! 
  If one is saving many MD objects to one file it is quite possible that the
  limit on the number of netCDF dimensions will be reached which will
  cause an exception to be thrown.
  
  For informational purposes (for the user) the value is returned by this 
  function. It is possible to change it--it is set by a define statement
  in netcdf.h; it requires recompiling netCDF though.
 */

int NetCDFWriter::GetMaxNumDimensions() const
{
	return NC_MAX_DIMS; 
}

/*! You can open a file by hand, but it will be opened for you if you
	forget.
	*/
void NetCDFWriter::Open()
{
	NetCDFStructure::Open( filename, NetCDFStructure::WRITE, ncID );
	bFileOpen = true;
}


void NetCDFWriter::Close()
{
	NetCDFStructure::Close( ncID );
	bFileOpen = false;
}

/*! This function recursively stores a Serializable object and
	all of its contents in a virtual directory format.
	*/
void NetCDFWriter::Put(const Serializable *serializable, const string &sItem)
{
	ASSERT( 0 != serializable );
	if ( 0 == serializable ) throw NetCDFError("NetCDFWriter","Put",
		"Passed a zero pointer to a serializable." );

	NetCDFStructure::CheckItemName( sItem );

	if ( !bFileOpen ) {
		//TRACE1( "NetCDFWriter::NetCDFWriter Saving object %s.\n", sItem.c_str());
		Open();
	}
	NetCDFStructure::DownDir( path, sItem );

	PutString(serializable->GetType(), "type");
	serializable->Save(this);

	NetCDFStructure::UpDir( path );
	if( NetCDFStructure::IsAtRoot( path ) ) {
		//TRACE1( "NetCDFWriter::NetCDFWriter Done saving object %s.\n",
		//	sItem.c_str());
		Close();
	}
}


void NetCDFWriter::PutBool( bool bOn, const string& sItem )
{
	int iBool = ( bOn ) ? 1 : 0;
	PutInt( iBool, sItem );
}


void NetCDFWriter::PutInt(int n, const string &sItem )
{
	if (!bFileOpen) Open();
	string sItemName = NetCDFStructure::AbsolutePath( path, sItem );

	BeginDefinition( ncID );
	int ncNID = DefineVariable( ncID, sItemName, NC_INT );
	EndDefinition( ncID );
	
	// Write n
	int status = nc_put_var_int(ncID, ncNID, &n);
	ASSERT( status == NC_NOERR );
	if (status != NC_NOERR) {
		throw NetCDFError("nc_put_var_int",sItemName,nc_strerror(status));}
}


void NetCDFWriter::PutDouble(double x, const string &sItem )
{
	if (!bFileOpen) Open();
	string sItemName = NetCDFStructure::AbsolutePath( path, sItem );

	BeginDefinition( ncID );
	int ncNID = DefineVariable( ncID, sItemName, NC_DOUBLE );
	EndDefinition( ncID );
	
	// Write x
	int status = nc_put_var_double(ncID, ncNID, &x);
	ASSERT( status == NC_NOERR );
	if (status != NC_NOERR) {
		throw NetCDFError("nc_put_var_double",sItemName,nc_strerror(status));}
}


void NetCDFWriter::PutString(const string &str, const string &sItem)
{
	if (!bFileOpen) Open();
	string sItemName = NetCDFStructure::AbsolutePath( path, sItem );

	vector<int> dims(1);
	dims[0] = str.size();

	if ( str.size() > 0 ) {
        	int ncTypeID = DefineArray( ncID, dims, sItemName, NC_CHAR );

		// Write string
		int status = nc_put_var_text(ncID, ncTypeID, str.c_str());
		ASSERT( status == NC_NOERR );
		if (status != NC_NOERR) {
			throw NetCDFError("nc_put_var_uchar",sItemName,nc_strerror(status));
		}

	} else {

		WriteZeroDimensions( dims, sItem);

	}
}


void NetCDFWriter::PutIntArray(const int *data, const vector<int> &dims,
							   const string& name)
{
	if (!bFileOpen) Open();
	bool isNonZero = CheckDims( (0==data), dims, name );

	string sAbsItem = NetCDFStructure::AbsolutePath( path, name );

	if ( isNonZero ) {

		int ncArrayID = DefineArray( ncID, dims, sAbsItem, NC_INT );

		// Write array
		int writeStatus = nc_put_var_int(ncID, ncArrayID, data);
		ASSERT( writeStatus == NC_NOERR );
		if ( writeStatus != NC_NOERR) {
			throw NetCDFError("nc_put_var_int",sAbsItem,nc_strerror(writeStatus));
		}
		
	} else {

		WriteZeroDimensions( dims, name );

	}
}


void NetCDFWriter::PutDoubleArray( const double *data, const vector<int> &dims,
								  const string &name )
{
	bool isNonZero = CheckDims( 0==data, dims, name );

	if (!bFileOpen) Open();
	string sAbsItem = NetCDFStructure::AbsolutePath( path, name );

	if ( isNonZero ) {

		int ncArrayID = DefineArray( ncID, dims, sAbsItem, NC_DOUBLE );

		// Write array
		int status = nc_put_var_double( ncID, ncArrayID, data);
		ASSERT( status == NC_NOERR );
		if (status != NC_NOERR) {
			throw NetCDFError("nc_put_var_double",sAbsItem,nc_strerror(status));
		}

	} else {

		WriteZeroDimensions( dims, name );

	}
}


////////////////////// PRIVATE FUNCTIONS /////////////////////////////////

void NetCDFWriter::BeginDefinition( int ncID )
{
	int status = nc_redef(ncID);
	ASSERT( status == NC_NOERR );
	if (status != NC_NOERR) {
		throw NetCDFError("nc_redef","BeginDefinition",nc_strerror(status));}
}


void NetCDFWriter::EndDefinition( int ncID )
{
	int status = nc_enddef(ncID);
	ASSERT( status == NC_NOERR );
	if (status != NC_NOERR) {
		throw NetCDFError("nc_enddef","EndDefinition",nc_strerror(status));}
}


int NetCDFWriter::DefineVariable( int ncID, const std::string& sItemName,
								 int type )
{
	int ncNID, zero = 0;
	nc_type ncType = static_cast<nc_type>( type );
	int status = nc_def_var(ncID, sItemName.c_str(), ncType, 0, &zero, &ncNID);
	ASSERT( status == NC_NOERR );
	if (status != NC_NOERR) {
		throw NetCDFError("nc_def_var",sItemName,nc_strerror(status));}

	return ncNID;
}

int NetCDFWriter::DefineDimension( int ncID, const std::string& sDimItem,
								  int size )
{
	int ncDimID;
	int status = nc_def_dim(ncID, sDimItem.c_str(), size, &ncDimID);
	ASSERT( status == NC_NOERR);

	// most likely (according to experience so far) reason to fail:
	if (status != NC_NOERR) {
		throw NetCDFError("nc_def_dim",sDimItem,nc_strerror(status));
	}

	return ncDimID;
}

int NetCDFWriter::DefineArray( int ncID, const vector<int>& dims,
							  const string sItem, int type )
{
	BeginDefinition( ncID );

	int* ncDimsID;
	try {
		ncDimsID = new int[ dims.size() ];
		if ( ncDimsID == 0 ) throw std::bad_alloc();

	} catch (std::bad_alloc) {
		ncDimsID = 0;
		ASSERT( ncDimsID != 0 );
		throw NetCDFError("NetCDFWriter","DefineArray",
			"Could not allocate a few integers.");
	}

	int ncArrayID;
	try {
		for(unsigned int i=0; i<dims.size(); ++i)
		{
			string nameDim = NetCDFStructure::DimMangle( sItem, i );
			ncDimsID[i] = DefineDimension( ncID, nameDim, dims[i] );
		}
		
		// Define the variable
		string nameData = NetCDFStructure::DataMangle( sItem );
		
		// nc_type is an enum.  We static cast so the NetCDF data type
		// won't show up in the header file.  I have a feeling this
		// routine will move to another class soon.
		nc_type ncType = static_cast<nc_type>( type );
		int defStatus = nc_def_var(ncID, nameData.c_str(), ncType, dims.size(),
			ncDimsID, &ncArrayID);
		ASSERT( defStatus == NC_NOERR );
		if (defStatus != NC_NOERR) {
			throw NetCDFError("nc_def_var",nameData,nc_strerror(defStatus));
		}
		
	} catch ( ... ) {
		delete[] ncDimsID;
		throw;
	}
	delete[] ncDimsID;

	EndDefinition( ncID );

	return ncArrayID;
}


void NetCDFWriter::WriteZeroDimensions( const vector<int>& dims,
									   const std::string& name )
{
	// Put dimensions into an integer array.
	int* piDims = new int[dims.size()];
	std::copy( dims.begin(), dims.end(), piDims );

	// Store that array in the file under a key name.
	vector<int> fakeDims( 1, dims.size() );
	PutIntArray( piDims, fakeDims, NetCDFStructure::FakeDimMangle(name) );

	delete[] piDims;
}


/*! If you call this with a shape which has zero size, it will throw
	an exception.
	*/
size_t NetCDFWriter::DimsToSize( const vector<int>& dims )
{
	vector<int>::const_iterator begin;

	size_t len = 1;
	for ( begin = dims.begin(); begin < dims.end(); begin++ ) {
		len *= *begin;
	}

	return len;
}


/*! Pass nonzero data and dimensions which are nonzero?  Good.
	Pass zero data and dimensions which have length zero?  Good, too.
	Pass dimensions vector of zero size? Baddy bad.
	Zero data and nonzero dimensions? Bad.
	Nonzero data and zero dimensions? Bad.
	*/
bool NetCDFWriter::CheckDims
	(
	bool isInvalidPointer,
	const vector<int>& dims,
	const string& name
	)
{
	ASSERT( !dims.empty() );
	if ( dims.empty() ) {
		throw NetCDFError( "NetCDFWriter", "PutIntArray",
			SprintString("Passed an empty shape array for \"%s\".",
			name.c_str() ) );
	}

	bool isNonZeroSize = ( DimsToSize( dims ) > 0 );
	bool isValidPointer = !isInvalidPointer;
	bool isGoodData = ( isNonZeroSize && isValidPointer );
	bool isZeroData = !isNonZeroSize;
	ASSERT( isGoodData || isZeroData );
	if ( !isGoodData && !isZeroData ) {
		if ( !isValidPointer )
			throw NetCDFError( "NetCDFWriter", "PutIntArray",
				SprintString("Passed a null pointer for the data \"%s\".",
				name.c_str()) );
		else if ( !isNonZeroSize )
			throw NetCDFError( "NetCDFWriter", "PutIntArray", SprintString(
			"Passed a nonzero data pointer for data \"%s\" of zero length.",
			name.c_str()) );
		else
			throw NetCDFError( "NetCDFWriter", "PutIntArray",SprintString(
				"This program logic is confused by the array \"%s\".",
				name.c_str()) );
	}
	
	return isNonZeroSize;
}
