// ProcessFactory.h interface for the ProcessFactory class
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef _PROCESSFACTORY_H_
#define _PROCESSFACTORY_H_

#ifndef SWIG
#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000
#endif

class ThreadImp; // Need know no more about it.
class RWLockImp;
class EggTimer;

//! Abstract factory to create process-related objects.
/**
 * This abstract factory creates Win32 or Unix specific objects for
 * handling processes.  It may be expanded to include other types
 * of objects, but I can't foresee a good name.
 *
 * This abstract factory is also an abstract class.  It is a singleton,
 * as suggested by the Gang of Four, but it does not instantiate itself.
 * That is,
 * you have to choose which base class to instantiate, for instance
 * WinProcessFactory or UnixProcessFactory.  You do that by calling
 * WinProcessFactory::Create() or UnixProcessFactory::Create()
 * early in the initialization of the code.  Without this, there
 * will be no way to instantiate threads and locks.
 * \author Drew Dolgert
 * \date 21 Feb 2000
 * \ingroup dmsystem
 */
class ProcessFactory {
public:
	//! Get an instance of a ProcessFactory.
	/*! This works from anywhere in your code.  There is only
		one global instance of a ProcessFactory. */
	static ProcessFactory *Instance() { return _instance; }
	//! Get an instance of a thread class.
	/*! You don't normally call this in your code.  It gets
		called by the Thread class who uses the ProcessFactory
		to decide on its implementation. */
	virtual ThreadImp* GetThread() = 0;
	//! Get an instance of a read-write lock.
	/*! You don't normally call this directly, either.
		\sa GetThread. */
	virtual RWLockImp* GetRWLock() = 0;
	virtual EggTimer*  GetEggTimer() = 0;
protected:
	ProcessFactory() {}
	static ProcessFactory *_instance;
};

#endif // _PROCESSFACTORY_H_
