// ProtocolFactory.h: interface for the ProtocolFactory class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef PROTOCOLFACTORY_H
#define PROTOCOLFACTORY_H

#include "DMFirstHeader.h"
#include <stack>

#include "DigitalMaterial.h"

class DMProtocol;

//! A singleton class that manages the creation and destruction of the protocols.
/*! The singleton is instantiated in the cpp file, so the user never cares 
about it. The ProtocolFactory creates and deletes the protocols used to 
communicate between processors. There is a global variable called dmProtocol 
that is a pointer to a DMProtocol and can be used from everywhere. By default 
dmProtocol is a LocalProtocol, where no inter-processor communication is 
enabled. If MPI_Init has been called before the ProtocolFactory is
instantiated, the current Protocol become an MPIProtocol. This will happen
automatically in Python. If using raw C++, you should first call MPI_Init(),
then ProtocolFactory::Instance()->SetProtocol(ProtocolFactory::World)
before any use of dmProtocol.

\sa DMProtocol. 
*/

class ProtocolFactory  
{
    //! The constructor (protected, feature of the singleton pattern).
    ProtocolFactory();
    //! The unique instance of the class (feature of the singleton pattern).
    static ProtocolFactory *_instance;

    //! A stack of protocols. 
    /*! If a class wants to use a specific protocol, it can set its favorite. Then
    before leaving, it should restore the previous one.
    */
    std::stack<DMProtocol *> protocolStack;
public:
    //! The function that allows to access the unique instance.
    static ProtocolFactory *Instance();
    //! The 2 possible communicators, for now: the MPI_World, and a local communicator.
    /*! It would be better to have them self register in the factory, like for the 
    serialization (see SerialFactory.h). It's not really necessary as long as we have
    only 2 communicators. However it is possible to define other communicators, 
    in the spirit of MPI, and extend the ProtocolFactory in order to deal with these
    cases.
    */
    enum Comm {World, Local};
    //! Set the current protocol to use the given communicator.
    void SetProtocol(Comm comm);
    //! Restore the provious protocol.
    void Reset();
    //! The destructor. Deletes all the remaining protocols.
    ~ProtocolFactory();
};

#endif // !defined PROTOCOLFACTORY_H
