// Rand.h interface for the Rand class
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef RAND_H
#define RAND_H

#include "DMFirstHeader.h"
#include <math.h>
#include <stdlib.h>
#include "DigitalMaterial.h"
#include "Serializable.h"

const int BIGPRIME = 899999963;

//! Ranmar pseudorandom number generator
/*! The algorithm is called Ranmar.  It is a Lagged Fibonnaci Generator
	proposed by Marsaglia and Zaman.
	
	I have two separate references for claimants to the original paper:
	G. Marsaglia, A. Zaman and W.-W Tsang, Stat Prob.
	Lett <b>9</b>, 35 (1990).
	Marsaglia, G.A., Toward a universal random number generator,
	Stat. Prob. Lett. <b>8,</b> 35 (1990).
	A reviews of it is in
	F. James, A review of pseudorandom number generators, Computer
	Physics Communications <b>60</b>, 329-44 (1990).
	A weakness in the algorithm is described in
	Coddington, P.D., Analysis of random number generators using
	Monte Carlo simulation, International Journal of Modern Physics,
	<b>C</b> 5, 547 (1994).  The online 
	<a href="http://nhse.cs.rice.edu/NHSEreview/96-2.html">NHSE Review by
	Paul Coddington</a> suggests using a much larger lag of (1279,1063).
  F. James, <CITE>Comp. Phys. Comm.</CITE> <STRONG>60</STRONG> (1990) p 329-344
	*/
class Rand : public Serializable
{
  static const double cd, cm;
	double c,u[97];
	int i97,j97 ;
	bool outputReady;
	double output;
	int iOriginalSeed;
public:
	//! The default creator seeds the generator always with the value 1.
	Rand();

	//! Constructor where you pass the rand seed.
#ifndef SWIG
	Rand(int seed);
#endif //SWIG

	//! Set the seed of the random number generator.
	void SetSeed(int seed) {
		ranmarin(abs(seed%BIGPRIME));
		outputReady=false;
	}

	int GetOriginalSeed( void ) const { return iOriginalSeed; }

	//! Return a number from the uniform distribution.
	double uniform()
	{
	  double uni;
	
	  uni=u[i97]-u[j97];
	  if (uni<0.0) uni+=1.0;
	  u[i97]=uni;
	  if (--i97<0) i97=96;
	  if (--j97<0) j97=96;
	  c-=cd;
	  if (c<0.0) c+=cm;
	  uni-=c;
	  if (uni<0.0) uni+=1.0;
	  return(uni);
	}

#ifndef SWIG
	//! For filling arrays
	void uniform(double d[],int n)
	{
		for(int i=0;i<n;i++)
		{
			double uni=u[i97]-u[j97];
			if (uni<0.0) uni+=1.0;
			u[i97]=uni;
			if (--i97<0) i97=96;
			if (--j97<0) j97=96;
			c-=cd;
			if (c<0.0) c+=cm;
			uni-=c;
			if (uni<0.0) uni+=1.0;
			d[i]=uni;
		};
	}
#endif //SWIG

	//! Gives all numbers weighted equally, min and max inclusive
	int discrete(int min, int max)
	{
		int ret= (int)(floor(uniform()*(max+1-min)+min));
//		assert(ret<=max && ret>=min);
		return ret;
	}


	//! Return a number from the gaussian distribution with spread sd.
	double gaussian(double sd)
	{
		double x,y,r,z;
		if(outputReady)
		{
			outputReady = false;
			return output*sd;
		}

		do
		{
			x=(uniform()*2.0)-1.0;
			y=(uniform()*2.0)-1.0;
			r=x*x+y*y;
		} while (r>=1.0);

		z=sqrt(-2.0*log(r)/r);
		output = x*z;
		outputReady=true;
		return y*z*sd;
	}

#ifndef SWIG
	/*! Serializable interface */
	//@{
	void Save( DMWriter *writer ) const;
	void Load( DMReader *reader );
#endif //SWIG
	const std::string &GetType() const;
	//@}


private:
	void ranmarin(int ijkl)
	{
	  int i,ii,j,jj,k,l,m ;
	  double s,t ;

      int ij=ijkl/30082;
      int kl=ijkl-30082*ij;

	  i=((ij/177)%177)+2 ;
		j=(ij%177)+2 ;
		k=((kl/169)%178)+1 ;
		l=kl%169 ;
		for (ii=0;ii<97;ii++) {
	    s=0.0 ;
			t=0.5 ;
			for (jj=0;jj<24;jj++) {
	      m=(((i*j)%179)*k)%179 ;
				i=j;
				j=k;
				k=m;
				l=(53*l+1)%169;
				if (((l*m)%64)>=32) s+=t;
				t*=0.5;
			}
			u[ii]=s;
		}
		c=362436.0/16777216.0;
		i97=96;
		j97=32;
	}
	static const std::string className;
};	

#endif
