/*
 * SimQA: Copyright (C) Jakob Schiotz 1996, 1997
 *
 * IBM POWER2 optimized vector library:
 * r = a * b * c
 * r = r * a * b
 */

#include "vectools.h"

/* r[] = a[] * b[] * c */
void vec_mul_mul_scal(double *res, const double *a, const double *b, double c, int n)
{
#ifdef AIX
#pragma disjoint (*res, *a, *b, c)
#endif
 
    int i, limit;
    double a0, a1, a2, a3;
    double b0, b1, b2, b3;

    limit = n-3;
    
    for (i = 0; i < limit; i+=4)
    {
	a0 = a[i];
	a1 = a[i+1];
	a2 = a[i+2];
	a3 = a[i+3];
	b0 = b[i];
	b1 = b[i+1];
	b2 = b[i+2];
	b3 = b[i+3];
	res[i] = a0 * b0 * c;
	res[i+1] = a1 * b1 * c;
	res[i+2] = a2 * b2 * c;
	res[i+3] = a3 * b3 * c;
    }
    for (i = n - n % 4; i < n; i++)
    {
	a0 = a[i];
	b0 = b[i];
	res[i] = a0 * b0 * c;
    }
}

/* r[] = a[] * b[] * c[] */
void vec_mul_mul(double *res, const double *a, const double *b, const double *c, int n)
{
#ifdef AIX
#pragma disjoint (*res, *a, *b, *c)
#endif
   
    int i, limit;
    double a0, a1, a2, a3;
    double b0, b1, b2, b3;
    double c0, c1, c2, c3;

    limit = n-3;
    
    for (i = 0; i < limit; i+=4)
    {
	a0 = a[i];
	a1 = a[i+1];
	a2 = a[i+2];
	a3 = a[i+3];
	b0 = b[i];
	b1 = b[i+1];
	b2 = b[i+2];
	b3 = b[i+3];
	c0 = c[i];
	c1 = c[i+1];
	c2 = c[i+2];
	c3 = c[i+3];
	res[i] = a0 * b0 * c0;
	res[i+1] = a1 * b1 * c1;
	res[i+2] = a2 * b2 * c2;
	res[i+3] = a3 * b3 * c3;
    }
    for (i = n - n % 4; i < n; i++)
    {
	a0 = a[i];
	b0 = b[i];
	c0 = c[i];
	res[i] = a0 * b0 * c0;
    }
}

/* r[] = r[] * a[] * b */
void vec_self_mul_mul_scal(double *res, const double *a, double b, int n)
{
#ifdef AIX
#pragma disjoint (*res, *a, b)
#endif
    
    int i, limit;
    double a0, a1, a2, a3;
    double s0, s1, s2, s3;

    limit = n-3;
    
    for (i = 0; i < limit; i+=4)
    {
	a0 = a[i];
	a1 = a[i+1];
	a2 = a[i+2];
	a3 = a[i+3];
	s0 = res[i];
	s1 = res[i+1];
	s2 = res[i+2];
	s3 = res[i+3];
	res[i] = s0 * a0 * b;
	res[i+1] = s1 * a1 * b;
	res[i+2] = s2 * a2 * b;
	res[i+3] = s3 * a3 * b;
    }
    for (i = n - n % 4; i < n; i++)
    {
	a0 = a[i];
	s0 = res[i];
	res[i] = s0 * a0 * b;
    }
}


/* r[] = a[] * b[] * c[offset[]] */
void vec_mul_mul_indir(double *res, const double *a, const double *b, const double *c,
		       const int *offset, int n)
{
#ifdef AIX
#pragma disjoint (*res, *a, *b, *c)
#endif
    
    int i, limit;
    double a0, a1, a2, a3;
    double b0, b1, b2, b3;
    double c0, c1, c2, c3;

    limit = n-3;
    
    for (i = 0; i < limit; i+=4)
    {
	a0 = a[i];
	a1 = a[i+1];
	a2 = a[i+2];
	a3 = a[i+3];
	b0 = b[i];
	b1 = b[i+1];
	b2 = b[i+2];
	b3 = b[i+3];
	c0 = c[offset[i]];
	c1 = c[offset[i+1]];
	c2 = c[offset[i+2]];
	c3 = c[offset[i+3]];
	res[i] = a0 * b0 * c0;
	res[i+1] = a1 * b1 * c1;
	res[i+2] = a2 * b2 * c2;
	res[i+3] = a3 * b3 * c3;
    }
    for (i = n - n % 4; i < n; i++)
    {
	a0 = a[i];
	b0 = b[i];
	c0 = c[offset[i]];
	res[i] = a0 * b0 * c0;
    }
}

/* r[] = r[] * a[] * b[offset[]] */
void vec_self_mul_mul_indir(double *res, const double *a, const double *b,
			    const int *offset, int n)
{
#ifdef AIX
#pragma disjoint (*res, *a, *b)
#endif
    
    int i, limit;
    double s0, s1, s2, s3;
    double a0, a1, a2, a3;
    double b0, b1, b2, b3;

    limit = n-3;
    
    for (i = 0; i < limit; i+=4)
    {
	s0 = res[i];
	s1 = res[i+1];
	s2 = res[i+2];
	s3 = res[i+3];
	a0 = a[i];
	a1 = a[i+1];
	a2 = a[i+2];
	a3 = a[i+3];
	b0 = b[offset[i]];
	b1 = b[offset[i+1]];
	b2 = b[offset[i+2]];
	b3 = b[offset[i+3]];
	res[i] = s0 * a0 * b0;
	res[i+1] = s1 * a1 * b1;
	res[i+2] = s2 * a2 * b2;
	res[i+3] = s3 * a3 * b3;
    }
    for (i = n - n % 4; i < n; i++)
    {
	s0 = res[i];
	a0 = a[i];
	b0 = b[offset[i]];
	res[i] = s0 * a0 * b0;
    }
}

