// SprintString.cpp implementation of the SprintString class
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "DMFirstHeader.h"
#include <stdexcept>
#include <new>
#include <stdarg.h>
#include <stdio.h>
#include "DigitalMaterial.h"
#include "SprintString.h"
#include "DMException.h"
#include "DMLastHeader.h"

const int SprintString::BufferCnt = 400;
char SprintString::threadUnsafeBuffer[BufferCnt];

/*!	This creates an STL string from a printf-like constructor.
	It uses temporary storage on the stack for small strings but
	dynamically allocates memory for larger strings.  If you use
	this for exceptions, avoid huge strings, especially if, for
	instance, the exception declares that you are out of memory.
	*/
SprintString::SprintString( const char* sMessage, ... )
{
	ASSERT( 0 != sMessage );
	if ( 0 == sMessage ) throw DMException("SprintString","SprintString",
		"Passed null pointer in constructor." );
	
	char buffer[ BufferCnt ];
	va_list marker;
	
#ifdef HAVE_VSNPRINTF
	// Try to write the string to a buffer on the stack.
	va_start( marker, sMessage );
	int writtenNotIncludingNull = vsnprintf( buffer, BufferCnt-1, sMessage, marker );
	va_end( marker );

	// If the string fit in the buffer,
	if ( writtenNotIncludingNull >= 0 ) {
		this->append( buffer, writtenNotIncludingNull );

	} else {
		// Write the string to a dynamically allocated buffer.
		char *sBuff;
		int bufferMultiple = 1;
		int messageCnt = strlen( sMessage ) + 1;
		do {
			int buffCnt = messageCnt + bufferMultiple * BufferCnt;
			bufferMultiple++;
                        
			try {
				sBuff = new char[ buffCnt ];
				} catch ( std::bad_alloc& ) {
				sBuff = 0;
				}
			ASSERT( 0 != sBuff );
			if ( 0 == sBuff ) {
				// If we can't allocate memory,
				//   then use our inviolable stash of memory to get the message out.
				va_start( marker, sMessage );
				vsnprintf( threadUnsafeBuffer, BufferCnt-1, sMessage, marker );
				va_end( marker );
				this->append( threadUnsafeBuffer );
				return;
			}

			va_start( marker, sMessage );
			writtenNotIncludingNull = vsnprintf( sBuff, buffCnt-1, sMessage, marker );
			va_end( marker );

		} while ( writtenNotIncludingNull < 0 );
		
		this->append( sBuff, writtenNotIncludingNull );
		delete[] sBuff;
	}
#else

	// Try to write the string to a buffer on the stack.
	va_start( marker, sMessage );
	int writtenNotIncludingNull = vsprintf( buffer, sMessage, marker );
	va_end( marker );

	ASSERT( writtenNotIncludingNull < BufferCnt );
	if ( writtenNotIncludingNull >= BufferCnt ) {
	  throw DMException( "SprintString", "SprintString", "Overwrote internal buffer.  Memory error.");
	}
	this->append( buffer, writtenNotIncludingNull );

#endif
}
