// Tree.h: interface for the Tree class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef TREE_H
#define TREE_H

#include "DMFirstHeader.h"
#include <vector>
#include "Shape.h"

template<class Type>
class Tree  
{
    typedef std::vector<Type *>::iterator leafIterator;
public:
    virtual ~Tree();
    int Size() const {return shape.GetSize();}
    int NumberOfBranches() const;
    int NumberOfLeaves() const {return flatList.size();}
    bool IsLeaf() const {return (branches == 0);}
    const Shape &GetShape() const {return shape.Global();}
    void SetShape(const Shape &shape);
    Type *GetBranch(int i);
    const Type *GetBranch(int i) const;
    virtual void RemoveBranch(Type *branchToRemove);
    virtual void Merge(const Tree<Type> *tree);
    virtual void Attach(Type *tree);
    void UpdateTree();
    virtual void Clear();
    leafIterator begin(){return flatList.begin();}
    leafIterator end() {return flatList.end();}
    Type *GetLeaf(int i){return flatList[i];}
    void SetUpperNode(Type *upperNode) {this->upperNode = upperNode;}
protected:
    Tree();
    virtual void Allocate(const Shape &newShape);
    virtual Type *LOAAllocator() = 0;
    Shape shape;
    std::vector<Type *> *branches;
    std::vector<Type *> flatList;
    Type *upperNode;
    Type *that;
};

// CONSTRUCTOR
template<class Type>
Tree<Type>::Tree()
{
    branches = 0;
    upperNode = 0;
    flatList.push_back(that);
}

// DESTRUCTOR
template<class Type>
Tree<Type>::~Tree()
{
    if(upperNode) upperNode->RemoveBranch(that);
    ASSERT(upperNode == 0);
    Clear();
}

template<class Type>
void Tree<Type>::Clear()
{
    if(branches)
    {
        for(std::vector<Type *>::reverse_iterator typ=branches->rbegin(); 
            typ != branches->rend(); ++typ) 
                delete *typ;
        delete branches;
        branches = 0;
    }
    shape = Shape(0);
    flatList.clear();
    flatList.push_back(that);
}

template<class Type>
void Tree<Type>::Allocate(const Shape &newShape)
{
    if(!shape.IsSimilar(newShape))
    {
        Clear();
        if(!newShape.IsLeaf())
        {
            branches = new std::vector<Type *>;
            flatList.clear();
            for(int i=0; i<newShape.GetNSubShapes(); ++i)
            {
                Type *t = LOAAllocator();
                t->SetShape(newShape.GetSubShape(i));
                Attach(t);
            }
        }
    }
    else
    {
        if(newShape.IsLeaf())
        {
            shape = newShape;
            UpdateTree();
        }
        else
        {
            ASSERT(branches);
            for(int i=0; i<branches->size(); ++i)
                (*branches)[i]->SetShape(newShape.GetSubShape(i));
        }
    }
}

template<class Type>
void Tree<Type>::SetShape(const Shape &shape)
{
    Allocate(shape);
}

template<class Type>
int Tree<Type>::NumberOfBranches( ) const
{
    if (0 != branches) return branches->size();
    return 0;
}

//! access the branches of a tree
template<class Type>
Type *Tree<Type>::GetBranch(int i)
{
    if(branches) return (*branches)[i];
    else throw DMException("Tree<Type>", "GetBranch", "attempt to access branches of a leaf"); // we may want to accept i=0 in this case. 
}

template<class Type>
const Type *Tree<Type>::GetBranch(int i) const
{
    if(branches) return (*branches)[i];
    else throw DMException("Tree<Type>", "GetBranch", "attempt to access branches of a leaf"); // we may want to accept i=0 in this case. 
}


template<class Type>
void Tree<Type>::Merge(const Tree<Type> *tree)
{
    ASSERT(branches); // the leaves are responsible for merging the data
    for(int i=0; i<branches->size(); ++i) (*branches)[i]->Merge(tree->GetBranch(i));
}

template<class Type>
void Tree<Type>::Attach(Type *tree)
{
    if(!branches)
    {
        if(shape.GetSize() > 0)
            throw DMException("Tree<Type>", "Attach", "Cannot attach an branch to a leaf array");
        Clear();
        branches = new std::vector<Type *>;
    }
    branches->push_back(tree);
    UpdateTree();
    tree->SetUpperNode(that);
}

template<class Type>
void Tree<Type>::RemoveBranch(Type *branchToRemove)
{
    if(!branches) throw DMException("Tree<Type>", "RemoveBranch", "Cannot remove a branch from a leaf");
    std::vector<Type *>::iterator found = std::find(branches->begin(), branches->end(),branchToRemove);
    if(found == branches->end()) throw DMException("Tree<Type>", "RemoveBranch", "branch to remove not found");
    (*found)->SetUpperNode(0);  // we also want to notify the branch that we cut our relations!
    branches->erase(found);
    UpdateTree();
}


template<class Type>
void Tree<Type>::UpdateTree()
{
    flatList.clear();
    if(branches)
    {
        shape = Shape(); // sets it as an empty composite
        for(std::vector<Type *>::iterator iter=branches->begin(); iter != branches->end(); ++iter)
        {
            for(std::vector<Type *>::iterator branch=(*iter)->flatList.begin(); branch!=(*iter)->flatList.end(); ++branch)
                flatList.push_back(*branch);
            shape.Attach((*iter)->shape);
        }
    }
    else
    {
        flatList.push_back(that);
        // shape is up to date for a leaf.
    }
    if(upperNode) upperNode->UpdateTree();
}

#endif // !defined TREE_H
