/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/
#include "StillingerWeber.h"

// This is a main function designed to test the StillingerWeberPotential by performing a simulation more or less the same as the one in the original SW paper (PRB 31, 5262 (1985)). ie it runs the same number of atoms in the same size box. It can check energy conservation and measure the average potential energy for a range of temperatures. The main difference is the choice of thermostat: we use Langevin dynamics while they used constant energy dynamics following a rescaling of the momenta to get the desired kinetic energy.


void main()
{

    cout.precision(16);
    try 
    {
#ifdef _WIN32
		WinProcessFactory::Create();
#else
		UnixProcessFactory::Create();
#endif
		int plotting      = 1;
		int drawRadius = 20;
		int stepsBetweenDraws = 40;
		
		double dt		= 0.005;  // same as SW paper
		double dtQM	= 0.04;
		double dtNEB      = 0.015;
		double dtContinuumMover = 0.04;
		double tolQM	= 1.0e-18;
		double tolNEB = 1.0e-9;
		double tolConjGrad = 1.0e-10;	
		int maxItConjGrad = 10;
		int maxQuickMinSteps = 60;		
		


		// simulation parameters
		double latticeConst = 5.430; // Angstronms for EDIP 2.5922152 for SW
		double cellLength = 3 * latticeConst; 
		double lengths[DIMENSION] = {cellLength,cellLength,cellLength};
		// knowing there are 216 atoms, and 8 atoms per conventional FCC unitcell in a diamond structure, that means 27=3 cubed conventional FCC cells


		double startTemp = 0.03;
		double endTemp = 0.15;
		double tempStep = 0.1;
		double friction  = 0.1;
		int nAveragingSteps = 1500;
		int nEquilSteps = 500;

		int coord;
		
		
		// create the main objects
		DynamicListOfAtoms *atoms  = new DynamicListOfAtoms();
		BoundaryConditions *bc = new SimplePeriodicBoundaryConditions(lengths);
		//Potential *potential    = new CutLennardJonesPotential();        
		//Potential *potential    = new StillingerWeberPotential();        
		Potential *potential    = new SiliconEDIPPotential();        
		double boundaryWidth = potential->GetCutoffDistance();
		NeighborLocator *neighborLocator = new SimpleNeighborLocator(potential->GetCutoffDistance());      
		//NeighborLocator *neighborLocator = new CellNeighborLocator(potential->GetCutoffDistance(),0.2*potential->GetCutoffDistance());      
		LangevinAtomsMover *langevin= new LangevinAtomsMover(potential,dt);
		//CompositeLangevinAtomsMover *langevin= new CompositeLangevinAtomsMover(potential,dt);
		VerletAtomsMover *verlet = new VerletAtomsMover(potential,dt);
		QuickMinAtomsMover *qm = new QuickMinAtomsMover(potential, dtQM, tolQM);
		ConjugateGradientMinimizer *conjGrad = new ConjugateGradientMinimizer(tolConjGrad,maxItConjGrad);
		conjGrad->SetLineMinimizer(0);
		SteepestDescentMinimizer *steepest = new SteepestDescentMinimizer(1.e-4,10000,0.005);
		steepest->Verbose();     
		
		
		// Initialize the atoms in a diamond structure (216 atoms, cubic cell, side 7.64614)
		DiamondLattice lattice(latticeConst);

		RectangularClusterInitializer cluster(lengths,&lattice);
		double clusterCenter[DIMENSION];
		for(coord=0;coord<DIMENSION;++coord)
			clusterCenter[coord] = lengths[coord]/2.;
		cluster.SetCenter(clusterCenter);
		cluster.Create(atoms);

		cout << "Number of atoms is " << atoms->GetNumber() << endl;

		atoms->SetBoundaryConditions(bc);
		atoms->SetNeighborLocator(neighborLocator);

		atoms->SetTypeName("mainAtoms");


		// Attach a PlotAtomsObserver
#ifndef _WIN32
		PlotAtomsObserver *plot = NULL; 
		if(plotting)
		{
			plot = new PlotAtomsObserver(lengths[0],lengths[1],drawRadius);
			atoms->AddObserver(plot);
			atoms->Notify();
		}
#endif // _WIN32		

		double averagePotentialEnergy;


		// loop through temperatures 
		for(double temperature=startTemp;temperature<=endTemp;temperature+=tempStep)
		{
		   

			langevin->SetFriction(friction);
			/*	double multiFriction[DIMENSION];
			for(int coord=0;coord<DIMENSION;++coord)
				multiFriction[coord] = friction;
				langevin->AddFriction("mainAtoms",multiFriction);*/
			langevin->SetTemperature(temperature);
			
			// move for 500 time steps (equilibration)
			langevin->SetNSteps(nEquilSteps);			
			langevin->Move(atoms);


			// Move the atoms for 1000 time steps, drawing every 50, also sampling for averages (later)
			langevin->SetNSteps(stepsBetweenDraws);
			verlet->SetNSteps(stepsBetweenDraws);
			int nMajorSteps = nAveragingSteps/stepsBetweenDraws;

			averagePotentialEnergy = 0.;

			for(int movidx=0;movidx<nMajorSteps;++movidx)
			{
				langevin->Move(atoms); 
				//verlet->Move(atoms); // for checking energy conservation
				atoms->Notify();
				// calculate some averages
				double potentialEnergy = potential->CalculateEnergy(atoms);
				double kineticEnergy = atoms->GetKineticEnergy();
				averagePotentialEnergy += potentialEnergy;
				cout << "Potential Energy per atom = " << potentialEnergy/atoms->GetNumber() << endl;
				//cout << "Total energy " << potentialEnergy + kineticEnergy << endl;
				
			}
			
			averagePotentialEnergy /= nMajorSteps*atoms->GetNumber();
			cout << "Temperature: " << temperature << " PE per atom: " << averagePotentialEnergy << endl;
		}
		
	
		delete atoms; 
		delete bc;
		delete neighborLocator;
		delete potential;
		delete langevin;
		delete qm;
		delete conjGrad;
		delete steepest;

	
    }
	catch(DMException exception)
    {
		std::cerr << exception;
    }
	catch(std::runtime_error error)
    {
		std::cerr << "std::runtime_error: " << error.what() << endl;
    }
	catch(...)
    {
		std::cerr << "Unknown error" << endl;
    }
	
	return 0;


}
