// AtomicSurfaceInitializer.h: interface for the AtomicSurfaceInitializer class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef ATOMICSURFACEINITIALIZER_H
#define ATOMICSURFACEINITIALIZER_H

#include "DMFirstHeader.h"
#include "DigitalMaterial.h"
#include "AtomsInitializer.h"

class ListOfAtoms;
class BravaisLatticeWithBasis;
class SkewPeriodicBoundaryConditions;

//! Initializes a surface---a layer of dynamic atoms over a layer of fixed atoms; alternatively (mode 1), it creates a single rectangular cluster with one side corresponding to the given surface orientation. 

/*!
  There are two different modes of operation. One might have different classes
  but a lot of code is the same. Maybe we will have a second class which 
  derives from the first...for now, one class, two modes.

  In mode 0:

  The spbc is set according to the surface orientation, two layers of
  atoms, one fixed, one free, are created and the top atoms have their z
  coordinate equal to 0.

  In mode 1: spbc is not changed but an error is thrown if the SuperCellVectors
  are not consistent with the orientation. A single rectangular cluster is
  created of given height and position in the z-direction

  */

#if DIMENSION == 3

class AtomicSurfaceInitializer : public AtomsInitializer
{
public:
	AtomicSurfaceInitializer( BravaisLatticeWithBasis *lattice, int millerIndices[DIMENSION], int mode = 0, SkewPeriodicBoundaryConditions *spbc = 0);
	virtual ~AtomicSurfaceInitializer();
	void SetMode(int mode);
	int GetMode() const { return mode; }
	void SetDepthOfFixedLayer(double depthOfFixedLayer);
	double GetDepthOfFixedLayer() const {return depthOfFixedLayer;}
	void SetDepthOfDynamicLayer(double depthOfDynamicLayer);
	double GetDepthOfDynamicLayer() const {return depthOfDynamicLayer;}
	void SetMillerIndices(const int millerIndices[DIMENSION]);
	const int *GetMillerIndices() const {return (const int *)millerIndices;}
	void SetMinimalLength(double minimalLength);
	double GetMinimalLength() const {return minimalLength;}

	void SetZCenter(double zCenter) { this->zCenter = zCenter; }
	double GetZCenter() const { return zCenter; }

	void ConfigureBoundaryConditions();

	const double *GetCenter() const {return center;}	//! Rough center (for graphics)
	double GetMaxSize() const;	//! Rough size (for graphics)

	void Create(ListOfAtoms *atoms);

	void SetBravaisLattice(BravaisLatticeWithBasis *lattice);
	BravaisLatticeWithBasis *GetBravaisLattice() const;

	bool SimplifiedVector(const int k,const int l, const int m) ;
private:
	void CreateModeZero(ListOfAtoms *atoms);
	void CreateModeOne(ListOfAtoms *atoms);
	BravaisLatticeWithBasis *lattice;
	int mode;
	double center[DIMENSION];
	double zCenter;
	double surfaceVectors[DIMENSION][DIMENSION];
	double superCellLength;
	double superCellWidth;
	double depthOfFixedLayer;
	double depthOfDynamicLayer;
	double minimalLength;
	int millerIndices[DIMENSION];
	void ResetPrivateData();
	SkewPeriodicBoundaryConditions *spbc;
	static const double TINY;
};

#endif // DIMENSION == 3

#endif // !defined(ATOMICSURFACEINITIALIZER_H)
