// ClusterInitializer.cpp: implementation of the ClusterInitializer class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "DMFirstHeader.h"
#include <math.h>
#include <vector>
#include <map>
#include "DigitalMaterial.h"

#include "ClusterInitializer.h"
#include "BravaisLatticeWithBasis.h"
#include "DMArray.h"
#include "TinyVector.h"
#include "ListOfAtoms.h"

#include "DMLastHeader.h"

using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

ClusterInitializer::ClusterInitializer(const BravaisLatticeWithBasis *bravais)
{
    
    for (int coord=0; coord<DIMENSION; coord++) clusterCenter[coord]=0.0;
    this->bravais = bravais;
}

ClusterInitializer::~ClusterInitializer()
{
	// cout << "ClusterInitializer deleted" << endl;
}

void ClusterInitializer::SetBravaisLattice(const BravaisLatticeWithBasis *bravais)
{
    this->bravais = bravais;
}

void ClusterInitializer::Create(ListOfAtoms *atoms)
{
    const int maxCells = SetMaxSize(); 
	
    const unsigned int nAtomsInCell = bravais->GetNAtomsPerCell();
    DMArray<double, DIMENSION> *positionsBuild = 
        new DMArray<double, DIMENSION>[nAtomsInCell];
    int *nAtomsInArrays = new int[nAtomsInCell];
    for(unsigned int i=0; i<nAtomsInCell; ++i)
    {
        nAtomsInArrays[i] = 0;
        positionsBuild[i].SetShape(maxCells);
    }

    int zeroCenter[DIMENSION];
    for (int coord=0; coord<DIMENSION; coord++) zeroCenter[coord] = 0;
    // Recursion to work in general dimension:
	
    MakeSlab(zeroCenter, DIMENSION, positionsBuild, nAtomsInArrays);	
    // we get rid of the unused space. Necessary before calling SetCartesianPositions()
    unsigned int atom;
    for(atom=0; atom<nAtomsInCell; ++atom)
    {
        ASSERT(nAtomsInArrays[atom] <= maxCells);
        positionsBuild[atom].SetShape(nAtomsInArrays[atom]);
    }

    // figure out the number of different types:
    const vector<string> atomTypes = bravais->GetAtomTypes();

    map<string, vector<int> > typesOccurence;
    for(atom=0; atom<atomTypes.size(); ++atom)
        typesOccurence[atomTypes[atom]].push_back(atom);

    map<string,vector<int> >::const_iterator typeIter;
#ifndef NDEBUG
    int tot =0;
    for(typeIter  = typesOccurence.begin();
        typeIter != typesOccurence.end(); ++typeIter) tot += typeIter->second.size();
    ASSERT(tot == atomTypes.size());
#endif //NDEBUG

    const int nTypes = typesOccurence.size();

    DMArray<double, DIMENSION> positions;
    if(nTypes != 1)
    {
        for(int type=0; type<nTypes; ++type)
            positions.Attach(new DMArray<double, DIMENSION>);
             
    }
    int type = 0;
    for(typeIter  = typesOccurence.begin();
        typeIter != typesOccurence.end(); ++typeIter)
    {
        const vector<int> &sameTypes = typeIter->second;
        for(unsigned int i=0; i<sameTypes.size(); ++i)
        {
            positions.GetLeaf(type)->Merge(&positionsBuild[sameTypes[i]]);
        }
        ++type;
    }
    
    // Set number of atoms and positions of atoms
    atoms->SetShape(positions.GetShape());
    atoms->SetCartesianPositions(&positions);

    ASSERT(atoms->NumberOfLeaves() == nTypes);
    type = 0;
    for(typeIter  = typesOccurence.begin();
        typeIter != typesOccurence.end(); ++typeIter)
    {
        atoms->GetLeaf(type++)->SetTypeName(typeIter->first);
    }
    
    delete [] positionsBuild;
}

void ClusterInitializer::SetCenter(double clusterCenter[DIMENSION])
{
    for (int coord=0; coord<DIMENSION; coord++) 
    {
        this->clusterCenter[coord] = clusterCenter[coord];
    }
}

void ClusterInitializer::MakeSlab(int center[DIMENSION],
                                  int subDimension,
                                  DMArray<double, DIMENSION> *positions,
                                  int *nAtomsInArrays)
{
    int newCenter[DIMENSION];
    for(int coord=subDimension; coord<DIMENSION; coord++)
        newCenter[coord] = center[coord];
    if (subDimension > 1)
    {
        newCenter[subDimension-1] = minSize[subDimension-1];
        for (int index = minSize[subDimension-1]; index <= maxSize[subDimension-1]; index++)
        {
            MakeSlab(newCenter, subDimension-1, positions, nAtomsInArrays);
            newCenter[subDimension-1]++;
        }
    }
    else	// One dimension left
    {
        double potentialPos[DIMENSION];
        double latticePotentialPos[DIMENSION];
        const int nAtomsPerCell = bravais->GetNAtomsPerCell();
        for (int index = minSize[0]; index <= maxSize[0]; index++)
        {
            newCenter[0] = index;
            for(int intraCell=0; intraCell<nAtomsPerCell; intraCell++)
            {
                for(int coord=0; coord<DIMENSION; ++coord)
                {
                    latticePotentialPos[coord] = newCenter[coord] +
                        bravais->GetAtomsCoordinatesInCell(intraCell)[coord];
                }
                bravais->CartesianCoordinatesFromLatticeCoordinates(potentialPos,
                                                            latticePotentialPos);
                if (Inside(potentialPos))
                {
                    for (int coord=0; coord<DIMENSION; coord++) 
                    {
                        positions[intraCell][nAtomsInArrays[intraCell]][coord] = 
                            potentialPos[coord];
                    }
                    nAtomsInArrays[intraCell]++;
                }
            }
        }
    }
    
}


double ClusterInitializer::GetMaxSize() const
{
    // Returns rough maximum size for graphics
    
    double max = 0;
    for (int coord=0; coord<DIMENSION; coord++)
    {
        double br2 = 0;	// Squared length of Bravais lattice vector
        for (int coord2=0; coord2<DIMENSION; coord2++)
        {
            double brComponent = bravais->GetLatticeVectors()[coord][coord2];
            br2 += brComponent * brComponent;
        }
        double extent = (double) (maxSize[coord] - minSize[coord])*sqrt(br2);
        if (extent > max) max = extent;
    }
    return max;
}
