// AtomsExtension.h: interface for the AtomsExtension class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef ATOMSEXTENSION_H
#define ATOMSEXTENSION_H


#include "DMFirstHeader.h"
#include "DigitalMaterial.h"
#include "DMArray.h"
#include "AttributeError.h"
#include <vector>
class ListOfAtoms;
class Potential;

/*

  This class is supposed to be an interface for cases where an LOA consists of
  a 'core' branch(es) plus a 'boundary' branch(es) whose positions are 
  parametrically determined by a set of parameter values.
  
  The main functions required in this interface are  (1) SetParameters--to
  which are passed the parameter values and the LOA. This function adjusts the
  appropriate atoms' positions according to the parameter values and (2) 
  CalculateForces(), to which are passed current parameter values, and a full
  list of atomic forces. The function calculates the generalize forces on the
  parameters.

  There are some choices which could be otherwise (maybe it would be better)
  (1) It is assumed that one AtomsExtension object can be used for many
  LOA objects (representing copies of the system). This sometimes requires
  extra overhead to maintain consistency---which s why we pass parameter values
  into CalculateForces--for example if another external object is being used
  such as a dislocation maker, functions like the gradient of displacement
  are needed to calculate the generalized force. The DislocationMaker needs
  to have its own parameters (which are some of the AtomsExtensions's
  parameters) set to values corresponding to the current LOA. we could avoid
  this by having an AtomsExtension object (and DislocationMaker etc) for each
  copy. It's not clear that this is a better situation however.


  \author nbailey
  \ingroup moleculardynamics
  \date October 2000

  (this comment written October 2001, however)
 */


class AtomsExtension
{
public:
	AtomsExtension(){}
	virtual void SetParameters(const double *parameters,ListOfAtoms *atoms)=0;
	virtual int GetNParameters() const=0;
	virtual int GetMainList() const=0;

	//! Calculates forces on atoms and atomic contribution to energy
	virtual double AtomicForcesEnergy(ListOfAtoms *atoms, const double *parameters, Potential *potential, DMArray<double,DIMENSION> *forces=0)=0;
	//! Pass in a force array and fill it in with parameter forces, using previously calculated atomic forces; also calculate mass matrix
	virtual double ParameterForcesEnergy(ListOfAtoms *atoms, const double *parameters, double *parameterForces, double *kineticMatrix, DMArray<double,DIMENSION> *forces, double atomicEnergy, double *parameterForceSq)=0;
	//! Increment parameter velocities from previously calculate mass matrix
	virtual void IncrementVelocitiesFromForces(const double *parameterForces, double *kineticMatrix, double *parameterVelocities, ListOfAtoms *atoms,double dt)=0;
	//! add contributions to total energy which are functions of parameters alone
	//virtual double CompleteEnergy(ListOfAtoms *atoms, const double *parameters, double atomicEnergy)=0;
	virtual double GetAtomicEnergy(ListOfAtoms *atoms, const double *parameters, int i, Potential *potential)=0;


	virtual double GetTotalForceSq()=0;

	virtual int GetQMBranch(ListOfAtoms *atoms)=0;
	

protected:
};



#endif // ATOMSEXTENSION_H
