// VerletThermalizeAtomsMover.cpp: implementation of the VerletThermalizeAtomsMover class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "DigitalMaterial.h"

#include "VerletThermalizeAtomsMover.h"
#include "ListOfAtoms.h"
#include "DMArray.h"
#include "SerialBuilder.h"


#include "DMLastHeader.h"

static SerialBuilder<VerletThermalizeAtomsMover> *_VerletThermcreator = 
    new SerialBuilder<VerletThermalizeAtomsMover>("VerletThermalizeAtomsMover");

const std::string VerletThermalizeAtomsMover::className = std::string("VerletThermalizeAtomsMover");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

VerletThermalizeAtomsMover::VerletThermalizeAtomsMover(Potential *potential, double temperature, double dt, int nSteps)
{
	this->potential = potential;
	this->temperature = temperature;
	this->dt = dt;
	this->nSteps = nSteps;

	nAtoms=0;
	forces = 0;
	newVelocities = 0;
}


void VerletThermalizeAtomsMover::Move(ListOfAtoms *atoms)	// Verlet
{
	if (nAtoms != atoms->GetNumber()) Allocate(atoms->GetNumber(),atoms->GetCartesianVelocities());

	// Thermalize

	double sig = sqrt(temperature/atoms->GetMass());

	// XXX Something bad is happening here!!!

	// Fill array, then set: makes boundary condition enforcement much more efficient
	// the way we're doing it so far.
	// Also, rand should allow filling of whole arrays...
	// XXX Mass = 1 
	DMArrayIterator<double, DIMENSION> newVeloIter(newVelocities);
	for(newVeloIter.Init(); !newVeloIter.IsDone(); newVeloIter.Next())
	{
		for (int coord=0; coord<DIMENSION; coord++)
		{
			newVeloIter.Current()[coord] = rand.gaussian(sig);
		}
	}
	atoms->SetCartesianVelocities(newVelocities);
	atoms->AdjustVelocities();

	// Run nSteps of Verlet
	atoms->IncrementCartesianPositions(atoms->GetCartesianVelocities(), 0.5*dt);

	for (int i=0; i<nSteps-1; i++)
	{
		potential->CalculateForces(atoms, forces);
		atoms->IncrementCartesianMomenta(forces,dt);
		atoms->AdjustVelocities();
		atoms->IncrementCartesianPositions(atoms->GetCartesianVelocities(),dt);
	}
	potential->CalculateForces(atoms, forces);
	atoms->IncrementCartesianMomenta(forces,dt);
	atoms->AdjustVelocities();
	atoms->IncrementCartesianPositions(atoms->GetCartesianVelocities(), 0.5*dt);

}

void VerletThermalizeAtomsMover::Allocate(int nAtoms, const DMArray<double, DIMENSION> *array)
{
	this->nAtoms = nAtoms;
	if(forces) delete forces;
	forces = new DMArray<double, DIMENSION> (array);
	if (newVelocities) delete newVelocities;
	newVelocities = new DMArray<double, DIMENSION> (array);
}



////////////////////////////////////////////////////////////////////////////////
/// METHODS RELATED TO SERIALIZATION
////////////////////////////////////////////////////////////////////////////////


void VerletThermalizeAtomsMover::Save(DMWriter *writer) const
{
	AtomsMover::Save(writer);	

	writer->PutDouble(dt, "dt");
	writer->PutInt(nSteps, "nSteps");

	writer->PutDouble(temperature, "temperature"); 
	writer->Put(&rand, "rand");
}

void VerletThermalizeAtomsMover::Load(DMReader *reader)
{
	AtomsMover::Load(reader);

	SetDt(reader->GetDouble("dt"));
	SetNSteps(reader->GetInt("nSteps"));

	SetTemperature(reader->GetDouble("temperature"));
	reader->Fill(&rand, "rand");
}
