// ExtendedNudgedElasticBand.h: interface for the ExtendedNudgedElasticBand class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef EXTENDEDNUDGEDELASTICBAND_H
#define EXTENDEDNUDGEDELASTICBAND_H


#include "DMFirstHeader.h"

#include "DMArray.h"
#include "AtomsExtension.h"
#include "NudgedElasticBandBarrierFinder.h"
#include "Minimizable.h"

class ListOfAtoms;
class Potential;
class BoundaryConditions;
class NeighborLocator;

/**
   This is a class derived from NudgedElasticBand which has some extra 
   functionality to deal with having some of the atoms constrained to move
   in accordance with external parameters, similar to 
   ExtendedQuickMinAtomsMover

   It was first written in 2000 but was only made a derivative of NEB
   in October 2001.

   In August 2002 both classes have been extensively restructured, with the
   aim of minimizing code duplication and maximizing clarity, ease of use
   and facilty to add new features (improved tangent estimates etc.)

   \author nbailey
   \date October 2001
   \ingroup molecular dynamics
   
 */

class ExtendedNudgedElasticBand  : public NudgedElasticBandBarrierFinder, public Minimizable
{
public:
	ExtendedNudgedElasticBand(int nReplicas,double dt,double *springConstant,BoundaryConditions *bc, Potential *potential,NeighborLocator *neighborLocator,double tol,AtomsExtension *atomsExtension, bool wantAbsoluteEnergies = false,bool usingClimbingImage = false);
	virtual ~ExtendedNudgedElasticBand();
	void SetParameters(const double *parameters1,const double *parameters2)
		{ 
			this->parameters1 = parameters1;
			this->parameters2 = parameters2;
		}

	//! Set the metric values (defaults to unity)
	void SetParameterMetric( const double *pMetric = 0);

	//! return pointers to the replica parameters (e.g. after Find())
	const double *GetParameters( int prmIdx) const { return parameters[prmIdx]; }



	//! A Minimizable must provide an ObjectiveFunction method
	double ObjectiveFunction(double *parameters);
	//! A Minimizable should (not enforced) provide a Gradient method
	double *Gradient(double *parameters); 
	//! Another Minimizable method: does NOT return ExtendedNEB::nParameters!
	int GetNParameters();
	//! A function to initialize the array of doubles that CG will use
	void SetCurrentPoint(double *currentPoint);		

	ListOfAtoms *GetSaddlePointConfiguration();
	const double *GetSaddleParameters() const { return saddleParameters; };


	void SetInitializeFromSaved(std::string **repFileList, std::string repParamFile) { NudgedElasticBandBarrierFinder::SetInitializeFromSaved(repFileList); this->repParamFile = repParamFile; }




protected:
#ifndef SWIG
	// protected member functions
	void Initialize();
	void QuickMinStep(ListOfAtoms *repAtoms, DMArray<double, DIMENSION> *repForces, int repIdx);
	//! Vector between successive replicas---has two more parameters than the equivalent base class versoin (which it calls)---to deal with boundary parameters.
	void CalculateLinkVector( int repIdxSecond, double(*linkVector)[DIMENSION], double &linkLength);
	
	void GetDotProducts(int repIdx, double (*linkVectorMinus)[DIMENSION], double (*linkVectorPlus)[DIMENSION], double *tangentLength, double *f_UDotTangent, double *linkPlusDotMinus);
    double ReplicaForceEnergy(int chainIdx);
	double AddSpringForce(int repIdx, double (*linkVectorMinus)[DIMENSION], double (*linkVectorPlus)[DIMENSION], double linkMinusLength, double linkPlusLength, double f_UDotTangent, double cosPhi);
	double FlipTangentialForce(int repIdx, double f_UDotTangent);

	void FreeMemoryNReplicas();
	void Allocate(int arrayLength,int nReplicas);
	//! helper function for ObjectiveFunction and Gradient
	void SetExternalParameters(double *parameters);
	

	// basic parameters
	//! number of parameters for boundary atoms
	int nParameters;        
	//! the sublist of free atoms (independent DOFs).
	int mainList;  
	//! the number of core atoms
	int nFreeAtoms;

	// NEB arrays: need to be allocated and freed carefully!
	double **parameters;
	double **parameterVelocities;
	double **parameterForces;
	double **kineticMatrix; // we don't need to store this for all replicas

	//! The metric components associated with parameters (assuming a diagonal metric tensor)
	double *parameterMetric;

	// pointers passed in from outside: not NEB's responsibility
	const double *parameters1, *parameters2;
	AtomsExtension *atomsExtension;

	double *saddleParameters;

	//! for when initFromSaved is true
	std::string repParamFile;

	//! for the purpose of Gradient (return value)
	double *gradientArray;
	int sizeGradientArray;
#endif // SWIG
};

#endif // !defined(EXTENDEDNUDGEDELASTICBAND_H)
