// NudgedElasticBandBarrierFinder.cpp: implementation of the NudgedElasticBandBarrierFinder class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "DMFirstHeader.h"
#include <fstream>
#include <iostream>
#include "blas.h"
#include "DigitalMaterial.h"
#include "NetCDFReader.h"
#include "NudgedElasticBandBarrierFinder.h"
#include "ListOfAtoms.h"
#include "Potential.h"
#include "BoundaryConditions.h"
#include "NeighborLocator.h"
#include "ExternalWork.h"

#include "DMException.h"
#include "DMLastHeader.h"


const int NudgedElasticBandBarrierFinder::STEPSBETWEENOUTPUT = 100;
const int NudgedElasticBandBarrierFinder::END_POINTS = 2;

#define MAXCOFPTS 10

using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////


NudgedElasticBandBarrierFinder::NudgedElasticBandBarrierFinder(int nReplicas,double dt,double *springConstant, BoundaryConditions *bc,Potential *potential,NeighborLocator *neighborLocator,double tol, ExternalWork *externalWork, bool wantAbsoluteEnergies, bool usingClimbingImage)
{
	this->NEB_ArrayLength = this->nReplicas = -1; // so that Allocate works properly the first time it's called
	this->dt = dt;
	this->bc = bc;
	this->potential = potential;
	this->nl = neighborLocator;
	this->atoms1 = 0;
	this->atoms2 = 0;
	this->externalWork = 0;
	this->tol = tol;
	this->externalWork = externalWork;
	this->wantAbsoluteEnergies = wantAbsoluteEnergies;
	this->usingClimbingImage = usingClimbingImage;
	this->increasedSpringForceRange = -1;
	this->incSpringFactor = 2.0;
	this->initializeFromSaved = false;
	this->repFileList = 0;
	this->verbosity = 0;
	this->numKinks = 0;

	// default method for computing the tangent is the original one
	this->TANGENT_METH = ORIG_T;
	// Return information:
	saddleCoord = 0.;
	barrierHeight = 0.;
	backwardBarrier = 0.;
	barrierError = 0.;
	chainLength = 0.;
	totalForceSq = 0;
	nIterations = 0;
	totalEnergyDrop = 0.0;

	
	//operational parameters
	// these values of the angles are what Enrique Batista suggests 
	cosPhiLarge = 0.8660254038; // corresponds to 30 degrees
	cosPhiSmall = 0.9961946981; // corresponds to 5 degrees
	
	maxIterations = 1000;
	itToSetClimbing = 10;
	
	maxEnergy = 0.;
	maxEnergyIdx = 0;
	energy1 = 0;
	springMethod = 0;
	
	// operational arrays
	tangent = 0;
	linkVector1 = 0;
	linkVector2 = 0;
	springForce = 0;
	forces = 0;
	replicas = 0;
	repNeighborLocators = 0;
	repBoundaryConditions = 0;
	barrierPoints = 0;
	Allocate(0,nReplicas);
	
	if(nReplicas > 0)
		for( int a = 0; a < nReplicas + 1; ++a )
		{
			this->springConstant[a] = springConstant[a];
			this->replicaForceSq[a] = 0.;
		}
	cof = new double[MAXCOFPTS]; // assuming we won't need more than this (we use 5 now)
}


NudgedElasticBandBarrierFinder::~NudgedElasticBandBarrierFinder()
{
	if(NEB_ArrayLength > 0)
		FreeMemoryNAtoms();
	if(nReplicas > 0)
		FreeMemoryNReplicas();
	
	delete [] cof;
}

void NudgedElasticBandBarrierFinder::Allocate(int nAtoms,int nReplicas)
{
	if(nReplicas != this->nReplicas)
    {
		if(this->nReplicas > 0)
			FreeMemoryNReplicas();
		if(nReplicas > 0)
		{
			ListOfAtoms **replicasMinusOne = new ListOfAtoms*[nReplicas + END_POINTS];
			replicas = replicasMinusOne + 1;  replicas[0] = 0;
			// replicas[0] will be used to check whether replicas and neighborLocators etc need to be deleted: it will only be nonzero when replicas have actually been created, in which case so will the NeighborLocators, force DMArrays and BoundaryConditions
			forces = new DMArray<double, DIMENSION>*[nReplicas];
			repBoundaryConditions = new BoundaryConditions*[nReplicas];
			repNeighborLocators = new NeighborLocator*[nReplicas];
			for(int a=0;a<nReplicas;a++)
			{      
				repBoundaryConditions[a] = bc->Copy();
				repNeighborLocators[a] = nl->Copy();
			}
			barrierPoints = new double[nReplicas + END_POINTS][2]; 
			profileIntegrand = new double[nReplicas + END_POINTS];
			springConstant = new double[nReplicas + 1];
			replicaForceSq = new double[nReplicas];
		}
		this->nReplicas = nReplicas;  
    }  
	
	if(this->NEB_ArrayLength != nAtoms)
    {
		if(this->NEB_ArrayLength > 0)
			FreeMemoryNAtoms();
		if(nAtoms > 0)
		{
			tangent = new double[nAtoms][DIMENSION];  
			linkVector1 = new double[nAtoms][DIMENSION];
			linkVector2 = new double[nAtoms][DIMENSION];
			springForce = new double[nAtoms][DIMENSION];      
		}     
		this->NEB_ArrayLength = nAtoms;
    }
	
}

void NudgedElasticBandBarrierFinder::FreeMemoryNAtoms()
{
  delete [] tangent;
  delete [] linkVector1;
  delete [] linkVector2;
  delete [] springForce;
}

void NudgedElasticBandBarrierFinder::FreeMemoryNReplicas()
{
	if(replicas[0])
    {
		for(int a=0;a<nReplicas;++a)
		{
			delete forces[a];
			delete replicas[a];
			delete repNeighborLocators[a];
			delete repBoundaryConditions[a];
		}
    }
	delete [] (replicas-1);
	delete [] forces;
	delete [] repBoundaryConditions;
	delete [] repNeighborLocators;
	delete [] barrierPoints;
	delete [] profileIntegrand;
	delete [] springConstant;
	delete [] replicaForceSq;
}

void NudgedElasticBandBarrierFinder::CreateReplicas()
{
	int repIdx, numberAtoms = atoms1->GetNumber();
	if(replicas[0])
		for( repIdx = 0; repIdx < nReplicas; ++repIdx )
		{
			delete replicas[repIdx];
			delete forces[repIdx];
		}

	for( repIdx = 0; repIdx < nReplicas; ++repIdx )
    {
		// Replicas are created when Find() is called and persist until Find is called again
		if(initializeFromSaved)
		{			
			NetCDFReader reader(*repFileList[repIdx]);
			replicas[repIdx] = dynamic_cast<ListOfAtoms*>(reader.Get("LOA"));
		}
		else
			replicas[repIdx] = atoms1->Copy();
		forces[repIdx] = new DMArray<double,DIMENSION>(atoms1->GetCartesianPositions()->GetShape());
		replicas[repIdx]->SetBoundaryConditions(repBoundaryConditions[repIdx]);
		replicas[repIdx]->SetNeighborLocator(repNeighborLocators[repIdx]);
    }

	replicas[-1] = atoms1;
	replicas[nReplicas] = atoms2;
	if(!initializeFromSaved)
	{
		// Get the difference vector (in configuration space)  between the two
		// given ListOfAtoms configurations:
		double fractionalSpacing = 1. / (nReplicas + 1);
		
		DMArray<double, DIMENSION> *configurationDiff = new DMArray<double, DIMENSION>(atoms1->GetShape());
		configurationDiff->Copy(atoms2->GetCartesianPositions());
		configurationDiff->Axpy(atoms1->GetCartesianPositions(),-1.);
		
		bc->DifferenceBoundaryConditions(configurationDiff);
		
		for( repIdx = 0; repIdx < nReplicas; ++repIdx)
			replicas[repIdx]->IncrementCartesianPositions(configurationDiff,(repIdx+1)*fractionalSpacing);
		
		delete configurationDiff;
	}
}


void NudgedElasticBandBarrierFinder::Initialize()
{
	if(NEB_ArrayLength != atoms1->GetNumber()) Allocate(atoms1->GetNumber(),nReplicas);
	// create the replicas as copies of atoms1
	CreateReplicas();
}

void NudgedElasticBandBarrierFinder::Find(ListOfAtoms *atoms1, ListOfAtoms *atoms2)
{
	if( !(atoms1->GetShape() == atoms2->GetShape()) )
		throw DMException("NudgedElasticBandBarrierFinder","Find","atoms1 and atoms2 have different shapes");
	
	this->atoms1 = atoms1; this->atoms2 = atoms2;
	// allocate if necessary, copy replicas, set straight-line configuration
	Initialize();
	
	// intialize the energies and the "distances" along the chain
	energy1 = 0; energy1 = ReplicaForceEnergy(0); // energy1 should be zero for first call
	barrierPoints[0][0] = 0.;
	if(wantAbsoluteEnergies) barrierPoints[0][1] = energy1;
	else barrierPoints[0][1] = 0;

	ResetMaxEnergy();
	for( int chainIdx = 1; chainIdx < nReplicas + END_POINTS; ++chainIdx )
	{
		barrierPoints[chainIdx][0] = chainIdx / (nReplicas + 1.);
		barrierPoints[chainIdx][1] = ReplicaForceEnergy( chainIdx );
		if(verbosity > 1)
			cout << "Initial energy, chain index " << chainIdx << ": " << barrierPoints[chainIdx][1] << endl;
		CheckMaxEnergy(barrierPoints[chainIdx][1],chainIdx);
	}
	

	// open log file
	ofstream progFile("progress.log"); progFile.precision(15);
	
	// reset climbing image (don't want to be stuck with value from last call)
	climbingReplica = -1;
	// reset limits of increased spring force
	minSprIdx = -1; maxSprIdx = -1;

	// this won't allow us to set the climbing image to be image 0
	if(itToSetClimbing < 0 && itToSetClimbing > -nReplicas && usingClimbingImage)
		StartClimbingImage( -itToSetClimbing);




	// Relax the composite system  using the quickmin technique:
	nIterations = 0;

	// Initial forces
	ResetMaxEnergy();
	int repIdx;
	for( repIdx = 0; repIdx < nReplicas; ++repIdx )
	{
		barrierPoints[repIdx+1][1] = ReplicaForceEnergy(repIdx + 1);
		CheckMaxEnergy(barrierPoints[repIdx+1][1],repIdx+1);
	}
	// calculate the corrections to forces from the NEB method
	totalForceSq = CalculateNudging();
	while(totalForceSq > nReplicas * tol && nIterations < maxIterations)
    {
		// set climbing image, if appropriate
		if(nIterations == itToSetClimbing && usingClimbingImage)
			StartClimbingImage( maxEnergyIdx - 1);

		
		// Update velocities, using quick-min procedure:
		for( repIdx = 0; repIdx < nReplicas; ++repIdx )
			QuickMinStep(replicas[repIdx], forces[repIdx], repIdx);


		// Calculate the gradient forces and the potential energies:
		ResetMaxEnergy();
		for( repIdx = 0; repIdx < nReplicas; ++repIdx )
		{
			barrierPoints[repIdx+1][1] = ReplicaForceEnergy(repIdx + 1);
			CheckMaxEnergy(barrierPoints[repIdx+1][1],repIdx+1);
		}
		// calculate the corrections to forces from the NEB method
		numKinks = 0;
		totalForceSq = CalculateNudging();		

		// write to progress file and increment counter      
		if(nIterations%STEPSBETWEENOUTPUT == 0) 
		{
			CalculateReturnInfo();
			progFile << nIterations << " " << totalForceSq << " " << barrierHeight << endl;	
		}
		++nIterations;
    }

	// restore spring forces for next call
	if( minSprIdx >= 0 )
		for( int sprIdx = minSprIdx; sprIdx <= maxSprIdx; ++sprIdx)
			springConstant[sprIdx] /= incSpringFactor;

	this->initializeFromSaved = false;

	// calculate distances along chain, potential energies etc.
	CalculateReturnInfo();  	
}


void NudgedElasticBandBarrierFinder::StartClimbingImage(int climbingReplica)
{
	this->climbingReplica = climbingReplica;

	if(verbosity > 0)
		cout << "Setting climbing image: " << climbingReplica << endl;
	// minus 1 to account for indexing from atoms1
	if( increasedSpringForceRange > 0 )
	{
		int sprIdx;
		minSprIdx = climbingReplica - increasedSpringForceRange +1;
		if(minSprIdx < 0) minSprIdx = 0;
		
		maxSprIdx = climbingReplica + increasedSpringForceRange;
		if(maxSprIdx > nReplicas) minSprIdx = nReplicas;
		for(sprIdx = minSprIdx; sprIdx <= maxSprIdx; ++sprIdx)
			springConstant[sprIdx] *= incSpringFactor;
	}
	
}


void NudgedElasticBandBarrierFinder::QuickMinStep(ListOfAtoms *repAtoms, DMArray<double, DIMENSION> *repForces,int repIdx)
{
	int coord;
	repAtoms->IncrementCartesianVelocities(repForces,dt);
	DMArrayIterator<double,DIMENSION> velIter(repAtoms->GetCartesianVelocities());
	DMArrayIterator<double,DIMENSION> forcesIter(repForces);
	
	for( velIter.Init(), forcesIter.Init();
		!velIter.IsDone();
		velIter.Next(), forcesIter.Next() )
	{
		double vDotF = 0;
		double *currentForce = forcesIter.Current();	
		double *currentVel = velIter.Current();
		for(coord=0;coord<DIMENSION;coord++)
			vDotF += currentVel[coord] * currentForce[coord]; 
		
		if(vDotF < 0)
			for(coord=0;coord<DIMENSION;coord++)
				currentVel[coord] = 0.;
		
	} 
	
	// Apply constraints and update positions by dt 
	repAtoms->AdjustVelocities();
	repAtoms->IncrementCartesianPositions(repAtoms->GetCartesianVelocities(),dt);
}


double NudgedElasticBandBarrierFinder::CalculateNudging()
{
	double linkMinusLength, linkPlusLength;	
	double (*linkVectorMinus)[DIMENSION], (*linkVectorPlus)[DIMENSION], (*linkVectorTemp)[DIMENSION];

	// reset force measures
	totalForceSq = 0;
	
	// Calculate the spring forces. Start by calculating the "minus"
	// linkVector for the first replica. In the loop we only need to calculate
	// the "plus" one because the minus one is the previous plus one.
	linkVectorMinus = linkVector1; linkVectorPlus = linkVector2;
	CalculateLinkVector(0, linkVectorMinus, linkMinusLength);
	barrierPoints[1][0] = linkMinusLength;

	//cout << "iteration: " << nIterations << endl;
	for( int repIdx = 0; repIdx < nReplicas; ++repIdx )
    {

		// calculate the difference i to i+1 (already have i-1 to i)
		CalculateLinkVector(repIdx+1, linkVectorPlus, linkPlusLength);
		barrierPoints[repIdx + 2][0] = barrierPoints[repIdx + 1][0] + linkPlusLength;
		
		// calculate tangent, dot product between it and potential force,
		// and angle between the two link vectors
		double cosPhi;
		double f_UDotTangent = CalculateTangent(repIdx, linkVectorMinus, linkVectorPlus, linkMinusLength, linkPlusLength, &cosPhi);
		if(repIdx != climbingReplica && cosPhi < 0.5 * (cosPhiLarge + cosPhiSmall)) // 0.931110051
			numKinks++;
		
		// now to add on the spring force to the total force---but not if the 
		// replica is the special 'climbingReplica'
		if(repIdx == climbingReplica)
			replicaForceSq[repIdx] = FlipTangentialForce(repIdx, f_UDotTangent);
		else
			replicaForceSq[repIdx] = AddSpringForce(repIdx, linkVectorMinus,linkVectorPlus, linkMinusLength, linkPlusLength, f_UDotTangent, cosPhi);

		totalForceSq += replicaForceSq[repIdx];

		// swap the pointers
		linkVectorTemp = linkVectorMinus;
		linkVectorMinus = linkVectorPlus;
		linkVectorPlus = linkVectorTemp;
		// save linkPlusLength
		linkMinusLength = linkPlusLength;
    } // end of loop over replicas
	
	return totalForceSq;
}




void NudgedElasticBandBarrierFinder::CalculateLinkVector(int repIdxSecond, double (*linkVector)[DIMENSION], double &linkLength)
{
	// Certain derived classes need the length of arrays (NEB_ArrayLength) to
	// be greater than the number of atoms in the list passed to this function
	// hence NEB_ArrayLength should not be used here to control the 
	// iteration--rather the length of the passed-in LOA should.
	int atomIdx;
	DMArrayIterator<double, DIMENSION> firstIter(replicas[repIdxSecond-1]->GetCartesianPositions());
	DMArrayIterator<double, DIMENSION> secondIter(replicas[repIdxSecond]->GetCartesianPositions());

	for( atomIdx = 0, firstIter.Init(), secondIter.Init();
		!firstIter.IsDone();
		 ++atomIdx, firstIter.Next(), secondIter.Next() )
	{
		const double *firstPos = firstIter.Current();
		const double *secondPos = secondIter.Current();

		for( int coord = 0; coord < DIMENSION; ++coord )
			linkVector[atomIdx][coord] = secondPos[coord] - firstPos[coord];
	} 
	ASSERT(atomIdx == atoms1->GetNumber());
	bc->DifferenceBoundaryConditions(linkVector,atomIdx);	
	
	linkLength = cblas_dnrm2(atomIdx*DIMENSION,(const double *)linkVector,1);
	ASSERT(linkLength >= 0.);
}


double NudgedElasticBandBarrierFinder::CalculateTangent(int repIdx, double (*linkVectorMinus)[DIMENSION], double (*linkVectorPlus)[DIMENSION], double linkMinusLength, double linkPlusLength, double *cosPhi)
{
	// the following method adds the two link vectors
	if( TANGENT_METH == ORIG_T )
	{
		memcpy(tangent, linkVectorPlus, NEB_ArrayLength*DIMENSION*sizeof(double));
		cblas_daxpy(NEB_ArrayLength*DIMENSION, 1., (double*)linkVectorMinus, 1, (double*)tangent, 1);
	}
	
	else if( TANGENT_METH == BISECT_T )
	{
		// alternative way of computing the tangent---add the unit link vectors
		// (divide by their lengths before adding)
		// so we're really bisecting the angle between them
		memcpy(tangent, linkVectorPlus, NEB_ArrayLength*DIMENSION*sizeof(double));
		cblas_dscal(NEB_ArrayLength * DIMENSION, 1./linkPlusLength, (double *)tangent,1);
		cblas_daxpy(NEB_ArrayLength * DIMENSION, 1./linkMinusLength,(double *)linkVectorMinus, 1, (double *)tangent, 1);
		// I'm making an assumption that the difference between coordinates in 
		// adjacent configurations will be small compared to the system size 
		// and since linkVectorPlus and linkVectorMinus have already had 
		// boundary conditions applied, I don't need to do boundary conditions
		// on their difference.
	}
	else if( TANGENT_METH == ENERGY_T )
	{
		// energy-dependent tangent from Henkelman and Jonsson
		// need the three energies 
		double energyCurr = barrierPoints[repIdx+1][1], energyPrev = barrierPoints[repIdx][1], energyNext = barrierPoints[repIdx+2][1];

		if(energyNext > energyCurr && energyCurr > energyPrev)
			memcpy(tangent, linkVectorPlus, NEB_ArrayLength*DIMENSION*sizeof(double));
		else if
			( energyNext < energyCurr && energyCurr < energyPrev)
			memcpy(tangent, linkVectorMinus, NEB_ArrayLength*DIMENSION*sizeof(double));
		else
		{
			// current image is a maximum or minimum
			double deltaEPlus = fabs(energyNext - energyCurr);
			double deltaEMinus = fabs(energyCurr - energyPrev);
			double deltaEMax = deltaEPlus, deltaEMin = deltaEMinus;
			if(deltaEMinus > deltaEPlus)
			{
				deltaEMax = deltaEMinus;
				deltaEMinus = deltaEPlus;
			}
			double weights[2] = {deltaEMax,deltaEMin};
			// weights[0] weights linkVectorPlus
			if( energyNext < energyPrev )
			{ weights[0] = deltaEMin; weights[1] = deltaEMax; }
			memcpy(tangent, linkVectorPlus, NEB_ArrayLength*DIMENSION*sizeof(double));
			cblas_dscal(NEB_ArrayLength * DIMENSION, weights[0], (double *)tangent,1);
			cblas_daxpy(NEB_ArrayLength * DIMENSION, weights[1],(double *)linkVectorMinus, 
						1, (double *)tangent, 1);
		}
	}
	
	double tangentLength, f_UDotTangent, linkPlusDotMinus;
	GetDotProducts(repIdx, linkVectorMinus, linkVectorPlus, &tangentLength, &f_UDotTangent, &linkPlusDotMinus);
	
	 // divide tangent by its length to return a unit vector
	 cblas_dscal(NEB_ArrayLength*DIMENSION, 1./tangentLength, (const double *)tangent,1);


	*cosPhi = linkPlusDotMinus / (linkMinusLength * linkPlusLength);

	return f_UDotTangent;
}

void NudgedElasticBandBarrierFinder::GetDotProducts(int repIdx, double (*linkVectorMinus)[DIMENSION], double (*linkVectorPlus)[DIMENSION], double *tangentLength, double *f_UDotTangent, double *linkPlusDotMinus)
{
	// length of tangent vector
	*tangentLength = cblas_dnrm2(NEB_ArrayLength*DIMENSION,(const double *)tangent,1);

	// dot product between tangent and potential forces
	*f_UDotTangent = 0.;
	DMArrayIterator<double, DIMENSION> forceIter(forces[repIdx]);
	int atIdx;
	for( atIdx  = 0, forceIter.Init(); !forceIter.IsDone(); forceIter.Next(), ++atIdx )
	{
		double *currentForce = forceIter.Current(); 
		for( int coord = 0; coord < DIMENSION; ++coord )
			*f_UDotTangent += currentForce[coord] * tangent[atIdx][coord];
	}  


	// dot product between consecutive link vectors
	*linkPlusDotMinus = cblas_ddot( NEB_ArrayLength * DIMENSION, (double*)linkVectorPlus, 1, (double *)linkVectorMinus, 1);
}



double NudgedElasticBandBarrierFinder::AddSpringForce(int repIdx, double (*linkVectorMinus)[DIMENSION], double (*linkVectorPlus)[DIMENSION], double linkMinusLength, double linkPlusLength, double f_UDotTangent, double cosPhi)
{
	// compute how much of perp spring force we want to add in
	


	// this function implements the original method for calculating the NEB
	// spring forces
	int atIdx; double repForceSq = 0.;

	// construct the spring forces as linear combination of linkVectors
	// using memcpy and BLAS
	// springForce = springConstant[repIdx+1] * linkVectorPlus
	//               - springConstant[repIdx] * linkVectorMinus
	if(springMethod == 0)
	{
		memcpy(springForce,linkVectorPlus, NEB_ArrayLength*DIMENSION*sizeof(double));
		cblas_dscal(NEB_ArrayLength*DIMENSION, springConstant[repIdx+1], (double*)springForce, 1);
		cblas_daxpy(NEB_ArrayLength*DIMENSION, - springConstant[repIdx], (double *)linkVectorMinus, 1, (double *)springForce, 1);
	}
	else
	{
		// the is a slightly different way of calculating the spring forces
		// this way the calculated forces are from the start parallel to the
		// tangent vector, so there's no possibility of adding back the perp
		// part of the spring force
		memcpy(springForce,tangent, NEB_ArrayLength*DIMENSION*sizeof(double));
		cblas_dscal(NEB_ArrayLength*DIMENSION, springConstant[repIdx+1] * linkPlusLength - springConstant[repIdx] * linkMinusLength, (double*)springForce, 1);
	}

	// dot product of spring force with tangent
	double f_SDotTangent = cblas_ddot(NEB_ArrayLength*DIMENSION, (double *)springForce, 1 , (double *)tangent,1);


	double unKink = GetUnKinkFactor(cosPhi, repIdx);

	// add the appropriate parts of spring force and subtract the appropriate
	// parts of the potential force
	DMArrayIterator<double, DIMENSION> forceIter(forces[repIdx]);
	for( atIdx = 0, forceIter.Init(); !forceIter.IsDone(); forceIter.Next(), ++atIdx )
	{
		double *currentForce = forceIter.Current();
		if(springMethod == 0)
			for( int coord = 0; coord < DIMENSION; ++coord )      
			{
				currentForce[coord] += 
					tangent[atIdx][coord] * (f_SDotTangent - f_UDotTangent )
					+ unKink * (springForce[atIdx][coord] - tangent[atIdx][coord] * f_SDotTangent ); 
				repForceSq += currentForce[coord] * currentForce[coord];
			}
		else
			for( int coord = 0; coord < DIMENSION; ++coord )      
			{
				currentForce[coord] += springForce[atIdx][coord] - f_UDotTangent * tangent[atIdx][coord] ;
				repForceSq += currentForce[coord] * currentForce[coord];
			}
	}
	
	return repForceSq;
}
		 
	


double NudgedElasticBandBarrierFinder::GetUnKinkFactor(double cosPhi, int repIdx) const
{
	double unKink = 0.0;
	if(cosPhi < cosPhiSmall)
		if(cosPhi > cosPhiLarge)
			unKink = 0.5 * (1.0 + cos(PI * (cosPhi - cosPhiLarge)/
									  (cosPhiSmall - cosPhiLarge)));
		else unKink = 1.0;
	
	if(unKink > 0.5)
		cerr << "Kinky path condition (phi = " << acos(cosPhi) << ") at image " << repIdx << " in NEB, at iteration " << nIterations << endl;
	return unKink;
}

double NudgedElasticBandBarrierFinder::FlipTangentialForce(int repIdx, double f_UDotTangent)
{
	int atIdx;
	double repForceSq = 0.;
	DMArrayIterator<double, DIMENSION> forceIter(forces[repIdx]);
	for(atIdx = 0, forceIter.Init();
		atIdx < NEB_ArrayLength;
		forceIter.Next(), ++atIdx )
	{
		double *currentForce = forceIter.Current();
		for(int coord=0;coord<DIMENSION;coord++)      
		{
			currentForce[coord] += - 2 * f_UDotTangent * tangent[atIdx][coord];
			repForceSq += currentForce[coord] * currentForce[coord];
		}
	}
	return repForceSq;
}



void NudgedElasticBandBarrierFinder::CalculateReturnInfo()
{
	chainLength = barrierPoints[nReplicas + 1][0];
	
	double barrHeight5 = barrHeight5 = CalculateBarrierHeight(5,barrierPoints);
	// with this order saddleCoord gets the 4-point value	
	barrierHeight = CalculateBarrierHeight(4,barrierPoints);
	totalEnergyDrop = barrierPoints[0][1] - barrierPoints[nReplicas + END_POINTS-1][1];
	backwardBarrier = barrierHeight + totalEnergyDrop;
	barrierError = fabs(barrHeight5 - barrierHeight);
}


double NudgedElasticBandBarrierFinder::ReplicaForceEnergy(int chainIdx)
{
	ListOfAtoms *atomsLoc;
	bool calcForces;
	if( chainIdx <= 0 )
	{ atomsLoc = atoms1; calcForces = false; }
	else if( chainIdx > nReplicas )
	{ atomsLoc = atoms2; calcForces = false; }
	else
	{ atomsLoc = replicas[chainIdx-1]; calcForces = true;}
	
	if(calcForces)
	{
		potential->CalculateForces(atomsLoc, forces[chainIdx-1]);
		atomsLoc->AdjustForces(forces[chainIdx-1]);
	}

	double energy = potential->CalculateEnergy(atomsLoc);
	if(externalWork)
		energy -= externalWork->CalculateWork(atomsLoc);

	// energy1 is set by first setting it to zero, then calling this function
	// on atoms1, and setting energy1 to the result
	if(!wantAbsoluteEnergies)
		energy -= energy1;

	return energy;
}



double NudgedElasticBandBarrierFinder::CalculateBarrierHeight(int nPoints, double (*barrierPoints)[2])
{
	double height;
	if( climbingReplica >= 0)
	{
		saddleCoord = barrierPoints[climbingReplica+1][0];
		height = barrierPoints[climbingReplica+1][1];
		return height - barrierPoints[0][1];
	}
	
	ASSERT( nPoints <= MAXCOFPTS );
	// Now get the nPoints highest points, to which we'll fit a parabola/cubic/quartic/whatever
	double (*point)[2]; 
	double *x,*y,*s,a1,a2,a3,q,Q,R,Z,theta,modR,phi,ff,b,discriminant;
	int a,signB,signR,numberPossible,p,l,k,j,i;;
	double lowest = 1.0e99, highest = -1.0e99,highestCoord;
	
	s = new double[nPoints];
	point = new double[nPoints][2];
	x = new double[nPoints-2];
	y = new double[nPoints-2];

	
	for(p=0;p<nPoints;++p)
    {
		point[p][0] = barrierPoints[p][0];
		point[p][1] = barrierPoints[p][1];
		if(point[p][1] < lowest) lowest = point[p][1];
		if(point[p][1] > highest) {highest = point[p][1];highestCoord = point[p][0];}
    }
	
	//This runs through the rest of barrierPoints, shifting the members of the point
	// array over whenever a new point is higher than the lowest of the old points
	// There's an asumption that the three/four/five highest points are consecutive
	
	int lessThanHighest = 0; // need at least one point further along the chaing than the highest
	// which is less than the highest
	for(l=nPoints;l<nReplicas+END_POINTS;++l)
    {

		if(barrierPoints[l][1] > lowest || lessThanHighest == 0)
		{
			for(p=0;p<nPoints-1;++p) // do the shifting
			{
				point[p][0] = point[p+1][0];
				point[p][1] = point[p+1][1];
			}
			point[nPoints-1][0] = barrierPoints[l][0]; // include the new point in the rightmost slot
			point[nPoints-1][1] = barrierPoints[l][1];
			if(barrierPoints[l][1] > highest)
			{
				highest = barrierPoints[l][1];
				highestCoord = point[p][0];
				lessThanHighest = 0;
			}
			else
			{
				++lessThanHighest;
			}
			lowest = highest;
			for(int p=0;p<nPoints;++p) // find the new lowest
				
			{
				if(point[p][1] < lowest) lowest = point[p][1];
			}
		}	
    }
	//assert(lowest == point[0][1] || lowest == point[nPoints-1][1]);
	
	
	
	// Fitting to polynomial: based on Numerical Recipes routine polcoe
	for (i=0;i<nPoints;++i)
		s[i]=cof[i]=0.0;
	
	s[nPoints-1] = -point[0][0];
	
	for (i=1;i<nPoints;++i) 
    {
		for (j=nPoints-1-i;j<nPoints-1;++j)
			s[j] -= point[i][0]*s[j+1];
		s[nPoints-1] -= point[i][0];
    }
	for (j=0;j<nPoints;++j)
    {
		phi=nPoints;
		for (k=nPoints-1;k>=1;--k)
			phi=k*s[k]+point[j][0]*phi;
		ff=point[j][1]/phi;
		b=1.0;
		for (k=nPoints-1;k>=0;--k)
		{
			cof[k] += b*ff;
			b=s[k]+point[j][0]*b;
		}
    }
	// outputting the coefficients: is this necessary? If so should not print it but set member variables as with everything else  
	/*for(int yy=0;yy<nPoints;++yy)
	  cout << yy << " " << cof[yy] << endl;*/
	// Now to get the maximum. Can do nPoints = 3,4 or 5 analytically
	
	switch(nPoints)
    {
    case 3:
		
		height = cof[0] - 0.25*cof[1]*cof[1]/cof[2]; 
		saddleCoord = -0.5 * cof[1]/cof[2];
		
		break;
		
    case 4:
		// a = 3*cof[3]; b = 2*cof[2]; c = cof[1]
		signB = 1;
		if(cof[2] < 0) signB = -1;
		discriminant = 4*cof[2]*cof[2] - 12*cof[1]*cof[3];
		if(discriminant < 0)
		{
			saddleCoord = point[0][0] - 1.0; // to generate bad fit error
			height = 0;
			break;
		}
		q = -0.5*(2*cof[2] + signB*sqrt(discriminant));
		x[0] = q/(3*cof[3]);
		x[1] = cof[1]/q;
		for(a=0;a<2;++a)
			y[a] = cof[0] + x[a] * (cof[1] + x[a]*(cof[2] + x[a]*cof[3]));
		if(y[0]>y[1])
		{
			saddleCoord = x[0];
			height = y[0];		
		}
		else
		{
			saddleCoord = x[1];
			height = y[1];	
		}
		break;
		
    case 5:
		// to use NR formula, scale by coefficient of x^3, which is 4*cof[4]
		a1 = 3*cof[3]/(4*cof[4]);
		a2 = 2*cof[2]/(4*cof[4]);
		a3 = cof[1]/(4*cof[4]);
		Q = (a1*a1-3*a2)/9;
		
		R = (2*a1*a1*a1 - 9*a1*a2 + 27*a3)/54;
		if(Q*Q*Q - R*R > 0)  // there are three real roots
			
		{
			theta = acos(R/sqrt(Q*Q*Q));
			numberPossible = 0;
			double xPossible[2];
			double yPossible[2];
			int index = 0;
			for(a=0;a<3;++a)
			{
				x[a] = -2*sqrt(Q) * cos((theta + a*2*PI)/3) - a1/3;
				y[a] = cof[0] + x[a] * (cof[1] + x[a]*(cof[2] + x[a]*(cof[3] + x[a]*cof[4])));
				if(2*cof[2] + x[a]*(6*cof[3] + x[a]*12*cof[4]) <0 && x[a] > 0)  //second derivative test, also, take only ones which occur at x>0
				{
					++numberPossible;
					xPossible[index] = x[a];
					yPossible[index] = y[a];
					++index;
				}
			}
			// Need to decide which is my maximum!
			assert(1 == numberPossible || 2 == numberPossible); // can't have more than two local maxima!
			saddleCoord = xPossible[0];
			height = yPossible[0];
			// only criterion left is: Which is closer to the beginning of the barrier profile?
			if(2 == numberPossible && xPossible[1] < xPossible[0])
			{
				saddleCoord = x[a];
				height = y[a];
			}				
		}
		else  // only one real root (obviously the right one)
		{
			signR = 1;
			modR = R;
			if(R<0) 
			{
				signR = -1;
				modR = -R;
			}
			Z = pow(sqrt(R*R - Q*Q*Q) + modR,(1.0/3.0));
			x[0] = -signR*(Z + Q/Z) - a1/3;
			saddleCoord = x[0];
			height = cof[0] + x[0] * (cof[1] + x[0]*(cof[2] + x[0]*(cof[3] + x[0]*cof[4])));
		}
		break;
		
    default: // nPoints < 3 or nPoints > 5
		saddleCoord = highestCoord;
		height = highest;
    }
	
	
	if(saddleCoord < point[0][0] ||  saddleCoord > point[nPoints-1][0]) // Bad fit
    {
		// maybe set a flag in this case without printing anything
		cerr << "Iteration " << nIterations << ": Bad fit!";
		cerr << "Using the highest point of the chain." << endl;
		saddleCoord = highestCoord;
		height = highest;
		
    }
	delete [] x;
	delete [] y;
	delete [] point;
	delete [] s;
	
	return height - barrierPoints[0][1];
}


ListOfAtoms *NudgedElasticBandBarrierFinder::GetSaddlePointConfiguration()
{
	if(replicas[0] == 0)
		throw DMException("NudgedElasticBandBarrierFinder","GetSaddlePointConfiguration","No replicas");

	if(saddleCoord == 0.)
		throw DMException("NudgedElasticBandBarrierFinder","GetSaddlePointConfiguration","saddleCoord has not been set---call Find()");


	int i,coord,l,index,nCurves = DIMENSION*atoms1->GetNumber(); 
	// nCurves is the number of DOFS, for each of which there is one interpolation to be done
	int nInterpPoints = 4; // number of points used for the interpolation
	double dy,value[DIMENSION];
	for(coord=0;coord<DIMENSION;++coord)value[coord] = 0.;
	double **values = new double*[nCurves];
	for(i=0;i<nCurves;++i)
		values[i] = new double[nReplicas+END_POINTS];
	DMArray<double, DIMENSION> *interpValues = new DMArray<double,DIMENSION>(atoms1->GetCartesianPositions()->GetShape());
	double *c = new double[nInterpPoints];
	double *d = new double[nInterpPoints];

	index = 0;
	for(i=0;i<atoms1->GetNumber();++i)
	{
		for(coord=0;coord<DIMENSION;++coord,++index)
		{
			double startValue = atoms1->GetCartesianPosition(i,coord);
			values[index][0] = 0.0;
			for(l=0;l<nReplicas;l++)
			{  
				value[coord] = replicas[l]->GetCartesianPosition(i,coord)- startValue;
				bc->DifferenceBoundaryConditions(&value,1);
				values[index][l+1] = value[coord];
			}
			value[coord] = atoms2->GetCartesianPosition(i,coord) - startValue;
			bc->DifferenceBoundaryConditions(&value,1);
			values[index][nReplicas+END_POINTS-1] = value[coord];
			
		}
    }

	int offset = 0;
	while(barrierPoints[offset][0] < saddleCoord) ++offset;
	offset -= 2; // Want to use two points on each side of saddleCoord
	if(offset < 0) offset = 0;
	if(offset > nReplicas+END_POINTS-nInterpPoints) offset = nReplicas+END_POINTS-nInterpPoints;

	// The following is adapted from Numerical Recipes routine polint:
	index = 0;
	for(int atom=0;atom<atoms1->GetNumber();++atom)
		for(coord=0;coord<DIMENSION;++coord,++index)
		{
			int i,m,ns=0;
			double den,dif,dift,ho,hp,w;
			
			dif = fabs(saddleCoord-barrierPoints[offset+0][0]);
			
			for (i=0;i<nInterpPoints;i++)
			{
				if ( (dift=fabs(saddleCoord-barrierPoints[offset+i][0])) < dif)
				{
					ns=i;
					dif=dift;
				}
				c[i]=values[index][i+offset];
				d[i]=values[index][i+offset];
			}
			
			(*interpValues)[atom][coord]=values[index][offset+ns--];
			for (m=1;m<nInterpPoints;m++)
			{
				for (i=0;i<nInterpPoints-m;i++) 
				{
					ho=barrierPoints[offset+i][0]-saddleCoord;
					hp=barrierPoints[offset+i+m][0]-saddleCoord;
					w=c[i+1]-d[i];
					if ( (den=ho-hp) == 0.0)
					{
						throw DMException("NudgedElasticBandBarrierFinder",
										  "GetSaddlePointConfiguration",
										  "Error in SaddleInterpolate");
					}
					den=w/den;
					d[i]=hp*den;
					c[i]=ho*den;
				}
				(*interpValues)[atom][coord] += (dy=(2*ns < (nInterpPoints-m) ? c[ns+1] : d[ns--]));
			}
		}

	ListOfAtoms *saddle = atoms1->Copy();
	saddle->IncrementCartesianPositions(interpValues,1.);

	delete d;
	delete c;
	delete interpValues;
	for(i=0;i<atoms1->GetNumber();++i)
		delete values[i];
	delete values;


	return saddle;
}

