//  Prefactor.cpp: implementation of the  Prefactor.cpp class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "DMFirstHeader.h"
#include <fstream>
#include <iostream>
#include "DigitalMaterial.h"
#include "DMArray.h"
#include "ListOfAtoms.h"
#include "Potential.h"
#include "Prefactor.h"
#include "AtomsExtension.h"
#include "blas.h"
#include "lapack.h"
#include "DMLastHeader.h"

#define SMALL_EVALUE 0.001

using namespace std ;


Prefactor::Prefactor(Potential *potential, double eps, AtomsExtension *atomsExtension)
{
	this->potential = potential;
	this->epsilon = eps;
	this->atomsExtension = atomsExtension;

	
	if(atomsExtension)
	{
		this->nParameters = atomsExtension->GetNParameters();
		invMassMatrix = new double[nParameters * nParameters];
	}
	else 
	{
		this->nParameters = 0;
		invMassMatrix = 0;
	}
}

Prefactor::~Prefactor()
{
	if(atomsExtension)
		delete invMassMatrix;
}

double Prefactor::CalculatePrefactor(ListOfAtoms *minimum,ListOfAtoms *saddle)
{
	int i, nDOFs, mainList = 0;	
	if(atomsExtension == 0)
		nDOFs = minimum->NumberGeneralizedCoordinates();
	else
	{
		mainList = atomsExtension->GetMainList();
		nDOFs = minimum->GetLeaf(mainList)->GetNumber() * DIMENSION + nParameters;
	}
	cout << "Number of independent Degrees of Freedom is " << nDOFs << endl; 

	
	// allocate memory for Hessian matrix and force array
	double *hessMin = new double[nDOFs * nDOFs];
	double *hessSadd = new double[nDOFs * nDOFs];
	double *reducedForcesPlus = new double[nDOFs];
	double *reducedForcesMinus = new double[nDOFs];

	cout << "Allocating forces" << endl;
	DMArray<double, DIMENSION> *forces = new DMArray<double,DIMENSION>(minimum->GetCartesianPositions()->GetShape());

	if(atomsExtension == 0)
	{
		CalculateHessian(nDOFs, hessMin, minimum, forces, reducedForcesPlus, reducedForcesMinus);
		CalculateHessian(nDOFs, hessSadd, saddle, forces, reducedForcesPlus, reducedForcesMinus);
	}
	else
	{
		cout << "First call to CalculateExtendedHessian" << endl;
		CalculateExtendedHessian(nDOFs, hessMin, minimum, parametersMin, forces, reducedForcesPlus, reducedForcesMinus);
		cout << "Second call to CalculateExtendedHessian" << endl;
		CalculateExtendedHessian(nDOFs, hessSadd, saddle, parametersSadd, forces, reducedForcesPlus, reducedForcesMinus);
	}
	
	// Check that the matrices are symmetric
	cout << "Checking symmetry of hessMin" << endl;
	int nAsymmetric = CheckSymmetry(nDOFs, hessMin, 1000 * epsilon);
	cout << "Checking symmetry of hessSadd" << endl;
	nAsymmetric = CheckSymmetry(nDOFs, hessSadd, 1000 * epsilon);

	if( nAsymmetric > 0 )
	   cerr << "Warning: nAsymmetric = " << nAsymmetric <<"; max asymmetry is " << this->maxAsymmetry << endl;

	// arrays for eigenvalues, Lapack parameters
	double *eigenvaluesMin = new double[nDOFs];
	double *eigenvaluesSadd = new double[nDOFs];
	int lRealWork = - 1;
	double tempWork;
	int n = nDOFs, lda = nDOFs, info, stringLength=1;
	char jobz = 'V', uplo = 'U';
	// Lapack calls: matrix diagonalization. First, workspace query
	dsyev_(&jobz,&uplo,&n,hessMin,&lda,eigenvaluesMin,&tempWork,&lRealWork,&info);
	lRealWork = (int) tempWork;
	double *realWork   = new double[lRealWork];
	dsyev_(&jobz,&uplo,&n,hessMin,&lda,eigenvaluesMin,realWork,&lRealWork,&info);
	dsyev_(&jobz,&uplo,&n,hessSadd,&lda,eigenvaluesSadd,realWork,&lRealWork,&info);
	
	if(realWork[0] > lRealWork)
		cerr << "Workspace not optimal: optimal amount is: " << realWork[0] << endl;

	fstream outFile("eigenvalues",ios::out);
	fstream outFile2("unstableMode",ios::out);
	// accumulate the factors to construct the prefactor
	double prefactor = 1.0;
	int numPosMin = 0, numPosSadd = 0, unstableIdx = -1;
	for( i = 0; i < nDOFs; ++i )
	{
		if(eigenvaluesMin[i] > SMALL_EVALUE)// need a justifiable number here!
		{
			++numPosMin;
			prefactor *= eigenvaluesMin[i];			
		}
		if(eigenvaluesSadd[i] > SMALL_EVALUE)
		{
			++numPosSadd;
			prefactor /= eigenvaluesSadd[i];
		}
		if(eigenvaluesSadd[i] < -SMALL_EVALUE)
			unstableIdx = i;
		outFile << eigenvaluesMin[i] << "    " << eigenvaluesSadd[i] << endl;
	}
	prefactor = sqrt(prefactor);
	
	cout << "Hessian at the minimum has " << numPosMin << " positive eigenvalues" << endl;
	cout << "Hessian at the saddle-point has " << numPosSadd << " positive eigenvalues" << endl;

	double saddForceSq, massGradientFactor  = 1.;
	

	if(atomsExtension)
		 massGradientFactor = CalculateMassGradientFactor(hessSadd, nDOFs, unstableIdx, reducedForcesPlus, &saddForceSq);
	else
		saddForceSq = cblas_ddot(nDOFs,reducedForcesPlus, 1, reducedForcesPlus, 1);

	for( i = 0; i < nDOFs; ++i )
	{
		outFile2 << hessSadd[unstableIdx*nDOFs+i] << endl;
	}

	cout << "Norm-squared of force at saddle point: " << saddForceSq << endl;

	delete realWork;
	delete eigenvaluesMin;
	delete eigenvaluesSadd;
	delete hessMin;
	delete hessSadd;
	delete forces;
	delete reducedForcesPlus;
	delete reducedForcesMinus;

	
	// setting massGradientFactor to 1---I think it is cancelled by the
	// grad_f x denominator when there are no zero eigenvalues
	cout << "massGradientFactor " << massGradientFactor << endl;
	massGradientFactor = 1.;
	// final factor 
	prefactor *= ( massGradientFactor / ( 2 * PI ) );
   	
	return prefactor;
}


void Prefactor::CalculateHessian(int nDOFs, double *hessian, ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces, double *reducedForcesPlus, double *reducedForcesMinus)
{
	memset(hessian, 0, nDOFs*nDOFs*sizeof(double));

	int index=0;
	for( int i = 0; i < nDOFs; ++i )
    {
		// 'i'th derivative of the forces
		atoms->IncrementGeneralizedCoordinate(i,epsilon);
		
		potential->CalculateForces(atoms,forces);
		atoms->CalculateGeneralizedForces(forces,reducedForcesPlus);
		
		// now subtract 2*epsilon to get the original value minus epsilon
		atoms->IncrementGeneralizedCoordinate(i,-2*epsilon); 
		
		potential->CalculateForces(atoms,forces);
		atoms->CalculateGeneralizedForces(forces,reducedForcesMinus);
		
		for( int j = 0; j < nDOFs; ++j, ++index)
		{
			double forceDeriv = 0.5*(reducedForcesPlus[j] - reducedForcesMinus[j])/epsilon;
			hessian[index] = -forceDeriv;
		}	  
		// Now add epsilon to get back the original value
		atoms->IncrementGeneralizedCoordinate(i,epsilon);	
    }



	// return the forces at the given point (minimum or saddle)
	potential->CalculateForces(atoms,forces);
	atoms->CalculateGeneralizedForces(forces,reducedForcesPlus);
		

}



void Prefactor::CalculateExtendedHessian(int nDOFs, double *hessian, ListOfAtoms *atoms, const double *parameters, DMArray<double, DIMENSION> *forces, double *reducedForcesPlus, double *reducedForcesMinus)
{
	int prmIdx, p1, p2, p3, p4, i, j;
	// array to store forces at current point
	double *currentForces = new double[nDOFs];
	// to store inverse sq root of the mass matrix
	double *invRootMassMatrix = new double[nParameters * nParameters];
	// arrays for mass matrix stuff
	double *parametersLoc = new double[nParameters];
	double *parameterForces = new double[nParameters];
	double *kineticMatrix = new double[nParameters*nParameters];
	double atomicEnergy, parameterForceSq;
	int mainList = atomsExtension->GetMainList();
	int nFreeAtoms = atoms->GetLeaf(mainList)->GetNumber();
	int nFreeAtomsDOFs = nFreeAtoms * DIMENSION;


	for(prmIdx = 0; prmIdx < nParameters; ++prmIdx )
		parametersLoc[prmIdx] = parameters[prmIdx];

	atomsExtension->SetParameters(parametersLoc,atoms);
	
	// calculate forces at current point, and mass matrix
	atomicEnergy = atomsExtension->AtomicForcesEnergy(atoms,parametersLoc, potential, forces);
	atomsExtension->ParameterForcesEnergy(atoms, parametersLoc, parameterForces, kineticMatrix, forces, atomicEnergy, &parameterForceSq);
	memcpy(currentForces, (*forces->GetLeaf(mainList))[0], nFreeAtomsDOFs * sizeof(double));
	memcpy(currentForces + nFreeAtomsDOFs, parameterForces, nParameters * sizeof(double)); 


	// calculate inverse square root of mass matrix. Assume all the free atoms
	// have mass 1--we only have to worry about the parameters.


	// first diagonalize kineticMatrix

	
	double *massEigenvalues= new double[nDOFs];
	int n = nParameters, lda = nParameters, info, stringLength=1;
	char jobz = 'V', uplo = 'U';
	int lRealWork = 3 * nParameters - 1;
	double *realWork   = new double[lRealWork];
	// Lapack call: matrix diagonalization
	dsyev_(&jobz,&uplo,&n,kineticMatrix,&lda,massEigenvalues,realWork,&lRealWork,&info);


	ASSERT(info == 0);

	for( prmIdx = 0; prmIdx < nParameters; ++prmIdx )
		massEigenvalues[prmIdx] = 1./sqrt(massEigenvalues[prmIdx]);

	memset(invRootMassMatrix, 0, nParameters * nParameters*sizeof(double));
	memset(invMassMatrix, 0, nParameters * nParameters*sizeof(double));
	
	// for each p1, p2 pair, sum over p3 to compute that matrix element
	for( p2 = 0; p2 < nParameters; ++p2) // column index of target matrix
		for( p1 = 0; p1 < nParameters; ++p1) // row index of target matrix
		{
			for( p3 = 0; p3 < nParameters; ++p3)
			{
				// massEigenvalues is now really "inv square root mass
				// eigenvalues"
				invRootMassMatrix[p1 + p2 * nParameters] += massEigenvalues[p3] * kineticMatrix[p1 + nParameters * p3] * kineticMatrix[p2 + nParameters * p3];
				invMassMatrix[p1 + p2 * nParameters ] += massEigenvalues[p3] * massEigenvalues[p3] * kineticMatrix[p1 + nParameters * p3] * kineticMatrix[p2 + nParameters * p3];
			}
			
		}

	// this computation has been checked with temporary code used to 
	// multiply the mass matrix by invRootMassMatrix twice to get the identity
	// 19/8/2003

	// now do the hessian   
	int index = 0;
	for( i = 0; i < nDOFs; ++i )
    {

		// 'i'th derivative of the forces
		if( i < nFreeAtomsDOFs)
		{
			int at = i/DIMENSION, cd = i%DIMENSION;
			atoms->IncrementCartesianPosition( at, cd, epsilon);
		}
		else
		{
			parametersLoc[i - nFreeAtomsDOFs] += epsilon;
			atomsExtension->SetParameters(parametersLoc,atoms);
		}

		atomicEnergy = atomsExtension->AtomicForcesEnergy(atoms, parametersLoc, potential, forces);
		atomsExtension->ParameterForcesEnergy(atoms, parametersLoc, parameterForces, kineticMatrix,forces, atomicEnergy, &parameterForceSq);
		memcpy(reducedForcesPlus, (*forces->GetLeaf(mainList))[0], nFreeAtomsDOFs * sizeof(double));
		memcpy(reducedForcesPlus + nFreeAtomsDOFs, parameterForces, nParameters * sizeof(double));	
		
		// now subtract 2*epsilon to get the original value minus epsilon
		if( i < nFreeAtomsDOFs)
		{
			int at = i/DIMENSION, cd = i%DIMENSION;
			atoms->IncrementCartesianPosition( at, cd, -2. * epsilon);
		}
		else
		{
			parametersLoc[i - nFreeAtomsDOFs] += -2. * epsilon;
			atomsExtension->SetParameters(parametersLoc,atoms);
		}
		
		atomicEnergy = atomsExtension->AtomicForcesEnergy(atoms, parametersLoc, potential, forces);
		atomsExtension->ParameterForcesEnergy(atoms, parametersLoc,parameterForces, kineticMatrix,forces, atomicEnergy, &parameterForceSq);
		memcpy(reducedForcesMinus, (*forces->GetLeaf(mainList))[0], nFreeAtomsDOFs * sizeof(double));
		memcpy(reducedForcesMinus + nFreeAtomsDOFs, parameterForces, nParameters * sizeof(double));
		

		for( j = 0; j < nDOFs; ++j, ++index)
		{
			double forceDeriv = 0.5 * (reducedForcesPlus[j] - reducedForcesMinus[j])/epsilon;
			hessian[index] = -forceDeriv;
		}	  
		
		
		// Now add epsilon to get back the original value
		if( i < nFreeAtomsDOFs)
		{
			int at = i/DIMENSION, cd = i%DIMENSION;
			atoms->IncrementCartesianPosition( at, cd, epsilon);
		}
		else
		{
			parametersLoc[i - nFreeAtomsDOFs] += epsilon;
			atomsExtension->SetParameters(parametersLoc,atoms);
		}

	}



	// pre- and post- multiply the Hessian by invRootMassMatrix
	double *tempHess = new double[nDOFs * nDOFs];
	memcpy(tempHess, hessian, nDOFs * nDOFs * sizeof(double));	
	
	// sides of H: nFreeAtomsDOFs x nParameters
	for( i = nFreeAtomsDOFs; i < nDOFs; ++i) // row of target
		for( j = 0; j < nFreeAtomsDOFs; ++j ) // column of target
		{
			ASSERT(fabs(tempHess[i + nDOFs * j] - hessian[i + nDOFs * j]) < 1.e-5);
			hessian[i + nDOFs * j] = 0.;


			for( p3 = 0; p3 < nParameters; ++p3)
				{
					hessian[i + nDOFs * j] += invRootMassMatrix[(i-nFreeAtomsDOFs) + nParameters * p3] * tempHess[(nFreeAtomsDOFs + p3) + nDOFs * j];
				}

			hessian[j + nDOFs * i] = hessian[i + nDOFs * j];
		}
		
	// bottom right nParameters x nParameters corner of H
	for( p1 = 0; p1 < nParameters; ++p1) // row of target
		for( p2 = 0; p2 < nParameters; ++p2) //column of target
		{
			int targetIdx = (nFreeAtomsDOFs + p1) + (nFreeAtomsDOFs + p2) * nDOFs;
			hessian[targetIdx] = 0.;

			for( p3 = 0; p3 < nParameters; ++p3)
				for( p4 = 0; p4 < nParameters; ++p4)
					hessian[targetIdx] += 
						invRootMassMatrix[p1 + p3 * nParameters ] * 
						tempHess[(p3 + nFreeAtomsDOFs)
								+ nDOFs * (p4 + nFreeAtomsDOFs)] * 
						invRootMassMatrix[ p4 + nParameters * p2];
			
		}

	delete tempHess;
 
	// copy the current forces into reducedForcesPlus, which is where
	// the calling function expects them
	memcpy(reducedForcesPlus, currentForces, nDOFs*sizeof(double));




	
	delete invRootMassMatrix;
	delete currentForces;
	delete massEigenvalues;
	delete realWork;
	
	delete parametersLoc;
	delete parameterForces;
	delete kineticMatrix;
}


int Prefactor::CheckSymmetry(int size, double *matrix, double tol)
{
	cerr.precision(12);
	int nAsymmetric = 0;
	maxAsymmetry = 0.0;
	for( int i = 0; i < size; ++i )
		for( int j = 0; j < i; ++j )
		{
			double  asymmetry = fabs(matrix[i*size + j] - matrix[j*size + i]);
			if(asymmetry > maxAsymmetry)
				maxAsymmetry = asymmetry;
			if(asymmetry > tol)// The difference should be of order the truncation error which goes as epsilon squared but a proper comparison would require an estimate of the coefficient which would mean for example reevaluating the derivative at half epsilon.
			{
				cout << "ASYM: (i,j)=" << i << "," << j << "; vals " << matrix[j + size * i] << "," << matrix[i + size *j];
				++nAsymmetric;
			}
		}
	return nAsymmetric;
}


double Prefactor::CalculateMassGradientFactor(double *hessSadd, int nDOFs, int unstableIdx, double *currentForces, double *saddForceSq)
{
	int nFreeAtomsDOFs = nDOFs - nParameters;
	*saddForceSq = 0.;
	double massGradientFactor = 0.;
	
	for( int i = 0; i < nFreeAtomsDOFs; ++i )
	{	
		massGradientFactor += hessSadd[unstableIdx*nDOFs+i] * hessSadd[unstableIdx*nDOFs+i];
		*saddForceSq += currentForces[i] * currentForces[i];
	}
	
	for( int prmIdx1 = 0; prmIdx1 < nParameters; ++prmIdx1 )
		for( int prmIdx2 = 0; prmIdx2 < nParameters; ++prmIdx2 )
		{
			int idx1 = nFreeAtomsDOFs + prmIdx1;
			int idx2 = nFreeAtomsDOFs + prmIdx2;
			massGradientFactor += hessSadd[unstableIdx*nDOFs+idx1] * invMassMatrix[prmIdx1 + nParameters * prmIdx2] * hessSadd[unstableIdx*nDOFs+idx2];
			
			*saddForceSq += currentForces[idx1] * invMassMatrix[prmIdx1 + nParameters * prmIdx2] * currentForces[idx2];
		}
	
   	massGradientFactor = sqrt(massGradientFactor);
	return massGradientFactor;
}
