// FixedInLine.cpp: implementation of the FixedInLine class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "DigitalMaterial.h"

#include "FixedInLine.h"
#include "DMArray.h"

#include "DMLastHeader.h"
#include "ListOfAtoms.h"


const std::string FixedInLine::className = std::string("FixedInLine");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////



FixedInLine::FixedInLine(const DMArray<double, DIMENSION> *positions, const double tangent[DIMENSION])
{
	fixedPositions = 0;
	SetFixedPositions(positions);
	for(int i = 0; i < DIMENSION; i++)
		totalForce[i] = 0.0;
	SetTangent(tangent);
}

FixedInLine::~FixedInLine()
{
	if(fixedPositions) delete fixedPositions;
}

Constraint *FixedInLine::Copy() const
{
	return new FixedInLine(fixedPositions,fixedLineTangent);
}

void FixedInLine::SetFixedPositions(const DMArray<double, DIMENSION> *positions)
{
	if(fixedPositions) delete fixedPositions;
	fixedPositions = new DMArray<double, DIMENSION> (positions);
	fixedPositions->Copy(positions);
}

void FixedInLine::SetTangent(const double tangent[DIMENSION])
{
	for(int coord=0;coord<DIMENSION;coord++)
		fixedLineTangent[coord] = tangent[coord];
}

void FixedInLine::AdjustPositions(DMArray<double, DIMENSION> *positions)
{
	double disp, newDisp, dot;


	if(!(positions->GetShape() == fixedPositions->GetShape()))
		throw DMException("FixedInLine","AdjustPositions","Shape of positions differs from that of fixedPositions");
	ASSERT(positions->GetShape() == fixedPositions->GetShape());
	for(int i=0; i < atoms->GetNumber(); i++){
		dot = 0;
		for(int coord=0;coord<DIMENSION;coord++){
			disp = (*positions)[i][coord] - (*fixedPositions)[i][coord];
			dot += disp*fixedLineTangent[coord];
		}
		for(int coord=0;coord<DIMENSION;coord++){
			disp = (*positions)[i][coord] - (*fixedPositions)[i][coord];
			newDisp = dot * fixedLineTangent[coord];
			(*positions)[i][coord] = (*fixedPositions)[i][coord] + newDisp;
		}
	}	  
}

void FixedInLine::AdjustForces(DMArray<double, DIMENSION> *forces)
{
	double dot;


	forces->Sum(totalForce);
	for(int i=0; i < atoms->GetNumber(); i++){
		for(int coord=0;coord<DIMENSION;coord++){
			dot += (*forces)[i][coord]*fixedLineTangent[coord];
		}
		for(int coord=0;coord<DIMENSION;coord++){
			(*forces)[i][coord] = dot*fixedLineTangent[coord];
		}
	}
}

double FixedInLine::GetTotalForce(int coord)
{
	return totalForce[coord];
}

int FixedInLine::NumberGeneralizedCoordinates()
{
	return 0;
}

void FixedInLine::SetGeneralizedCoordinates(double *generalizedCoordinates)
{

}


void FixedInLine::GetGeneralizedCoordinates(double *generalizedCoordinates)
{

}


void FixedInLine::IncrementGeneralizedCoordinate(int n,double inc)
{
	
}


void FixedInLine::CalculateGeneralizedForces(DMArray<double, DIMENSION> *forces, double *generalizedForces)
{

}

void FixedInLine::DerivativesWRTGeneralizedCoordinates(int atom, double (*gradient)[DIMENSION])
{

}
