// FixedInPlane.cpp: implementation of the FixedInPlane class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "DigitalMaterial.h"

#include "FixedInPlane.h"
#include "DMArray.h"

#include "DMLastHeader.h"
#include "ListOfAtoms.h"


const std::string FixedInPlane::className = std::string("FixedInPlane");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////



FixedInPlane::FixedInPlane(const DMArray<double, DIMENSION> *positions, const double normal[DIMENSION])
{
	fixedPositions = 0;
	SetFixedPositions(positions);
	for(int i = 0; i < DIMENSION; i++)
		totalForce[i] = 0.0;
	SetNormal(normal);
}

FixedInPlane::~FixedInPlane()
{
	if(fixedPositions) delete fixedPositions;
}

Constraint *FixedInPlane::Copy() const
{
	return new FixedInPlane(fixedPositions,fixedPlaneNormal);
}

void FixedInPlane::SetFixedPositions(const DMArray<double, DIMENSION> *positions)
{
	if(fixedPositions) delete fixedPositions;
	fixedPositions = new DMArray<double, DIMENSION> (positions);
	fixedPositions->Copy(positions);
}

void FixedInPlane::SetNormal(const double normal[DIMENSION])
{
	for(int coord=0;coord<DIMENSION;coord++)
		fixedPlaneNormal[coord] = normal[coord];
}

void FixedInPlane::AdjustPositions(DMArray<double, DIMENSION> *positions)
{
	double disp, newDisp, dot;


	if(!(positions->GetShape() == fixedPositions->GetShape()))
		throw DMException("FixedInPlane","AdjustPositions","Shape of positions differs from that of fixedPositions");
	ASSERT(positions->GetShape() == fixedPositions->GetShape());
	for(int i=0; i < atoms->GetNumber(); i++){
		dot = 0;
		for(int coord=0;coord<DIMENSION;coord++){
			disp = (*positions)[i][coord] - (*fixedPositions)[i][coord];
			dot += disp*fixedPlaneNormal[coord];
		}
		for(int coord=0;coord<DIMENSION;coord++){
			disp = (*positions)[i][coord] - (*fixedPositions)[i][coord];
			newDisp = disp - dot * fixedPlaneNormal[coord]; 
			(*positions)[i][coord] = (*fixedPositions)[i][coord] + newDisp;
		}
	}	  
}

void FixedInPlane::AdjustForces(DMArray<double, DIMENSION> *forces)
{
	double dot;


	forces->Sum(totalForce);
	for(int i=0; i < atoms->GetNumber(); i++){
		for(int coord=0;coord<DIMENSION;coord++){
			dot += (*forces)[i][coord]*fixedPlaneNormal[coord];
		}
		for(int coord=0;coord<DIMENSION;coord++){
			(*forces)[i][coord] -= dot*fixedPlaneNormal[coord];
		}
	}
}

double FixedInPlane::GetTotalForce(int coord)
{
	return totalForce[coord];
}

int FixedInPlane::NumberGeneralizedCoordinates()
{
	return 0;
}

void FixedInPlane::SetGeneralizedCoordinates(double *generalizedCoordinates)
{

}


void FixedInPlane::GetGeneralizedCoordinates(double *generalizedCoordinates)
{

}


void FixedInPlane::IncrementGeneralizedCoordinate(int n,double inc)
{
	
}


void FixedInPlane::CalculateGeneralizedForces(DMArray<double, DIMENSION> *forces, double *generalizedForces)
{

}

void FixedInPlane::DerivativesWRTGeneralizedCoordinates(int atom, double (*gradient)[DIMENSION])
{

}
