// SingleDirectionRigidBody.cpp: implementation of the SingleDirectionRigidBody class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "DigitalMaterial.h"

#include "SingleDirectionRigidBody.h"
#include "DMArray.h"
#include "ListOfAtoms.h"


#include "DMLastHeader.h"

const std::string SingleDirectionRigidBody::className = std::string("SingleDirectionRigidBody");


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////


SingleDirectionRigidBody::SingleDirectionRigidBody(int constrainedCoord,double weight):weightFactor(weight)
{
	this->constrainedCoord = constrainedCoord;
	ASSERT(constrainedCoord < DIMENSION);
}

SingleDirectionRigidBody::~SingleDirectionRigidBody()
{
	
}

Constraint *SingleDirectionRigidBody::Copy() const
{
	SingleDirectionRigidBody *newConstraint = new SingleDirectionRigidBody(constrainedCoord,weightFactor);
	return newConstraint;
}


void SingleDirectionRigidBody::AdjustForces(DMArray<double, DIMENSION> *forces)
{
	double average = 0.;
	DMArrayIterator<double, DIMENSION> forcesIter(forces);

	for(forcesIter.Init(); !forcesIter.IsDone(); forcesIter.Next())
	  average += forcesIter.Current()[constrainedCoord];


	average /= (double) forcesIter.Size();

	for(forcesIter.Init(); !forcesIter.IsDone(); forcesIter.Next())
	  forcesIter.Current()[constrainedCoord] = average;
}

void SingleDirectionRigidBody::AdjustForceIncrements(DMArray<double, DIMENSION> *forceIncrements)
{
	AdjustForces(forceIncrements);
}

void SingleDirectionRigidBody::AdjustVelocities(DMArray<double, DIMENSION> *velocities)
{
	AdjustForces(velocities);
}


void SingleDirectionRigidBody::InitializeCOMAndDisplacements()
{
	// assumes all atoms have same mass!
	DMArrayIterator<double, DIMENSION> posIter (atoms->GetCartesianPositions());
	constrainedCenterOfMass = 0.;
	
	for(posIter.Init();!posIter.IsDone();posIter.Next())
		constrainedCenterOfMass += posIter.Current()[constrainedCoord];
	

	constrainedCenterOfMass /= (atoms->GetNumber() * atoms->GetMass());
	
	displacements = new DMArray<double, DIMENSION>(atoms->GetCartesianPositions()->GetShape());

	double shift[DIMENSION];
	for(int coord=0;coord<DIMENSION;++coord)
		shift[coord] = 0.;
	shift[constrainedCoord] = -constrainedCenterOfMass;

	displacements->Copy(atoms->GetCartesianPositions());
	displacements->Increment(shift);

	// only the constrainedCoord coordinates of the positions stored in displacements are important!!!
}

/*!
  Somewhat clumsy method for dealing with the other coordinates of the
  positions: unless new information is supplied (eg in the
  SetGeneralizedCoordiantes function) we copy the current values into the local
  array so that they are unchanged when we use the local array in 
  SetCartesianPositions. We have to do this since since is no version of 
  SetCarteisanPositions that lets us set only the constrainedCoord components.


 */

void SingleDirectionRigidBody::SetCartesianPositionsFromCOM(double *restOfPositions)
{
	if(!centerOfMassIsDefined)
	{
		InitializeCOMAndDisplacements();
		return;
	}
	// easiest way to set the positions: add the COM to the displacements
	// and set the positions using the displacements
	int coord;
	double shift[DIMENSION];
	for(coord=0;coord<DIMENSION;++coord)
		shift[coord] = 0.;
	shift[constrainedCoord] = constrainedCenterOfMass;

	displacements->Increment(shift);
	// set the other components to the current values from the atoms
	if(!restOfPositions) // if not being supplied with infor for the y and z components, keep them at their current values
	{

		for(coord=0;coord<DIMENSION;++coord)
		{
			if(coord != constrainedCoord)
			{
				for(int i=0;i<atoms->GetNumber();++i)
					(*displacements)[i][coord] = atoms->GetCartesianPosition(i,coord);
			}
		}
	}
	else
	{
		DMArrayIterator<double, DIMENSION> posIter(displacements);
		int index = 0;
		for(posIter.Init();!posIter.IsDone();posIter.Next())
		{
			// split the loop to avoid constrainedCoord
			for(coord=0;coord<constrainedCoord;++coord)
				posIter.Current()[coord] = restOfPositions[index++];

			for(coord=constrainedCoord+1;coord<DIMENSION;++coord)
				posIter.Current()[coord] = restOfPositions[index++];
		}
	}
	
	atoms->SetCartesianPositions(displacements);

	// then need to subtract the COM from the (constrainedCoord) displacements again.
	shift[constrainedCoord] *= -1.;
	displacements->Increment(shift);
}



int SingleDirectionRigidBody::NumberGeneralizedCoordinates()
{
	return 1 + (DIMENSION - 1)*atoms->GetNumber();
}

void SingleDirectionRigidBody::SetGeneralizedCoordinates(double *generalizedCoordinates)
{
	if(!centerOfMassIsDefined)
		InitializeCOMAndDisplacements();	// to have displacements be defined

	constrainedCenterOfMass = generalizedCoordinates[constrainedCoord]*weightFactor;


	SetCartesianPositionsFromCOM(generalizedCoordinates+1);
	
}

void SingleDirectionRigidBody::GetGeneralizedCoordinates(double *generalizedCoordinates)
{
	if(!centerOfMassIsDefined)
		InitializeCOMAndDisplacements();	

	generalizedCoordinates[constrainedCoord] = constrainedCenterOfMass/weightFactor;

	int coord, index = 1;
	for(int i=0;i<atoms->GetNumber();++i)
	{
		// split the loop to avoid constrainedCoord
		for(coord=0;coord<constrainedCoord;++coord)
			generalizedCoordinates[index++] = atoms->GetCartesianPosition(i,coord);
		
		for(coord=constrainedCoord+1;coord<DIMENSION;++coord)
			generalizedCoordinates[index++] = atoms->GetCartesianPosition(i,coord);
	}
	
}


void SingleDirectionRigidBody::IncrementGeneralizedCoordinate(int n,double inc)
{
	if(n==0)
    {
		constrainedCenterOfMass += inc*weightFactor;
		SetCartesianPositionsFromCOM();		
    }
	// this can't work if IncrementCartesianPosition itself uses the constraints
	else
    {
		int atom = (n-1)/(DIMENSION-1);
		int coord = (n-1)%(DIMENSION-1);
		if(coord >= constrainedCoord) ++ coord;
		
		atoms->IncrementCartesianPosition(atom,coord,inc);
    }
}

void SingleDirectionRigidBody::CalculateGeneralizedForces(DMArray<double, DIMENSION> *forces, double *generalizedForces)
{
	AdjustForces(forces);

	generalizedForces[0] = (*forces)[0][constrainedCoord] * atoms->GetNumber()*weightFactor;

	int index = 1;
	int coord;
	DMArrayIterator<double, DIMENSION> forcesIter(forces);
	for(forcesIter.Init();!forcesIter.IsDone();forcesIter.Next())
	{
		// split the loop to avoid constrainedCoord
		for(coord=0;coord<constrainedCoord;++coord)
				generalizedForces[index++] = forcesIter.Current()[coord];

		for(coord=constrainedCoord+1;coord<DIMENSION;++coord)
				generalizedForces[index++] = forcesIter.Current()[coord];
	}
}

void SingleDirectionRigidBody::DerivativesWRTGeneralizedCoordinates(int atom, double (*gradient)[DIMENSION])
{
	// all atoms, in particular that with number "atom", move in the 
	// cosntrainedCoord direction when that DOF is moved, but they don't move 
	// in any other direction
	int coord;
	for(coord=0;coord<DIMENSION;++coord)
		gradient[0][coord] = 0.;

	gradient[0][constrainedCoord] = weightFactor;
	

	// when the DOFs corresponding to the other components of individual atoms are changed, only the corresponding atom moves!
	int index=1;
	for(int i=0;i<atoms->GetNumber();++i)
	{
		for(coord=0;coord<DIMENSION;++coord)
			gradient[index++][coord] = 0.;
	}
	
	// the degrees of freedom corresponding to atom "atom", the gradient is 
	// diagonal, making sure to skip the constrained coord as usual
	index = 1 + atom*(DIMENSION-1);
	for(coord=0;coord<constrainedCoord;++coord)
		gradient[index++][coord] = 1.;
	for(coord=constrainedCoord;coord<DIMENSION;++coord)
		gradient[index++][coord] = 1.;
}

