// ArrayIO.h: interface for the ArrayIO class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef ARRAYIO_H
#define ARRAYIO_H

#include "DMFirstHeader.h"
#include <map>
#include "DMArray.h"
#include "DMProtocol.h"


template<class Type, int N>
void DMArray<Type, N>::Load(DMReader *reader)
{
    Shape *newShape = static_cast<Shape *> (reader->Get("shape"));
    this->SetShape(*newShape);
    const int leaves = newShape->NumberOfLeaves();
    std::vector<int> dims;

    for(int leaf=0; leaf<leaves; ++leaf)
    {
        SprintString name("Data%i",leaf);
        Type *tmp;
        reader->GetArray(dims, name, tmp);
        ASSERT(dims.size() == 2);
        ASSERT(dims[1] == N);
        memcpy(cache.ptrs[leaf],tmp, dims[0]*N*sizeof(Type));
        delete []tmp;
    }
    delete newShape;
}


template<class Type, int N>
void DMArray<Type, N>::Save(DMWriter *writer) const
{
    writer->Put(&shape, "shape");
    int n = N;
    std::vector<int> v(2,n); // v = [N,N]
    const std::vector<int> &leavesSize = shape.GetLeavesSize();
    for(unsigned int leaf=0; leaf<leavesSize.size(); ++leaf)
    {
        SprintString name("Data%i",leaf);
        v[0] = leavesSize[leaf];
        Type *data;
        data = reinterpret_cast<Type *>(cache.ptrs[leaf]);
        writer->PutArray(data, v, name);
    }
}


template<class Type, int N>
const DMArray<Type, N> *Gather(const DMArray<Type, N> *array, const DMArray<int> *ids = 0)
{
#ifdef DMPARALLEL
    const DMArray<int> *globIds = 0;
    if(ids)
    {
        globIds = Gather(ids);   // we could avoid having two communications. Not worth.
    }

    const int myProc = dmProtocol->GetProcessorNumber();
    const int nProcs = dmProtocol->GetNumberOfProcessors();
    DMProtocol::Buffers *buffers = dmProtocol->NewBuffers();

    array->Save(buffers->GetSendBuffer(0));
    DMArray<Type, N> *tmp = new DMArray<Type, N>();

    for(int proc=1; proc < nProcs; ++proc) tmp->Save(buffers->GetSendBuffer(proc));

    buffers->AllToAll(); // kind of a waste because communication only with proc 0  

    DMArray<Type, N> *gathered;
    if(myProc == 0) gathered = new DMArray<Type, N>();
    for(int proc2=0; proc2 < nProcs; ++proc2)
    {
        tmp->Load(buffers->GetRecvBuffer(proc2));
        if(myProc == 0) gathered->Merge(tmp);
    }
    delete tmp;
    delete buffers;
    if(globIds)  // implicitely myProc == 0
    {
        gathered->Sort(globIds);
        delete globIds;
    }
    if(myProc == 0) return gathered;
    else return 0;
#endif // DMPARALLEL
    return array;
}


template<class Type, int N>
void DMArray<Type, N>::Sort(const DMArray<int> *ids)
{
    ASSERT(ids->GetShape() == shape);
    const int nLeaves = NumberOfLeaves();
    for(int leaf=0; leaf != nLeaves; ++leaf)
    {
        const DMArray<int> *currentIds = ids->GetLeaf(leaf);

        const int thisSize = currentIds->Size();
        Type (*tmp)[N] = new Type[thisSize][N];
        Type (*current)[N] = cache.ptrs[leaf];
        // we have to use this trick in case there are holes in the ID vector
        std::map<int,int> sorted;
        for(int j=0; j<thisSize; ++j)
        {
            const int id = (*currentIds)[j];
            sorted[id] = j;
        }
        int i=0;
        for(std::map<int,int>::const_iterator it=sorted.begin();
            it != sorted.end(); ++it)
        {
            for(int j=0; j<N; ++j) tmp[i][j] = current[it->second][j]; 
            ++i;
        }
        memcpy(current, tmp, N*sizeof(Type)*thisSize);
        delete [] tmp;
    }
}





//! The default behavior is throwing an error
template<class Type>
std::string TypeToString(const Type &type)
{
    throw DMException("::", "TypeToString", "No conversion defined for this type");
}

template<>
inline std::string TypeToString(const double &type)
{
    return "double";
}

template<>
inline std::string TypeToString(const int &type)
{
    return "int";
}


template<class Type, int N>
const std::string &DMArray<Type, N>::GetType() const
{
    return className;
}


#endif // !defined ARRAYIO_H
