// DMarrayIterator.h: an iterator for DMarrays
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef _DMARRAYITERATOR
#define _DMARRAYITERATOR

#include "DMArray.h"

// ######################################################################
// DMArrayIterator Declaration
// ######################################################################


/*! 
  The iterator is the standard way to traverse a DMArray, because indexing
  with ints is slower. At this time, the iterator has the interface of the 
  Design Pattern. When the iterator is constructed, it copies the
  DMArray's cached list of pointers to the data. As we usual iterator, 
  if the container (DMArray) is changed, the iterator becomes invalidated.
 \ingroup dmsystem
*/

template <class Type, int N=1>
class DMArrayIterator
{
public:
	DMArrayIterator(const DMArray<Type, N> *dmArray);
	~DMArrayIterator() {}
	void Init() ;
	void Next();
	typename ReturnType<Type, N>::IndexOp Current() const {return ReturnType<Type, N>::convert(*current);}
	bool IsDone() const {return done;}
	int Size() const {return size;}
	const Index& GetIndex() const{ return index;}
protected:
	Type (*current)[N];
	Index index;
	unsigned int numberOfLists;
	unsigned int currentLimit;
	int size;
	bool done;
	const std::vector<int> sizes;
	const std::vector<Type (*)[N]> ptrs;
};


// #############################################################################
// ITERATOR Definitions
// #############################################################################



template<class Type, int N>
DMArrayIterator<Type, N>::DMArrayIterator(const DMArray<Type, N> *dmArray) : 
  sizes(dmArray->cache.sizes),ptrs(dmArray->cache.ptrs), size(dmArray->GetShape().GetSize())
{
	numberOfLists = sizes.size();
	done = (size == 0 ? true : false);
}

template<class Type, int N>
void DMArrayIterator<Type, N>::Init()
{
	current = ptrs[0];
	index = 0;
	currentLimit = sizes[0];
	if(currentLimit == 0) Next();
	done = (size == 0 ? true : false);
}


template<class Type, int N>
void DMArrayIterator<Type, N>::Next()
{
	if(++index.inLeaf < currentLimit)
		++current;
	else
	{
		index.inLeaf = 0;
		do
		{
			++index.leaf;
			if(index.leaf == numberOfLists)
			{
				done = true;
				currentLimit = 1; // to make sure to get out of the while loop
			}
			else 
			{
				current = ptrs[index.leaf];
				currentLimit = sizes[index.leaf];
			}
		}
		while(currentLimit == 0);
	}
}



#endif // ! def DMARRAYITERATOR
