// DynamicListOfAtoms.cpp: implementation of the DynamicListOfAtoms class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "DigitalMaterial.h"

#include "DynamicListOfAtoms.h"
#include "Shape.h"
#include "Constraint.h"

#include "DMWriter.h"
#include "DMReader.h"
#include "SerialBuilder.h"

#include "DMException.h"
#include "DMLastHeader.h"

using namespace std;

static SerialBuilder<DynamicListOfAtoms> *_DLOAcreator = 
    new SerialBuilder<DynamicListOfAtoms>("DynamicListOfAtoms");

const string DynamicListOfAtoms::className = string("DynamicListOfAtoms");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

DynamicListOfAtoms::DynamicListOfAtoms(const Shape shape, 
                                           const BoundaryConditions* bc,
                                           NeighborLocator *neighborLocator): 
                                           ListOfAtoms(0, bc, neighborLocator), 
                                           velocities(&identities)
{
    mass = 1.;
    Allocate(shape); // we passed a Shape(0) to list of atoms. 
}


DynamicListOfAtoms::~DynamicListOfAtoms()
{
    /* it appears that cutting the relation with the parent is the first 
       thing to do. We should not wait for ~ListOfAtoms to do it because
       the parent will want to access our velocities, which will already
       be deleted) */
    SetNeighborLocator(0);
    if(upperNode) upperNode->RemoveBranch(this); 
    ASSERT(!upperNode);
    Clear();
}

ListOfAtoms *DynamicListOfAtoms::LOAAllocator()
{
    return new DynamicListOfAtoms();
}


void DynamicListOfAtoms::Allocate(const Shape &newShape)
{
    ListOfAtoms::Allocate(newShape);
    if(newShape.IsLeaf())
    {
         velocities.SetShape(newShape);
    }
    ASSERT(positions.GetShape() == newShape);
}

void DynamicListOfAtoms::IncrementCartesianMomenta(
                           const DMArray<double, DIMENSION> *forces, double dt)
{
    if(branches)
    {
        int numOfLists = branches->size();
        for(int i=0; i<numOfLists; ++i)
            (*branches)[i]->IncrementCartesianMomenta(forces->GetBranch(i), dt);
    }
    else
    {
        this->WriteLock();
        velocities.Axpy(forces,dt/mass);
        this->WriteRelease();
    }
}


double DynamicListOfAtoms::GetKineticEnergy() const
{
    if(branches)
    {
        double Ekin = 0;
        for(std::vector<ListOfAtoms *>::iterator loa=branches->begin(); 
            loa != branches->end(); ++loa)
            Ekin += (*loa)->GetKineticEnergy();
        return Ekin;
    }
    else
    {
        // XXX It's not clear how much we gain using a blas routine here.
        DMArrayIterator<double, DIMENSION> veloIter(&velocities);
        double Kin = 0.;
        int coord;
        for(veloIter.Init(); !veloIter.IsDone(); veloIter.Next())
            for(coord=0; coord < DIMENSION;++coord)
                Kin += veloIter.Current()[coord]*veloIter.Current()[coord];
            
            return Kin*mass*0.5;
    }
}


void DynamicListOfAtoms::AdjustVelocities()
{
    constraint->AdjustVelocities(&velocities);
    if(branches) 
    {
        int numOfLists = branches->size();
        for(int lis=0; lis<numOfLists; ++lis) (*branches)[lis]->AdjustVelocities();
    }
}



void DynamicListOfAtoms::Save(DMWriter *writer) const
{
    ListOfAtoms::Save(writer);
    writer->PutInt(1, "DLOAversion");
    writer->PutDouble(mass, "mass");
    if(!branches) 
    {   // we gather the velocities on processor 0. On the other procs, "gathered" == 0.
        const DMArray<double, DIMENSION> *gathered = Gather(&velocities,&identities);
        if(gathered)
        {
            writer->Put(gathered, "velocities");
            if(gathered != &velocities) delete gathered;
        }
    }
}

void DynamicListOfAtoms::Load(DMReader *reader)
{
    ListOfAtoms::Load(reader);
    int version;
    try
    {
        version = reader->GetInt("DLOAversion");
    } catch ( DMException )
    {
        version = 0;
    }
    if(version > 0) this->mass = reader->GetDouble("mass");
    if(!branches)
    {
        DMArray<double, DIMENSION> *tmp = 
            tmp = dynamic_cast<DMArray<double, DIMENSION> *>(reader->Get("velocities"));
        if(!tmp)
            throw DMException("DynamicListOfAtoms","Load", "Cannot read velocities");
        velocities.SetShape(tmp->GetShape());
        velocities.Copy(tmp);
        delete tmp;
    }
}


void DynamicListOfAtoms::Clear()
{
    ListOfAtoms::Clear();
    velocities.SetShape(0 );
}

void DynamicListOfAtoms::RemoveBranch(ListOfAtoms *atomsToRemove)
{
    velocities.RemoveBranch(const_cast<DMArray<double, DIMENSION> *>
        (atomsToRemove->GetCartesianVelocities()));
    ListOfAtoms::RemoveBranch(atomsToRemove);
}

ListOfAtoms *DynamicListOfAtoms::Copy() const
{
    DynamicListOfAtoms *newListOfAtoms;
    if(GetShape().IsLeaf())
    {
        newListOfAtoms = new DynamicListOfAtoms(positions.GetShape(), bc);
        newListOfAtoms->SetCartesianPositions(&positions);
        newListOfAtoms->SetCartesianVelocities(&velocities);
        newListOfAtoms->SetMass(mass);
    }
    else
    {
        newListOfAtoms = new DynamicListOfAtoms();
        for(unsigned int list=0;list<branches->size();++list)
            newListOfAtoms->Attach((*branches)[list]->Copy());
    }
    newListOfAtoms->SetTypeName(typeName);
    newListOfAtoms->SetConstraint(constraint->Copy());
    // NeighborLocator, BoundaryConditions
    return newListOfAtoms;
}



void DynamicListOfAtoms::Merge(const ListOfAtoms *atoms)
{
    const DynamicListOfAtoms *dloa = dynamic_cast<const DynamicListOfAtoms *>(atoms);
    if(!dloa) throw DMException("DynamicListOfAtoms","Merge",
        "attempt to merge a non DynamicListOfAtoms");
    ListOfAtoms::Merge(atoms);
    velocities.Merge(&dloa->velocities);
}

void DynamicListOfAtoms::Attach(ListOfAtoms *atoms)
{
    DynamicListOfAtoms *dloa = dynamic_cast<DynamicListOfAtoms *>(atoms);
    if(!dloa) throw DMException("DynamicListOfAtoms","Attach",
        "attempt to attach a non DynamicListOfAtoms");
    ListOfAtoms::Attach(atoms);
    velocities.Attach(&dloa->velocities);
}
