// DynamicListOfAtoms.h: interface for the DynamicListOfAtoms class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef DYNAMICLISTOFATOMS_H
#define DYNAMICLISTOFATOMS_H

#include "DigitalMaterial.h"

#include "ListOfAtoms.h"

#include "DMArray.h"

class DynamicListOfAtoms : public ListOfAtoms
{
public:
    //! Create a DynamicListOfAtoms (ie atoms with positions and velocities only) of a given shape
#ifndef SWIG
    DynamicListOfAtoms(const Shape shape = Shape(0), 
                         const BoundaryConditions *bc=0,
                         NeighborLocator *neighborLocator=0);
#endif //SWIG
    virtual ~DynamicListOfAtoms();
    
    
#ifndef SWIG
    const DMArray<double, DIMENSION> *GetCartesianVelocities() const {return &velocities;}
    
    void SetCartesianVelocities(const DMArray<double, DIMENSION> *newVelocities)
    { velocities.Copy(newVelocities); }	
#endif
    
    //! Shouldn't use this in MD! Too slow.
    virtual double GetCartesianVelocity(int atomNumber, int coord) const {return velocities[atomNumber][coord];}
    
    virtual double GetKineticEnergy() const;
    
    // The Set and Increment array functions use memcpy and BLAS routines
    // and hence should be quite fast
    
    
    virtual void SetCartesianVelocity(int atomNumber, int coord, double newVelocity)
    { velocities[atomNumber][coord] = newVelocity; }
    
    
    virtual void ScaleCartesianVelocities(double scaleFactor=1.0){velocities *= scaleFactor; }
    // version for vector scaleFactor:
    virtual void ScaleCartesianVelocities(double scaleFactor[DIMENSION]){velocities *= scaleFactor; }
    
    virtual void IncrementCartesianVelocities(const DMArray<double, DIMENSION> *shift, double scaleFactor=1.0)
    {
        this->WriteLock();
        velocities.Axpy(shift,scaleFactor);
        this->WriteRelease();
    }
    // version for vector scaleFactor:
    virtual void IncrementCartesianVelocities(const DMArray<double, DIMENSION> *shift, double scaleFactor[DIMENSION])
    {
        this->WriteLock();
        velocities.Axpy(shift,scaleFactor);
        this->WriteRelease();
    }
    
    
    virtual void IncrementCartesianVelocity(int atomNumber, int coord, double deltaVelocity)
    { velocities[atomNumber][coord] += deltaVelocity; }
    
    virtual void IncrementCartesianMomenta(const DMArray<double, DIMENSION> *forces, double dt=1.0);
    
    void SetMass(double mass) {
		this->mass = mass;    if(branches) {
			int numOfLists = branches->size();
			for(int i=0; i<numOfLists; ++i)
				(*branches)[i]->SetMass(mass);
		}
	}
    double GetMass() {return mass;}
    
    virtual void Save(DMWriter *writer) const;
    virtual void Load(DMReader *reader);
    
    const std::string &GetType() const
	{
		return className;
	}

    
    virtual void Attach(ListOfAtoms *atoms);

    virtual void Merge(const ListOfAtoms *atoms);

    void AdjustVelocities();
    
    virtual ListOfAtoms *Copy() const;
    
    virtual void Clear();
    virtual void RemoveBranch(ListOfAtoms *atomsToRemove);
    
protected:
    
    DMArray<double, DIMENSION> velocities;
    
    double mass;
    
    //! Returns a new DynamicListOfAtoms
    virtual ListOfAtoms *LOAAllocator();
    
    virtual void Allocate(const Shape &newShape);
    
    static const std::string className;
    
};

#endif // !defined(DYNAMICLISTOFATOMS_H)
