// ListOfAtoms.cpp: implementation of the ListOfAtoms class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "DigitalMaterial.h"
#include "ListOfAtoms.h"
#include "CompositeNeighborLocator.h"  //XXXTC This should be redesigned such that we don't need this file
#include "FreeBoundaryConditions.h"
#include "Constraint.h"
#include "DMArray.h"
#include "DMException.h"
#include "AttributeError.h"

#include "SerialBuilder.h"
#include "SprintString.h" // for serialization

#include "ListOfAtomsObserver.h"

#include "DMLastHeader.h"

using namespace std;

static SerialBuilder<ListOfAtoms> *_LOAcreator = 
    new SerialBuilder<ListOfAtoms>("ListOfAtoms");

static SerialBuilder<DMArray<double,DIMENSION> > *_Arraycreator = 
    new SerialBuilder<DMArray<double, DIMENSION> >
        (SprintString("DMArray<double,%i>",DIMENSION));


template<> const std::string DMArray<double, DIMENSION>::className = std::string("DMArray<double") + SprintString(",%i>",DIMENSION);

template<> const std::string DMArray<double, 1>::className = std::string("DMArray<double,1>");

template<> const std::string DMArray<int, DIMENSION>::className = std::string("DMArray<int") + SprintString(",%i>",DIMENSION);

template<> const std::string DMArray<int, 1>::className = std::string("DMArray<int,1>");


const string ListOfAtoms::className("ListOfAtoms");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

ListOfAtoms::ListOfAtoms(const Shape shape,
                         const BoundaryConditions *bc,
                         NeighborLocator *neighborLocator):positions(&identities)
{
    branches = 0;
    upperNode = 0;
    flatList.push_back(this);

    Allocate(shape);
    noNeighborLocator = new NoNeighborLocator;
    freeBC            = new FreeBoundaryConditions;
    constraint        = new Constraint;
    constraint->atoms = this;
    
    // Initialize the Read-Write Lock;
    this->InitializeLock();
    this->WriteLock();

    this->bc = (bc == 0 ? freeBC: bc);
    this->neighborLocator = (neighborLocator == 0 ? noNeighborLocator : neighborLocator);
    this->neighborLocator->SetPositionArray(&positions); 
    this->neighborLocator->SetBoundaryConditions(this->bc);

    typeName = "Non-typed atoms";
    this->WriteRelease();

}

ListOfAtoms::~ListOfAtoms() 
{
    SetNeighborLocator(0);
    if(upperNode) upperNode->RemoveBranch(this);
    ASSERT(upperNode == 0);
    Clear();

    delete noNeighborLocator;
    delete freeBC;
    delete constraint;
    this->CleanupLock();
}



///////////////////////////////////////////////////////////////////////////////
/// METHODS RELATED TO THE POSITIONS ARRAY
///////////////////////////////////////////////////////////////////////////////


void ListOfAtoms::SetCartesianPositions(const DMArray<double, DIMENSION> *newPositions)
{
    if (newPositions != &positions)
        positions.Copy(newPositions);
    AdjustPositions();
    bc->EnforceBoundaryConditions(&positions);
    neighborLocator->UpdatePosition();
}

void ListOfAtoms::SetCartesianPosition(int atomNumber, int coord, double newPosition)
{ 
    positions[atomNumber][coord] = newPosition; 
    bc->EnforceBoundaryConditions(positions[atomNumber]);
    neighborLocator->UpdatePosition();
}

void ListOfAtoms::IncrementCartesianPositions(const DMArray<double, DIMENSION> *shift,
                                              double scaleFactor)
{
    this->WriteLock();
    positions.Axpy(shift,scaleFactor);
    AdjustPositions();
    bc->EnforceBoundaryConditions(&positions);
    this->WriteRelease();
    neighborLocator->UpdatePosition();
}

void ListOfAtoms::IncrementCartesianPositions(const DMArray<double, DIMENSION> *shift,
                                              double scaleFactor[DIMENSION])
{
    this->WriteLock();
    positions.Axpy(shift,scaleFactor);
    AdjustPositions();
    bc->EnforceBoundaryConditions(&positions);
    this->WriteRelease();
    neighborLocator->UpdatePosition();
}


void ListOfAtoms::IncrementCartesianPosition(int atomNumber,
                                             int coord,
                                             double deltaPosition)
{
    positions[atomNumber][coord] += deltaPosition; 
    bc->EnforceBoundaryConditions(positions[atomNumber]);
    neighborLocator->UpdatePosition();
}


///////////////////////////////////////////////////////////////////////////////
/// METHODS RELATED TO UNDEFINED VELOCITIES
///////////////////////////////////////////////////////////////////////////////

const DMArray<double, DIMENSION> *ListOfAtoms::GetCartesianVelocities() const
{ 
    throw AttributeError("ListOfAtoms","GetCartesianVelocities");
}


double ListOfAtoms::GetCartesianVelocity(int atomNumber, int coord) const 
{ 
    throw AttributeError("ListOfAtoms","GetCartesianVelocity");
}

double ListOfAtoms::GetKineticEnergy() const 
{ 
    throw AttributeError("ListOfAtoms","GetKineticEnergy");
}

void ListOfAtoms::SetCartesianVelocities(const DMArray<double, DIMENSION> *newVelocities)
{ 
    throw AttributeError("ListOfAtoms","SetCartesianVelocities");
}

void ListOfAtoms::SetCartesianVelocity(int atomNumber, int coord, 
                                       double newVelocity)
{ 
    throw AttributeError("ListOfAtoms","SetCartesianVelocity");
}

void ListOfAtoms::IncrementCartesianVelocities(const DMArray<double, DIMENSION> *shift,
                                               double scaleFactor)
{ 
    throw AttributeError("ListOfAtoms","IncrementCartesianVelocities");
}

void ListOfAtoms::IncrementCartesianVelocities(const DMArray<double, DIMENSION> *shift,
                                               double scaleFactor[DIMENSION])
{ 
    throw AttributeError("ListOfAtoms","IncrementCartesianVelocities");
}

void ListOfAtoms::IncrementCartesianVelocity(int atomNumber, int coord,
                                             double deltaVelocity)
{ 
    throw AttributeError("ListOfAtoms","IncrementCartesianVelocity");
}


void ListOfAtoms::ScaleCartesianVelocities(double scaleFactor)
{ 
    throw AttributeError("ListOfAtoms","IncrementCartesianVelocities");
}

void ListOfAtoms::ScaleCartesianVelocities(double scaleFactor[DIMENSION])
{ 
    throw AttributeError("ListOfAtoms","ScaleCartesianVelocities");
}




///////////////////////////////////////////////////////////////////////////////
////  ADJUSTMENT OF CONSTRAINTS
///////////////////////////////////////////////////////////////////////////////

// it's still unclear how a composite LOA can have it's own constraint in a consistent way
void ListOfAtoms::SetConstraint(Constraint *constraint)
{
    delete this->constraint;
    this->constraint = constraint;
    constraint->atoms = this; // this should be the only place the LOA pointer in constraint is set!
}

void ListOfAtoms::AdjustPositions()
{
    constraint->AdjustPositions(&positions);
    if(branches) 
    {
        int numOfLists = branches->size();
        for(int lis=0; lis<numOfLists; ++lis) (*branches)[lis]->AdjustPositions();
    }
}

void ListOfAtoms::AdjustForces(DMArray<double, DIMENSION> *forces)
{
    constraint->AdjustForces(forces);
    if(branches)
    {
        int numOfLists = branches->size();
        for(int list=0; list<numOfLists; ++list)
            (*branches)[list]->AdjustForces(forces->GetBranch(list));
    }
}

void ListOfAtoms::AdjustForceIncrements(DMArray<double, DIMENSION> *forces)
{
    constraint->AdjustForceIncrements(forces);
    if(branches)
    {
        int numOfLists = branches->size();
        for(int lis=0; lis<numOfLists; ++lis)
            (*branches)[lis]->AdjustForceIncrements(forces->GetBranch(lis));
    }
}


///////////////////////////////////////////////////////////////////////////////
////  GENERALIZED COORDINATES
///////////////////////////////////////////////////////////////////////////////


int ListOfAtoms::NumberGeneralizedCoordinates()
{
    if(branches)
    {
        int nGenCoords = 0;
        for(unsigned int list=0; list<branches->size(); ++list)
            nGenCoords += (*branches)[list]->NumberGeneralizedCoordinates();  
        return nGenCoords;
    }
    else
        return constraint->NumberGeneralizedCoordinates();
}

void ListOfAtoms::SetGeneralizedCoordinates(double *generalizedCoordinates)
{
    if(branches)
    {
        int nGenCoords = 0;
        for(unsigned int list=0; list<branches->size(); ++list)
        {
            (*branches)[list]->
                SetGeneralizedCoordinates(generalizedCoordinates + nGenCoords);
            nGenCoords += (*branches)[list]->NumberGeneralizedCoordinates(); 
        }
    }
    else
        constraint->SetGeneralizedCoordinates(generalizedCoordinates);
}

void ListOfAtoms::GetGeneralizedCoordinates(double *generalizedCoordinates)
{
    if(branches)
    {
        int nGenCoords = 0;
        for(unsigned int list=0; list<branches->size(); ++list)
        {
            (*branches)[list]->
                GetGeneralizedCoordinates(generalizedCoordinates + nGenCoords);
            nGenCoords += (*branches)[list]->NumberGeneralizedCoordinates(); 
        }
    }
    else
        constraint->GetGeneralizedCoordinates(generalizedCoordinates);
} 


void ListOfAtoms::IncrementGeneralizedCoordinate(int n,double inc)
{
    if(branches)
    {
        int nGenCoords = 0,list=0,numOfLists = branches->size();
        while(nGenCoords + (*branches)[list]->NumberGeneralizedCoordinates() - 1 < n)
    /* nDOFs + (*branches)[list]->NumberGeneralizedCoordinates() - 1 is the largest 
       index of a genCoord up to and including those of (*branch)[list]. So if this 
       largest value is smaller than n, we need to consider the next branch, 
       so add this one to the total */
        {
            nGenCoords += (*branches)[list]->NumberGeneralizedCoordinates();
            ++list;
        }
        (*branches)[list]->IncrementGeneralizedCoordinate(n-nGenCoords,inc);
    }
    else
        constraint->IncrementGeneralizedCoordinate(n,inc);
}

void ListOfAtoms::CalculateGeneralizedForces(DMArray<double, DIMENSION> *forces,
                                             double *generalizedForces)
{
    if(branches)
    {
        int nGenCoords = 0,list=0,numOfLists = branches->size();
        for(list=0;list<numOfLists;++list)
        {
            (*branches)[list]->CalculateGeneralizedForces(forces->GetBranch(list),
                generalizedForces+nGenCoords); 
            nGenCoords += (*branches)[list]->NumberGeneralizedCoordinates(); 
        }
    }
    else
        constraint->CalculateGeneralizedForces(forces,generalizedForces); 
}


void ListOfAtoms::DerivativesWRTGeneralizedCoordinates(int atom, 
                                                       double (*gradient)[DIMENSION])
{
    if(branches)
    {
        int subIndex = 0, nAtoms = 0, branchToUse = 0, subtotalGenCoords=0,genCoord;
        // Only the branch which contains atom number "atom" will give non zero gradients
        
        // 1. find which branch atom is in and get the total number of 
        //    Generalized Coordinates in branches up to that branch
        while( (nAtoms += (*branches)[branchToUse]->GetNumber()) < atom)
        {
            subtotalGenCoords += (*branches)[branchToUse]->NumberGeneralizedCoordinates();
            ++branchToUse;
        }
        
        // 2. set all elements of the gradient array corresponding to atoms not in 
        //    that branch to zero. First the branches before the actual branch of atom
        for(genCoord=0;genCoord<subtotalGenCoords;++genCoord)
            for(int coord=0;coord<DIMENSION;++coord)
                gradient[genCoord][coord] = 0.;
            
        // skip the branch with atom
        subtotalGenCoords += (*branches)[branchToUse]->NumberGeneralizedCoordinates();
        int nGenCoords = NumberGeneralizedCoordinates();
        for(genCoord=subtotalGenCoords;genCoord<nGenCoords;++genCoord)
            for(int coord=0;coord<DIMENSION;++coord)
                gradient[genCoord][coord] = 0.;
               
        // 3. pass the index within the branch of atom to the function in that branch
        subIndex = atom - nAtoms;
        (*branches)[branchToUse]->DerivativesWRTGeneralizedCoordinates(
                                  subIndex, gradient + subtotalGenCoords);
    }
    else
        constraint->DerivativesWRTGeneralizedCoordinates(atom, gradient);
}


////////////////////////////////////////////////////////////////////////////////
/// BOUNDARY CONDITIONS AND NEIGHBOR LOCATOR
////////////////////////////////////////////////////////////////////////////////

void ListOfAtoms::SetBoundaryConditions(const BoundaryConditions *bc)
{
    this->bc = bc;
    if(branches)
        for(vector<ListOfAtoms *>::iterator iter = branches->begin();
            iter != branches->end(); ++iter)
        {
                (*iter)->SetBoundaryConditions(bc);
        }
    else
        bc->EnforceBoundaryConditions(&positions);

    neighborLocator->SetBoundaryConditions(bc);
    neighborLocator->UpdatePosition();
}

void ListOfAtoms::SetNeighborLocator(NeighborLocator *neighborLocator)
{
    if(neighborLocator == NULL) 
    {
        this->neighborLocator = noNeighborLocator;
    }
    else
    {
        this->neighborLocator = neighborLocator;
        neighborLocator->SetBoundaryConditions(bc);
        neighborLocator->SetPositionArray(GetCartesianPositions());
    }
    // if this listOfAtoms belongs to a composite and the parent has a
    // composite NeighborLocator, the latter has to be notified.
    if(upperNode)
    {
        CompositeNeighborLocator *parentsNL =
            dynamic_cast<CompositeNeighborLocator *> (upperNode->GetNeighborLocator());
        if (parentsNL) parentsNL->SetChildren();
    }
}


///////////////////////////////////////////////////////////////////////////////////////
/// METHODS RELATED TO THE TREE STRUCTURE AND TREE MANAGEMENT
///////////////////////////////////////////////////////////////////////////////////////


void ListOfAtoms::Merge(const ListOfAtoms *atoms)
{
    if(branches) // the leaves are responsible for merging the data (because of ID)
    {
        for(unsigned int i=0; i<branches->size(); ++i)
            (*branches)[i]->Merge(atoms->GetBranch(i));
    }
    else 
    {
        positions.Merge(&atoms->positions);
        // the IDs of the merged atoms is shifted w.r.t. their value in the original LOA.
        const int shift = GetShape().GetSize();
        const int start = identities.Size();
        identities.Merge(&atoms->identities);
        const int end   = identities.Size();
        for(int i = start; i<end; ++i) identities[i] += shift;
    }
    UpdateTree();
}


void ListOfAtoms::Attach(ListOfAtoms *atoms)
{
    if(!branches)
    {
        if(GetNumber() > 0)
            throw DMException("ListOfAtoms", "Attach", "Cannot attach a branch to a leaf");
        Clear();
        branches = new std::vector<ListOfAtoms *>;
    }
    branches->push_back(atoms);

    positions.Attach (&atoms->positions) ;
    identities.Attach(&atoms->identities);

    neighborLocator->UpdatePosition();
    atoms->SetUpperNode(this);
    UpdateTree();
}



void ListOfAtoms::RemoveBranch(ListOfAtoms *atomsToRemove)
{
    if(!branches)
        throw DMException("ListOfAtoms", "RemoveBranch", 
                          "Cannot remove a branch from a leaf");
    vector<ListOfAtoms *>::iterator found;
    found = find(branches->begin(), branches->end(),atomsToRemove);
    if(found == branches->end()) 
        throw DMException("ListOfAtoms", "RemoveBranch", "branch to remove not found");
    (*found)->SetUpperNode(0);  // we also want to notify the branch that we cut our relations!
    branches->erase(found);
    UpdateTree();

    positions. RemoveBranch(&atomsToRemove->positions);
    identities.RemoveBranch(&atomsToRemove->identities);

    neighborLocator->UpdatePosition();
}



ListOfAtoms *ListOfAtoms::GetBranch(int i)
{
    if(branches) return (*branches)[i];
    else throw DMException("ListOfAtoms", "GetBranch", 
                           "attempt to access branches of a leaf");
}

const ListOfAtoms *ListOfAtoms::GetBranch(int i) const
{
    if(branches) return (*branches)[i];
    else throw DMException("ListOfAtoms", "GetBranch", 
                           "attempt to access branches of a leaf");
}


ListOfAtoms *ListOfAtoms::LOAAllocator()
{
    return new ListOfAtoms();
}


void ListOfAtoms::Clear()
{
    if(branches)
    {
        for(vector<ListOfAtoms *>::reverse_iterator typ=branches->rbegin(); 
            typ != branches->rend(); ++typ) 
                delete *typ;
        delete branches;
        branches = 0;
    }
    flatList.clear();
    flatList.push_back(this);

    positions.SetShape(0);
    identities.SetShape(0);
}


void ListOfAtoms::Allocate(const Shape &newShape)
{
    // if the new tree structure is different, we throw out all our data
    if(!GetShape().IsSimilar(newShape))
    {
        Clear();
        if(!newShape.IsLeaf())
        {
            branches = new vector<ListOfAtoms *>;
            flatList.clear();
            for(int i=0; i<newShape.GetNSubShapes(); ++i)
            {
                ListOfAtoms *branch = LOAAllocator(); // virtual.
                branch->SetShape(newShape.GetSubShape(i));
                Attach(branch);
            }
        }
    }
    else
    {
        if(newShape.IsLeaf())
        {
            positions.SetShape(newShape);
            identities.SetShape(newShape);
            for(int i=0; i<newShape.GetSize(); ++i) identities[i] = i;
            UpdateTree(); // this notifies the parent arrays
        }
        else
        {
            ASSERT(branches);
            for(unsigned int i=0; i<branches->size(); ++i)
                (*branches)[i]->SetShape(newShape.GetSubShape(i));
        }
    }
    
    ASSERT(identities.GetShape() == newShape);
}


void ListOfAtoms::UpdateTree()
{
    flatList.clear();
    if(branches)
    {
        for(vector<ListOfAtoms *>::iterator iter=branches->begin(); 
            iter != branches->end(); ++iter)
        {
            flatList.insert(flatList.end(),
                           (*iter)->flatList.begin(), (*iter)->flatList.end());
        }
    }
    else
    {
        flatList.push_back(this);
    }
    if(upperNode) upperNode->UpdateTree();
}


void ListOfAtoms::RemoveAtoms(const set<int> & atomNumbers)
{
    // user should be aware that atom numbering is screwed up after calling this routine!!!
    set<Index> atomsIndex; //we use int in LOA and Index in Arrays
    const vector<int> &leavesSize = GetShape().GetLeavesSize();
    for(set<int>::const_iterator atom=atomNumbers.begin();atom != atomNumbers.end(); atom++)
    {
        int tmp = *atom;
        int currentLeaf = 0;
        while(tmp >= leavesSize[currentLeaf])
        {
            tmp -= leavesSize[currentLeaf++];
        }
        atomsIndex.insert(Index(currentLeaf, tmp));
    }
    const DMArray_AnyType::siblingContainer& siblings = positions.GetSiblings();
    DMArray_AnyType::siblingContainer::const_iterator arrayIt;
    for(arrayIt = siblings.begin(); arrayIt != siblings.end(); ++arrayIt)
	(*arrayIt)->RemoveElements(atomsIndex);

    neighborLocator->UpdatePosition();
    if(upperNode) upperNode->UpdateTree();
}


////////////////////////////////////////////////////////////////////////////////
/// METHODS RELATED TO THE SERIALIZATION
////////////////////////////////////////////////////////////////////////////////

void ListOfAtoms::Save(DMWriter *writer) const
{
    writer->PutInt(1,"LOAversion");
    writer->PutString(typeName, "atomType");
    if(branches)
    {
        writer->PutInt(branches->size(), "NumberOfSubLists");
        for(unsigned int i=0; i<branches->size(); ++i)
        {
            SprintString name("subLOA%i",i);
            writer->Put((*branches)[i], name);
        }
    }
    else
    {
        writer->PutInt(-1, "NumberOfSubLists"); // to declare it's a leaf
        // save positions:
        // we gather the array on processor 0. On the other procs, "gathered" == 0.
        const DMArray<double, DIMENSION> *gathered = Gather(&positions,&identities);
        if(gathered)
        {
            writer->Put(gathered, "positions");
            if(gathered != &positions) delete gathered;
        }
    }
}

void ListOfAtoms::Load(DMReader *reader)
{
    Clear();
    int version;
    try
    {
        version = reader->GetInt("LOAversion");
    } catch (DMException)
    {
        version = 0;
    }
    if(version > 0) this->typeName = reader->GetString("atomType");
    int nSubLists = reader->GetInt("NumberOfSubLists");
    if(nSubLists >= 0)
    {
        branches = new std::vector<ListOfAtoms *>;
        for(int i=0; i<nSubLists; ++i)
        {
            SprintString name("subLOA%i",i);
            Attach(dynamic_cast<ListOfAtoms *>(reader->Get(name)));
        }
    }
    else
    {
        DMArray<double, DIMENSION> *tmp = 
            dynamic_cast<DMArray<double, DIMENSION> *>(reader->Get("positions"));
        if(!tmp)
            throw DMException("ListOfAtoms", "Load", "Cannot read positions");
        positions.SetShape(tmp->GetShape());
        positions.Copy(tmp);
        delete tmp;
        identities.SetShape(positions.GetShape());
        const int thisSize = identities.Size();
        for(int i=0; i<thisSize; ++i) identities[i] = i;

    }
    // no more data to read. The derived classes will read their part if the list leaf.
}


const std::string &ListOfAtoms::GetType() const
{
    return className;
}



ListOfAtoms *ListOfAtoms::Copy() const
{
    // creates new object, then passes back a pointer to it. 
    // This way the calling function, if it has a base class pointer, 
    // doesn't need to know the actual type to make a copy of it
    ListOfAtoms *newListOfAtoms = new ListOfAtoms(identities.GetShape());
    newListOfAtoms->SetTypeName(typeName);
    newListOfAtoms->SetConstraint(constraint->Copy());
        
    // Things like the pointers to NeighborLocator and BoundaryConditions
    // are to be dealt with differently, since they have to point to copies 
    // of the existing NeighborLocator etc.
    // This should be taken care of in the function where the copies are being made.
    
    return newListOfAtoms;
}


////////////////////////////////////////////////////////////////////////////////
/// METHODS RELATED TO THE SUBJECT/OBSERVER PATTERN
////////////////////////////////////////////////////////////////////////////////

bool ListOfAtoms::AddObserver(ListOfAtomsObserver *observer)
{
    return observers.AddObserver(observer);
}

bool ListOfAtoms::RemoveObserver(ListOfAtomsObserver *observer)
{
    return observers.RemoveObserver(observer);
}

void ListOfAtoms::Notify(int notifyLevel)
{
    for (unsigned int i=0; i< observers.size(); ++i)
	{
		if(observers.at(i)->GetNotifyLevel() <= notifyLevel)
			observers.at(i)->Update(this);
}
}
