// CellNeighborLocator.cpp: implementation of the CellNeighborLocator class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/
#include "DMFirstHeader.h"
#include <iostream>
#include <math.h>

#include "DigitalMaterial.h"

#include "DMArray.h"
#include "LeafToLeafIterator.h"
#include "CellNeighborLocator.h"
#include "SimpleNeighborLocator.h"
#include "CellNeighborList.h"
#include "BoundaryConditions.h"
#include "SerialBuilder.h"

#include "DMLastHeader.h"

using namespace std;


static SerialBuilder<CellNeighborLocator> *_cellNLcreator = 
	new SerialBuilder<CellNeighborLocator>("CellNeighborLocator");

const string CellNeighborLocator::className = string("CellNeighborLocator");

const int CellNeighborLocator::MAXCELLS = 100000000;

CellNeighborLocator::CellNeighborLocator(double cutoffDistance, double drift) :
     NeighborLocator(cutoffDistance), maxCellsForSimpleNeighborLocator(40), maxCellsForCellNeighborList(100000)
{
    nAtoms = 0; 
    storage = 0;
    Storage = 0;
    cellListGlobal = 0;
    cellList = 0;
    typedCellList = 0;
    numberOfAtomsInCell = 0;
    rankToIndex = 0;
    this->drift = 0.0;

    totalNumberOfCells = 0;
    numberOfNonEmptyCells = 0;

    oldPositions = new DMArray<double, DIMENSION>();
    nextLevelNeighborLocator = 0;
    cellsPositions = new DMArray<double, DIMENSION>();

    SetCutoffDistance(cutoffDistance);
    SetDrift(drift);
    
    this->maxDisplacement = 0.;
}

CellNeighborLocator::~CellNeighborLocator()
{
    if(storage)             delete [] storage;
    if(Storage)             delete [] Storage;
    if(cellListGlobal)      delete [] cellListGlobal;
    if(cellList)            delete [] cellList;
    if(typedCellList)       delete [] typedCellList;
    if(numberOfAtomsInCell) delete [] numberOfAtomsInCell;
    if(rankToIndex)         delete [] rankToIndex;
    if(oldPositions)        delete oldPositions;
    if(nextLevelNeighborLocator) delete nextLevelNeighborLocator;
    delete cellsPositions;
}

void CellNeighborLocator::SetCutoffDistance(double cutoff)
{
    double oldCutoff = cutoffDistance;
    NeighborLocator::SetCutoffDistance(cutoff);
    cellSize = cutoffDistance + 2.*drift;
    if((oldCutoff < cutoffDistance) && (positions)) UpdateMyData();
}

void CellNeighborLocator::SetDrift(double drift)
{
    double oldDrift = this->drift;
    this->drift = drift;
    cellSize = cutoffDistance + 2.*drift;
    if((oldDrift < drift) && (nAtoms > 0) && !DataIsValid()) UpdateMyData();
}

void CellNeighborLocator::UpdatePosition()
{   // Check and see whether the criterion for rebuilding cell list has been satisfied, if so rebuild the cell list.
    if(!DataIsValid()) UpdateMyData();  // reminder: UpdateData is virtual 
}

void CellNeighborLocator::Allocate(int nAtoms)
{
    this->nAtoms = nAtoms;
    if(oldPositions) delete oldPositions;
    oldPositions = new DMArray<double, DIMENSION> (positions->GetShape()); // copy real positions
    if(storage) delete [] storage;
    if(Storage) delete [] Storage;
    storage   = new int[nAtoms];
    Storage   = new Index[nAtoms];
    cellIndexOfPositions.SetShape(positions->GetShape());
    ASSERT(cellIndexOfPositions.GetShape() == positions->GetShape());
}


bool CellNeighborLocator::DataIsValid()
{
    if(positions->Size() == 0) return true;
    if(!(positions->GetShape() == oldPositions->GetShape())) return false;

    this->maxDisplacement = sqrt(MaximumDisplacementSquared(positions));
    if (maxDisplacement > drift) return false;
    else return true;
}

void CellNeighborLocator::UpdateMyData()
{
    int newNAtoms = positions->Size();
    if(!(positions->GetShape() == cellIndexOfPositions.GetShape()))
        Allocate(newNAtoms);

    oldPositions->Copy(positions);
    if(positions->Size() > 0) MakeCellList();	

    this->maxDisplacement = 0.;
}

void CellNeighborLocator::SetBoundaryConditions(const BoundaryConditions *bc)
{
    if(this->bc != bc)
    {
        NeighborLocator::SetBoundaryConditions(bc);
        if (positions) UpdateMyData(); // Rebuild everything
    }
}


bool CellNeighborLocator::HalfNeighbors(int atomNumber, int *neighbors, int& numNeighbors, 
            double (*displacements)[DIMENSION], double* displacementsSquared, int& spacesLeftInArrays)
{
    numNeighbors = 0;
    if(cutoffDistance == 0.) return true;
    int candidatesCount = 0;

    // Find out the cell number of the cell that atom atomNumber is in
    int cellIndexOfThisAtom = cellIndexOfPositions[atomNumber];
    if(nextLevelNeighborLocator)
    {
        int cellOccupancy, numberOfNeighborCells;
        int slots = spacesLeftInArrays;
        // we need all the neighboring cells, because HalfNeighbors should return the 
        // atoms with a higher index that atomNumber
        bool append = nextLevelNeighborLocator->Neighbors(cellIndexOfThisAtom, neighbors, 
            numberOfNeighborCells, displacements, displacementsSquared, slots);
        if(!append) return false;
        // we'll need the neighbors array later
        int *neighborCells = new int[numberOfNeighborCells];
        memcpy(neighborCells, neighbors, numberOfNeighborCells*sizeof(int)); 
        for(int nbCell = 0; nbCell<numberOfNeighborCells; ++nbCell)
        {
            cellOccupancy = numberOfAtomsInCell[neighborCells[nbCell]];
            int *thisList = cellListGlobal[neighborCells[nbCell]];
            for(int index=cellOccupancy-1; index>=0; --index)
            {
                if(candidatesCount >= spacesLeftInArrays) return false;
                if(thisList[index] > atomNumber)
                    neighbors[candidatesCount++] = thisList[index];
            }
        }
        delete [] neighborCells;
    }
    // then we add half the atoms in the same cell
    int atomsInThisCell = numberOfAtomsInCell[cellIndexOfThisAtom];
    int *atomPtr = cellListGlobal[cellIndexOfThisAtom] + atomsInThisCell - 1;
    //we probably won't have enough space.
    if(candidatesCount + atomsInThisCell > spacesLeftInArrays) return false;
    // insert half of the atoms which are located in this cell
    while(*atomPtr != atomNumber) 
        neighbors[candidatesCount++] = *(atomPtr--);
    const double *center = (*positions)[atomNumber];
    PruneNeighborCandidates(center,candidatesCount,neighbors,positions,
        displacements,displacementsSquared, cutoffDistance2, numNeighbors);
    spacesLeftInArrays -= numNeighbors; 
    return true;    
}

bool CellNeighborLocator::Neighbors(int atomNumber, int *neighbors, int& numNeighbors, 
            double (*displacements)[DIMENSION], double* displacementsSquared, int& spacesLeftInArrays)
{
    numNeighbors = 0;
    if(cutoffDistance == 0.) return true;
    int candidatesCount = 0;

    // Find out the cell number of the cell that atom atomNumber is in
    int cellIndexOfThisAtom = cellIndexOfPositions[atomNumber];
    if(nextLevelNeighborLocator)
    {
        int cellOccupancy, numberOfNeighborCells = 0;
        int slots = spacesLeftInArrays;
        bool append = nextLevelNeighborLocator->Neighbors(cellIndexOfThisAtom, neighbors, 
            numberOfNeighborCells, displacements, displacementsSquared, slots);
        if(!append) return false;
        int *neighborCells = new int[numberOfNeighborCells];
        memcpy(neighborCells, neighbors, numberOfNeighborCells*sizeof(int)); // we need the neighbors array below
        for(int nbCell = 0; nbCell<numberOfNeighborCells; ++nbCell)
        {
            cellOccupancy = numberOfAtomsInCell[neighborCells[nbCell]];
            if(candidatesCount + cellOccupancy > spacesLeftInArrays) return false;
            memcpy(neighbors+candidatesCount, cellListGlobal[neighborCells[nbCell]],
                cellOccupancy*sizeof(int));
            candidatesCount += cellOccupancy;
        }
        delete [] neighborCells;
    }
    // then we add the atoms in the same cell
    int *atomPtr = cellListGlobal[cellIndexOfThisAtom];
    int *end = atomPtr+numberOfAtomsInCell[cellIndexOfThisAtom];
    if(candidatesCount + numberOfAtomsInCell[cellIndexOfThisAtom] > spacesLeftInArrays) return false;
    while(atomPtr != end) 
    {
        if(*atomPtr != atomNumber) neighbors[candidatesCount++] = *atomPtr;
        ++atomPtr;
    }
    const double *center = (*positions)[atomNumber];
    PruneNeighborCandidates(center,candidatesCount,neighbors,positions,
        displacements,displacementsSquared, cutoffDistance2, numNeighbors);
    spacesLeftInArrays -= numNeighbors; 
    return true;
}

/*bool CellNeighborLocator::Neighbors(Index atomNumber, Index *neighbors, int& numNeighbors, 
            double (*displacements)[DIMENSION], double* displacementsSquared, int& spacesLeftInArrays)
{
    numNeighbors = 0;
    if(cutoffDistance == 0.) return true;
    int maxNumberOfNeighbors = 0;
    bool append = true;

    // Find out the cell number of the cell that atom atomNumber is in
    int cellIndexOfThisAtom = cellIndexOfAtom[atomNumber];
    int iTmp = spacesLeftInArrays;       // we'll overwrite the arrays, so spacesLeftInArrays shouldn't change
    int cellOccupancy, numberOfNeighborCells = 0;
    if(nextLevelNeighborLocator)
    {
        int nbCell;
        append = nextLevelNeighborLocator->Neighbors(cellIndexOfThisAtom, neighbors, 
            numberOfNeighborCells, displacements, displacementsSquared, iTmp);
        int *neighborCells = new int[numberOfNeighborCells];
        memcpy(neighborCells, neighbors, numberOfNeighborCells*sizeof(int)); // we need the neighbors array below
        iTmp = spacesLeftInArrays; // we have to make sure there's enough space to store all the neighbor candidates.
        for(nbCell = 0; append && nbCell<numberOfNeighborCells; ++nbCell)
        {
            cellOccupancy = numberOfAtomsInCell[neighborCells[nbCell]];
            iTmp -= cellOccupancy;
            append &= (iTmp >= 0);
            if(append) memcpy(neighbors+maxNumberOfNeighbors, cellListGlobal[neighborCells[nbCell]],
                cellOccupancy*sizeof(Index));
            maxNumberOfNeighbors += cellOccupancy;
        }
        delete [] neighborCells;
    }
    // then we add the atoms in the same cell
    int *atomPtr = cellListGlobal[cellIndexOfThisAtom];
    int *end = atomPtr+numberOfAtomsInCell[cellIndexOfThisAtom];
    iTmp -= numberOfAtomsInCell[cellIndexOfThisAtom];
    if(iTmp < 0) append = false; //we probably won't have enough space.
    else
        // insert half of the atoms which are located in this cell
        while(atomPtr != end) 
        {
            if(*atomPtr != atomNumber) neighbors[maxNumberOfNeighbors++] = *atomPtr;
            ++atomPtr;
        }
    if(!append) return false;
    const DMVector<double, DIMENSION> &center = (*positions)[atomNumber];
    PruneNeighborCandidates(center,maxNumberOfNeighbors,neighbors,positions->GetExtended(),
        displacements,displacementsSquared, cutoffDistance2, numNeighbors);
    spacesLeftInArrays -= numNeighbors; 
    return true;
}
*/
bool CellNeighborLocator::Neighbors(int atomNumber, int type1, int type2, int *neighbors, 
                int& numNeighbors, double (*displacements)[DIMENSION], double* displacementsSquared, 
                int& spacesLeftInArrays)
{
    numNeighbors = 0;
    if(cutoffDistance == 0.) return true;

    int candidatesCount = 0;
    // Find out the cell number of the cell that atom atomNumber is in
    int cellIndexOfThisAtom = (*cellIndexOfPositions.GetLeaf(type1))[atomNumber];
    if(nextLevelNeighborLocator)
    {
        int cellOccupancy, numberOfNeighborCells = 0;
        int slots = spacesLeftInArrays;
        bool append = nextLevelNeighborLocator->Neighbors(cellIndexOfThisAtom, neighbors, 
            numberOfNeighborCells, displacements, displacementsSquared, slots);
        if (!append) return false;
        // we'll need the neighbors array later
        int *neighborCells = new int[numberOfNeighborCells];
        memcpy(neighborCells, neighbors, numberOfNeighborCells*sizeof(int));
        // we have to make sure there's enough space to store all the neighbor candidates.
        for(int nbCell = 0; nbCell<numberOfNeighborCells; ++nbCell)
        {
            int thiscell = neighborCells[nbCell];
            cellOccupancy = typedCellList[thiscell][type2].second;
            if(candidatesCount + cellOccupancy > spacesLeftInArrays) return false;
            for(int nb=0; nb<cellOccupancy; ++nb)
            {
                neighbors[candidatesCount+nb] = 
                    typedCellList[thiscell][type2].first[nb].inLeaf;
                int ll = typedCellList[thiscell][type2].first[nb].leaf;
                ASSERT(ll == type2);
            }
//          memcpy(neighbors+candidatesCount, typedCellList[thiscell][type2].first,
//              cellOccupancy*sizeof(int));
            candidatesCount += cellOccupancy;
        }
        delete [] neighborCells;
    }
    // then we add the atoms in the same cell
    Index *atomPtr = typedCellList[cellIndexOfThisAtom][type2].first;
    int cellOccupancy = typedCellList[cellIndexOfThisAtom][type2].second;
    Index *end = atomPtr + cellOccupancy;
    if(candidatesCount + cellOccupancy > spacesLeftInArrays) return false; //we probably won't have enough space.
    // insert half of the atoms which are located in this cell
    if(type1 != type2) while(atomPtr != end) neighbors[candidatesCount++] = (atomPtr++)->inLeaf;
    else
        while(atomPtr != end)
        {
            if(atomPtr->inLeaf != atomNumber) neighbors[candidatesCount++] = atomPtr->inLeaf;
            ++atomPtr;
        }
    const double *center = (*(positions->GetLeaf(type1)))[atomNumber];
    PruneNeighborCandidates(center, candidatesCount, neighbors, positions->GetLeaf(type2),
        displacements, displacementsSquared, cutoffDistance2, numNeighbors);
    spacesLeftInArrays -= numNeighbors;
    return true;
}

bool CellNeighborLocator::HalfNeighbors(int atomNumber, int type1, int type2, int *neighbors, 
                int& numNeighbors, double (*displacements)[DIMENSION], double* displacementsSquared, 
                int& spacesLeftInArrays)
{
    numNeighbors = 0;
    if(cutoffDistance == 0.) return true;
    if(type2 < type1) return true;
    int candidatesCount = 0;
    // Find out the cell number of the cell that atom atomNumber is in
    int cellIndexOfThisAtom = cellIndexOfPositions[Index(type1,atomNumber)];
    if(nextLevelNeighborLocator)
    {
        int cellOccupancy, numberOfNeighborCells = 0;
        int slots = spacesLeftInArrays;
        bool append = nextLevelNeighborLocator->Neighbors(cellIndexOfThisAtom, neighbors, 
            numberOfNeighborCells, displacements, displacementsSquared, slots);
        if (!append) return false;
        // we'll need the neighbors array later
        int *neighborCells = new int[numberOfNeighborCells];
        memcpy(neighborCells, neighbors, numberOfNeighborCells*sizeof(int));
        // we have to make sure there's enough space to store all the neighbor candidates.
        for(int nbCell = 0; nbCell<numberOfNeighborCells; ++nbCell)
        {
            int thiscell = neighborCells[nbCell];
            cellOccupancy = typedCellList[thiscell][type2].second;
            if(candidatesCount + cellOccupancy > spacesLeftInArrays) return false;
            for(int nb=0; nb<cellOccupancy; ++nb)
            {
                int cand = typedCellList[thiscell][type2].first[nb].inLeaf;
                if(type1 != type2 || cand > atomNumber)
                    neighbors[candidatesCount++] = cand;
            }
        }
        delete [] neighborCells;
    }
    // then we add the atoms in the same cell
    Index *atomPtr = typedCellList[cellIndexOfThisAtom][type2].first;
    int cellOccupancy = typedCellList[cellIndexOfThisAtom][type2].second;
    Index *end = atomPtr + cellOccupancy;
    if(candidatesCount + cellOccupancy > spacesLeftInArrays) return false; //we probably won't have enough space.
    // insert half of the atoms which are located in this cell
    if(type1 != type2) while(atomPtr != end) neighbors[candidatesCount++] = (atomPtr++)->inLeaf;
    else
        while(atomPtr != end)
        {
            if(atomPtr->inLeaf > atomNumber) neighbors[candidatesCount++] = atomPtr->inLeaf;
            ++atomPtr;
        }
    const double *center = (*(positions->GetLeaf(type1)))[atomNumber];
    PruneNeighborCandidates(center, candidatesCount, neighbors, positions->GetLeaf(type2),
        displacements, displacementsSquared, cutoffDistance2, numNeighbors);
    spacesLeftInArrays -= numNeighbors;
    return true;
}

/*bool CellNeighborLocator::Neighbors(Index atomNumber, int type2, int *neighbors, int& numNeighbors, double (*displacements)[DIMENSION], double* displacementsSquared, int& spacesLeftInArrays)
{
    numNeighbors = 0;
    if(cutoffDistance == 0.) return true;
    int maxNumberOfNeighbors = 0;
    bool append = true;

    // Find out the cell number of the cell that atom atomNumber is in
    int cellIndexOfThisAtom = cellIndexOfAtom[atomNumber];
    int iTmp = spacesLeftInArrays;       // we'll overwrite the arrays, so spacesLeftInArrays shouldn't change
    int cellOccupancy, numberOfNeighborCells = 0;
    if(nextLevelNeighborLocator)
    {
        int nbCell;
        append = nextLevelNeighborLocator->Neighbors(cellIndexOfThisAtom, neighbors, 
            numberOfNeighborCells, displacements, displacementsSquared, iTmp);
        int *neighborCells = new int[numberOfNeighborCells];
        memcpy(neighborCells, neighbors, numberOfNeighborCells*sizeof(int)); // we need the neighbors array below
        iTmp = spacesLeftInArrays; // we have to make sure there's enough space to store all the neighbor candidates.
        for(nbCell = 0; append && nbCell<numberOfNeighborCells; ++nbCell)
        {
            int thiscell = neighborCells[nbCell];
            cellOccupancy = cellListWithTypesSize[thiscell][type2];
            iTmp -= cellOccupancy;
            append &= (iTmp >= 0);
            if(append) memcpy(neighbors+maxNumberOfNeighbors, cellListWithTypes[thiscell][type2], cellOccupancy*sizeof(int));
            maxNumberOfNeighbors += cellOccupancy;
        }
        delete [] neighborCells;
    }
    // then we add the atoms in the same cell
    Index *atomPtr = cellListWithTypes[cellIndexOfThisAtom][type2];
    Index *end = cellListWithTypes[cellIndexOfThisAtom][type2] + cellListWithTypesSize[cellIndexOfThisAtom][type2];
    iTmp -= cellListWithTypesSize[cellIndexOfThisAtom][type2];
    if(iTmp < 0) append = false; //we probably won't have enough space.
    else
        // insert half of the atoms which are located in this cell
        if(type1 != type2) while(atomPtr != end) neighbors[maxNumberOfNeighbors++] = atomPtr++->inLeaf;
        else
        while(atomPtr != end)
        {
            if(atomPtr->inLeaf != atomNumber) neighbors[maxNumberOfNeighbors++] = atomPtr->inLeaf;
            ++atomPtr;
        }
    if(!append) return false;
    const DMVector<double, DIMENSION> &center = (*(positions->GetLeaf(type1)))[atomNumber];
    PruneNeighborCandidates(center,maxNumberOfNeighbors,neighbors,positions,
        displacements,displacementsSquared, cutoffDistance2, numNeighbors);
    spacesLeftInArrays -= numNeighbors; 
    return true;    
}
*/
/** 
    Strategy: 
  We need to find all the cells within a distance \a D = \a cutoff + 0.5 \a diag( \a cell) + \a drift. Where \a diag(\a cell) 
  is the diagonal of the cells. Then we test all the atoms that are registered in these cells.
  
  To get the neighboring cells:

  1. Check whether \a center belongs to a nonempty cell (ie which means it is registered in nextLevelNeighborLocator).

  2. Check whether we can actually use the nextLevelNeighborLocator->Neighbor(int, ...) function. This is true if the 
  cutoff distance we pass is smaller than cutoffDistance + drift + 0.5*diag(cell).

  3. If we cannot, then simply call nextLevelNeighborLocator->Neighbor(center, ..., cutoff=cellCutoff)
  */
bool CellNeighborLocator::Neighbors(const double *center, int *neighbors, int& numNeighbors, 
                double (*displacements)[DIMENSION], double* displacementsSquared, 
                int& spacesLeftInArrays, double cutoff)
{
    numNeighbors = 0;
    if(positions->Size() == 0) return true;
    // if we didn't pass a cutoff, then it is assumed to be cutoffDistance.
    if(cutoff<0.) cutoff = cutoffDistance;
    // we look in which cell "center" stands. insideBox is false if it is out of the box.
    bool insideBox = true;
    int centerCoord[DIMENSION];
    for(int coord=0; coord<DIMENSION; ++coord)
    {
        if(center[coord] < minExtent[coord])
        {
            centerCoord[coord] = -1;
            insideBox = false;
        }
        else 
        {
            centerCoord[coord] = (int) ((center[coord]-minExtent[coord])/cellSize);
            insideBox &= (centerCoord[coord] < numberOfCells[coord] );
        }
    }
    int candidatesCount = 0;

    if(nextLevelNeighborLocator)
    {
        // if we use the neighborLocator of the cells, here is the cutoff distance we need:
        double cellCutoff = cutoff + 0.5*GetCellDiagonal() + this->maxDisplacement;
        // this is the rank of the cell in which "center" is. 
        int cellRank = (insideBox ? CellCoordinatesToRank(centerCoord) : 0);
        // decide which kind of Neighbors function we'll use for the nextlevelNBLoc
        bool useSimpleNeighborFunction = false;
        if(insideBox  && (rankToIndex[cellRank] != -1) && (cutoff <= cutoffDistance ))
            useSimpleNeighborFunction = true;

        int numberOfNeighborCells;
        bool append;
        if(useSimpleNeighborFunction)
        {
            int cellInd = rankToIndex[cellRank];
            neighbors[0] = cellInd;  // put this cell in the list of "good" cell
            displacementsSquared[0] = 0;
            memset(displacements[0], 0, DIMENSION*sizeof(double));
            
            int slots = spacesLeftInArrays - 1;
            append = nextLevelNeighborLocator->Neighbors(cellInd, neighbors+1, 
                numberOfNeighborCells, displacements+1, displacementsSquared+1, slots);
            ++numberOfNeighborCells;   // because of the first cell
        }
        else
        {
            int slots = spacesLeftInArrays;
            append = nextLevelNeighborLocator->Neighbors(center, neighbors, 
                numberOfNeighborCells, displacements, displacementsSquared, 
                slots, cellCutoff);
        }
        if (!append) return false;
        // among the returned cells, some are unuseful.
        PruneNeighborCells(neighbors, numberOfNeighborCells, displacements,
                           cutoff, useSimpleNeighborFunction);
        // Now copy the index of the neighboring cells
        vector<int> neighborCells(neighbors, neighbors+numberOfNeighborCells);
        for(int nbCell = 0; nbCell<numberOfNeighborCells; ++nbCell)
        {
            int cellOccupancy = numberOfAtomsInCell[neighborCells[nbCell]];
            if(candidatesCount + cellOccupancy > spacesLeftInArrays) return false;
            memcpy(neighbors+candidatesCount, cellListGlobal[neighborCells[nbCell]], 
                    cellOccupancy*sizeof(int));
            candidatesCount += cellOccupancy;
        }
    }
    else
    {
        int numberOfAtomsInThisCell = numberOfAtomsInCell[0]; // we have only one cell
        if(numberOfAtomsInThisCell > spacesLeftInArrays)
            return false;
        memcpy(neighbors, cellListGlobal[0], numberOfAtomsInThisCell*sizeof(int));
        candidatesCount = numberOfAtomsInThisCell;
    }

    // now Prune and keep only the neighborCandidates which are true neighbors
    PruneNeighborCandidates(center, candidatesCount, neighbors, positions,
        displacements, displacementsSquared, cutoff*cutoff, numNeighbors);
    spacesLeftInArrays -= numNeighbors;
    return true;
}


void CellNeighborLocator::MakeCellList()
{// Make cell list for current level of hierarchy. The method is directly based on J. Schiotz's method
    int cell, coord;
    unsigned int nTypes = positions->NumberOfLeaves();
// 1. Determine the edges of a rectangle enclosing all positions. 
//    Sets minExtent[DIMENSION] and maxExtent[DIMENSION]
    SetBoundingBox(positions);

// 2. Determine the number of cells in each direction by rounding down. 
    for (coord = 0; coord < DIMENSION; coord++)
    {
        double tmp = (maxExtent[coord] - minExtent[coord])/cellSize;
        ASSERT(tmp > 0.);
		// need to check here if MAXCELLS is exceeded because if positions end 
		// up with large finite values tmp could be larger than the maximum
		// allowable int, and we get a random (possibly negative) value 
		// for numberOfCells[coord]
		if(tmp > MAXCELLS)
			throw DMException("CellNeighborLocator","MakeCellList","MAXCELLS exceeded in coordinate");
        numberOfCells[coord] = (int) ceil(tmp);
    }
// 3. Compute the total number of cells
    int newTotalNumberOfCells = numberOfCells[0];
    for (coord = 1; coord < DIMENSION; coord++)
        newTotalNumberOfCells *= numberOfCells[coord];

	// still need to check the total number of cells hasn't exceeded MAXCELLS
	if(newTotalNumberOfCells > MAXCELLS)
		throw DMException("CellNeighborLocator","MakeCellList","MAXCELLS exceeded");

// 4.Reallocate if the allocated space isn't good
    if(newTotalNumberOfCells != totalNumberOfCells)
    {
        if(cellListGlobal)
        {
            delete [] numberOfAtomsInCell;
            delete [] cellListGlobal;
            delete [] cellList;
            delete [] rankToIndex;
        }
        totalNumberOfCells = newTotalNumberOfCells;
        cellListGlobal = new int*[totalNumberOfCells];
        cellList = new Index*[totalNumberOfCells];
        numberOfAtomsInCell = new int[totalNumberOfCells];
        rankToIndex = new int[totalNumberOfCells];
    }
    memset(numberOfAtomsInCell, 0, totalNumberOfCells*sizeof(int));
    

// 5. Goal of this section: determine the number of atoms in each cells, deduce how many cells are not empty, 
// and set the pointers to the beginning of the cell lists (ignoring types).
    int rankOfCell;                     // rank of cell that atom i is in
    int cellCoordInt[DIMENSION]; // coordinates of cell that atom i is in
    ASSERT(cellIndexOfPositions.GetShape() == positions->GetShape());
    DMArrayIterator<int> cellIndexIter(&cellIndexOfPositions);
    DMArrayIterator<double, DIMENSION> posIter(positions);
    for(cellIndexIter.Init(), posIter.Init(); !posIter.IsDone(); posIter.Next(), cellIndexIter.Next())
    {
        // Find out the cell coordinates of the cell that atom i is in -> cellCoordInd[DIMENSION]
        for (coord = 0; coord < DIMENSION; coord++)
        {
            cellCoordInt[coord] = 
                (int) ((posIter.Current()[coord]-minExtent[coord])/cellSize);
        }
        // Find out the cell number of the cell that atom i is in
        rankOfCell = CellCoordinatesToRank(cellCoordInt);
        ASSERT(rankOfCell < totalNumberOfCells && rankOfCell >=0);
        ++numberOfAtomsInCell[rankOfCell];  // one more atom in cell this cell
        cellIndexIter.Current() = rankOfCell; // store which cell this atom is in.
    }
    int newNumberOfNonEmptyCells = 0;
    int *nextPtr = storage;
    Index *NextPtr = Storage;
    for(cell=0; cell<totalNumberOfCells; ++cell)
        if(numberOfAtomsInCell[cell] == 0) rankToIndex[cell] = -1;
        else
        {
            rankToIndex[cell] = newNumberOfNonEmptyCells;   // here we determine the transformation from rank to index (see discussion in .h)
            cellListGlobal[newNumberOfNonEmptyCells] = nextPtr;
            cellList[newNumberOfNonEmptyCells] = NextPtr;
            numberOfAtomsInCell[newNumberOfNonEmptyCells] = numberOfAtomsInCell[cell];
            nextPtr += numberOfAtomsInCell[cell];
            NextPtr += numberOfAtomsInCell[cell];
            ++newNumberOfNonEmptyCells;
        }

    ASSERT(nextPtr-storage == positions->Size());
// from now, we know the number of non empty cells and we know the conversion from the rank to the index of the cells.

// 6. now we can delete/reallocate the cell list with types (with the correct size).
    numberOfNonEmptyCells = newNumberOfNonEmptyCells;
    if(typedCellList) delete [] typedCellList;
    typedCellList = new TypedAtomsInCell[numberOfNonEmptyCells];

// 7. Fill the cell list
    int *counter = new int[numberOfNonEmptyCells];
    memset(counter, 0, numberOfNonEmptyCells*sizeof(int));
    int globalIndex=0;
    DMArrayIterator<int> atomIter(&cellIndexOfPositions);
    LeafIndex nextType = 0;
    for(atomIter.Init(); !atomIter.IsDone(); atomIter.Next())
    {
        cell = rankToIndex[atomIter.Current()];
        atomIter.Current() = cell;  // from here cellIndexOfAtom is the really the index of the cell (until now it was the rank) 
        cellListGlobal[cell][counter[cell]++] = globalIndex++;
        const Index &ind = atomIter.GetIndex();
        while(ind.leaf >= nextType)
        {
            for(int cellCnt=0; cellCnt<numberOfNonEmptyCells; ++cellCnt)
            {
                Index *Begin = (nextType == 0 ? Begin = cellList[cellCnt] : 
                        typedCellList[cellCnt][nextType-1].first +
                            typedCellList[cellCnt][nextType-1].second);
                typedCellList[cellCnt][nextType] = make_pair(Begin, 0);
            }
            nextType++;
        }
        typedCellList[cell][ind.leaf].first[typedCellList[cell][ind.leaf].second++] = ind;
    }
    while(nextType < nTypes)
    {
        for(int cellCnt=0; cellCnt<numberOfNonEmptyCells; ++cellCnt)
            typedCellList[cellCnt][nextType] = 
                make_pair(typedCellList[cellCnt][nextType-1].first,0);
        nextType++;
    }
//  9. Checks
#ifndef NDEBUG
    for(cell=0; cell<numberOfNonEmptyCells; ++cell)
    {
        int tmp = 0;
        for(int type=0; type<nTypes; ++type)
        {
            tmp += typedCellList[cell][type].second;
        }
        ASSERT(tmp == numberOfAtomsInCell[cell]);
        ASSERT(counter[cell] = numberOfAtomsInCell[cell]);
    }
#endif // NDEBUG
    delete [] counter;

// 10. Build the next level neighbor locator
    if (nextLevelNeighborLocator) delete nextLevelNeighborLocator;

    if (numberOfNonEmptyCells == 1)
        nextLevelNeighborLocator = 0;
    else
    { 
        // In order to determine the neighboring cells, we build a listofatoms 
        // where the position of the atoms is at the center of the cells.
        cellsPositions->SetShape(numberOfNonEmptyCells);

        for(coord=0; coord<DIMENSION; ++coord) cellCoordInt[coord] = 0;
        int j=0;  // j is incremented only if cells are non empty
        for (int cell=0; cell<totalNumberOfCells; ++cell)  
        {
            if(rankToIndex[cell] != -1)
            {
                for (coord = 0; coord < DIMENSION; coord++)
                {
                    (*cellsPositions)[j][coord] = 
                        minExtent[coord] + (0.5 + cellCoordInt[coord]) * cellSize;
                }
                ++j;
            }
            ++cellCoordInt[DIMENSION-1];
            for(coord=DIMENSION-1; coord>=1; --coord)
                if(cellCoordInt[coord] == numberOfCells[coord])
                {
                    cellCoordInt[coord] = 0;
                    ++cellCoordInt[coord-1];
                }
        }
        ASSERT(j == numberOfNonEmptyCells);
        
        // create next level neighbor locator for the cells.
        // The cutoff distance is given by cutoff + 2 drift + the diagonal of the cells.
        const double nextCutDist = (1. + sqrt((double)DIMENSION)) * cellSize;
        // more efficient (in PruneNeighborCells) if we use a non-zero drift
        const double nextDrift   = 1.0e-6*nextCutDist;

        // if the number of cell is small, we'd rather use a simpleNeighborLocator
        if(cellsPositions->Size() < maxCellsForSimpleNeighborLocator) { 
			nextLevelNeighborLocator = new SimpleNeighborLocator(nextCutDist);
		}
        else if(cellsPositions->Size() < maxCellsForCellNeighborList) {
            nextLevelNeighborLocator = new CellNeighborList(nextCutDist, nextDrift);
		}
        else {
            nextLevelNeighborLocator = new CellNeighborLocator(nextCutDist, nextDrift);
		}

        nextLevelNeighborLocator->SetBoundaryConditions(bc);
        nextLevelNeighborLocator->SetPositionArray(cellsPositions);
    }
}


void CellNeighborLocator::GetCellCoordinates(int atomNumber, 
                                             int cellCoord[DIMENSION])  
{
    int coord;
    for (coord = 0; coord < DIMENSION; coord++)
    {
        cellCoord[coord] = (int)(((*oldPositions)[atomNumber][coord] - 
                                  minExtent[coord])/cellSize);
    }
}

int CellNeighborLocator::CellCoordinatesToRank(int* cellCoord) const
{
    // Example: The cell coordinates (c1,c2,c3) corresponds to the cell number
    // c1*n2*n3 + c2*n3 + c3.  This can be written as ((c1)*n2 + c2)*n3 + c3, 
    // which can be arrived at using the procedure:
    //
    // (1) N = c1;
    // (2) N = N*n2 + c2;
    // (3) N = N*n3 + c3;
    //
    // (2) and (3) can be iterated, while (1) can be done separately.
    
    int cellNumber = cellCoord[0];
    
    for (int coord = 1; coord < DIMENSION; coord++)
        cellNumber = cellNumber*numberOfCells[coord] + cellCoord[coord];
    
    return cellNumber;
}

void CellNeighborLocator::SetBoundingBox(const DMArray<double, DIMENSION> *positions)   
// We should write something more efficient
{
    int coord;
    DMArrayIterator<double, DIMENSION> posIter(positions);
    posIter.Init();
    if(!posIter.IsDone()) 
        for(coord=0; coord<DIMENSION; ++coord)
            maxExtent[coord] = minExtent[coord] = posIter.Current()[coord];
    else throw DMException("CellNeighborLocator", "SetBoundingBox", "no atom!");
    for(posIter.Next(); !posIter.IsDone(); posIter.Next())
    {
        for(coord=0; coord<DIMENSION; ++coord)
        {
            if(posIter.Current()[coord] < minExtent[coord]) minExtent[coord] = posIter.Current()[coord];
            if(posIter.Current()[coord] > maxExtent[coord]) maxExtent[coord] = posIter.Current()[coord];
        }
    }
    // to avoid having atoms right on the bounding box, we slightly expand minExtent
    // and maxExtent in each direction.
    for(coord=0; coord<DIMENSION; ++coord)
    {
        minExtent[coord] -= 1.0e-10 * cellSize;
        maxExtent[coord] += 1.0e-10 * cellSize;
    }
}


double CellNeighborLocator::MaximumDisplacementSquared
                (const DMArray<double, DIMENSION> *positions) const
{
     // diff doesn't need to be a sibling, since it's temporary
    DMArray<double, DIMENSION> *diff =
        new DMArray<double, DIMENSION> (positions->GetShape());
    diff->Copy(positions);
    diff->Axpy(oldPositions, -1.);

    bc->DifferenceBoundaryConditions(diff);

    double maxDispSq = 0.;
    DMArrayIterator<double, DIMENSION> diffIter(diff);
    for (diffIter.Init(); !diffIter.IsDone(); diffIter.Next())
    {
        double dr2 = 0.;
        const double (&curr)[DIMENSION] = diffIter.Current();
        for (int coord = 0; coord < DIMENSION; ++coord)
        {
            dr2 += curr[coord] * curr[coord];
        }
        if (dr2 > maxDispSq) maxDispSq = dr2;
    }
    delete diff;
    return maxDispSq;
}

NeighborLocator *CellNeighborLocator::Copy() const
{
    CellNeighborLocator *newNeighborLocator = 
        new CellNeighborLocator(cutoffDistance,drift);
    return newNeighborLocator;
}



void CellNeighborLocator::PruneNeighborCells(int *neighbors,
                                             int &cellsAround,
                                             double (*cellDisp)[DIMENSION],
                                             double cutoff,
                                             bool centerIsACell)
{
    return ;
    // if centerIsACell, there's no point working with a cutoff bigger than cutoffDistance 
    if(centerIsACell) ASSERT(cutoff <= this->cutoffDistance) ;
    unsigned int goodCellCnt = 0;
    double amplitude;
    if(centerIsACell) amplitude = 0.5 * cellSize + cutoff + 2. * maxDisplacement;
    else
    {
        if(2.*cutoff < cellSize) amplitude = 0.5 * cellSize + maxDisplacement;
        else amplitude = cutoff + maxDisplacement;
    }

    const double increment = 0.5 * cellSize;
    //  Check each given cell
    for(int neighborCell=0; neighborCell < cellsAround; ++neighborCell)
    {
        bool goodNeighbor = false;
        int coord;
        // first check: if each component of the displacement vector is smaller than 
        // amplitude + increment, then the cell can be accepted.
        bool ok = true;
        for(coord=0; coord<DIMENSION; ++coord)
        {
            ok &= (fabs(cellDisp[neighborCell][coord]) <= amplitude + increment);
        }
        if(ok) goodNeighbor = true;
        else
        {
            // second check: try to find a corner of the cell that is located  
            // within a square of half side amplitude 
            int delta[DIMENSION];
            for(coord=0; coord<DIMENSION; ++coord) delta[coord] = -1;
            while(delta[DIMENSION-1] < 2)
            {
                double corner[DIMENSION];
                for(coord=0; coord<DIMENSION; ++coord)
                {
                    corner[coord] = 
                        cellDisp[neighborCell][coord] + delta[coord] * increment;
                }
                bc->DifferenceBoundaryConditions(&corner,1);
                bool goodCorner = true;
                for(coord=0; coord<DIMENSION; ++coord)
                    if(fabs(corner[coord]) > amplitude)
                    {
                         goodCorner = false;
                         break;
                    };
                if(goodCorner)
                {
                    goodNeighbor = true;
                    break;
                }
                delta[0] += 2;
                for(coord=0; coord<DIMENSION-1; ++coord)
                {
                    if(delta[coord] > 1) 
                    {
                        delta[coord] = -1;
                        delta[coord+1] += 2;
                    }
                }
            }
        }
        if(goodNeighbor) neighbors[goodCellCnt++] = neighbors[neighborCell];
    }
//    if(cellsAround != goodCellCnt) cout << "Removed " << 
//        (cellsAround - goodCellCnt) << " cells over " << cellsAround << endl;
    cellsAround = goodCellCnt;

}

////////////////////////////////////////////////////////////////////////////////
/// METHODS RELATED TO SERIALIZATION
////////////////////////////////////////////////////////////////////////////////

void CellNeighborLocator::Save(DMWriter *writer) const
{
	NeighborLocator::Save(writer);
	writer->PutDouble(drift, "drift");
}

void CellNeighborLocator::Load(DMReader *reader)
{
	NeighborLocator::Load(reader);
	SetDrift(reader->GetDouble("drift"));
}
