#ifndef CELL_NEIGHBOR_LOCATOR_H
#define CELL_NEIGHBOR_LOCATOR_H
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/
#include "DMFirstHeader.h"
#include <map>
#include <vector>
#include <math.h>
#include "DigitalMaterial.h"


#include "blas.h"
#include "NeighborLocator.h"
#include "DMArray.h"
/** This neighbor locator builds and stores a cell list. It first determines 
the size of a box that encloses all the atoms, then it decomposes this box into
 identical cells of size \a minCellWidth  (given by \a cutoffDistance + 
2 \a drift). Then each cell stores the index of the atoms it contains. The next
 step is to determine which cells are the neighbor of a given cell. To do so,
 we treat the center of the cells as atoms and use an upper level 
NeighborLocator (which NeighborLocator we use is given by the number of cells.
 See comment below).

This class does not store a list of pairs of atoms. It means it is generally 
slower than a class that stores one but it may be useful if memory limitations
 do not allow to store all the pairs.

The cell list is considered up to date as long as no atom moves more than a
 distance \a drift away since the time we built the cell list.

Some details about the implementation: the cell neighbor locator is generic
 enough such that it can work for all kind of boundary conditions. The idea is
 to identify neighboring cells using another neighborlocator: we build a 
temporary list of atoms where the position of the atoms is the center of the
 cells and we call an upper level neighborlocator. The neighboring cells are
 thus obtained hierarchically. There is however is a little problem in the case
 many cells are empty: then you do not want to set a next level neighborlocator
 because it is a big overhead. So we only build the next level listofatoms with
 non empty cells. The difficulty with this approach is that the index of the 
cells is not the index of their center in the next level listofatoms: we need
 to translate between the two sets of indices. In order to distinguish them, we
 call "index" of a (non empty) cell its number on the next level LOA and 
 its "rank" the
 integer number we can deduce from its coordinates.

**/


class CellNeighborLocator : public NeighborLocator
{
public:
	CellNeighborLocator(double cutoffDistance = 1., double drift = 1.);
	~CellNeighborLocator();
#ifndef SWIG
	virtual bool HalfNeighbors(int atomNumber,
                               int *neighbors,
                               int& numNeighbors,
                               double (*displacements)[DIMENSION],
                               double *displacementsSquared,
                               int& spacesLeftInArrays);

	virtual bool Neighbors(int atomNumber,
                           int *neighbors,
                           int& numNeighbors,
                           double (*displacements)[DIMENSION],
                           double *displacementsSquared,
                           int& spacesLeftInArrays);

	virtual bool Neighbors(int atomNumber,
                           int type1,
                           int type2,
                           int* neighbors, 
                           int& numNeighbors,
                           double (*displacements)[DIMENSION],
                           double* displacementsSquared,
                           int& spacesLeftInArrays);

	virtual bool HalfNeighbors(int atomNumber,
                               int type1,
                               int type2,
                               int* neighbors,
                               int& numNeighbors,
                               double (*displacements)[DIMENSION],
                               double* displacementsSquared,
                               int& spacesLeftInArrays);

	//! returns the atoms in the ListOfAtoms whose distance to center is less than the cutoffDistance 
	virtual bool Neighbors(const double *center,
                           int *neighbors,
                           int& numNeighbors,
                           double (*displacements)[DIMENSION],
                           double *displacementsSquared,
                           int& spacesLeftInArray,
                           double cutoff=-1);
#endif //SWIG
	virtual void UpdatePosition();
	virtual void SetCutoffDistance(double cutoff);
	virtual void SetDrift(double drift);
	double GetDrift() const {return drift;}
	void SetBoundaryConditions(const BoundaryConditions *bc) ;

	double GetCellDiagonal() const {return sqrt(double(DIMENSION))*cellSize;}

	NeighborLocator *Copy() const;

#ifndef SWIG
	void Save(DMWriter *writer) const;
	void Load(DMReader *reader);
	//! Get the string name of the class for saving or loading.
	const std::string &GetType() const
	{
		return className;
	}



protected:
	virtual bool DataIsValid();
	virtual void UpdateMyData();

	double drift;
	double cellSize;
	int nAtoms;
	void Allocate(int nAtoms); 
	// The following is needed in derived classes
	DMArray<double, DIMENSION> *oldPositions;
	DMArray<double, DIMENSION> *cellsPositions;
	NeighborLocator* nextLevelNeighborLocator;
    //! Given a set of neighboring cells, remove the irrelevant ones.
/** In the hierarchy of neighbor locators, (Half)Neighbors returns the neighbors
cells based on their separation distance. However it isn't sufficient, 
because we'll consider neighbor cell that are irrelevant (for example, the 
the 2nd layer of neighbor cells are probably within the cutoffDistance
of the nextlevel neighborLocator, whereas we know that there are no neighboring
atoms in it, since the cell size is cutoffPlus2Drift).
So this routine is useful to remove some of the irrelevant neighbors cells, which
will save a lots of distance check for the atoms. 
*/
	void PruneNeighborCells(int *neighbors,
                            int &cellsAround,
                            double (*cellDisp)[DIMENSION],
                            double cutoff,
                            bool centerIsCell);
/*	class CellList
	{
	public:
		CellList();
		virtual ~CellList();
		const DMArray<double, DIMENSION> *GetCellsCenter() {return cellsCenter;}
		void Build(const DMArray<double, DIMENSION> *positions);
	private:
		std::vector<std::vector<int *> > cellList; // cellList[cell][TypeIndex] is a pointer to the list of atoms of type TypeIndex in cell "cell"
		int *storage; // is allocated to reserve the space for the cell list
		int storagelength;
		DMArray<double, DIMENSION> *cellsCenter;
		double boxSize[DIMENSION];
		double lowerCorner[DIMENSION];
		NeighborLocator* cellsNeighborLocator;
		const int maxCellsForSimpleNeighborLocator;
		const int maxCellsForCellNeighborList;
	}; */

	typedef std::map<LeafIndex, std::pair<Index *, int> > TypedAtomsInCell;
	TypedAtomsInCell *typedCellList;
//! a NumberOfNonEmptyCells x nTypes array of int
//	int ** cellListWithTypesSize;
//! a TotalNumberOfCells array of int *. Actually only the first numberOfNonEmptyCells are used.
	int **cellListGlobal;
	Index **cellList;
//! a TotalNumberOfCells array of int *. Actually only the first numberOfNonEmptyCells are used.
	int *numberOfAtomsInCell;
//! The rank of a cell can be deduced from its center. Its index is its number in the cellListAsLOA 
	int *rankToIndex;
	int totalNumberOfCells;
	int numberOfNonEmptyCells;

private:

/** In order to get the neighbor cells of a given cell, we use a neighborlocator on the cells.
In general, it's better to build a NeighborList for the cells themselves, such that calls to Neighbors(center, ...) 
can we treated efficiently. However a there are only a few cells, it's simpler to use a \SimpleNeighborLocator
and if there are really many cells, we'll use CellNeighborLocator. The criteria are resp. 
numberOfBonds < maxCellsForSimpleNeighborLocator and numberOfBonds > maxCellsForNeighborList.
**/
	const int maxCellsForSimpleNeighborLocator;
	const int maxCellsForCellNeighborList;
	void SetBoundingBox(const DMArray<double, DIMENSION> *positions);
	double minExtent[DIMENSION];
	double maxExtent[DIMENSION];

// variables related to the cell list
	void MakeCellList();
	int *storage;
	Index *Storage;

	static const std::string className;


protected:
	static const int MAXCELLS;

	//! the index is the global index at the begining of the MakeCellList function. Then it is the index wrt cellListAsLOA
	DMArray<int> cellIndexOfPositions;
private:
    //! Extract the coordinates of the cell atom \a atomNumber belongs to
    void GetCellCoordinates(int atomNumber, int cellCoord[DIMENSION]);
    //! Return the rank of a cell coordinates
	int CellCoordinatesToRank(int* cellCoord) const;
    //! The number of cells in each direction
	int numberOfCells[DIMENSION];
    //! Compute the largest displacement of the atoms since the last update
	double MaximumDisplacementSquared(const DMArray<double, DIMENSION> *positions) const;
    //! The largest displacement of the atoms since the last update 
    double maxDisplacement;
#endif //SWIG
};

#endif // CELL_NEIGHBOR_LOCATOR_H
