// NeighborList.cpp: implementation of the NeighborList class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "DMFirstHeader.h"
#include <assert.h>
#include <math.h>
#include <string>
#include <vector>

#include "DigitalMaterial.h"

#include "NeighborList.h"
#include "ListOfAtoms.h"

#include "DMLastHeader.h"

const std::string NeighborList::className = std::string("NeighborList");

NeighborList::NeighborList(double cutdist, double drift):NeighborLocator(cutdist)
{
	assert(DIMENSION == 3);

	// negligible memory expenditure to duplicate such information...

	this->cutdist = cutdist;
	this->cutlist = cutdist+2*drift;

	// Minimal cell width is the cutoff distance

	cellwidth = cutlist;

	// some initialization

	nAtoms = 0;
	oldPositions = 0;
	cell = 0;
	cellmax = 0;
	cellcnt = 0;
	ix = 0;
	nblist = 0;
	nnblo = 0;
	nnbhi = 0;
	initialUpdate = true;
	ndouble = 7;
	nfloat = ndouble * 2;

	// Global arrays for local temporary data.  A large block of memory
	// is allocated, it is then divided into parts (dTmp[n]) each holding
	// nAtoms numbers.  Since the data is contiguous it can be used for
	// arrays that are any multiple of the number of atoms, eg. if a 
	// temporary 3-vector is needed for each atom.  The space can also be
	// used for single precision numbers (fTmp[n]) or integers (iTmp[n]).

	// EMTPotential has similar temporary arrays, but in order to avoid 
	// unforseeable design problems, I will not be merging the two memory
	// spaces.  Hopefully these will be used so efficiently that the
	// overhead in memory is justifiable.

	int i;

	for (i = 0; i < MAXTMP; i++)
	{
		dTmp[i] = NULL;
	}
	for (i = 0; i < 2*MAXTMP; i++)
	{
		fTmp[i] = NULL;
		iTmp[i] = NULL;
	}
}

NeighborList::~NeighborList()
{
	if (oldPositions != NULL) delete oldPositions;
	if (cell != NULL) delete [] cell;
	if (cellmax != NULL) delete [] cellmax;
	if (cellcnt != NULL) delete [] cellcnt;
	if (ix != NULL) delete [] ix;
	if (nblist != NULL) delete [] nblist;
	if (nnblo != NULL) delete [] nnblo;
	if (nnbhi != NULL) delete [] nnbhi;
	if (dTmp[0] != NULL) delete [] dTmp[0];

	// cout << "NeighborList deleted" << endl;
}

void NeighborList::Allocate(int nAtoms, const DMArray<double, DIMENSION> *positions)
{
	this->nAtoms = nAtoms;

	// Memory allocations for oldPositions, ix, nnblo, nnbhi	
	if(oldPositions) 
		delete oldPositions;
	oldPositions = new DMArray<double, DIMENSION>(positions);
	if(ix)
		delete [] ix;
	ix = new int[nAtoms][DIMENSION];
	if(nblist)
		delete [] nblist;
	nblist = new int[nAtoms][MAXNBR];
	if(nnblo)
		delete [] nnblo;
	nnblo = new int[nAtoms];
	if(nnbhi)
		delete [] nnbhi;
	nnbhi = new int[nAtoms];

	// Memory allocations for the temporary arrays

	int i;

	// Allocate temp data as a contiguous block

	// C style memory allocation in EMTPotential.cpp changed to C++ style

	if(dTmp[0])
		delete [] dTmp[0];
	dTmp[0] = new double[ndouble * nAtoms];
	fTmp[0] = (float *) dTmp[0];
	iTmp[0] = (int *) dTmp[0];

	for (i = 1; i < ndouble; i++)
		dTmp[i] = dTmp[i-1] + nAtoms;
	for (i = ndouble; i < MAXTMP; i++)
		dTmp[i] = NULL;
	for (i = 1; i < nfloat; i++)
	{
		fTmp[i] = fTmp[i-1] + nAtoms;
		iTmp[i] = iTmp[i-1] + nAtoms;
	}
	for (i = nfloat; i < 2*MAXTMP; i++)
	{
		fTmp[i] = NULL;
		iTmp[i] = NULL;
	}

	// Build the first copy of the neighbor list.  This would use
	// the oldPositions array that has been freshly updated.
	MakeNeighborList(positions);
}

void NeighborList::UpdatePosition()
{
	int newnAtoms = positions->Size();
	if(newnAtoms == 0) return;
	if (this->nAtoms != newnAtoms)
		Allocate(newnAtoms,positions);
	// Check and see whether the criterion for rebuilding the neighbor list
	// has been satisfied, if so rebuild the neighbor list
	else CheckCriteria(positions);
}

// This function generates a fresh cell list

void NeighborList::MakeCellList(const DMArray<double, DIMENSION> *positions)
{
	int i, coord;

	// Find the max and min positions along each direction, and then
	// enlarge the extense by a scale factor, say 1.1?

	double scaleFactor = 1.00000001;
	double extension;

	SetMaximumMinimumPosition(positions);

	for (coord = 0; coord < DIMENSION; coord++)
	{
		extension = 0.5 * ( scaleFactor - 1.0 ) * (max[coord] - min[coord]);
		max[coord] += extension;
		min[coord] -= extension;
	}

   // Get number of cells (rounding down)
   // Jakob's Schiotz's code handles general parallelepipeds, but we assume
   // a rectangular box with L[i] != L[j] but box edges meet at 90 deg at
   // all vertices

   for (coord = 0; coord < DIMENSION; coord++)
   {
      ncell[coord] = (int) floor((max[coord] - min[coord])/cellwidth);
	  if(ncell[coord] == 0) ncell[coord] = 1;
      icw[coord] = (double) ncell[coord];
   }
   ntotcell = ncell[0] * ncell[1] * ncell[2];

   // Allocate/re-allocate space for the cells

   if (cell != NULL) delete [] cell;
	cell = new int*[ntotcell];
	if (cellmax != NULL) delete [] cellmax;
   cellmax = new int[ntotcell];
	if (cellcnt != NULL) delete [] cellcnt;
   cellcnt = new int[ntotcell];

	// Setup cells

	// First empty the cells

	for (i = 0; i < ntotcell; i++)
		cellcnt[i] = cellmax[i] = 0;

	int x, y, z, c;				// Cell coordinates
// Not used?	int cand;						// Candidate for inclusion on neighbor list

	// First, we count how many atoms are in each cell

	DMArrayIterator<double, DIMENSION> posIter(positions);
//	for (i = 0; i < nAtoms; i++)
	double *pos;
	for(i=0, posIter.Init(); !posIter.IsDone(); posIter.Next(), ++i)
	{
		// Which cell does the atom belong to?
		// Jakob Schiotz uses scaled coordinates to find x, y, z
		pos = posIter.Current();

		x = (int) floor(ncell[0]*(pos[0] - min[0])/(max[0] - min[0]));
		y = (int) floor(ncell[1]*(pos[1] - min[1])/(max[1] - min[1]));
		z = (int) floor(ncell[2]*(pos[2] - min[2])/(max[2] - min[2]));

		// Remainder of Jakob Schiotz's code to find which cell atom i
		// belongs to.  We probably won't need this segment, but we will leave
		// it for the time being.

		assert(x >= 0 && x < ncell[0]);
		assert(y >= 0 && y < ncell[1]);
		assert(z >= 0 && z < ncell[2]);

		// Jakob Schiotz defines the cell number of atom i to be 
		// CELLNUM(x,y,z) = ((z * ncell[1] + y) * ncell[0] + x)
		// This is the sequential access index of the cell c starting from
		// cell[0][0]

		c = ((z * ncell[1] + y) * ncell[0] + x);

		// Store which cell atom i belongs to

		ix[i][0] = x;
		ix[i][1] = y;
		ix[i][2] = z;

		// And count one more atom in that cell

		cellmax[c]++;
	}
/*	Jakob Schiotz has a small check routine here to make sure that no atoms
	are left out

	int test = 0;

	for (i = 0; i < ntotcell; i++)
		test += cellmax[i];
	assert(test == nAtoms);
*/
	// Make the cell list without wasting memory

	cell[0] = iTmp[3];			// I have no idea why 3...

	// Let each cell start right after the end of the previous one

	for (i = 1; i < ntotcell; i++)
		cell[i] = cell[i-1] + cellmax[i-1];
	assert((cell[ntotcell-1] - cell[0]) + cellmax[ntotcell-1] == nAtoms);

	// Assign atoms to cells

	for (i = 0; i < nAtoms; i++)
	{
		c = ((ix[i][2] * ncell[1] + ix[i][1]) * ncell[0] + ix[i][0]);
		cell[c][cellcnt[c]++] = i;
	}

/* Yet another check by Jakob Schiotz

	for (i = 0; i < ntotcell; i++)
		assert(cellmax[i] == cellcnt[i]);
*/

	// Now cell[n] points to cellmax[n] atoms that belong to that cell
}

//	MakeNeighborListWihtoutCellList(): DEBUG purposes only

void NeighborList::MakeNeighborListWithoutCellList()
{
	int i, j, inbr;
	double r0, r1, r2;
	double dr0, dr1, dr2;
	double sqcutdist = cutdist * cutdist;

	for (i = 0; i < nAtoms; i++)
	{
		nnblo[i] = 0;				// Reset number of neighbors for atom i to 0
		nnbhi[i] = 0;				// Reset number of neighbors for atom i to 0

		r0 = (*oldPositions)[i][0];	// Store coordinates of this atom in
		r1 = (*oldPositions)[i][1];	// scalar variables for efficiency
		r2 = (*oldPositions)[i][2];

		for (int j = 0; j < i; j++)
		{
			dr0 = r0 - (*oldPositions)[j][0];
			dr1 = r1 - (*oldPositions)[j][1];
			dr2 = r2 - (*oldPositions)[j][2];

			// Test if candidate is inside cutoff

			if (dr0*dr0 + dr1*dr1 + dr2*dr2 <= sqcutdist)
			{
				nblist[i][nnblo[i]] = j;
         		nnblo[i]++;
			}
		}
	}

	// Fold nblist[i][inbr] to get the full neighbor list

	for (i = 0; i < nAtoms; i++)
	{
		for (inbr = 0; inbr < nnblo[i]; inbr++)
		{
			j = nblist[i][inbr];
			nblist[j][nnblo[j]+nnbhi[j]] = i;
			nnbhi[j]++;
		}
	}
}

// This function makes half neighbor lists for all atoms, i.e. it only
// includes neighbors with a lower number than the atom whose list is being
// generated.  A cell algorithm is being used.  Space is divided in cells with
// a cell size no less than the cutoff distance for NB list generation.
// When the neighbor list of an atom is generated, only atoms in the same cell
// or in neighboring cells are considered.

void NeighborList::MakeNeighborList(const DMArray<double, DIMENSION> *positions)
{
	oldPositions->Copy(positions);
		
	// Generate the cell list

	MakeCellList(positions);

	// Constructing neighbor lists

	// We loop over neighboring cells in each direction, wrapping if we
	// exceed the end of the computational box.  If there is only one cell
	// in a given direction, no looping is done (lowlim = highlim = 0).  If
	// there are two cells, the other is included in the loop
	// (lowlim = -1; highlim = 0).  If there are three or more cells, one
	// cell in each direction is included ((lowlim = -1; highlim = 1).

	int xlowlim, xhighlim;	// Limits of loops over neighboring cells
	int ylowlim, yhighlim;
	int zlowlim, zhighlim;

	xlowlim = ncell[0] > 1 ? -1 : 0;
	ylowlim = ncell[1] > 1 ? -1 : 0;
	zlowlim = ncell[2] > 1 ? -1 : 0;
	xhighlim = ncell[0] > 2 ? 1 : 0;
	yhighlim = ncell[1] > 2 ? 1 : 0;
	zhighlim = ncell[2] > 2 ? 1 : 0;

	// Generate neighbor list for the atoms one by one, but consider all
	// potential neighbors at once

	int i, c, cand;
	int xx, yy, zz;				// Cell coordinates of candidate neighbor
	double r0, r1, r2;			// Coordinates
	double dr0, dr1, dr2;		// displacements
	double sqcutdist = cutlist * cutlist;
										// squared cut off with buffer zone
										// Guarantees that any atom within cutdist of
										// of specified atom will be included, but
										// a few atoms outside of cutdist will be
										// included

	// The only data to be stored from this function is nblist[i][inbr] =
	// atom index of neighbor inbr of atom i, nnblo[i] = number of neighbors
	// of atom i with index < i, nnbhi[i] = number of neighbors of atom i
	// with index > i

	DMArrayIterator<double, DIMENSION> posIter(positions);
	i = 0;
	for(posIter.Init(); !posIter.IsDone(); posIter.Next(), ++i)
//	for (i = 0; i < nAtoms; i++)
	{
		nnblo[i] = 0;			// Reset number of neigbhors for atom i to 0
		nnbhi[i] = 0;			// Reset number of neigbhors for atom i to 0

		// MakeNeighborList() is never called before CheckCriteria() is called.
	
		r0 = posIter.Current()[0];	// Store coordinates of this atom in
		r1 = posIter.Current()[1];	// scalar variables for efficiency
		r2 = posIter.Current()[2];

		// Loop over this and neighboring cells

		for (int j = xlowlim; j <= xhighlim; j++)
		{
			for (int k = ylowlim; k <= yhighlim; k++)
			{
				for (int l = zlowlim; l <= zhighlim; l++)
				{
					// Get coordinates of neighboring cells

					xx = ix[i][0] + j;
					yy = ix[i][1] + k;
					zz = ix[i][2] + l;

					// Wrap coordinates (a kind of periodic boundary condition
					while (xx < 0)
						xx += ncell[0];
					while (xx >= ncell[0])
						xx -= ncell[0];
					while (yy < 0)
						yy += ncell[1];
					while (yy >= ncell[1])
						yy -= ncell[1];
					while (zz < 0)
						zz += ncell[2];
					while (zz >= ncell[2])
						zz -= ncell[2];
					assert(xx >= 0 && xx < ncell[0] && 
							yy >= 0 && yy < ncell[1] &&
							zz >= 0 && zz < ncell[2]);

					// This is the cell number

					c = ((zz * ncell[1] + yy) * ncell[0] + xx);

					// Now add all atoms with lower "serial number" to
					// list of candidates

					for (int m = 0; m < cellcnt[c] && (cand = cell[c][m]) < i; m++)
					{
						// In the original segment of the code, the displacements
						// are stored in rnb[VECLIM][DIMENSION] until this runs out
						// of space.  Then this array is sent out of the loop over
						// cells for Jakob Schiotz's optimized vector operations to
						// compute sqdist[VECLIM], which is used to pick out the
						// correct neighbors, but afterwhich the (useful) data stored
						// here is discarded.  There does not seem to be any savings
						// in terms of the "if" statements with Jakob Schiotz's way
						// of doing things.

						dr0 = r0 - (*positions)[cand][0]; //(*oldPositions)[cand][0];
						dr1 = r1 - (*positions)[cand][1]; //(*oldPositions)[cand][1];
						dr2 = r2 - (*positions)[cand][2]; //(*oldPositions)[cand][2];

						// DEBUG: Behavior of MD not the same as when the
						// entire list of atoms is returned, but the system
						// remain controlled.

						// Test if candidate is within cutoff

						if (dr0*dr0 + dr1*dr1 + dr2*dr2 <= sqcutdist)
						{
							// Found a valid neighbor: (1) set 
							// nblist[i][current_inbr] = cand, and (2) increment
							// nnblo[i]

							nblist[i][nnblo[i]] = cand;
							nnblo[i]++;
						}

/*	DEBUG: Uncomment to return everybody

						// DEBUG: This returns every atom with j < i
						// To compare with the code without neighbor lists
						// Does not work either!  So something is wrong with
						// the cell list

						nblist[i][nnblo[i]] = cand;
						nnblo[i]++;

						// End of DEBUG
*/
					}
				}
			}
		}
	}

	// Fold nbrlist[i][inbr] to get the full neighbor list

	int j;

	for (i = 0; i < nAtoms; i++)
	{
		for (int inbr = 0; inbr < nnblo[i]; inbr++)
		{
			j = nblist[i][inbr];
			nblist[j][nnblo[j]+nnbhi[j]] = i;
			nnbhi[j]++;
		}
	}

	for(i=0; i<nAtoms; i++) if(nnblo[i] + nnbhi[i] < MAXNBR)
	{
		int INeedThisNumberOfNeighbors = nnblo[i] + nnbhi[i];
		assert(INeedThisNumberOfNeighbors <= MAXNBR);
	}
}
// This function checks the criterion for rebuilding the neighbor list
// if criterion for rebuilding neighbor list has been met, CheckCriteria will
// (1) copy data in positions to oldPositions, (2) call MakeCellList to
// generate an updated copy of the cell list information need by
// MakeNeighborList(), and (3) return true.  Otherwise it would do nothing,
// and return false.

void NeighborList::CheckCriteria(const DMArray<double, DIMENSION> *positions)
{
	double maxmoved, skin;
	
	maxmoved = max_abs_vec_diff(positions, oldPositions);

	skin = 0.25 * (cutlist - cutdist) * (cutlist - cutdist);

	// Check to see whether there is a need to rebuild neighbor list.
	// If so, call MakeNeighborList() to rebuild neighbor list, and
	// update oldPositions for next comparison

	if (maxmoved > skin)
	{
		MakeNeighborList(positions);
	}
}

// This function will return from the nblist array the list of neighbors of
//	atom atomNumber, the indices of which is to be appended to the stack
// that *neighbors points to, as well as append the corresponding list of
// (oldPositions[i][coord] - oldPositions[inbr][coord]) to the stack that
// (*displacement)[DIMENSION] points to.  The (int&) nbatch is the current
// position of the stack, i.e. the point we start appending to, and ARRAYLIM
// is the maximum size of the stack.

// An implicit assumption made here is that oldPositions is current when
// this function is called, i.e. positions is not changed between the last
// time CheckCriteria (which updates oldPositions to positions) and this
// function call.  I am wary of this, because there is not built-in control
// to prevent an user from actually messing things up.

bool NeighborList::HalfNeighbors(	int atomNumber,
												int* neighbors, 
												int& numNeighbors,
												double (*displacements)[DIMENSION], 
												double* displacementsSquared,
                     					int& spacesLeftInArrays)
{
	bool append;

	// Check to see that there is enough room in the stacks to accomodate
	// the neighbor list of atom atomNumber

	int j, inbr, coord;
	double dr[DIMENSION];
	double dr2;
	double sqcutdist = cutdist * cutdist;

	// Reset number of Half neighbors to zero

	numNeighbors = 0;

	if (nnblo[atomNumber] < spacesLeftInArrays)
	{
		// proceed to append to neighbors and displacements

		for (inbr = 0; inbr < nnblo[atomNumber]; inbr++)
		{
			// Second round of pruning, only return those neighbor atoms
			// within cutdist of atom atomNumber

			dr2 = 0.0;
			j = nblist[atomNumber][inbr];

			for (coord = 0; coord < DIMENSION; coord++)
			{
				dr[coord] =	(*positions)[j][coord] - (*positions)[atomNumber][coord];	// dr is position of new atom wrt main atom!!!
				dr2 += dr[coord] * dr[coord];
			}

			if (dr2 <= sqcutdist)
			{
				*neighbors = j;
				neighbors++;
				*displacementsSquared = dr2;
				displacementsSquared++;
			
				for (coord = 0; coord < DIMENSION; coord++)
					*(displacements[0]+coord) = dr[coord];
				displacements++;

				// increment stack index and number of half neighbors
				numNeighbors++;
				spacesLeftInArrays--;
			}
		}
		append = true;
	}
	else
		append = false;

	return append;
}

// Once the stack is filled, there is no way to tell atomNumber from
// neighbors or displacement, so it is the responsibility of the routine
// calling HalfNeighbors to keep track of this index.

// This function will return from the nblist array the full list of neighbors 
// of atom atomNumber, the indices of which is to be appended to the stack
// that *neighbors points to, as well as append the corresponding list of
// (oldPositions[i][coord] - oldPositions[inbr][coord]) to the stack that
// (*displacement)[DIMENSION] points to.  The (int&) nbatch is the current
// position of the stack, i.e. the point we start appending to, and ARRAYLIM
// is the maximum size of the stack.

bool NeighborList::Neighbors(	int atomNumber,
										int* neighbors, 
										int& numNeighbors,
										double (*displacements)[DIMENSION],
										double* displacementsSquared,
										int& spacesLeftInArrays)
{
	bool append;

	// Check to see that there is enough room in the stacks to accomodate
	// the neighbor list of atom atomNumber

	int j, inbr, coord;
	double dr[DIMENSION];
	double dr2;
	double sqcutdist = cutdist * cutdist;

	// Reset number of neighbors to zero

	numNeighbors = 0;

	if (nnblo[atomNumber] + nnbhi[atomNumber] < spacesLeftInArrays)
	{
		// proceed to append to neighbors and displacements
		for (inbr = 0; inbr < nnblo[atomNumber] + nnbhi[atomNumber]; inbr++)
		{
			dr2 = 0.0;
			j = nblist[atomNumber][inbr];
			for (coord = 0; coord < DIMENSION; coord++)
			{
				dr[coord] = (*positions)[j][coord] - (*positions)[atomNumber][coord];	// dr is position of new atom wrt main atom!!!
				dr2 += dr[coord] * dr[coord];
			}
			if (dr2 <= sqcutdist)
			{
				*neighbors = j;
				neighbors++;
				*displacementsSquared = dr2;
				displacementsSquared++;
				for (int coord = 0; coord < DIMENSION; coord++)
					*(displacements[0]+coord) = dr[coord];
				displacements++;
				// increment stack index and number of half neighbors
				numNeighbors++;
				spacesLeftInArrays--;
			}
		}
		append = true;
	}
	else
		append = false;

	return append;
}

int NeighborList::MaxNumberOfNeighbors(int atomNumber)
{
	return (nnblo[atomNumber] + nnbhi[atomNumber]);
}

int NeighborList::MaxNumberOfHalfNeighbors(int atomNumber)
{
	return (nnblo[atomNumber]);
}

double NeighborList::max_abs_vec_diff(const DMArray<double, DIMENSION> *positions, const DMArray<double, DIMENSION> *oldPositions)
{
	assert (DIMENSION == 3);

	double d00, d01, d02, d10, d11, d12;
	double sum0, sum1, maxsum;
	int limit, i;

	limit = nAtoms - 2;
	maxsum = 0.0;

	DMArrayIterator<double, DIMENSION> posIter(positions), oldPosIter(oldPositions);
	posIter.Init();
	oldPosIter.Init();
	for(i=0; i < limit; i += 2)
	{
		d00 = posIter.Current()[0] - oldPosIter.Current()[0];
		d01 = posIter.Current()[1] - oldPosIter.Current()[1];
		d02 = posIter.Current()[2] - oldPosIter.Current()[2];
		posIter.Next();
		oldPosIter.Next();
		d10 = posIter.Current()[0] - oldPosIter.Current()[0];
		d11 = posIter.Current()[1] - oldPosIter.Current()[1];
		d12 = posIter.Current()[2] - oldPosIter.Current()[2];
		posIter.Next();
		oldPosIter.Next();
		sum0 = d00 * d00 + d01 * d01 + d02 * d02;
		sum1 = d10 * d10 + d11 * d11 + d12 * d12;
		if (sum0 > maxsum)
			maxsum = sum0;
		if (sum1 > maxsum)
			maxsum = sum1;
	}
	while(!posIter.IsDone())
	{
		d00 = posIter.Current()[0] - oldPosIter.Current()[0];
		d01 = posIter.Current()[1] - oldPosIter.Current()[1];
		d02 = posIter.Current()[2] - oldPosIter.Current()[2];
		posIter.Next();
		oldPosIter.Next();
		sum0 = d00 * d00 + d01 * d01 + d02 * d02;
		if (sum0 > maxsum)
			maxsum = sum0;
	}
	return maxsum;
}

void NeighborList::SetMaximumMinimumPosition(const DMArray<double, DIMENSION> *positions)
{
	int coord;
	DMArrayIterator<double, DIMENSION> posIter(positions);
	posIter.Init();
	double *pos = posIter.Current();
	for(coord=0; coord<DIMENSION; ++coord)
	{
		max[coord] = pos[coord];
		min[coord] = pos[coord];
	}
	for(posIter.Next(); !posIter.IsDone(); posIter.Next())  // we start from the 2nd
	{
		pos = posIter.Current();
		for(coord=0; coord<DIMENSION; ++coord)
		{
			if(max[coord] < pos[coord]) max[coord] = pos[coord];
			if(min[coord] > pos[coord]) min[coord] = pos[coord];
		}
	}
}


NeighborLocator *NeighborList::Copy() const
{
  NeighborList *newNeighborLocator = new NeighborList(cutdist,cutlist);
  
  return newNeighborLocator;
}
