// EnergyObserver.cpp: implementation of the EnergyObserver class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/
#include "DigitalMaterial.h"
#include "EnergyObserver.h"
#include "Potential.h"
#include "ListOfAtoms.h"
#include "DMLastHeader.h"
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

EnergyObserver::EnergyObserver(Potential *potential, int notifyLevel) : ListOfAtomsObserver(notifyLevel)
{
	this->potential = potential;
	nAtoms = 0;
	Reset();
}

EnergyObserver::~EnergyObserver()
{

}

void EnergyObserver::Update(ListOfAtoms *atoms)
{
	if (atoms->GetNumber() != nAtoms)
	{
		nAtoms = atoms->GetNumber();
		Reset();
	}

	nSampled ++;

	potentialEnergy = potential->CalculateEnergy(atoms);
	summedPotentialEnergy += potentialEnergy;

	// XXX Should pass in a kineticEnergy class
	// which can worry about masses and center-of-mass and rotations
	// Probably the kineticEnergy class should also be responsible for computing
	// the temperature...

	kineticEnergy = atoms->GetKineticEnergy();
	
	summedKineticEnergy += kineticEnergy;
	summedSquaredKineticEnergy += kineticEnergy * kineticEnergy;

	summedSquaredEnergy += (kineticEnergy+potentialEnergy)*(kineticEnergy+potentialEnergy);
}

void EnergyObserver::SetPotential(Potential *potential)
{
	this->potential = potential;
	Reset();
}

void EnergyObserver::Reset()
{
	nSampled = 0;

	kineticEnergy = 0;
	summedKineticEnergy = 0;
	summedSquaredKineticEnergy = 0;
	summedSquaredEnergy = 0;

	potentialEnergy = 0;
	summedPotentialEnergy = 0;
}

int EnergyObserver::GetNSampled() const
{
	return nSampled;
}

void EnergyObserver::SetNSampled(int nSampled)
{
	// This is a bit weird... We'll scale the summed quantities down

	if (nSampled==0) Reset();
	else 
	{
		double ratio = double(nSampled)/double(this->nSampled);
		summedKineticEnergy *= ratio;
		summedPotentialEnergy *= ratio;
		this->nSampled = nSampled;
	}
}

double EnergyObserver::GetKineticTemperature() const 
{
	if (nSampled == 0) return 0.;
	else return summedKineticEnergy/(nSampled*nAtoms*DIMENSION*0.5);
}

double EnergyObserver::GetMeanSquaredKineticEnergyFluctuation() const 
{
	if (nSampled <=1) return 0.;
	else 
	{
		double KEBar = summedKineticEnergy/nSampled;
		return (summedSquaredKineticEnergy - KEBar*KEBar*nSampled)/(nSampled-1);
	}
}

double EnergyObserver::GetMeanSquaredEnergyFluctuation() const 
{
	if (nSampled <=1) return 0.;
	else 
	{
		double EBar = (summedKineticEnergy+summedPotentialEnergy)/nSampled;
		return (summedSquaredEnergy - EBar*EBar*nSampled)/(nSampled-1);
	}
}

// (Delta T)^2 = k_B T^2 [2/(3 N k_B) - 1/C] where N is the number of atoms and C denotes the
// specific heat of the cluster as a whole

// From O.H. Nielsen, J.P. Sethna, P. Stoltze, K.W. Jacobsen, and J.K. Norskov,
// "Melting a Copper Cluster: Critical Droplet Theory"
// Europhysics Letters 26, 51-56 (1994) 			
double EnergyObserver::GetKineticSpecificHeat() const
{

	if (nSampled <= 1) return 0.;
	else 
	{
		double deltaTSquared = GetMeanSquaredKineticEnergyFluctuation() / (nAtoms*nAtoms*DIMENSION*DIMENSION*0.25);
		double inverseTotalSpecificHeat = 1.0/(0.5*DIMENSION*nAtoms) 
			- deltaTSquared/(GetKineticTemperature()*GetKineticTemperature());
		return 1.0 / (nAtoms * inverseTotalSpecificHeat);
	}
}
