// HomogeneousDomainSubdivision.cpp: implementation of the HomogeneousDomainSubdivision class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "DMFirstHeader.h"
#include <vector>


#include "DigitalMaterial.h"

#include "HomogeneousDomainSubdivision.h"

#include "DMProtocol.h"

#include "DMException.h"
#include "DMArray.h"
#include "DMLastHeader.h"

using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

HomogeneousDomainSubdivision::HomogeneousDomainSubdivision(
                                   const int cpuLayout[DIMENSION],
                                   int myDomain,
                                   double bufferLayer)
                                  :DomainSubdivision(cpuLayout, myDomain)
{
    this->bufferLayer = bufferLayer;
    
    int myCoord[DIMENSION];
    this->DomainToCpuCoord(myDomain, myCoord);
    
    int coord;
    for(coord=0; coord<DIMENSION; ++coord) this->lowerCorner[coord] = 0;
    for(coord=0; coord<DIMENSION; ++coord) this->upperCorner[coord] = 0;
    

    TinyVector<int,DIMENSION> diffCoord(-1);
    while(diffCoord[DIMENSION-1] < 2)
    {
        bool ok = true;
        for(coord=0; coord<DIMENSION; ++coord) 
        {   // with one proc along one direction, we must be careful
            ok &= (cpuLayout[coord] > 1 || diffCoord[coord] == 0); 
        }
        if(ok)
        {
            TinyVector<int,DIMENSION> otherProcCoord(myCoord);
            otherProcCoord -= diffCoord;
            for(coord=0; coord<DIMENSION; ++coord)
            {
                if(otherProcCoord[coord] < 0)
                    otherProcCoord[coord] += cpuLayout[coord];
                else if(otherProcCoord[coord] >= cpuLayout[coord])
                    otherProcCoord[coord] -= cpuLayout[coord];
            }
            const int otherProc = CpuCoordToDomain(otherProcCoord);
            if (otherProc != myDomain)
            {
                neighborDomains.push_back(otherProc);
                differenceCpuCoordNeighborDomain.push_back(diffCoord);
            }
        }
        diffCoord[0]++;
        for(coord=0; coord<DIMENSION-1; ++coord)
        {
            if(diffCoord[coord] == 2)
            {
                diffCoord[coord] = -1;
                diffCoord[coord+1]++;
            }
        }
    }
}

HomogeneousDomainSubdivision::~HomogeneousDomainSubdivision()
{
}

void HomogeneousDomainSubdivision::Update(const DMArray<double, DIMENSION> *positions)
{
    this->SetBoundingBox(positions);
}

int HomogeneousDomainSubdivision::WhichDomain(const double point[DIMENSION]) const
{
    int cpu[DIMENSION];
    for(int coord = 0; coord<DIMENSION; ++coord)
    {
        cpu[coord] = (int)((point[coord]-lowerCorner[coord])/cpuBoxSize[coord]);
        if(cpu[coord] >= cpuLayout[coord]) cpu[coord] = cpuLayout[coord]-1;
        else if(cpu[coord] < 0) cpu[coord] = 0;
    }
    return CpuCoordToDomain(cpu);
}


bool HomogeneousDomainSubdivision::PointNearDomain(const double point[DIMENSION],
                                                   int domain) const
{
    int cpuCoord[DIMENSION];
    DomainToCpuCoord(domain, cpuCoord);
    ASSERT(CpuCoordToDomain(cpuCoord) == domain);
    bool nearDomain = true;
    for(int coord=0; coord<DIMENSION; ++coord)
    {
        double thisLowerCorner = lowerCorner[coord] + cpuCoord[coord]*cpuBoxSize[coord];
        nearDomain &= ( (thisLowerCorner < point[coord]+bufferLayer) &&
            (thisLowerCorner+cpuBoxSize[coord] > point[coord]-bufferLayer) );
    }
    return nearDomain;
}

void HomogeneousDomainSubdivision::FindPointsNearDomain(
                                                        const DMArray<double, DIMENSION> *positions, 
                                                        vector<Index> *interestingAtomsToSend) const 
{
#ifndef NDEBUG  // I'm not 100% sure the if clause is stripped out in Release mode
    if(positions->Size() != 0) ASSERT(myDomain == WhichDomain((*positions)[0]));
#endif // NDEBUG
    DMArrayIterator<double, DIMENSION> posIter(positions);
    // loop over all the positions in my domain
    for(posIter.Init(); !posIter.IsDone(); posIter.Next())
    {
        bool nearBy[DIMENSION][2];
        // check whether the current position is in the lower and/or upper buffer layer
        for(int coord=0; coord<DIMENSION; ++coord)
        {
            ASSERT(posIter.Current()[coord] >= myLowerCorner[coord]);
            ASSERT(posIter.Current()[coord] <= myUpperCorner[coord]);
            nearBy[coord][0] = (posIter.Current()[coord]-myLowerCorner[coord] < bufferLayer);
            nearBy[coord][1] = (myUpperCorner[coord]-posIter.Current()[coord] < bufferLayer);
        }
        // loop over the neighboring processors
        for(unsigned int proc=0; proc < differenceCpuCoordNeighborDomain.size(); ++proc)
        {
            bool nearDomain = true;
            for(int coord=0; coord<DIMENSION; ++coord)
                if(differenceCpuCoordNeighborDomain[proc][coord] == -1)
                    nearDomain &= nearBy[coord][1];
                else if(differenceCpuCoordNeighborDomain[proc][coord] == 1)
                    nearDomain &= nearBy[coord][0];
                else ASSERT(differenceCpuCoordNeighborDomain[proc][coord] == 0);
                
                if(nearDomain) interestingAtomsToSend[neighborDomains[proc]].push_back(posIter.GetIndex());
        }
    }
}

int HomogeneousDomainSubdivision::CpuCoordToDomain(const int cpuCoord[DIMENSION]) const
{
    int domain = cpuCoord[0];
    for(int coord=1; coord<DIMENSION; ++coord)
        domain = domain*cpuLayout[coord] + cpuCoord[coord];
    return domain;
}

void HomogeneousDomainSubdivision::DomainToCpuCoord(int domain, 
                                                    int cpuCoord[DIMENSION]) const
{
    for(int coord=DIMENSION-1; coord>=0; --coord)
    {
        cpuCoord[coord] = domain % cpuLayout[coord];
        domain /= cpuLayout[coord];
    }
}



void HomogeneousDomainSubdivision::SetBoundingBox(
                                                  const DMArray<double, DIMENSION> *positions)
{
    int coord;
    DMArrayIterator<double, DIMENSION> posIter(positions);
    posIter.Init();
    if(!posIter.IsDone()) for(coord=0; coord<DIMENSION; ++coord)
        lowerCorner[coord] = upperCorner[coord] = posIter.Current()[coord];
    // otherwise we have currently no atom on this process and we simply keep the same 
    // lower/upperCorner we had before.
    for(posIter.Next(); !posIter.IsDone(); posIter.Next())
        for(coord=0; coord<DIMENSION; ++coord)
        {
            if(posIter.Current()[coord] < lowerCorner[coord])
                lowerCorner[coord] = posIter.Current()[coord];
            if(posIter.Current()[coord] > upperCorner[coord])
                upperCorner[coord] = posIter.Current()[coord];
        }
        
        // in order to have only one communication, we condense the min and max search into one max.
        double tmp[2*DIMENSION];
        for(coord=0; coord<DIMENSION; ++coord) tmp[coord] = -lowerCorner[coord];
        for(coord=0; coord<DIMENSION; ++coord) tmp[DIMENSION+coord] = upperCorner[coord];
        
        double tmp2[2*DIMENSION];
        
        dmProtocol->AllReduce(tmp, tmp2, 2*DIMENSION, DMProtocol::MAX);
        
        for(coord=0; coord<DIMENSION; ++coord) lowerCorner[coord] = -tmp2[coord];
        for(coord=0; coord<DIMENSION; ++coord) upperCorner[coord] = tmp2[DIMENSION+coord];
        double extension[DIMENSION];
        for(coord=0; coord<DIMENSION; ++coord)
            extension[coord] = 1.0e-12*(upperCorner[coord]-lowerCorner[coord]);
        for(coord=0; coord<DIMENSION; ++coord) lowerCorner[coord] -= extension[coord];
        for(coord=0; coord<DIMENSION; ++coord) upperCorner[coord] += extension[coord];
        
        int myCoord[DIMENSION];
        DomainToCpuCoord(myDomain, myCoord);
        for(coord=0; coord<DIMENSION; ++coord)
        {
            this->cpuBoxSize[coord] = 
                (upperCorner[coord] - lowerCorner[coord])/cpuLayout[coord];
            if(this->cpuBoxSize[coord] < bufferLayer  && this->cpuLayout[coord] > 1)
            {
                char error[] = "the bufferLayer (=%f) is bigger than the size of the domain along direction %i (%f)";
                throw DMException("HomogeneousDomainSubdivision", "SetBoundingBox",
                    SprintString(error, bufferLayer, coord, this->cpuBoxSize[coord]));
            }
            this->myLowerCorner[coord] = 
                lowerCorner[coord] + myCoord[coord]*cpuBoxSize[coord];
            this->myUpperCorner[coord] = 
                this->myLowerCorner[coord] + cpuBoxSize[coord];
        }
}

