// Migrator.cpp: implementation of the Migrator class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "DMFirstHeader.h"
#include <vector>
#include <set>
#include "DigitalMaterial.h"
#include "SprintString.h"
#include "Synchronizer.h"
#include "DMArray.h"
#include "Migrator.h"
#include "DomainSubdivision.h"

#include "DMLastHeader.h"

using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Migrator::Migrator(DomainSubdivision *domain,
                   DMArray<double, DIMENSION> *positions, 
                   DMProtocol *protocol)
{
	// because of a bug in swig1.3.6, the supposedly default argument 
	// can be screwed up, and 0 be used instead of dmProtocol. (TC).
	if(0 == protocol) protocol = dmProtocol;

    this->domain    = domain;
    this->positions = positions;
    this->protocol  = protocol;
    this->nProcs    = protocol->GetNumberOfProcessors();
    this->myProc    = protocol->GetProcessorNumber();
    this->buffers   = protocol->NewBuffers();
}

Migrator::~Migrator()
{
    delete buffers;
}

void Migrator::Migrate()
{
    int proc;
    int leafInd;
    const int numberOfLeaves = this->positions->NumberOfLeaves();

#ifndef NDEBUG
    // we check that the number of siblings is the same on all 
    // processors and for every leaf. 
    int *nSiblings = new int[numberOfLeaves];
    for(leafInd=0; leafInd < numberOfLeaves; ++leafInd)
    {
        nSiblings[leafInd] = positions->GetLeaf(leafInd)->GetSiblings().size();
    }
    int *nSiblingsMax = new int[numberOfLeaves];
    protocol->AllReduce(nSiblings, nSiblingsMax, numberOfLeaves, DMProtocol::MAX);
    for(leafInd = 0; leafInd < numberOfLeaves; ++leafInd)
    {
        ASSERT(nSiblings[leafInd] == nSiblingsMax[leafInd]);
    }
    delete [] nSiblingsMax;
    delete [] nSiblings;
#endif // NDEBUG

    TRACE("Migrating");

    DMArray_AnyType::siblingContainer::const_iterator arrayIt;

    // 1. We ask the domain subdivision to update its internal variables 
    //    (like the bounding box size), given the current positions
    domain->Update(this->positions);

    // 2. 1. We operate on the arrays leaf by leaf. Remove and send the atoms that 
    //       are out of our domain.
    for(leafInd=0; leafInd < numberOfLeaves; ++leafInd)
    {
        DMArray<double, DIMENSION> *leafPos = this->positions->GetLeaf(leafInd);
        const DMArray_AnyType::siblingContainer& siblings = leafPos->GetSiblings();
    
    // 2. 2. Figure out which atoms have to be sent where
        set<Index> toRemove;
        vector<Index> *toProc = new vector<Index>[nProcs];
    
        DMArrayIterator<double, DIMENSION> posIter(leafPos);
        for(posIter.Init(); !posIter.IsDone(); posIter.Next())
        {
            const int belongsTo = domain->WhichDomain(posIter.Current());
            if( belongsTo != myProc )
            {
                toProc[belongsTo].push_back(posIter.GetIndex());
                toRemove.insert(posIter.GetIndex());
            }
        }
     // 2. 3. We tell the receiving processors how many atoms we are sending to them
        for(proc=0; proc<nProcs; ++proc)
        {
            if (myProc == proc) continue;
            buffers->GetSendBuffer(proc)->PutInt(toProc[proc].size(), "");
        }
     // 2. 4. Remove the migrating atoms from the arrays and put them into the send buffers
        for(arrayIt = siblings.begin(); arrayIt != siblings.end(); ++arrayIt)
        {
            for(int proc=0; proc<nProcs; ++proc)
            {
                if(myProc == proc) continue;
                (*arrayIt)->GetSynchronizationKit()->
                    WriteItems(toProc[proc], buffers->GetSendBuffer(proc));
            }
            (*arrayIt)->RemoveElements(toRemove);
        }
        delete [] toProc;
    }

    // 3. Communicate the buffers to everybody
    buffers->AllToAll();
    
    // 4. The reverse procedure. Again, loop over the leaves of positions:
    for(leafInd = 0; leafInd < numberOfLeaves; ++leafInd)
    {
        DMArray<double, DIMENSION> *leafPos = this->positions->GetLeaf(leafInd);
        const DMArray_AnyType::siblingContainer& siblings = leafPos->GetSiblings();
    
        const int oldSizeOfLeaf      = leafPos->Size();
        int newSizeOfLeaf            = oldSizeOfLeaf;
        vector<Index> *toAddFromProc = new vector<Index>[nProcs];
    // 4. 1. Learn how many atoms we receive from where.
        for(proc=0; proc<nProcs; ++proc)
        {
            if(myProc == proc) continue;
            const int toAdd = buffers->GetRecvBuffer(proc)->GetInt("");
            for(int i=0; i<toAdd; ++i)
                toAddFromProc[proc].push_back(Index(0,newSizeOfLeaf++));
        }
    // 4. 2. Expand all the siblings and read to data
        for(arrayIt = siblings.begin(); arrayIt != siblings.end(); ++arrayIt)
        {
            (*arrayIt)->AddElementsAtEnd(newSizeOfLeaf - oldSizeOfLeaf);
            for(proc=0; proc<nProcs; ++proc)
            {
                if(myProc == proc) continue;
                (*arrayIt)->GetSynchronizationKit()->
                    ReadItems(toAddFromProc[proc],buffers->GetRecvBuffer(proc));
            }
        }
        delete [] toAddFromProc;
    }
    // 5. Rebuild the synchronizer.
    positions->GetSynchronizationKit()->GetSynchronizer()->Rebuild(domain);
    TRACE("done migrate");
}


void Migrator::MigrateBack()
{
    int proc;
    int leafInd;
    const int numberOfLeaves = this->positions->NumberOfLeaves();

#ifndef NDEBUG
    // we check that the number of siblings is the same on all 
    // processors and for every leaf. 
    int *nSiblings = new int[numberOfLeaves];
    for(leafInd=0; leafInd < numberOfLeaves; ++leafInd)
    {
        nSiblings[leafInd] = positions->GetLeaf(leafInd)->GetSiblings().size();
    }
    int *nSiblingsMax = new int[numberOfLeaves];
    protocol->AllReduce(nSiblings, nSiblingsMax, numberOfLeaves, DMProtocol::MAX);
    for(leafInd = 0; leafInd < numberOfLeaves; ++leafInd)
    {
        ASSERT(nSiblings[leafInd] == nSiblingsMax[leafInd]);
    }
    delete [] nSiblingsMax;
    delete [] nSiblings;
#endif // NDEBUG

    TRACE("Migrating");

    DMArray_AnyType::siblingContainer::const_iterator arrayIt;

    // 1. We ask the domain subdivision to update its internal variables 
    //    (like the bounding box size), given the current positions
    domain->Update(this->positions);

    // 2. 1. We operate on the arrays leaf by leaf. Remove and send the atoms that 
    //       are out of our domain.
    for(leafInd=0; leafInd < numberOfLeaves; ++leafInd)
    {
        DMArray<double, DIMENSION> *leafPos = this->positions->GetLeaf(leafInd);
        const DMArray_AnyType::siblingContainer& siblings = leafPos->GetSiblings();
    
    // 2. 2. Figure out which atoms have to be sent where
        set<Index> toRemove;
        vector<Index> *toProc = new vector<Index>[nProcs];
    
        DMArrayIterator<double, DIMENSION> posIter(leafPos);
        for(posIter.Init(); !posIter.IsDone(); posIter.Next())
        {
            const int belongsTo = 0;//domain->WhichDomain(posIter.Current());
            if( belongsTo != myProc )
            {
                toProc[belongsTo].push_back(posIter.GetIndex());
                toRemove.insert(posIter.GetIndex());
            }
        }
     // 2. 3. We tell the receiving processors how many atoms we are sending to them
        for(proc=0; proc<nProcs; ++proc)
        {
            if (myProc == proc) continue;
            buffers->GetSendBuffer(proc)->PutInt(toProc[proc].size(), "");
        }
     // 2. 4. Remove the migrating atoms from the arrays and put them into the send buffers
        for(arrayIt = siblings.begin(); arrayIt != siblings.end(); ++arrayIt)
        {
            for(int proc=0; proc<nProcs; ++proc)
            {
                if(myProc == proc) continue;
                (*arrayIt)->GetSynchronizationKit()->
                    WriteItems(toProc[proc], buffers->GetSendBuffer(proc));
            }
            (*arrayIt)->RemoveElements(toRemove);
        }
        delete [] toProc;
    }

    // 3. Communicate the buffers to everybody
    buffers->AllToAll();
    
    // 4. The reverse procedure. Again, loop over the leaves of positions:
    for(leafInd = 0; leafInd < numberOfLeaves; ++leafInd)
    {
        DMArray<double, DIMENSION> *leafPos = this->positions->GetLeaf(leafInd);
        const DMArray_AnyType::siblingContainer& siblings = leafPos->GetSiblings();
    
        const int oldSizeOfLeaf      = leafPos->Size();
        int newSizeOfLeaf            = oldSizeOfLeaf;
        vector<Index> *toAddFromProc = new vector<Index>[nProcs];
    // 4. 1. Learn how many atoms we receive from where.
        for(proc=0; proc<nProcs; ++proc)
        {
            if(myProc == proc) continue;
            const int toAdd = buffers->GetRecvBuffer(proc)->GetInt("");
            for(int i=0; i<toAdd; ++i)
                toAddFromProc[proc].push_back(Index(0,newSizeOfLeaf++));
        }
    // 4. 2. Expand all the siblings and read to data
        for(arrayIt = siblings.begin(); arrayIt != siblings.end(); ++arrayIt)
        {
            (*arrayIt)->AddElementsAtEnd(newSizeOfLeaf - oldSizeOfLeaf);
            for(proc=0; proc<nProcs; ++proc)
            {
                if(myProc == proc) continue;
                (*arrayIt)->GetSynchronizationKit()->
                    ReadItems(toAddFromProc[proc],buffers->GetRecvBuffer(proc));
            }
        }
        delete [] toAddFromProc;
    }
    // 5. Rebuild the synchronizer.
    positions->GetSynchronizationKit()->GetSynchronizer()->Rebuild(domain);
    TRACE("done migrate");
}
