// ParallelNeighborLocator.cpp: implementation of the ParallelNeighborLocator class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/
#include "DigitalMaterial.h"

#include "ParallelNeighborLocator.h"
#include "DMException.h"

#include "DMLastHeader.h"

#ifdef DMPARALLEL

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

ParallelNeighborLocator::ParallelNeighborLocator(Migrator *m,
                                                 NeighborLocator *neighborLocator):
NeighborLocator(0.) 
{
    if (m == 0)
        throw DMException("ParallelCellNeighborList", "ParallelCellNeighborList",
						  "No migrator");
    migrator = m;
    this->neighborLocator = neighborLocator;
}

ParallelNeighborLocator::~ParallelNeighborLocator()
{
}

void ParallelNeighborLocator::SetPositionArray(const DMArray<double, DIMENSION> *positions) 
{
    this->positions = positions;
    positions->GetSynchronizationKit()->Init();	
    migrator->Migrate();                   /* Changes atoms */
    positions->GetSynchronizationKit()->Synchronize();
    neighborLocator->SetPositionArray(positions->GetSynchronizationKit()->GetExtended());
}


void ParallelNeighborLocator::UpdatePosition()
{
    if(!DataIsValid()) UpdateMyData();
    else positions->GetSynchronizationKit()->Synchronize();
}

void ParallelNeighborLocator::UpdateMyData()
{
    migrator->Migrate();                   /* Changes atoms */
    positions->GetSynchronizationKit()->Synchronize();
    // Do NOT call atoms->SetCartesianPositions() as it would cause a loop.
    neighborLocator->UpdateMyData();
}

bool ParallelNeighborLocator::DataIsValid()
{
    int needUpdateHere = (neighborLocator->DataIsValid() ? 0: 1);
    int needUpdateSomewhere;
    
    dmProtocol->AllReduce(&needUpdateHere, &needUpdateSomewhere, 1, DMProtocol::MAX);
    if (needUpdateSomewhere) return false;
    else return true;
}


bool ParallelNeighborLocator::Neighbors(int atomNumber,
                                        int * neighbors, 
                                        int& numNeighbors,
                                        double (*displacements)[DIMENSION],
                                        double* displacementsSquared,
                                        int& spacesLeftInArrays)
{
    return neighborLocator->Neighbors(atomNumber, neighbors, numNeighbors, 
        displacements, displacementsSquared, spacesLeftInArrays);
}

bool ParallelNeighborLocator::HalfNeighbors(int atomNumber,
                                            int * neighbors, 
                                            int& numNeighbors,
                                            double (*displacements)[DIMENSION],
                                            double* displacementsSquared,
                                            int& spacesLeftInArrays)
{
    return neighborLocator->HalfNeighbors(atomNumber, neighbors, numNeighbors, 
        displacements, displacementsSquared, spacesLeftInArrays);
}


/*bool ParallelNeighborLocator::HalfNeighbors(Index atomNumber,
                                            Index* neighbors, 
                                            int& numNeighbors,
                                            double (*displacements)[DIMENSION],
                                            double* displacementsSquared,
                                            int& spacesLeftInArrays)
{
    return neighborLocator->HalfNeighbors(atomNumber, neighbors, numNeighbors, 
        displacements, displacementsSquared, spacesLeftInArrays);
}
*/


bool ParallelNeighborLocator::Neighbors(const double *center,
                                        int* neighbors, 
                                        int& numNeighbors,
                                        double (*displacements)[DIMENSION],
                                        double* displacementsSquared,
                                        int& spacesLeftInArrays, double cutoff)
{
    return neighborLocator->Neighbors(center, neighbors, numNeighbors, 
        displacements, displacementsSquared, spacesLeftInArrays, cutoff);
}


bool ParallelNeighborLocator::Neighbors(int atomNumber,
                                        int type1,
                                        int type2,
                                        int* neighbors,
                                        int& numNeighbors,
                                        double (*displacements)[DIMENSION],
                                        double* displacementsSquared,
                                        int& spacesLeftInArrays) 
{
    const int nTypes = positions->NumberOfLeaves();
    ASSERT(type1 < nTypes);
    ASSERT(type2 < nTypes);
    bool append = neighborLocator->Neighbors(atomNumber, type1, type2,
        neighbors, numNeighbors, displacements, displacementsSquared,
        spacesLeftInArrays);
    if(!append) return false;
    int numGhosts;
    int ghostType2 = type2 + nTypes; // ghosts come after reals
    append = neighborLocator->Neighbors(atomNumber, type1, ghostType2,
        neighbors+numNeighbors, numGhosts, displacements+numNeighbors,
        displacementsSquared+numNeighbors, spacesLeftInArrays);
    const int shift = positions->GetShape().GetLeavesSize()[type2];
    for(int gh=0; gh<numGhosts; ++gh)
        neighbors[numNeighbors+gh] += shift;
    numNeighbors += numGhosts;
    return append;
}

bool ParallelNeighborLocator::HalfNeighbors(int atomNumber,
                                            int type1,
                                            int type2,
                                            int* neighbors,
                                            int& numNeighbors,
                                            double (*displacements)[DIMENSION],
                                            double* displacementsSquared,
                                            int& spacesLeftInArrays) 
{
    const int nTypes = positions->NumberOfLeaves();
    ASSERT(type1 < nTypes);
    ASSERT(type2 < nTypes);
    if(type2 < type1) numNeighbors = 0;
    else
    {
        bool append = neighborLocator->HalfNeighbors(atomNumber, type1, type2,
            neighbors, numNeighbors, displacements, displacementsSquared,
            spacesLeftInArrays);
        if(!append) return false;
    }
    int numGhosts;
    int ghostType2 = type2 + nTypes; // ghosts come after reals
    bool append = neighborLocator->HalfNeighbors(atomNumber, type1, ghostType2,
        neighbors+numNeighbors, numGhosts, displacements+numNeighbors,
        displacementsSquared+numNeighbors, spacesLeftInArrays);
    const int shift = positions->GetShape().GetLeavesSize()[type2];
    for(int gh=0; gh<numGhosts; ++gh)
        neighbors[numNeighbors+gh] += shift;
    numNeighbors += numGhosts;
    return append;
}

#endif // DMPARALLEL
