// ParallelNeighborLocator.h
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef PARALLELNEIGHBORLOCATOR_H
#define PARALLELNEIGHBORLOCATOR_H

#include "DigitalMaterial.h"
#include "NeighborLocator.h"
#include "DMArray.h"
#include "Parallel/Migrator.h"
#include "Parallel/Synchronizer.h"

#ifdef DMPARALLEL

class ParallelNeighborLocator : public NeighborLocator
{
public:
	ParallelNeighborLocator(Migrator *m, NeighborLocator *neighborLocator); 

	virtual ~ParallelNeighborLocator();

	void SetPositionArray(const DMArray<double, DIMENSION> *positions);

#ifndef SWIG
//! Neighbors determines the index of the points located within the cutoff distance to point # atomNumber
/*! 
	On exit (if the function returns true) neighbors contains the numNeighbors neighbors of point atomNumber, 
	there are  p
	\param atomNumber the index of the point whose neighbors are needed.
	\param neighbors on exit contains the index of the neighbors
	\param numNeighbors on exit contains the number of neighbors
	\param displacements on exit contains the vector binding point \a atomNumber and its neighbors
	\param displacementsSquared on exit contains the squared distances
	\param spacesLeftInArray on entry tells the avaliable space in the arrays. If it's too small the function return false, otherwise
	\a spacesLeftInArray is decreased by \a numNeighbors.
*/
	bool Neighbors(int atomNumber,
                   int * neighbors, 
                   int& numNeighbors,
                   double (*displacements)[DIMENSION],		// Displacement returns place of other guy wrt main guy
                   double* displacementsSquared,
                   int& spacesLeftInArrays);

//! HalfNeighbors, when looped over all atoms, is guaranteed to return each pair within the cutoff exactly once.
/*!
  This function is called HalfNeighbors because on exit neighbors contains only part of the neighbors, but if it
  is called for all the points then we are sure we get all the pairs. Warning: the neighbors
  which are returned by HalfNeighbors aren't necessarily the neighbors with smaller index than 
  \a atomNumber.
*/
	bool HalfNeighbors(int atomNumber,
                       int * neighbors, 
                       int& numNeighbors,
                       double (*displacements)[DIMENSION],	// Displacement returns place of other guy wrt main guy
                       double* displacementsSquared,
                       int& spacesLeftInArrays);

/*	bool HalfNeighbors(Index atomNumber,
							Index* neighbors, 
							int& numNeighbors,
							double (*displacements)[DIMENSION],	// Displacement returns place of other guy wrt main guy
							double* displacementsSquared,
							int& spacesLeftInArrays) ;
*/

//! Returns all atoms within the cutoff distance to center (cutoff=-1 means this->cutoffDistance)
	bool Neighbors(const double *center,
                   int* neighbors, 
                   int& numNeighbors,
                   double (*displacements)[DIMENSION],
                   double* displacementsSquared,
                   int& spacesLeftInArrays,
                   double cutoff=-1.);

//! For Lists of Atoms made of several leaves, we need to know the neighbors in other leaves.
/*! Here atomNumber refers to the index in leaf \a type1. On exit, \a neighbors returns the index of the neighbors
with respect to the leaf \a type2.  */
    bool Neighbors(int atomNumber,
                   int type1,
                   int type2,
                   int* neighbors,
                   int& numNeighbors,
                   double (*displacements)[DIMENSION],
                   double* displacementsSquared,
                   int& spacesLeftInArrays);

    bool HalfNeighbors(int atomNumber,
                       int type1,
                       int type2,
                       int* neighbors,
                       int& numNeighbors,
                       double (*displacements)[DIMENSION],
                       double* displacementsSquared,
                       int& spacesLeftInArrays);

#endif //SWIG
	void UpdatePosition();
	
	void SetBoundaryConditions(const BoundaryConditions *bc) {neighborLocator->SetBoundaryConditions(bc);}
	void SetCutoffDistance(double cutoff) {neighborLocator->SetCutoffDistance(cutoff);}
	double GetCutoffDistance() const {return neighborLocator->GetCutoffDistance();} 
	void SetDrift(double drift) { neighborLocator->SetDrift(drift);}
	double GetDrift() { return neighborLocator->GetDrift();}


	NeighborLocator *Copy() const {return new ParallelNeighborLocator(migrator, neighborLocator->Copy());}

protected:
	NeighborLocator *neighborLocator;

//! Test whether the internal data has to be updated.
	bool DataIsValid();
//! Recalculate all the data (ie. update CellList, NeighborList, etc...)
	void UpdateMyData() ;

private:
	Migrator *migrator;

};

#endif // DMPARALLEL

#endif // PARALLELNEIGHBORLOCATOR_H
