// Synchronizer.cpp: implementation of the Synchronizer class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "DMFirstHeader.h"
#include <string>

#include "Synchronizer.h"
#include "DomainSubdivision.h"
#include "NotEnoughMemory.h"
#include "DMLastHeader.h"

using namespace std;


#ifdef DMPARALLEL
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Synchronizer::Synchronizer(DMArray<double, DIMENSION> *positions, 
                           DMProtocol *protocol)
{
	// because of a bug in swig1.3.6, the supposedly default argument 
	// can be screwed up, and 0 be used instead of dmProtocol. (TC).
	if(0 == protocol) protocol = dmProtocol;
    this->protocol = protocol;
    this->buffers  = protocol->NewBuffers();
    this->nProcs   = this->protocol->GetNumberOfProcessors();
    this->myProc   = this->protocol->GetProcessorNumber();

    this->interestingAtomsToSend = new vector<Index>[nProcs];
    this->interestingAtomsToRecv = new vector<Index>[nProcs];
    this->sizeOfLeavesToSend     = new vector<int>[nProcs];
    this->sizeOfLeavesToRecv     = new vector<int>[nProcs];

    this->sizeToSend = new int[nProcs];
    this->sizeToRecv = new int[nProcs];
    this->positions  = positions;
    this->numberOfLeaves = 0;
    if(positions != 0) SetPositionsArray(positions);
}

Synchronizer::~Synchronizer()
{
    delete [] this->interestingAtomsToSend;
    delete [] this->interestingAtomsToRecv;
    delete [] this->sizeOfLeavesToSend;
    delete [] this->sizeOfLeavesToRecv;
    delete [] this->sizeToSend;
    delete [] this->sizeToRecv;
    delete buffers;
}


void Synchronizer::SetPositionsArray(DMArray<double, DIMENSION> *positions)
{
    ASSERT(positions);
    this->positions = positions;
    // for now, assume the ghosts have a size 0.
    // It'll be set in the Rebuild function, called by Migrator
    const Shape &posShape = positions->GetShape();
    vector<int> tmp(posShape.NumberOfLeaves(),0);
    this->ghostShape = Shape(posShape.GetTreeStructure(), tmp);
    const DMArray_AnyType::siblingContainer& siblings = positions->GetSiblings();
    DMArray_AnyType::siblingContainer::const_iterator arrayIt;
    for(arrayIt = siblings.begin(); arrayIt != siblings.end(); ++arrayIt)
    {
        (*arrayIt)->GetSynchronizationKit()->SetSynchronizer(this);
    }
}

void Synchronizer::Synchronize(DMArray_AnyType *array) const
{	
    //	TRACE1("Synchronizing on proc %i\n", myProc);
    
    // 1. for each processor, write the interesting entries of the array
    // for this processor (their index is in interestingAtomsToSend[proc])
    
    int proc;
    for(proc=0; proc<nProcs; ++proc)
        array->GetSynchronizationKit()->
        WriteItems(interestingAtomsToSend[proc],buffers->GetSendBuffer(proc));
    
    // 2. make the all to all communication
    buffers->AllToAll();
    // 3. make sure the ghost array have the right shape
    DMArray_AnyType *ghostArray = array->GetSynchronizationKit()->GetUntypedGhosts();
    ASSERT(0 != ghostArray);
    ghostArray->SetShape(ghostShape); // nothing will happen if the shapes are the same
    //4. copy the data from the received buffer to the ghost array
    
    for(proc=0; proc<nProcs; ++proc)
        ghostArray->GetSynchronizationKit()->
        ReadItems(interestingAtomsToRecv[proc], buffers->GetRecvBuffer(proc));
    //	TRACE("done Synchronize on proc %i\n", myProc);
}

void Synchronizer::SynchronizeAndAdd(DMArray<double, DIMENSION> *array) const
{	
    //	TRACE1("Synchronizing on proc %i\n", myProc);
    // 1. make sure the ghost array have the right shape
    DMArray_AnyType *ghostArray = array->GetSynchronizationKit()->GetUntypedGhosts();
    ASSERT(0 != ghostArray);
    ghostArray->SetShape(ghostShape); // nothing will happen if the shapes are the same

    // 2. for each processor, write the interesting entries of the ghostArray
    // for this processor (their index is in interestingAtomsToRecv[proc])
	// We have switched array for ghostArray and Send with Recv
	// because we are now copying ghost atoms and adding them to real atoms!
	// Perhaps the name of the variable should be changed.
    
    int proc;
    for(proc=0; proc<nProcs; ++proc)
        ghostArray->GetSynchronizationKit()->
        WriteItems(interestingAtomsToRecv[proc],buffers->GetSendBuffer(proc));
    
    // 2. make the all to all communication
    buffers->AllToAll();
    //4. copy the data from the received buffer to the ghost array
    
    for(proc=0; proc<nProcs; ++proc)
        array->GetSynchronizationKit()->
        AddItems(interestingAtomsToSend[proc], buffers->GetRecvBuffer(proc));
    //	TRACE("done Synchronize on proc %i\n", myProc);
}



void Synchronizer::Rebuild(const DomainSubdivision *domain)
{
    // this function should be called by the migrator.
    int proc;
    this->numberOfLeaves = positions->NumberOfLeaves();
    // 2. Clear the interesting atoms to send/recv.
    for(proc=0; proc<nProcs; ++proc) interestingAtomsToSend[proc].clear();
    for(proc=0; proc<nProcs; ++proc) interestingAtomsToRecv[proc].clear();
    
    // 3. Ask the domainSubdivision to fill the vectors with the intersting atoms to send
    domain->FindPointsNearDomain(positions, interestingAtomsToSend);
    // 4. Compute sizeOfLeavesToSend
    for(proc=0; proc<nProcs; ++proc)
    {
        sizeOfLeavesToSend[proc].clear();
        sizeOfLeavesToSend[proc].resize(numberOfLeaves, 0);
        for(vector<Index>::iterator iter=interestingAtomsToSend[proc].begin();
        iter != interestingAtomsToSend[proc].end(); ++iter)
            sizeOfLeavesToSend[proc][iter->leaf]++;
    }
    
    // 5. Compute sizeToSend
    for(proc=0; proc<nProcs; ++proc)
    {
        sizeToSend[proc] = 0;
        for(int leaf=0; leaf<numberOfLeaves; ++leaf)
            sizeToSend[proc] += sizeOfLeavesToSend[proc][leaf];
    }
    
    // 6. Communicate the size of leaves to the other processors (alltoall). ->Get the size of vectors to receive
    vector<int> dims;
    dims.push_back(numberOfLeaves);
    for(proc=0; proc<nProcs; ++proc)
//        buffers->GetSendBuffer(proc)->PutIntArray(sizeOfLeavesToSend[proc].begin(), dims, "");
        buffers->GetSendBuffer(proc)->PutIntArray(&sizeOfLeavesToSend[proc][0], dims, "");

    buffers->AllToAll();

    for(proc=0; proc<nProcs; ++proc)
    {
        int *tmp = buffers->GetRecvBuffer(proc)->GetIntArray(dims,"");
        ASSERT(dims.size() == 1);
        ASSERT(dims[0] == numberOfLeaves);
        sizeOfLeavesToRecv[proc] = vector<int>(tmp, tmp+numberOfLeaves);
        delete [] tmp;
        sizeToRecv[proc] = 0;
        for(int leaf=0; leaf<numberOfLeaves; ++leaf)
            sizeToRecv[proc] += sizeOfLeavesToRecv[proc][leaf];
    }
    
    // 7. Figure out the length of the ghost arrays
    this->sizeOfGhostLeaves.clear();
    this->sizeOfGhostLeaves.resize(numberOfLeaves, 0); // reinitialize the vector
    for(proc=0; proc<nProcs; ++proc)
    {
        for(int leaf=0; leaf<numberOfLeaves; ++leaf)
        {
            Index index(leaf,sizeOfGhostLeaves[leaf]);
            for(int i=0; i<sizeOfLeavesToRecv[proc][leaf]; ++i)
            {
                interestingAtomsToRecv[proc].push_back(index);
                index.inLeaf++;
            }
            sizeOfGhostLeaves[leaf] += sizeOfLeavesToRecv[proc][leaf];
        }
    }
    
    ghostShape = Shape(positions->GetShape().GetTreeStructure(),sizeOfGhostLeaves);
    //	WE SHOULD PROBABLY SET THE SHAPE OF THE GHOSTS OF THE SYNCHRONIZED ARRAYS
    
    const DMArray_AnyType::siblingContainer& siblings = positions->GetSiblings();
    DMArray_AnyType::siblingContainer::const_iterator arrayIt;
    for(arrayIt = siblings.begin(); arrayIt != siblings.end(); ++arrayIt)
    {
        DMArray_AnyType *ghostArray =
            (*arrayIt)->GetSynchronizationKit()->GetUntypedGhosts();
        
        if(ghostArray) ghostArray->SetShape(ghostShape);
    }


}

#endif // DMPARALLEL

