// CutLennardJonesPotential.cpp: implementation of the CutLennardJonesPotential class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/
#include "DigitalMaterial.h"
#include <iostream>
#include "CutLennardJonesPotential.h"

#include "ListOfAtoms.h"
#include "DMArray.h"
#include "IndexError.h"
#include "NotEnoughMemory.h"
#include "NeighborLocator.h"
#include "SerialBuilder.h"
#include "BoundaryConditions.h"

#include "DMLastHeader.h"

using namespace std;

#define LEN_NBR_ARRAYS 1500


static SerialBuilder<CutLennardJonesPotential> *_CLJcreator = 
    new SerialBuilder<CutLennardJonesPotential>("CutLennardJonesPotential");

const string CutLennardJonesPotential::className = string("CutLennardJonesPotential");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////


CutLennardJonesPotential::CutLennardJonesPotential() : cut1(2.41308778858241),
													   r2cut1(cut1 * cut1),
													   cut2(2.7),
													   CLJ_alpha(0.01257815434637538131),
													   CLJ_beta(-0.18714033296465953868),
													   CLJ_gamma(0.05134165513433732),
													   CLJ_delta(-0.00352137552361713868)
{
	// cut1 marks the boundary between the standard LJ part and the polynomial part
	// cut2 is the actual cutoff, greater than which separation the potential is zero
	//r2cut1 = cut1*cut1;
	SetCutoffDistance(cut2);
	nPairs = 0;

	displacementsSq = 0;
	displacements = 0;
	neighbors = 0;
	AllocateForNeighborLocator(100); // a rough estimate of the expected number of neighbors
}

CutLennardJonesPotential::~CutLennardJonesPotential()
{
	if(neighbors)
	{
		delete [] neighbors;
		delete [] displacements;
		delete [] displacementsSq;
	}
}

void CutLennardJonesPotential::AllocateForNeighborLocator(int nPairs)
{
	// nPairs is playing the role of EMTPotential's arrayLim here
	this->nPairs = nPairs;
	
	if(neighbors) 
	{	
		delete [] neighbors;
		delete [] displacements;
		delete [] displacementsSq;
	}

	if(nPairs > 0)
	{
		neighbors       = new (std::nothrow) int[nPairs];
		displacements   = new (std::nothrow) double[nPairs][DIMENSION];
		displacementsSq = new (std::nothrow) double[nPairs];
		
		ASSERT(neighbors);
		ASSERT(displacements);
		ASSERT(displacementsSq);
		
		if(0 == neighbors)
			throw NotEnoughMemory("CutLennardJonesPotential",
						"AllocateForNeighborLocator",
						SprintString("Problem allocating %d doubles", nPairs));
		if(0 == displacements)
			throw NotEnoughMemory("CutLennardJonesPotential",
						"AllocateForNeighborLocator",
						SprintString("Problem allocating %d doubles", nPairs*DIMENSION));
		if(0 == displacementsSq)
			throw NotEnoughMemory("CutLennardJonesPotential",
						"AllocateForNeighborLocator",
						SprintString("Problem allocating %d integers", nPairs));
	}
	else
	{
		neighbors       = 0;
		displacements   = 0;
		displacementsSq = 0;
	}
}


void CutLennardJonesPotential::CalculateForces(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces)
{
	int coord;
	int numberHalfNeighbors;
	forces->Zero();
	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	ASSERT(neighborLocator);
	int nAtoms = atoms->GetNumber();
	double dummyForce[DIMENSION]; // where we send the force contribution to a ghost atom


	int atomIdx1, atomIdx2;
	int fastNeighbors[LEN_NBR_ARRAYS];
	int numbersOfNeighbors[LEN_NBR_ARRAYS];
	double fastDisplacements[LEN_NBR_ARRAYS][DIMENSION];
	double fastDisplacementsSq[LEN_NBR_ARRAYS];
	double forceR[LEN_NBR_ARRAYS];
	int batchSize;
	int batchIdx;					// used for distributing results from stack
    int slotsLeft;				// used for accumulating stack with HalfNeighbors()


	for (atomIdx1 = 0; atomIdx1 < nAtoms;)
    {
        // Collect up to LEN_NBR_ARRAYS displacements and process them in a
        // vectorized manner
		
		atomIdx2 = atomIdx1;			
		slotsLeft = LEN_NBR_ARRAYS;		// Nothing in the vector yet
		
		bool accumulate = true;
		
		while (accumulate && (atomIdx1 < nAtoms) && (atomIdx1-atomIdx2 < LEN_NBR_ARRAYS))
		{
			accumulate = 	neighborLocator->
				HalfNeighbors(	atomIdx1,
							fastNeighbors+(LEN_NBR_ARRAYS-slotsLeft),
							numbersOfNeighbors[atomIdx1-atomIdx2],
							fastDisplacements+(LEN_NBR_ARRAYS-slotsLeft),
							fastDisplacementsSq+(LEN_NBR_ARRAYS-slotsLeft),
							slotsLeft);
			if (accumulate) atomIdx1++;
		}
		batchSize = LEN_NBR_ARRAYS - slotsLeft;	// This is how many slots we have used
		ASSERT(batchSize >= 0);

		for( batchIdx = 0; batchIdx < batchSize; ++batchIdx )
		{
			if (fastDisplacementsSq[batchIdx] < r2cut1) 
			{
				double r2Inv = 1.0/ fastDisplacementsSq[batchIdx];
				double r4Inv = r2Inv*r2Inv;
				double r8Inv = r4Inv*r4Inv;
				double r14Inv = r8Inv*r4Inv*r2Inv;
				forceR[batchIdx] = 24.0*(2.0*r14Inv - r8Inv);
			}
			else // if (displacementsSq[batchIdx] < r2cut2)
				forceR[batchIdx] = (- CLJ_gamma*2. + - 4.*CLJ_delta*fastDisplacementsSq[batchIdx]);
			
		}


		int numAtomsDone = atomIdx1 - atomIdx2;
        batchIdx = 0;
 
		double *forceI, *forceJ;
		for ( ; atomIdx2 < atomIdx1; atomIdx2++)
        {
			forceI = (*forces)[atomIdx2];
			numberHalfNeighbors = numbersOfNeighbors[numAtomsDone-atomIdx1+atomIdx2];
			for (int jx = 0; jx < numberHalfNeighbors; ++jx)
			{
				int j = fastNeighbors[batchIdx];

				if( j < nAtoms)
					forceJ = (*forces)[j];
				else
					forceJ = dummyForce;

				for (coord=0; coord<DIMENSION; coord++)
				{
					forceI[coord] += -fastDisplacements[batchIdx][coord]*forceR[batchIdx];
					forceJ[coord] += +fastDisplacements[batchIdx][coord]*forceR[batchIdx];
				}
				batchIdx++;
			}
			
        }
        ASSERT(atomIdx1 == atomIdx2);
	} // close atomIdx1 loop


/*



	for(i=0;i<nAtoms;i++)
    {	
		int spacesLeftInArrays = nPairs;
		bool arrayNotTooFull = neighborLocator->HalfNeighbors(i,neighbors,
			numberHalfNeighbors, displacements,displacementsSq,spacesLeftInArrays);

		while(!arrayNotTooFull)
		{
			AllocateForNeighborLocator(2*nPairs);
			spacesLeftInArrays = nPairs; // nPairs has been doubled
			arrayNotTooFull = neighborLocator->HalfNeighbors(i,neighbors,
				numberHalfNeighbors, displacements,displacementsSq,spacesLeftInArrays);
		}
		forceI = (*forces)[i];
		double forceR;
		for(jx=0;jx<numberHalfNeighbors;jx++)
		{
			j=neighbors[jx];
			if(j<nAtoms) forceJ = (*forces)[j];
			else forceJ = dummyForce;
			if (displacementsSq[jx] < r2cut1) 
			{
				double r2Inv = 1.0/ displacementsSq[jx];
				double r4Inv = r2Inv*r2Inv;
				double r8Inv = r4Inv*r4Inv;
				double r14Inv = r8Inv*r4Inv*r2Inv;
				forceR = 24.0*(2.0*r14Inv - r8Inv);
			}
			else // if (displacementsSq[jx] < r2cut2)
				forceR = (- CLJ_gamma*2. + - 4.*CLJ_delta*displacementsSq[jx]);
			
			// these are the only two cases since we've already checked that we're inside the outer cutoff
			// Note that the displacements given by NeighborsHalf are of particles j wrt particles i
			for (coord=0; coord<DIMENSION; coord++)
			{
				forceI[coord] += -displacements[jx][coord]*forceR;
				forceJ[coord] += +displacements[jx][coord]*forceR;
			}
			
		}	// Finished with atom j, given by index jx
	} // end of i loop
*/
}

/**

   Calculates the force on an atom at the origin due to an atom at 
   displacement. Boundary conditions are assumed to have already been applied,
   and displacementSq is assumed to be the square of the displacement, and 
   within the bounds of the cutoff distance. Does not zero the array force.

 */

double CutLennardJonesPotential::CalculatePairForceEnergy(double displacement[DIMENSION], double displacementSq, double *force)
{
	ASSERT(displacementSq < cut2 * cut2);
	double pairEnergy = 0;

	double forceR = 0.;
	if (displacementSq < r2cut1) 
	{
		double r2Inv = 1.0/ displacementSq;
		double r6Inv = r2Inv * r2Inv * r2Inv;
		double r8Inv = r6Inv * r2Inv;
		
		double r12Inv = r6Inv*r6Inv;
		pairEnergy = 4*(r12Inv - r6Inv) + CLJ_alpha;
		
		double r14Inv = r12Inv*r2Inv;
		forceR = 24.0*(2.0*r14Inv - r8Inv);// note factor of 4 replaced
	}
	else
	{
		forceR = (- CLJ_gamma*2. + - 4.*CLJ_delta*displacementSq);
		pairEnergy = (CLJ_beta + CLJ_gamma*displacementSq + CLJ_delta*displacementSq*displacementSq);
	}

	for ( int coord = 0; coord < DIMENSION; ++coord )
		force[coord] += -displacement[coord]*forceR;
	
	return pairEnergy;
}


void CutLennardJonesPotential::CalculateForcesBetweenTypes(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces, int type1, int type2) 
// the force array isn't initialized in this function
{
	int i,j,coord,jx;
	int numNeighbors;

	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	ASSERT(neighborLocator);
	int spacesLeftInArrays;
	double *forceI, *forceJ;
	int nAtoms1 = atoms->GetLeaf(type1)->GetNumber();
	for(i=0;i<nAtoms1;i++)
    {	
		spacesLeftInArrays = nPairs;
		bool arrayNotTooFull = neighborLocator->Neighbors(i, type1, type2,
			neighbors,numNeighbors, displacements,displacementsSq,spacesLeftInArrays);
		while(!arrayNotTooFull)
		{
			AllocateForNeighborLocator(2*nPairs);
			spacesLeftInArrays = nPairs; // nPairs has been doubled
			arrayNotTooFull = neighborLocator->Neighbors(i, type1, type2,
				neighbors,numNeighbors, displacements,displacementsSq,spacesLeftInArrays);
		}

		forceI = (*(forces->GetLeaf(type1)))[i];
		for(jx=0;jx<numNeighbors;jx++)
		{
			j=neighbors[jx];
			forceJ = (*forces->GetLeaf(type2))[j];
			double forceR;
			
			if (displacementsSq[jx] < r2cut1) 
			{
				double r2Inv = 1.0/ displacementsSq[jx];
				double r4Inv = r2Inv*r2Inv;
				double r8Inv = r4Inv*r4Inv;
				double r14Inv = r8Inv*r4Inv*r2Inv;
			    forceR = 24.0*(2.0*r14Inv - r8Inv);// note factor of 4 replaced
			}
			else // if (displacementsSq[jx] < r2cut2)
				forceR = (- CLJ_gamma*2. + - 4.*CLJ_delta*displacementsSq[jx]);
			
			// these are the only two cases since we've already checked that we're inside the outer cutoff
			// Note that the displacements given by NeighborsHalf are of particles j wrt particles i
			for (coord=0; coord<DIMENSION; coord++)
			{
				forceI[coord] += -displacements[jx][coord]*forceR;
				forceJ[coord] += +displacements[jx][coord]*forceR;
			}
		}	// Finished with atom j, given by index jx
	} // end of i loop
}


double CutLennardJonesPotential::CalculateEnergy(ListOfAtoms *atoms)
{
	double potential = 0.0;
	int numberHalfNeighbors;
	int i,jx;
	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	ASSERT(neighborLocator);
	int nAtoms = atoms->GetNumber();
	for(i=0;i<nAtoms;i++)
    {
		int spacesLeftInArrays = nPairs;
		bool arrayNotTooFull = neighborLocator->HalfNeighbors(i,neighbors,
			numberHalfNeighbors, displacements,displacementsSq,spacesLeftInArrays);
		while(!arrayNotTooFull)
		{
			AllocateForNeighborLocator(2*nPairs);
			spacesLeftInArrays = nPairs; // nPairs has been doubled
			arrayNotTooFull = neighborLocator->HalfNeighbors(i,neighbors,
				numberHalfNeighbors, displacements,displacementsSq,spacesLeftInArrays);
		}
		for(jx=0;jx<numberHalfNeighbors;jx++)
		{
			double pairEnergy;
			
			if(displacementsSq[jx] < r2cut1)
			{
				double r2inv = 1.0/displacementsSq[jx];
				double r4inv = r2inv*r2inv;
				double r6inv = r4inv*r2inv;
				double r12inv = r6inv*r6inv;
				pairEnergy = 4*(r12inv - r6inv) + CLJ_alpha;
			}
			else
			{
			  pairEnergy = (CLJ_beta + CLJ_gamma*displacementsSq[jx] + CLJ_delta*displacementsSq[jx]*displacementsSq[jx]);
			}
			if(neighbors[jx] >= nAtoms) // means this pair is between a real and a ghost atom
				pairEnergy /= 2;
			potential += pairEnergy;
		}	// Finished with atom j, given by index jx
    }
	return potential;
}

double CutLennardJonesPotential::CalculateEnergyAndForces(ListOfAtoms *atoms,DMArray<double, DIMENSION> *forces)
{
	double potential = 0.0;
	int i,j,jx,coord;
	int numberHalfNeighbors;

	// zero the forces:
	forces->Zero();

	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	ASSERT(neighborLocator);

	int nAtoms = atoms->GetNumber();
	for(i=0;i<nAtoms;i++)
	{
		int spacesLeftInArrays = nPairs;
		bool arrayNotTooFull = neighborLocator->HalfNeighbors(i,neighbors,
			numberHalfNeighbors, displacements,displacementsSq,spacesLeftInArrays);
		while(!arrayNotTooFull)
		{
			AllocateForNeighborLocator(2*nPairs);
			spacesLeftInArrays = nPairs; // nPairs has been doubled
			arrayNotTooFull = neighborLocator->HalfNeighbors(i,neighbors,
				numberHalfNeighbors, displacements,displacementsSq,spacesLeftInArrays);
		}    
		for(jx=0;jx<numberHalfNeighbors;jx++)
		{
			j=neighbors[jx];
			double forceR;

			if (displacementsSq[jx] < r2cut1) 
			{
				double r2inv = 1.0/ displacementsSq[jx];
				double r6inv = r2inv*r2inv*r2inv;
				double r8inv = r2inv*r6inv;
				double r12inv = r6inv*r6inv;
				double r14inv = r6inv*r6inv*r2inv;		
				forceR = 24.0*(2.0*r14inv - r8inv); // note factor of 4 replaced
				potential += 4*(r12inv - r6inv) + CLJ_alpha;
			}
			else
			{
				forceR = (- CLJ_gamma*2. + - 4.*CLJ_delta*displacementsSq[jx]);
				potential += (CLJ_beta + CLJ_gamma*displacementsSq[jx]
								+ CLJ_delta*displacementsSq[jx]*displacementsSq[jx]);
			}
			// these are the only two cases since we've already checked that we're inside the outer cutoff
			
			// Note that the displacements given by NeighborsHalf are of particles j wrt particles i
			// (position i is center of universe, so displacements are rj - ri)
			for (coord=0; coord<DIMENSION; coord++)
			{
				(*forces)[i][coord] += -displacements[jx][coord]*forceR;
				(*forces)[j][coord] += +displacements[jx][coord]*forceR;
			}
			
			
		}	// Finished with atom j, given by index jx
    } // end of i loop
	return potential;
}

double CutLennardJonesPotential::CalculateAtomicEnergy(ListOfAtoms *atoms, int i)
{  // here the energy of an atom is half the energy in each of its bonds.
    if(i >= atoms->GetNumber() || i<0)
		throw IndexError("CutLennardJonesPotential","CalculateAtomicEnergy",
						 "No such atom.");

	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	int numberOfNeighbors;
	int spacesLeftInArrays = nPairs;

	bool arrayNotTooFull = neighborLocator->Neighbors(i,neighbors,
		numberOfNeighbors, displacements,displacementsSq,spacesLeftInArrays);
	while(!arrayNotTooFull)
	{
		AllocateForNeighborLocator(2*nPairs);
		spacesLeftInArrays = nPairs; // nPairs has been doubled
		arrayNotTooFull = neighborLocator->Neighbors(i,neighbors,
			numberOfNeighbors, displacements,displacementsSq,spacesLeftInArrays);
	}
    
	double dummy[DIMENSION];
	double potential = 0.0;
	for(int jx=0;jx<numberOfNeighbors;jx++)
		potential += CalculatePairForceEnergy(displacements[jx], displacementsSq[jx], dummy);
	
	return potential/2.;   // since we have split energy of bonds among the atoms.
}

double CutLennardJonesPotential::AtomicForcesEnergy(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces, int i)
{
	ASSERT(atoms && forces);

	int coord,jx,j;
	int numberOfNeighbors,nAtoms = atoms->GetNumber();
	int spacesLeftInArrays = nPairs;
	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	ASSERT(neighborLocator);
	bool arrayNotTooFull = neighborLocator->Neighbors(i,neighbors, numberOfNeighbors, displacements,displacementsSq,spacesLeftInArrays);
	while(!arrayNotTooFull)
	{
		AllocateForNeighborLocator(2*nPairs);
		spacesLeftInArrays = nPairs; // nPairs has been doubled
		arrayNotTooFull = neighborLocator->Neighbors(i,neighbors, numberOfNeighbors, displacements,displacementsSq,spacesLeftInArrays);
	}    
	
	double potentialEnergy = 0.;
	double tempForce[DIMENSION]; // where we store the force returned by CalculatePairForce temporarily
	double dummyForce[DIMENSION]; // for ghost atoms
	double *forceI, *forceJ;
	

	if(i<nAtoms)
		forceI = (*forces)[i];
	else forceI = dummyForce;

	for(jx=0;jx<numberOfNeighbors;++jx)
	{
		for ( coord = 0; coord < DIMENSION; ++coord )
			tempForce[coord] = 0.;
		
		potentialEnergy += CalculatePairForceEnergy(displacements[jx],displacementsSq[jx],tempForce);
		j = neighbors[jx];
		if(j < nAtoms)
			forceJ = (*forces)[j];
		else forceJ = dummyForce;
		
		for ( coord = 0; coord < DIMENSION; ++coord )
		{
			forceI[coord] += 0.5 * tempForce[coord];
			forceJ[coord] -= 0.5 * tempForce[coord];
		}
	}
	return 0.5 * potentialEnergy;
}


void CutLennardJonesPotential::CalculateStressTensor(ListOfAtoms *atoms, int i, double stress[DIMENSION*(DIMENSION+1)/2])
{
	for( int p = 0; p < (DIMENSION+1)*DIMENSION/2; ++p )
		stress[p] = 0.;

	StressTensorKernel(atoms, i, stress);

	// Add the dynamic part and scale with the atomic volume
	// Finally add the dynamic part and scale with the atomic volume
	double vol = 1.,lengthScale = GetLengthScale(); 
	if(DIMENSION == 2)
		vol = 0.5*sqrt(3.) * lengthScale * lengthScale;
	else if(DIMENSION == 3)
		vol = lengthScale*lengthScale*lengthScale/sqrt(2.);
	else 
		for(int coord=0;coord<DIMENSION;++coord)
			vol *= lengthScale;

	double invvol = 1./vol;
	double mass = atoms->GetMass();
	const double *veloCurrent = (*(atoms->GetCartesianVelocities()))[i] ;
	for(int alpha=0; alpha<DIMENSION; ++alpha)
		for(int beta=alpha; beta<DIMENSION; ++beta)
		{
			int ii = stresscomp[alpha][beta];
			stress[ii] -= veloCurrent[alpha]*veloCurrent[beta]*mass;
			stress[ii] *= invvol;
		}
}

void CutLennardJonesPotential::CalculateStressTensor(ListOfAtoms *atoms, DMArray<double, DIMENSION*(DIMENSION+1)/2> *stress)
{
	int i, nAtoms = atoms->GetNumber();
	stress->Zero();
	
	// the main work is done by StressTensorKernel which deals with one atom at a time
	DMArrayIterator<double,(DIMENSION+1)*DIMENSION/2> stressIter(stress); 
	for( i = 0, stressIter.Init(); i < nAtoms; ++i, stressIter.Next() )
	{
		double *current = stressIter.Current();
		StressTensorKernel(atoms,i, current);
	}


	// Add the dynamic part and scale with the atomic volume
	double vol = 1.,lengthScale = GetLengthScale(); 
	if(DIMENSION == 2)
		vol = 0.5*sqrt(3.) * lengthScale * lengthScale;
	else if(DIMENSION == 3)
		vol = lengthScale*lengthScale*lengthScale/sqrt(2.);
	else 
		for(int coord=0;coord<DIMENSION;++coord)
			vol *= lengthScale;

	double invvol = 1./vol;
	double mass = atoms->GetMass();
	
	DMArrayIterator<double, DIMENSION> velIter(atoms->GetCartesianVelocities());
	for( velIter.Init(), stressIter.Init(); !velIter.IsDone(); velIter.Next(), stressIter.Next() )
	{
		const double *veloCurrent = velIter.Current();
		double *stressCurrent = stressIter.Current();
		for(int alpha=0; alpha<DIMENSION; ++alpha)
			for(int beta=alpha; beta<DIMENSION; ++beta)
			{
				int ii = stresscomp[alpha][beta];
				stressCurrent[ii] -= veloCurrent[alpha]*veloCurrent[beta]*mass;
				stressCurrent[ii] *= invvol;
			}
		
	}
}



void CutLennardJonesPotential::StressTensorKernel(ListOfAtoms *atoms, int i, double stress[(DIMENSION+1)*DIMENSION/2])
{
	int j,jx,ii,alpha,beta;
	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	int spacesLeftInArrays = nPairs, numberOfNeighbors,nAtoms = atoms->GetNumber();
	bool arrayNotTooFull = neighborLocator->Neighbors(i,neighbors, numberOfNeighbors, displacements,displacementsSq,spacesLeftInArrays);

	while(!arrayNotTooFull)
	{
		AllocateForNeighborLocator(2*nPairs);
		spacesLeftInArrays = nPairs; // nPairs has been doubled
		arrayNotTooFull = neighborLocator->Neighbors(i,neighbors, numberOfNeighbors, displacements,displacementsSq,spacesLeftInArrays);
	}
	ASSERT(arrayNotTooFull);

	for(jx=0;jx<numberOfNeighbors;jx++)
	{
		j=neighbors[jx];
		double forceR;
		if (displacementsSq[jx] < r2cut1) 
		{
			double r2Inv = 1.0/ displacementsSq[jx];
			double r4Inv = r2Inv*r2Inv;
			double r8Inv = r4Inv*r4Inv;
			double r14Inv = r8Inv*r4Inv*r2Inv;
			forceR = 24.0*(2.0*r14Inv - r8Inv);
		}
		else // if (displacementsSq[jx] < r2cut2)
			forceR = (- CLJ_gamma*2. + - 4.*CLJ_delta*displacementsSq[jx]);
		
		for(alpha=0; alpha<DIMENSION; ++alpha)
			for(beta=alpha; beta<DIMENSION; ++beta)
			{
				ii = stresscomp[alpha][beta];
				stress[ii] += - 0.5 * forceR*displacements[jx][alpha]*displacements[jx][beta];
			}
	}

}


////////////////////////////////////////////////////////////////////////////////
/// METHODS RELATED TO SERIALIZATION
////////////////////////////////////////////////////////////////////////////////

void CutLennardJonesPotential::Save(DMWriter *writer) const
{
	Potential::Save(writer);

	writer->PutDouble(cut1, "cut1");
	writer->PutDouble(cut2, "cut2");
	writer->PutDouble(r2cut1, "r2cut1");
	writer->PutDouble(CLJ_alpha, "CLJ_alpha");
	writer->PutDouble(CLJ_beta, "CLJ_beta");
	writer->PutDouble(CLJ_gamma, "CLJ_gamma");
	writer->PutDouble(CLJ_delta, "CLJ_delta");
	
}

void CutLennardJonesPotential::Load(DMReader *reader)
{
	Potential::Load(reader);
	// We don't change these parameters. We're only saving them in case 
	// somebody does in the future, and in the future after that wants to 
	// re examine old data in which case. The function CheckPotentialParameter
	// throws an error if the double retrieved via the reader using the string
	// passed is not equal to the double passed

	CheckPotentialParameter(reader, cut1, "cut1");
	CheckPotentialParameter(reader, cut2, "cut2");
	CheckPotentialParameter(reader, r2cut1, "r2cut1");
	CheckPotentialParameter(reader, CLJ_alpha, "CLJ_alpha");
	CheckPotentialParameter(reader, CLJ_beta, "CLJ_beta");
	CheckPotentialParameter(reader, CLJ_gamma, "CLJ_gamma");
	CheckPotentialParameter(reader, CLJ_delta, "CLJ_delta");
}
