// EMT.cpp  --  Implements the EMT potential for multiple elements.
// When done, this will make EMTPotential.{h,cpp} obsolete.
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/
#include "EMT.h"
#include "EMTDefaultParameterProvider.h"
#include "ListOfAtoms.h"
#include "LeafToLeafIterator.h"
#include "NeighborLocator.h"
#include "vectools.h"
#include "mass.h"

#if DIMENSION == 3  // Only works in 3D

using namespace std;

const string EMT::className = string("EMT");

const int EMT::BUFLEN = 10000;  // The batch--buffer size.  Avoid powers of 2

EMT::EMT(ListOfAtoms *listofatoms, EMTParameterProvider *prov)
{
	atoms = listofatoms;
	if (prov == 0) {
		ownProvider = 1;
		provider = new EMTDefaultParameterProvider();
	}
	else {
		ownProvider = 0;
		provider = prov;
	}
	InitParameters();

	/* Initialize the internal variables. */
	Allocate(true);
}

EMT::~EMT()
{
	if (ownProvider) delete provider;
	DeAllocate();
}

void EMT::InitParameters()
{
	// Extract the elements from the list of atoms, and set up an array
	// of EMT parameters.
	vector<string> elements;

	// Extract the element names from the list of atoms.
	for (LeafToLeafIterator<ListOfAtoms> iter(atoms); !iter.Done(); iter.Next())
		elements.push_back(iter.Current()->GetTypeName());
	nleaves = elements.size();
	
	// Get the EMT parameters
	parameters.clear();
	for (int i = 0; i < nleaves; i++)
		parameters.push_back(provider->GetParameters(&(elements[i])));

	nelements = provider->GetNumberOfElements();
				
	// Calculate the quantities that can only be calculated, once the
	// elements in the simulations are known.
	provider->CalcGammaEtc();
	cutoffdistance = provider->GetCutoffDistance();
	cutoffslope = provider->GetCutoffSlope();
	chi = provider->GetChi();
}

void EMT::DeAllocate() 
{
	for (unsigned int i = 0; i < sigma1.size(); i++) {
		delete sigma1[i];
		delete sigma2[i];
		sigma1[i] = 0;
		sigma2[i] = 0;
	}
	delete Ec;
	delete Eas;
	delete radius;
	delete dEds;
}
	
void EMT::Allocate(bool first)
{
	/* Create all variables as siblings to the positions */
	const DMArray<double, DIMENSION> *pos = atoms->GetCartesianPositions();
	
	if (!first)
		DeAllocate();
	/* Allocate a sigma1 and a sigma2 for all combinations of elements */

	if (sigma1.size() != nelements)
		sigma1.resize(nelements);
	if (sigma2.size() != nelements)
		sigma2.resize(nelements);
	int i;
	for (i = 0; i < nelements; i++) {
		sigma1[i] = new DMArray<double>(pos);
		sigma2[i] = new DMArray<double>(pos);
	}
	/* The remaining data are simple vectors */
	Ec = new DMArray<double>(pos);
	Eas = new DMArray<double>(pos);
	radius = new DMArray<double>(pos);
	dEds = new DMArray<double>(pos);
	
	// Some arrays do not have to be reallocated if the number of atoms changes, only if the number
	// of leave change (which is not allowed).
//	if (first) {
//		me.Allocate(nleaves, nleaves);
//		other.Allocate(nleaves, nleaves);
//		rnb.Allocate(nleaves, nleaves);
//		dist2.Allocate(nleaves, nleaves);
//	}

	// Count the length of the longest leaf
	maxleaflen = 0;
	for (LeafToLeafIterator<DMArray<double> > iter(Ec); !iter.Done(); iter.Next()) {
		int llen = iter.Current()->Size();
		if (llen > maxleaflen)
			maxleaflen = llen;
	}
}

DMArray<double, DIMENSION> *EMT::AllocateForces() 
{
	return new DMArray<double, DIMENSION>(atoms->GetCartesianPositions());
}

void EMT::CheckListOfAtoms(ListOfAtoms *atoms)
{
	if (atoms != this->atoms)
		throw DMException("EMT", "CalculateSigmas", "Cannot use a new list of atoms");
}

double EMT::CalculateEnergy(ListOfAtoms *atoms)
{
	CheckListOfAtoms(atoms);
	CalculateSigmas();
	return CalculateEnergyAfterSigmas();
}

void EMT::CalculateForces(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces)
{
	// Most of the work in CalculateEnergyAfterSigmas() needs to be done anyway.
	CalculateEnergyAndForces(atoms, forces);
}

double EMT::CalculateEnergyAndForces(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces)
{
	double e;
	CheckListOfAtoms(atoms);
	CalculateSigmas();
	e = CalculateEnergyAfterSigmas();
	CalculateForcesAfterEnergy(forces);
	return e;
}

void EMT::CalculateSigmas()
{
	// TinyMatrix<int> nbatch(nleaves, nleaves);
	int me[BUFLEN];
	int other[BUFLEN];
	double rnb[BUFLEN][3];
	double dist2[BUFLEN];
	NeighborLocator *nblocator = atoms->GetNeighborLocator();
	
	// Clear the sigma arrays.
	for (int i = 0; i < nelements; i++) {
		sigma1[i]->Zero();
		sigma2[i]->Zero();
	}

#if 0
	cerr << "BEGIN TEST" << endl;
	DMArray<double> *s = sigma1[0];
	cerr << "GetLeaf loop" << endl;
	for (int i = 0; i < s->NumberOfLeaves(); i++)
		cerr << s->GetLeaf(i)->Size() << endl;
	cerr << "Leaf to leaf iterator" << endl;
	LeafToLeafIterator<DMArray<double> > si(s);
	for (; !si.Done(); si.Next())
		cerr << si.Current()->Size() << endl;
	cerr << "END TEST" << endl;
#endif
	
// 	// Mark the batch pools as empty
// 	for (int i = 0; i < nleaves; i++)
// 		for (int j = 0; j < nleaves; j++) {
// 			nbatch[i][j] = 0;
// 		}
	
	// Loop over all bonds.  First, double loop over leaves
	LeafToLeafIterator<ListOfAtoms> leaf_i(atoms);
	vector<const emt_parameters *>::iterator param_i = parameters.begin();
	int leafnum_i = 0;
	while (!leaf_i.Done()) {
		int elem_i = (*param_i)->index;   // The element of the atoms on this leaf.
		int natoms_i = leaf_i.Current()->GetNumber();

		LeafToLeafIterator<ListOfAtoms> leaf_j = leaf_i;
		vector<const emt_parameters *>::iterator param_j = param_i;
		int leafnum_j = leafnum_i;
		while (!leaf_j.Done()) {
			int elem_j = (*param_j)->index;   // The element of the atoms on this leaf.

			// An empty batch pool
			int nbatch = 0;
			int *me_p = me;
			int *other_p = other;
			double (*rnb_p)[3] = rnb;
			double *dist2_p = dist2;
			int spaceleft = BUFLEN;
			bool redo = 0;
			//double dummy = nblocator->GetDrift();
			
			// Loop over atoms in leaf_i
			for (int n_i = 0; n_i < natoms_i; ++n_i) {
				int numnb = 0;
				bool spaceok;
				spaceok = nblocator->Neighbors(n_i, leafnum_i, leafnum_j, other_p, numnb, rnb_p,
											   dist2_p, spaceleft);
				assert (spaceok || !redo); // Should not fail in redo.
				redo = 0;
				// OBS: IF leafnum_i == leafnum_j, remove half the neighbors !!
				if (spaceok && leafnum_i == leafnum_j) {
					int i = 0;
					for (int j = 0; j < numnb; j++)	{
						if (other_p[j] < n_i) {
							// Keep it
							if (j != i) {
								other_p[i] = other_p[j];
								rnb_p[i][0] = rnb_p[j][0];
								rnb_p[i][1] = rnb_p[j][1];
								rnb_p[i][2] = rnb_p[j][2];
								dist2_p[i] = dist2_p[j];
							}
							i++;
						} else {
							// Drop it
							spaceleft++;
						}
					}
					numnb = i;
				}
				if (spaceok) {
					// The call to Neighbors succeeded
					nbatch += numnb;
					other_p += numnb;
					rnb_p += numnb;
					dist2_p += numnb;
					while (numnb-- > 0) {
						*(me_p++) = n_i;
					}
				} else {
					// Batch pool is full
					sigma_batch(nbatch, **param_i, **param_j, me, other, rnb, dist2,
								sigma1[elem_j]->GetLeaf(leafnum_i),
								sigma1[elem_i]->GetLeaf(leafnum_j),
								sigma2[elem_j]->GetLeaf(leafnum_i),
								sigma2[elem_i]->GetLeaf(leafnum_j));
#if 0
					for (int k = 0; k < nelements; k++) {
						cerr << "Element " << k << ":  ";
						double max = -10000.0;
						int nm = sigma1[k]->Size();
						for (int i = 0; i < nm; i++) {
							double c = (*sigma1[k])[i];
							if (c > max) max = c;
						}
						cerr << "Max(sigma1) = " << max << endl;
					}
#endif					
					// Empty batch pool
					nbatch = 0;
					me_p = me;
					other_p = other;
					rnb_p = rnb;
					dist2_p = dist2;
					spaceleft = BUFLEN;
					// And redo the failed Neighbors call
					--n_i;
					redo = 1;   // Just used in assert.
				}
			}
			// Empty the batch pool
			sigma_batch(nbatch, **param_i, **param_j, me, other, rnb, dist2,
						sigma1[elem_j]->GetLeaf(leafnum_i),
						sigma1[elem_i]->GetLeaf(leafnum_j),
						sigma2[elem_j]->GetLeaf(leafnum_i),
						sigma2[elem_i]->GetLeaf(leafnum_j));
			
#if 0
			for (int k = 0; k < nelements; k++) {
				cerr << "Element " << k << ":  ";
				double max = -10000.0;
				int nm = sigma1[k]->Size();
				for (int i = 0; i < nm; i++) {
					double c = (*sigma1[k])[i];
					if (c > max) max = c;
				}
				cerr << "Max(sigma1) = " << max << endl;
			}
#endif					
			// Update counters and pointers in inner loop over leaves
			leaf_j.Next();
			++param_j;
			++leafnum_j;
		}  // while (!leaf_j.Done())
		
		// Update counters and pointers in outher loop over leaves
		leaf_i.Next();
		++param_i;
		++leafnum_i;
	}  // while (!leaf_i.Done()) 
}

void EMT::sigma_batch(const int n, const emt_parameters selfparam, const emt_parameters otherparam,
					  const int *self, const int *other, const double (*rnb)[3], const double *dist2,
					  DMArray<double> *sigma1self, DMArray<double> *sigma1other,
					  DMArray<double> *sigma2self, DMArray<double> *sigma2other) 
{
	double dist[BUFLEN];
    double wght[BUFLEN];
    double arg[BUFLEN], ex1[BUFLEN];
    double dsigma1s[BUFLEN], dsigma2s[BUFLEN];
    double ds1o[BUFLEN], ds2o[BUFLEN];
    double *dsigma1o, *dsigma2o;
    double cutslope, cutslopecutdist;
    double other_eta2betaseq, self_eta2betaseq;
    double other_kappaoverbeta, self_kappaoverbeta;
    double other_kappaseq, self_kappaseq;
	int zs, zo;
    int i;
    
    /* Get EMT parameters */
    zs = selfparam.index;
    zo = otherparam.index;
    cutslope = cutoffslope;
    cutslopecutdist = cutoffslope * cutoffdistance;
    other_eta2betaseq = otherparam.eta2 * Beta * otherparam.seq;
    self_eta2betaseq = selfparam.eta2 * Beta * selfparam.seq;
    other_kappaoverbeta = otherparam.kappa / Beta;
    self_kappaoverbeta = selfparam.kappa / Beta;
    other_kappaseq = otherparam.kappa * otherparam.seq;
    self_kappaseq = selfparam.kappa * selfparam.seq;
	
    /* dist = sqrt(dist2),  distances between atoms */
    vsqrt(dist, dist2, &n);

    /* Calculate weight factor (cutoff function) */
    /* ex1 = exp(cutslope * dist - cutslopecutdist) */
    vec_mul_scal_add_scal(arg, dist, cutslope, -cutslopecutdist, n);
    vexp(ex1, arg, &n);
    /* wght = 1 / (ex1 + 1.0) */
    vec_add_scal(arg, ex1, 1.0, n);
    vrec(wght, arg, &n);

    /* Contribution to sigma1 */
    /* ex1 = exp(-otherparam.eta2 * dist + other_eta2betaseq) */
    vec_mul_scal_add_scal(arg, dist, -otherparam.eta2, other_eta2betaseq, n);
    vexp(ex1, arg, &n);
    /* dsigma1s = wght * ex1 */
    vec_mul(dsigma1s, wght, ex1, n);
    if (zs == zo)
		dsigma1o = dsigma1s;
    else
    {
		/* ex1 = exp(-selfparam.eta2 * dist + self_eta2betaseq) */
		vec_mul_scal_add_scal(arg, dist, -selfparam.eta2, self_eta2betaseq, n);
		vexp(ex1, arg, &n);
		/* dsigma1o = wght * ex1 */
		vec_mul(ds1o, wght, ex1, n);
		dsigma1o = ds1o;
    }

    /* Contribution to sigma2 */
    /* ex1 = exp(-other_kappaoverbeta * dist + other_kappaseq) */
    vec_mul_scal_add_scal(arg, dist, -other_kappaoverbeta, other_kappaseq, n);
    vexp(ex1, arg, &n);
    /* dsigma2s = wght * ex1 */
    vec_mul(dsigma2s, wght, ex1, n);
    if (zs == zo)
		dsigma2o = dsigma2s;
    else
    {
		/* ex1 = exp(-self_kappaoverbeta * dist + self_kappaseq) */
		vec_mul_scal_add_scal(arg, dist, -self_kappaoverbeta, self_kappaseq,
							  n);
		vexp(ex1, arg, &n);
		/* dsigma2o = wght * ex1 */
		vec_mul(ds2o, wght, ex1, n);
		dsigma2o = ds2o;
    }

    /* Distribute contributions to sigma1 and sigma2 */
	int s1selfsize = sigma1self->Size();
	int s1othersize = sigma1other->Size();
	assert((sigma2self->Size() == s1selfsize) && (sigma2other->Size() == s1othersize));
    for (i = 0; i < n; i++)
    {
		assert(self[i] < s1selfsize);
		assert(other[i] < s1othersize);
#if 0
		cerr << "  (s,o): " << self[i] << "," << other[i] << " rnb: (" <<
			rnb[i][0] << "," << rnb[i][1] << "," <<	rnb[i][2] << ")";
		cerr << " dsigma1s: " << dsigma1s[i] <<
			"  dsigma1o: " << dsigma1o[i] << "  dsigma2s: " << dsigma2s[i] <<
			"  dsigma2o: " << dsigma2o[i] << endl;
#endif
		(*sigma1self)[self[i]] += dsigma1s[i];
		(*sigma1other)[other[i]] += dsigma1o[i];
		(*sigma2self)[self[i]] += dsigma2s[i];
		(*sigma2other)[other[i]] += dsigma2o[i];
    }
}

double EMT::CalculateEnergyAfterSigmas()
{
	double totalenergy = 0.0;
	
	// Allocate temporary data.  We need to be able to hold the largest leaf.
	double *tmp = new double[6*maxleaflen];
	double *sigma = tmp;
	double *tmp1 = tmp + maxleaflen;
	double *tmp2 = tmp + 2*maxleaflen;
	double *ex1 = tmp + 3*maxleaflen;
	double *ex2 = tmp + 4*maxleaflen;
	double *radius = tmp + 5*maxleaflen;
	double cnst;
	
	// Loop over leafs, and calculate the energies for each leaf.
	int leafnum = 0;
	LeafToLeafIterator<ListOfAtoms> leaf_i(atoms);
	LeafToLeafIterator<DMArray<double> > leaf_Ec(Ec);
	LeafToLeafIterator<DMArray<double> > leaf_Eas(Eas);
	LeafToLeafIterator<DMArray<double> > leaf_radius(this->radius);
	LeafToLeafIterator<DMArray<double> > leaf_dEds(dEds);
	vector<const emt_parameters *>::iterator param_i = parameters.begin();
	while (!leaf_i.Done()){
		ListOfAtoms *thisleaf = leaf_i.Current();
		int natoms = thisleaf->GetNumber();
		assert(natoms <= maxleaflen);
		int zs = (*param_i)->index; // The element of these atoms
		
		// Calculate combinations of EMT parameters
		double emkappaseq = (*param_i)->kappa * (*param_i)->seq;
		double eminv12gamma1 = 1.0 / (12.0 * (*param_i)->gamma1);
		double emseq = (*param_i)->seq;
		double emkappa = (*param_i)->kappa;
		double emlambda = (*param_i)->lambda;
		double eme0lambdalambda = (*param_i)->e0 * (*param_i)->lambda * (*param_i)->lambda;
		double eme0lambdalambdaseq = (*param_i)->e0 * (*param_i)->lambda * (*param_i)->lambda * (*param_i)->seq;
		double em6V0kappa = 6.0 * (*param_i)->V0 * (*param_i)->kappa;
		double eminvbetaeta2 = 1.0 / ((*param_i)->eta2 * Beta);
		double emlambdaseq = (*param_i)->lambda * (*param_i)->seq;
		double eme0lambda = (*param_i)->e0 * (*param_i)->lambda;
		double em6V0 = 6.0 * (*param_i)->V0;
		double emneghalfV0overgamma2 = -0.5 * (*param_i)->V0 / (*param_i)->gamma2;
		double eme0 = (*param_i)->e0;
		
		// Calculate the total sigma1
		int i;
		for (i = 0; i < natoms; i++)
			sigma[i] = 0.0;
		int zo;
		for (zo = 0; zo < nelements; zo++) {
#if 0
			DMArrayIterator<double> sig_iter(sigma1[zo]);
			sig_iter.Init();
			double chizszo = (*chi)[zs][zo];
			for (int i = 0; i < natoms; i++) {
				double tmp2 = sig_iter.Current();
				double dsig = chizszo * tmp2;
				sigma[i] += dsig;
				sig_iter.Next();
			}
#else
			DMArray<double> *s1 = sigma1[zo]->GetLeaf(leafnum);
			double chizszo = (*chi)[zs][zo];
			for (int i = 0; i < natoms; i++) {
				sigma[i] += chizszo * (*s1)[i];
			}
#endif
		}
		for (i = 0; i < natoms; i++) {
			if (sigma[i] < 1.0e-9)
				sigma[i] = 1.0e-9;  // Prevents trouble from isolated atoms.
		}
				
		// radius = emseq - eminvbetaeta2 * log(sigma1[is] * eminv12gamma1); 
		vec_mul_scal(tmp1, sigma, eminv12gamma1, natoms);
		vlog(tmp2, tmp1, &natoms);
		vec_mul_scal_add_scal(radius, tmp2, -eminvbetaeta2, emseq, natoms);
		
		// ex1 = exp(-emkappa * radius + emkappaseq); 
		vec_mul_scal_add_scal(tmp1, radius, -emkappa, emkappaseq, natoms);
		vexp(ex1, tmp1, &natoms);
		// ex2 = exp(-emlambda * radius + emlambdaseq); 
		vec_mul_scal_add_scal(tmp1, radius, -emlambda, emlambdaseq, natoms);
		vexp(ex2, tmp1, &natoms);
#ifdef CALCNBAR
		// nbar = emneq * ex1;
#error NBAR not implemented
		//vec_mul_scal_add_scal(nbar, ex1, emneq, 0.0, natoms);
#endif
		// tmp2 = (-eme0lambdalambda * radius + eme0lambdalambdaseq) * ex2 - em6V0kappa* ex1
		vec_dEds(tmp2, radius, ex2, ex1, -eme0lambdalambda, eme0lambdalambdaseq, -em6V0kappa, natoms);
		// dEds = -eminvbetaeta2 * tmp2 / sigma
		vrec(tmp1, sigma, &natoms);
		vec_self_mul_mul_scal(tmp2, tmp1, -eminvbetaeta2, natoms);
		
		// Ec = (*param_i)->e0 * (1 + (*param_i)->lambda * (radius - seq)) * ex2 
		cnst = (*param_i)->e0 * (1.0 - (*param_i)->lambda * (*param_i)->seq);
		vec_mul_scal_add_scal_mul(tmp1, radius, eme0lambda, cnst, ex2, natoms);
		
		// We have dumped Ec in tmp1 and dEds in tmp2, now copy it to the DMArray.
		DMArrayIterator<double> Ec_iter(leaf_Ec.Current());
		DMArrayIterator<double> dEds_iter(leaf_dEds.Current());
		DMArrayIterator<double> rad_iter(leaf_radius.Current());
		Ec_iter.Init();
		dEds_iter.Init();
		rad_iter.Init();
		for (i = 0; i < natoms; i++) {
			double ec = tmp1[i] - eme0;
			totalenergy += ec;
			Ec_iter.Current() = ec;
			Ec_iter.Next();
			dEds_iter.Current() = tmp2[i];
			dEds_iter.Next();
			rad_iter.Current() = radius[i];
			rad_iter.Next();
		}
		
		// Calculate the total sigma2
		for (i = 0; i < natoms; i++)
			sigma[i] = 0.0;
		for (zo = 0; zo < nelements; zo++) {
			DMArrayIterator<double> sig_iter(sigma2[zo]->GetLeaf(leafnum));
			sig_iter.Init();
			for (int i = 0; i < natoms; i++) {
				sigma[i] += (*chi)[zs][zo] * sig_iter.Current();
				sig_iter.Next();
			}
		}
		for (i = 0; i < natoms; i++) {
			if (sigma[i] < 1.0e-9)
				sigma[i] = 1.0e-9;  // Prevents trouble from isolated atoms.
			//cerr << "   " << i << "  sigma2 = " << sigma[i] << endl;
		}
		
		// Eas = 6 * (*param_i)->V0 * ex1 - 0.5 * (*param_i)->V0 * sigma2 / gamma2 
		vec_dbl_mul_scal_add(tmp1, ex1, em6V0, sigma, emneghalfV0overgamma2, natoms);			      
		// We have dumped Eas in tmp1, now copy it to the DMArray.
		DMArrayIterator<double> Eas_iter(leaf_Eas.Current());
		Eas_iter.Init();
		for (i = 0; i < natoms; i++) {
			totalenergy += tmp1[i];
			Eas_iter.Current() = tmp1[i];
			Eas_iter.Next();
		}
		
		// Update iterators and counters
		leaf_i.Next();
		leaf_Ec.Next();
		leaf_Eas.Next();
		leaf_radius.Next();
		leaf_dEds.Next();
		++param_i;
		++leafnum;
	}

	delete[] tmp;
	return totalenergy;
}

void EMT::CalculateForcesAfterEnergy(DMArray<double, DIMENSION> *forces)
{
	// Buffers
	int me[BUFLEN];
	int other[BUFLEN];
	double rnb[BUFLEN][3];
	double dist2[BUFLEN];
	double dEdss[BUFLEN];
	double dEdso[BUFLEN];
	NeighborLocator *nblocator = atoms->GetNeighborLocator();

	// Set forces to zero
	forces->Zero();
	
	// Loop over all bonds.  First, double loop over leaves
	LeafToLeafIterator<ListOfAtoms> leaf_i(atoms);
	LeafToLeafIterator<DMArray<double> > leaf_dEds_i(dEds);
	vector<const emt_parameters *>::iterator param_i = parameters.begin();
	int leafnum_i = 0;
	while (!leaf_i.Done()) {
		int elem_i = (*param_i)->index;   // The element of the atoms on this leaf.
		int natoms_i = leaf_i.Current()->GetNumber();
		DMArray<double> *dEds_i = leaf_dEds_i.Current();
		
		LeafToLeafIterator<ListOfAtoms> leaf_j = leaf_i;
		LeafToLeafIterator<DMArray<double> > leaf_dEds_j = leaf_dEds_i;
		vector<const emt_parameters *>::iterator param_j = param_i;
		int leafnum_j = leafnum_i;
		while (!leaf_j.Done()) {
			int elem_j = (*param_j)->index;   // The element of the atoms on this leaf.
			DMArray<double> *dEds_j = leaf_dEds_j.Current();
			
			// An empty batch pool
			int nbatch = 0;
			int *me_p = me;
			int *other_p = other;
			double (*rnb_p)[3] = rnb;
			double *dist2_p = dist2;
			double *dEdss_p = dEdss;
			double *dEdso_p = dEdso;
			int spaceleft = BUFLEN;
			bool redo = 0;
			
			// Loop over atoms in leaf_i
			for (int n_i = 0; n_i < natoms_i; ++n_i) {
				int numnb = 0;
				bool spaceok;
				spaceok = nblocator->Neighbors(n_i, leafnum_i, leafnum_j, other_p, numnb, rnb_p,
											   dist2_p, spaceleft);
				assert (spaceok || !redo); // Should not fail in redo.
				redo = 0;
				// If leafnum_i == leafnum_j, remove half the neighbors !!
				if (spaceok && leafnum_i == leafnum_j) {
					int i = 0;
					for (int j = 0; j < numnb; j++)	{
						if (other_p[j] < n_i) {
							// Keep it
							if (j != i) {
								other_p[i] = other_p[j];
								rnb_p[i][0] = rnb_p[j][0];
								rnb_p[i][1] = rnb_p[j][1];
								rnb_p[i][2] = rnb_p[j][2];
								dist2_p[i] = dist2_p[j];
							}
							i++;
						} else {
							// Drop it
							spaceleft++;
						}
					}
					numnb = i;
				}
				if (spaceok) {
					// The call to Neighbors succeeded
					for (int i = 0; i < numnb; i++) {
						*(me_p++) = n_i;
						*(dEdss_p++) = (*dEds_i)[n_i];
						*(dEdso_p++) = (*dEds_j)[other_p[i]];
					}
					nbatch += numnb;
					other_p += numnb;
					rnb_p += numnb;
					dist2_p += numnb;
				} else {
					// Batch pool is full
					force_batch(nbatch, **param_i, **param_j, me, other, rnb, dist2,
								dEdss, dEdso, forces->GetLeaf(leafnum_i),
								forces->GetLeaf(leafnum_j));
					// Empty batch pool
					nbatch = 0;
					me_p = me;
					other_p = other;
					rnb_p = rnb;
					dist2_p = dist2;
					dEdss_p = dEdss;
					dEdso_p = dEdso;
					spaceleft = BUFLEN;
					// And redo the failed Neihbbors call
					--n_i;
					redo = 1;   // Just used in assert.
				}
			}
			// The rest of the batch pool
			force_batch(nbatch, **param_i, **param_j, me, other, rnb, dist2,
						dEdss, dEdso, forces->GetLeaf(leafnum_i),
						forces->GetLeaf(leafnum_j));
			
			// Update counters and pointers in inner loop over leaves
			leaf_j.Next();
			leaf_dEds_j.Next();
			++param_j;
			++leafnum_j;
		}  // while (!leaf_j.Done())
		
		// Update counters and pointers in outher loop over leaves
		leaf_i.Next();
		leaf_dEds_i.Next();
		++param_i;
		++leafnum_i;
	}  // while (!leaf_i.Done()) 
}


//			force_batch(nbatch, **param_i, **param_j, me, other, rnb, dist2,
//						dEdss, dEdso, force->GetLeaf(leafnum_i),
//						force->GetLeaf(leafnum_j));

void EMT::force_batch(int n, const emt_parameters selfparam, const emt_parameters otherparam,
					  const int *self, const int *other, const double (*rnb)[3],
					  const double *dist2, const double *dEdss, const double *dEdso,
					  DMArray<double, DIMENSION> *forceself, DMArray<double, DIMENSION> *forceother)
{
	int i, j;
	int zs, zo;
    double dist[BUFLEN];
    double arg[BUFLEN];
    double ex1[BUFLEN];
    double wght[BUFLEN];
    double dwdr[BUFLEN];
    double dsigma1dr_o[BUFLEN];
    double dsigma2dr_o[BUFLEN];
    double ds1dr_s[BUFLEN];
    double ds2dr_s[BUFLEN];
    double df[BUFLEN];
    double *dsigma1dr_s;
    double *dsigma2dr_s;
    double dfx;
    double cutslope, cutslopecutdist, other_eta2betaseq, other_kappaoverbeta;
    double other_kappaseq, self_eta2betaseq, self_kappaoverbeta;
    double self_kappaseq, cnst_s, cnst_o;
    double chizszo, chizozs;

	assert(n <= BUFLEN);
#if 0
	for (i = 0; i < n; i++)
		cerr << i << " self:" << self[i] << " other:" << other[i] << " rnb:(" << rnb[i][0] << "," <<
			rnb[i][1] << "," << rnb[i][2] << ") dist2:" << dist2[i] << " dEdss:" << dEdss[i] <<
			" dEdso:" << dEdso[i] << endl;
#endif
	
    /* Get EMT parameters */
	zs = selfparam.index;
	zo = otherparam.index;
	chizszo = (*chi)[zs][zo];
	chizozs = (*chi)[zo][zs];
    cutslope = cutoffslope;
    cutslopecutdist = cutoffslope * cutoffdistance;
    other_eta2betaseq = otherparam.eta2 * Beta * otherparam.seq;
    other_kappaoverbeta = otherparam.kappa / Beta;
    other_kappaseq = otherparam.kappa * otherparam.seq;
    self_eta2betaseq = selfparam.eta2 * Beta * selfparam.seq;
    self_kappaoverbeta = selfparam.kappa / Beta;
    self_kappaseq = selfparam.kappa * selfparam.seq;

	/* calculate distances */
    vsqrt(dist, dist2, &n);

    /* Calculate cutoff function (weight factor) */
    /* ex1 = exp(cutslope * dist - cutslopecutdist) */
    vec_mul_scal_add_scal(arg, dist, cutslope, -cutslopecutdist, n);
    vexp(ex1, arg, &n);
    /* wght = 1 / (ex1 + 1.0) */
    vec_add_scal(arg, ex1, 1.0, n);
    vrec(wght, arg, &n);

    /* Calculate derivative of the cutoff function. */
    /* dwdr = - cutslope * wght * (1 - wght) */
    vec_diffweight(dwdr, wght, -cutslope, n);

    /* Derivatives of sigma */
    
    /* ex1 = exp(-otherparam.eta2 * dist + other_eta2betaseq) */
    vec_mul_scal_add_scal(arg, dist, -otherparam.eta2, other_eta2betaseq, n);
    vexp(ex1, arg, &n);
    /* dsigma1dr_o = (dwdr + wght * -otherparam.eta2) * ex1 */
    vec_mul_scal_add_mul(dsigma1dr_o, wght, -otherparam.eta2, dwdr, ex1, n);

    /* ex1 = exp(-other_kappaoverbeta * dist + other_kappaseq) */
    vec_mul_scal_add_scal(arg, dist, -other_kappaoverbeta, other_kappaseq, n);
    vexp(ex1, arg, &n);
    /* dsigma2dr_o = (dwdr + wght * -other_kappaoverbeta) * ex1 */
    vec_mul_scal_add_mul(dsigma2dr_o, wght, -other_kappaoverbeta, dwdr, ex1,
			 n);

    if (zs == zo)
    {
		dsigma1dr_s = dsigma1dr_o;
		dsigma2dr_s = dsigma2dr_o;
    }
    else
    {
		/* ex1 = exp(-selfparam.eta2 * dist + self_eta2betaseq) */
		vec_mul_scal_add_scal(arg, dist, -selfparam.eta2, self_eta2betaseq, n);
		vexp(ex1, arg, &n);
		/* dsigma1dr_s = (dwdr + wght * -selfparam.eta2) * ex1 */
		vec_mul_scal_add_mul(ds1dr_s, wght, -selfparam.eta2, dwdr, ex1, n);
		dsigma1dr_s = ds1dr_s;
		
		/* ex1 = exp(-self_kappaoverbeta * dist + self_kappaseq) */
		vec_mul_scal_add_scal(arg, dist, -self_kappaoverbeta, self_kappaseq,
							  n);
		vexp(ex1, arg, &n);
		/* dsigma2dr_o = (dwdr + wght * -self_kappaoverbeta) * ex1 */
		vec_mul_scal_add_mul(ds2dr_s, wght, -self_kappaoverbeta, dwdr, ex1, n);
		dsigma2dr_s = ds2dr_s;
    }

    /* arg = 1.0 / dist */
    vrec(arg, dist, &n);

    cnst_s = -0.5 * selfparam.V0 * chizszo / selfparam.gamma2;
    cnst_o = -0.5 * otherparam.V0 * chizozs / otherparam.gamma2;
    /* df = arg * (dsigma1dr_o * dEdss * chizszo + cnst_s * dsigma2dr_o +
       (dsigma1dr_s * dEdso * chizozs + cnst_o * dsigma2dr_s)  */
    vec_df(df, arg, dsigma1dr_o, dEdss, chizszo, dsigma2dr_o, cnst_s,
	   dsigma1dr_s, dEdso, chizozs, dsigma2dr_s, cnst_o, n);

    
    /* Distribute force contributions */
    for (i = 0; i < n; i++)
	{
		// cerr << self[i] << " " << other[i] << "  " << df[i] << endl;
		double *fs = (*forceself)[self[i]];
		double *fo = (*forceother)[other[i]];
		for (j = 0; j < 3; j++)
		{
			dfx = df[i] * rnb[i][j];
			fs[j] += dfx;
			fo[j] -= dfx;
		}
	}
}




#endif // DIMENSION == 3
