// Johnson1964IronPotential.h: interface for the Johnson1964IronPotential class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef JOHNSON1964IRONPOTENTIAL_H
#define JOHNSON1964IRONPOTENTIAL_H

#include "DMFirstHeader.h"

#include "DigitalMaterial.h"
#include "Potential.h"

//! Pair potential devised by Johnson in 1964 for alpha-Iron

/*!

  This is a pair potential for alpha-phase Iron, a BCC metal. It includes first and second neighbors, and consists of matched  cubics with parameters chosen to fit elastic constants.

  Reference: R. A. Johnson, Physical Review, Vol 134, A1329 (1964)

  Units are Angstroms for length and eV for energy

  \date February 5, 2001
  \author Nicholas Bailey

 */

class Johnson1964IronPotential : public Potential
{
public:
	Johnson1964IronPotential();
	virtual ~Johnson1964IronPotential();
	void CalculateForces(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces);
	void CalculateForcesBetweenTypes(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces, int type1, int type2) ; // the force array isn't initialized in this function

	double CalculateEnergy(ListOfAtoms *atoms);
	double CalculateEnergyAndForces(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces);
	double CalculateAtomicEnergy(ListOfAtoms *atoms, int i);

	double GetLengthScale() const
	{
		return 2.86;
	}

	double GetEnergyCurvatureScale() const
	{
		return 6.*J1964_param5*(J1964_param6 - 2.6);
		// second derivative at r = 2.6, which is roughly where the minimum is.
	}


	void Save(DMWriter *writer) const;
	void Load(DMReader *reader);
	//! Get the string name of the class for saving or loading.
	const std::string &GetType() const
	{
		return className;
	}

 protected:
#ifndef SWIG	
	//! Allocate the \a neighbors, \a displacements, and \a displacementsSq arrays
	void AllocateForNeighborLocator(int nPairs);
	//! a rough (over-)estimation of the number of neighbors
	int         nPairs;
	double      *displacementsSq;
	double      (*displacements)[DIMENSION];
	int         *neighbors;
	double		cutoffDistance; // needed directly in the formulas
	double		innerCutoff1Sq;
	double		innerCutoff2Sq;


	// parameters for the Johnson potential
	const double J1964_param1;
	const double J1964_param2;
	const double J1964_param3;
	const double J1964_param4;
	const double J1964_param5;
	const double J1964_param6;
	const double J1964_param7;
	const double J1964_param8;
	const double J1964_param9;
	const double J1964_param10;
	const double J1964_param11;
	const double J1964_param12;
	
private:
	static const std::string className;

#endif // SWIG 
};

#endif // !defined(JOHNSON1964IRONPOTENTIAL_H
