// SiliconEDIPPotential.cpp: implementation of the
// SiliconEDIPPotential class.
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/
//////////////////////////////////////////////////////////////////////
#include "DigitalMaterial.h"

#include "SiliconEDIPPotential.h"
#include "ListOfAtoms.h"
#include "DMArray.h"
#include "IndexError.h"
#include "NotEnoughMemory.h"
#include "NeighborLocator.h"
#include "SerialBuilder.h"

#include "DMLastHeader.h"

using namespace std;


static SerialBuilder<SiliconEDIPPotential> *_EDIPcreator = 
    new SerialBuilder<SiliconEDIPPotential>("SiliconEDIPPotential");


const string SiliconEDIPPotential::className = string("SiliconEDIPPotential");

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////


SiliconEDIPPotential::SiliconEDIPPotential() :  EDIP_A(7.9821730),
												EDIP_B(1.5075463),
												EDIP_Q0(312.1341346), 
												EDIP_a(3.1213820),
												EDIP_c(2.5609104),
												EDIP_alpha(3.1083847),
												EDIP_beta(0.0070975),
												EDIP_gamma(1.1247945),
												EDIP_eta(0.2523244),
												EDIP_lambda(1.4533108),
												EDIP_mu(0.6966326),
												EDIP_rho(1.2085196),
												EDIP_sigma(0.5774108),
												u1(-0.165799),
												u2(32.557),
												u3(0.286198),
												u4(0.66)
{
	SetCutoffDistance(EDIP_a);
	
	nAtoms = 0;

	neighbors = NULL;
	displacements = NULL;
	displacementsSq = NULL;

}

SiliconEDIPPotential::~SiliconEDIPPotential()
{
	if(neighbors)
	{
		delete [] neighbors;
		delete [] displacements;
		delete [] displacementsSq;
	}
}

void SiliconEDIPPotential::Allocate(int nAtoms)
{
   	// nAtoms is playing the role of EMTPotential's arrayLim here
	this->nAtoms = nAtoms;	
	
	if(neighbors) 
	{	
		delete [] neighbors;
		delete [] displacements;
		delete [] displacementsSq;
	}

	if(nAtoms > 0)
	{
		neighbors       = new (std::nothrow) int[nAtoms];
		displacements   = new (std::nothrow) double[nAtoms][DIMENSION];
		displacementsSq = new (std::nothrow) double[nAtoms];
		
		ASSERT(neighbors);
		ASSERT(displacements);
		ASSERT(displacementsSq);
		
		if(0 == neighbors)
			throw NotEnoughMemory("SiliconEDIPPotential",
								  "Allocate",
								  "Allocating nAtoms doubles");
		if(0 == displacements)
			throw NotEnoughMemory("SiliconEDIPPotential",
								  "Allocate",
								  "Allocating nAtoms*DIMENSION doubles");
		if(0 == displacementsSq)
			throw NotEnoughMemory("SiliconEDIPPotential",
								  "Allocate",
								  "Allocating nAtoms doubles");
	}
	else
	{
		neighbors       = NULL;
		displacements   = NULL;
		displacementsSq = NULL;
	}
	
}

void SiliconEDIPPotential::CalculateForces(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces)
{
	CalculateForcesAndEnergy(atoms, forces);
}


double SiliconEDIPPotential::CalculateForcesAndEnergy(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces)
{
	ASSERT(atoms && forces);
	// need the following otherwise might not enter the loop!
	if(nAtoms != atoms->GetNumber()) Allocate(atoms->GetNumber());
	forces->Zero();
	double potentialEnergy = 0;
	for(int i=0;i<nAtoms;i++)
	{
		potentialEnergy += AtomicForcesEnergy(atoms, forces, i);
	}
	return potentialEnergy;
}

double SiliconEDIPPotential::AtomicForcesEnergy(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces, int i)
{
	double potentialEnergy = 0.;
	ASSERT(atoms && forces);

	int coord,jx,kx,j,k;
	int numberNeighbors;
	bool arrayNotTooFull;
	if(nAtoms != atoms->GetNumber()) Allocate(atoms->GetNumber());
	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	ASSERT(neighborLocator);
	int spacesLeftInArrays;
	double *forceI, *forceJ, *forceK;
	double dummyForce[DIMENSION]; // where we send the force 
	// contribution to a ghost atom

	spacesLeftInArrays = nAtoms;
	arrayNotTooFull	= neighborLocator->Neighbors(i,neighbors,numberNeighbors, 
												 displacements,displacementsSq,spacesLeftInArrays);
	ASSERT(arrayNotTooFull);
	
	forceI = (*forces)[i];
	
	// evaluate coordination and its derivatives, and r1
	double Zi = 0.; // coordination of atom i
	double *dZidr1 = new double[numberNeighbors];
	double *r1 = new double[numberNeighbors];
	double *expFactor2 = new double[numberNeighbors];
	double *expFactor3 = new double[numberNeighbors];
	
	for(jx=0;jx<numberNeighbors;++jx)
	{
		r1[jx] = sqrt(displacementsSq[jx]);
		expFactor2[jx] = exp(EDIP_sigma / ( r1[jx] - EDIP_a ) );
		expFactor3[jx] = exp(EDIP_gamma / ( r1[jx] - EDIP_a ) );
		if(r1[jx] < EDIP_c)
		{
			Zi += 1.;
			dZidr1[jx] = 0.;
		}
		else 
		{
			double x = (r1[jx] - EDIP_c)/(EDIP_a - EDIP_c); 
			double x3 = x*x*x;
			double x4 = x3 * x;
			Zi += exp(EDIP_alpha/(1 - 1./x3));
			dZidr1[jx] = -3 * EDIP_alpha * Zi / ((1 - 1./x3) * (1 - 1./x3) * x4 * (EDIP_a - EDIP_c) );
		}
	} // end loop over neighbors to calculate Z, dZdr1, r1, expFactor	
	
	double pZ = exp(-EDIP_beta * Zi * Zi);
	double dpdZ = -2. * EDIP_beta * Zi * pZ;
	
	// "2 body" contributions---not really 2-body but they look it
	for(jx=0;jx<numberNeighbors;jx++)
	{
		j=neighbors[jx];
		if(j<nAtoms) forceJ = (*forces)[j];
		else forceJ = dummyForce;
		
		// first we get the forces which come from the explicit r_ij
		// dependence of the 2-body term
		
		// forceR is -(dE/dr)/r. The minus sign mostly cancels minus signs that come from the derivatives
		double forceR = EDIP_A * ( 
			pow(EDIP_B/r1[jx], EDIP_rho) * EDIP_rho / r1[jx]
			+ (pow(EDIP_B/r1[jx], EDIP_rho) - pZ) 
			* (EDIP_sigma/((r1[jx]-EDIP_a)*(r1[jx]-EDIP_a)))
			)
			* expFactor2[jx] / r1[jx];
		
		double pairEnergy = EDIP_A * (pow(EDIP_B/r1[jx], EDIP_rho) - pZ)
			* expFactor2[jx];
		potentialEnergy += pairEnergy;
		// Note that the displacements given by Neighbors are of particles j wrt particles i
		for (coord=0; coord<DIMENSION; coord++)
		{
			forceI[coord] += -displacements[jx][coord]*forceR;
			forceJ[coord] += +displacements[jx][coord]*forceR;
		}
		
		// because the "2-body" terms actually depend through Z on the 
		// positions of other atoms, we have contributions from this 
		// dependence to the forces on them (and i and j) too.
		for(int kk=0;kk<numberNeighbors;++kk)
		{
			if(dZidr1[kk] != 0.)
			{
				k=neighbors[kk]; 
				if(k<nAtoms) forceK = (*forces)[k];
				else forceK = dummyForce;
				double x_forceR = EDIP_A * dpdZ * expFactor2[jx] * dZidr1[kk] /r1[kk];
				for (coord=0; coord<DIMENSION; coord++)
				{
					forceI[coord] += -displacements[kk][coord] * x_forceR;
					forceK[coord] += displacements[kk][coord] * x_forceR; 
				}
			}
		}
		
		
		// "3 body" contributions
		for(kx=0;kx<jx;++kx)	
		{
			k=neighbors[kx];
			if(k<nAtoms) forceK = (*forces)[k];
			else forceK = dummyForce;
			// calculate the angle
			double cosTheta = 0;
			for(coord=0;coord<DIMENSION;++coord)
				cosTheta += displacements[jx][coord] *
					displacements[kx][coord];
			double r_ij = r1[jx];
			double r_ik = r1[kx];
			cosTheta /= (r_ij * r_ik);
			
			double threeBodyRadial = expFactor3[jx] * expFactor3[kx];
			//double threeBodyRadial = exp(EDIP_gamma*(1./(r_ij-EDIP_a)+1./(r_ik-EDIP_a)) );
			// tau function
			double tauZ = u1 + u2 * (u3 * exp(-u4 * Zi) - exp(-2 * u4 * Zi));
			double dtaudZ = u2 * (- u3 * u4 * exp(-u4 * Zi) + 2 * u4 * exp(-2 * u4 * Zi));
			double trigFactor = cosTheta + tauZ;
			double trigFactor2 = trigFactor * trigFactor;
			
			// Q function
			double QZ = EDIP_Q0 * exp(-EDIP_mu * Zi);
			double dQdZ = -EDIP_mu * QZ;
			
			double threeBodyAngular = EDIP_lambda * (1 - exp(-QZ * trigFactor2) + EDIP_eta * QZ * trigFactor2);
			
			// have enough for the energy now
			potentialEnergy += threeBodyRadial * threeBodyAngular;

			// the following saves some writing
			double expEtaSum = exp(-QZ * trigFactor2) + EDIP_eta;
			

			double dV3_drij_over_rij = - threeBodyRadial * threeBodyAngular
				* EDIP_gamma / ((r_ij - EDIP_a) * (r_ij - EDIP_a) * r_ij);
			double dV3_drik_over_rik = - threeBodyRadial * threeBodyAngular
				* EDIP_gamma / ((r_ik - EDIP_a) * (r_ik - EDIP_a) * r_ik);
			double dV3_dCosTheta = threeBodyRadial * EDIP_lambda * 2 * QZ * trigFactor * expEtaSum;
			
			// we have extra dependence on distances through Z 
			// (note Z does not depend on angles)
			double dV3_dZ = threeBodyRadial * EDIP_lambda * expEtaSum * (trigFactor2 * dQdZ + 2 * QZ * trigFactor * dtaudZ);
			
			// for getting forces on each of the three atoms need the derivatives of cosTheta wrt their positions
			double dCosThetadr_i[DIMENSION],dCosThetadr_j[DIMENSION],dCosThetadr_k[DIMENSION];
			
			for(coord=0; coord<DIMENSION; coord++)
			{
				dCosThetadr_i[coord] = cosTheta*(displacements[jx][coord]/displacementsSq[jx] + displacements[kx][coord]/displacementsSq[kx]) - (displacements[jx][coord]+displacements[kx][coord])/(r_ij*r_ik);
				dCosThetadr_j[coord] = displacements[kx][coord]/(r_ij*r_ik) - cosTheta*displacements[jx][coord]/displacementsSq[jx];
				dCosThetadr_k[coord] = displacements[jx][coord]/(r_ij*r_ik) - cosTheta*displacements[kx][coord]/displacementsSq[kx];
				
				forceI[coord] -= dV3_drij_over_rij * -displacements[jx][coord] + dV3_drik_over_rik * -displacements[kx][coord] + dV3_dCosTheta * dCosThetadr_i[coord];
				forceJ[coord] -= dV3_drij_over_rij * displacements[jx][coord] + dV3_dCosTheta * dCosThetadr_j[coord];
				forceK[coord] -= dV3_drik_over_rik * displacements[kx][coord] + dV3_dCosTheta * dCosThetadr_k[coord];
			}

			// we now need to calculate contributions for atom i and all 
			// its neighbors which come purely through the Z dependence
			for(int kk=0;kk<numberNeighbors;++kk)
			{
				if(dZidr1[kk] != 0.)
				{
					k=neighbors[kk]; 
					if(k<nAtoms) forceK = (*forces)[k];
					else forceK = dummyForce;
					double x_forceR = - dV3_dZ * dZidr1[kk] / r1[kk];
					for (coord=0; coord<DIMENSION; coord++)
					{
						forceI[coord] += -displacements[kk][coord] * x_forceR;
						forceK[coord] += displacements[kk][coord] * x_forceR; 
					}
				}
			}
			
			
			
		} // finished with atom k, given by index kx
		
		
	}	// Finished with atom j, given by index jx
	delete [] dZidr1;
	delete [] r1;
	delete [] expFactor2;
	delete [] expFactor3;
	return potentialEnergy;
}


/*!
  The EDIP potential energy for a set of atoms must be calculated atom by
  atom, so we simply put the single-atom code in the function which returns 
  the atomic energy, and call this function repeatedly to get the total energy.
  The overhead of the function call should be low compared to the work of
  calculating the atomic energy.

 */

double SiliconEDIPPotential::CalculateEnergy(ListOfAtoms *atoms)
{	
	ASSERT(atoms);
	if(nAtoms != atoms->GetNumber()) Allocate(atoms->GetNumber());
	double potentialEnergy = 0.0;
	for(int i=0;i<nAtoms;i++)
		potentialEnergy += CalculateAtomicEnergy(atoms, i);
		
	return potentialEnergy;
}


double SiliconEDIPPotential::CalculateAtomicEnergy(ListOfAtoms *atoms, int i)
{
	ASSERT(atoms);
	if(nAtoms != atoms->GetNumber()) Allocate(atoms->GetNumber());

	double potentialEnergy = 0.0;
	int numberNeighbors;
	bool arrayNotTooFull;
	int j,jx,k,kx;
	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	ASSERT(neighborLocator);
	int spacesLeftInArrays;

	spacesLeftInArrays = nAtoms;
	arrayNotTooFull	= neighborLocator->Neighbors(i,neighbors,numberNeighbors, 
												 displacements,displacementsSq,spacesLeftInArrays);
	ASSERT(arrayNotTooFull);

	double Zi = 0.; // coordination of atom i
	double *r1 = new double[numberNeighbors];
	double *expFactor2 = new double[numberNeighbors];
	double *expFactor3 = new double[numberNeighbors];
	
	for(jx=0;jx<numberNeighbors;++jx)
	{
		r1[jx] = sqrt(displacementsSq[jx]);
		expFactor2[jx] = exp(EDIP_sigma / ( r1[jx] - EDIP_a ) );
		expFactor3[jx] = exp(EDIP_gamma / ( r1[jx] - EDIP_a ) );
		
		if(r1[jx] < EDIP_c)
			Zi += 1.;
		else 
		{
			double x = (r1[jx] - EDIP_c)/(EDIP_a - EDIP_c); 
			Zi += exp(EDIP_alpha/(1 - 1./(x*x*x)));
		}
	}
	ASSERT(Zi <= numberNeighbors);
	double pZ = exp(-EDIP_beta * Zi * Zi);
	
	for(jx=0;jx<numberNeighbors;++jx)
	{
		j=neighbors[jx];
		// because we got the neighbors and not just the half neighbors, we need the following if to avoid double counting the pair part.
		double pairEnergy = EDIP_A * (pow(EDIP_B/r1[jx], EDIP_rho) - pZ)
			* expFactor2[jx];
		potentialEnergy += pairEnergy;
		
		// 3 body contributions
		for(kx=0;kx<jx;++kx)	
		{
			k=neighbors[kx];
			// radial part
			double r_ij = r1[jx];
			double r_ik = r1[kx];
			double threeBodyRadial = expFactor3[jx] * expFactor3[kx];
			
			// angular part; first calculate the angle
			double cosTheta = 0;
			for(int coord=0;coord<DIMENSION;++coord)
				cosTheta += displacements[jx][coord]*displacements[kx][coord];
			
			cosTheta /= (r_ij * r_ik);
			
			// tau function
			double tauZ = u1 + u2 * (u3 * exp(-u4 * Zi) - exp(-2 * u4 * Zi));
			// Q function
			double QZ = EDIP_Q0 * exp(-EDIP_mu * Zi);
			
			double trigFactor = cosTheta + tauZ;
			double trigFactor2 = trigFactor * trigFactor;
	

			double threeBodyAngular = EDIP_lambda * (1 - exp(-QZ * trigFactor2) + EDIP_eta * QZ * trigFactor2);
			
			potentialEnergy += threeBodyRadial * threeBodyAngular;
			
		} // finished with atom k, given by index kx

		
	}	// Finished with atom j, given by index jx
    
	delete [] r1;
	delete [] expFactor2;
	delete [] expFactor3;
	return potentialEnergy;
	
}

/*!
  As in the force calculations, we separate the contributions from the explicit
  dependence on r_ij and r_ik from the contributions which arise through the
  Z-depdence of the potential

 */

void SiliconEDIPPotential::CalculateStressTensor(ListOfAtoms *atoms, int i, double stress[DIMENSION*(DIMENSION+1)/2])
{
	if(nAtoms != atoms->GetNumber()) Allocate(atoms->GetNumber());
	for( int p = 0; p < (DIMENSION+1)*DIMENSION/2; ++p )
		stress[p] = 0.;

	StressTensorKernel(atoms, i, stress);

	// Add the dynamic part and scale with the atomic volume
	double latticeConst = (4./sqrt(3.)) * GetLengthScale();
	double invvol = 8./(latticeConst * latticeConst * latticeConst); // 2.59 is equilibrium lattice constant. 8 atoms per unit cell in a diamond lattice. Assuming D=3 here.
	double mass = atoms->GetMass();
	const double *veloCurrent = (*(atoms->GetCartesianVelocities()))[i] ;
	for(int alpha=0; alpha<DIMENSION; ++alpha)
		for(int beta=alpha; beta<DIMENSION; ++beta)
		{
			int ii = stresscomp[alpha][beta];
			stress[ii] -= veloCurrent[alpha]*veloCurrent[beta]*mass;
			stress[ii] *= invvol;
		}

}

void SiliconEDIPPotential::CalculateStressTensor(ListOfAtoms *atoms, DMArray<double, DIMENSION*(DIMENSION+1)/2> *stress)
{
	if(nAtoms != atoms->GetNumber()) Allocate(atoms->GetNumber());
	int i;
	stress->Zero();
	
	// the main work is done by StressTensorKernel which deals with one atom at a time
	DMArrayIterator<double,(DIMENSION+1)*DIMENSION/2> stressIter(stress); 
	for( i = 0, stressIter.Init(); i < nAtoms; ++i, stressIter.Next() )
	{
		double *current = stressIter.Current();
		StressTensorKernel(atoms,i, current);
	}


	// Add the dynamic part and scale with the atomic volume
	double latticeConst = (4./sqrt(3.)) * GetLengthScale();
	double invvol = 8./(latticeConst * latticeConst * latticeConst); // assuming diamond lattice (8 atoms per unit cell) and DIMENSION = 3 of course.
	double mass = atoms->GetMass();
	
	DMArrayIterator<double, DIMENSION> velIter(atoms->GetCartesianVelocities());
	for( velIter.Init(), stressIter.Init(); !velIter.IsDone(); velIter.Next(), stressIter.Next() )
	{
		const double *veloCurrent = velIter.Current();
		double *stressCurrent = stressIter.Current();
		for(int alpha=0; alpha<DIMENSION; ++alpha)
			for(int beta=alpha; beta<DIMENSION; ++beta)
			{
				int ii = stresscomp[alpha][beta];
				stressCurrent[ii] -= veloCurrent[alpha]*veloCurrent[beta]*mass;
				stressCurrent[ii] *= invvol;
			}
		
	}
}



void SiliconEDIPPotential::StressTensorKernel(ListOfAtoms *atoms, int i, double stress[(DIMENSION+1)*DIMENSION/2])
{
	int jx,kx,ii,coord,alpha,beta;
	
	NeighborLocator *neighborLocator = atoms->GetNeighborLocator();
	int spacesLeftInArrays, numberNeighbors;
	spacesLeftInArrays = nAtoms;

	bool arrayNotTooFull = neighborLocator->Neighbors(i,neighbors,numberNeighbors, 
													  displacements,displacementsSq,spacesLeftInArrays);
	ASSERT(arrayNotTooFull);

	
	// evaluate coordination and its derivatives
	double Zi = 0.; // coordination of atom i
	double *dZidr1 = new double[numberNeighbors];
	double *r1 = new double[numberNeighbors];
	double *expFactor2 = new double[numberNeighbors];
	double *expFactor3 = new double[numberNeighbors];
	
	for(jx=0;jx<numberNeighbors;++jx)
	{
		r1[jx] = sqrt(displacementsSq[jx]);
		expFactor2[jx] = exp(EDIP_sigma / ( r1[jx] - EDIP_a ) );
		expFactor3[jx] = exp(EDIP_gamma / ( r1[jx] - EDIP_a ) );
		
		if(r1[jx] < EDIP_c)
		{
			Zi += 1.;
			dZidr1[jx] = 0.;
		}
		else 
		{
			double x = (r1[jx] - EDIP_c)/(EDIP_a - EDIP_c); 
			double x3 = x*x*x;
			double x4 = x3 * x;
			Zi += exp(EDIP_alpha/(1 - 1./(x3)));
			dZidr1[jx] = 3 * EDIP_alpha * exp(EDIP_alpha/(1 - 1./x3))/ ((1 - 1./x3) * (1 - 1./x3) * x4) / (EDIP_a - EDIP_c);
		}
	} // end initial loop over neighbors to calculate Z and dZdr1	
		
		double pZ = exp(-EDIP_beta * Zi * Zi);
		double dpdZ = -2. * EDIP_beta * Zi * pZ;


	for(jx=0;jx<numberNeighbors;jx++)
	{
		// pair contributions
		// forceR is -(dE/dr)/r. The minus sign mostly cancels minus signs that come from the derivatives
		double forceR = EDIP_A * ( 
			pow(EDIP_B/r1[jx], EDIP_rho) * EDIP_rho / r1[jx]
			+ (pow(EDIP_B/r1[jx], EDIP_rho) - pZ) 
			* (EDIP_sigma/((r1[jx]-EDIP_a)*(r1[jx]-EDIP_a)))
			)
			* expFactor2[jx] / r1[jx];
		
		
		for(alpha=0; alpha<DIMENSION; ++alpha)
			for(beta=alpha; beta<DIMENSION; ++beta)
			{
				ii = stresscomp[alpha][beta];
				
				stress[ii] += - forceR*displacements[jx][alpha]*displacements[jx][beta];
				// leaving out the factor 0.5 because in the usual pair
				// potential case this divides the contribution from a 
				// bond between the two atoms but here there are
				// separate and different terms for each of the atoms
				// anyway
			}


		// because the "2-body" terms actually depend through Z on the 
		// positions of other atoms, we have contributions from this 
		// dependence to the forces on them (and i and j) too.
		for(int kk=0;kk<numberNeighbors;++kk)
		{
			if(dZidr1[kk] != 0.)
			{
				double x_forceR = EDIP_A * dpdZ * expFactor2[jx] * dZidr1[kk] /r1[kk];
				for(alpha=0; alpha<DIMENSION; ++alpha)
					for(beta=alpha; beta<DIMENSION; ++beta)
					{
						ii = stresscomp[alpha][beta];
						stress[ii] +=  -x_forceR * displacements[kk][alpha]*displacements[kk][beta];
						
					}
			}
		}
		



		// three-body contributions
		for(kx=0;kx<jx;++kx)	
		{
			// calculate the angle
			double cosTheta = 0;
			for(coord=0;coord<DIMENSION;++coord)
				cosTheta += displacements[jx][coord] *
					displacements[kx][coord];
			double r_ij = r1[jx];
			double r_ik = r1[kx];
			cosTheta /= (r_ij * r_ik);
			
			double threeBodyRadial = expFactor3[jx] * expFactor3[kx];
			// tau function
			double tauZ = u1 + u2 * (u3 * exp(-u4 * Zi) - exp(-2 * u4 * Zi));
			double dtaudZ = u2 * (- u3 * u4 * exp(-u4 * Zi) + 2 * u4 * exp(-2 * u4 * Zi));
			double trigFactor = cosTheta + tauZ;
			double trigFactor2 = trigFactor * trigFactor;
			
			// Q function
			double QZ = EDIP_Q0 * exp(-EDIP_mu * Zi);
			double dQdZ = -EDIP_mu * QZ;
			
			double threeBodyAngular = EDIP_lambda * (1 - exp(-QZ * trigFactor2) + EDIP_eta * QZ * trigFactor2);
			// the following saves some writing
			double expEtaSum = exp(-QZ * trigFactor2) + EDIP_eta;
			
			double dV3_drij_over_rij = - threeBodyRadial * threeBodyAngular
				* EDIP_gamma / ((r_ij - EDIP_a) * (r_ij - EDIP_a) * r_ij);
			double dV3_drik_over_rik = - threeBodyRadial * threeBodyAngular
				* EDIP_gamma / ((r_ik - EDIP_a) * (r_ik - EDIP_a) * r_ik);
			double dV3_dCosTheta = threeBodyRadial * EDIP_lambda * 2 * QZ * trigFactor * expEtaSum;
			
			// we have extra dependence on distances through Z 
			// (note Z does not depend on angles)
			double dV3_dZ = threeBodyRadial * EDIP_lambda * expEtaSum * (trigFactor2 * dQdZ + 2 * QZ * trigFactor * dtaudZ);
			

			double dV3_dr_ij[DIMENSION],dV3_dr_ik[DIMENSION], dCosTheta_drij[DIMENSION], dCosTheta_drik[DIMENSION];
			for(coord=0;coord<DIMENSION;++coord)
			{
				dCosTheta_drij[coord] = displacements[kx][coord]/(r_ij*r_ik) - cosTheta*displacements[jx][coord]/(r_ij*r_ij);
				dCosTheta_drik[coord] = displacements[jx][coord]/(r_ij*r_ik) - cosTheta*displacements[kx][coord]/(r_ik*r_ik)
;
				dV3_dr_ij[coord] = dV3_drij_over_rij * displacements[jx][coord] + dV3_dCosTheta * dCosTheta_drij[coord];
				dV3_dr_ik[coord] = dV3_drik_over_rik * displacements[kx][coord] + dV3_dCosTheta * dCosTheta_drik[coord];

			}

			for(alpha=0; alpha<DIMENSION; ++alpha)
				for(beta=alpha; beta<DIMENSION; ++beta)
				{
					ii = stresscomp[alpha][beta];
					
					stress[ii] += dV3_dr_ij[alpha]*displacements[jx][beta] + dV3_dr_ik[alpha]*displacements[kx][beta];
				}

			// we now need to calculate contributions which come from the Z 
			// dependence
			for(int kk=0;kk<numberNeighbors;++kk)
			{
					if(dZidr1[kk] != 0.)
					{
						double x_forceR = - dV3_dZ * dZidr1[kk] / r1[kk];
						for(alpha=0; alpha<DIMENSION; ++alpha)
							for(beta=alpha; beta<DIMENSION; ++beta)
							{
								ii = stresscomp[alpha][beta];
								stress[ii] +=  -x_forceR * displacements[kk][alpha]*displacements[kk][beta];
								
							}
					}
			} // end Z-dep loop over neighbors kk (given second and third body)
		} // end loop over kx third body
	} // end loop over jx second body
	
	
}



////////////////////////////////////////////////////////////////////////////////
/// METHODS RELATED TO SERIALIZATION
////////////////////////////////////////////////////////////////////////////////


void SiliconEDIPPotential::Save(DMWriter *writer) const
{
	Potential::Save(writer);

	writer->PutDouble(EDIP_A, "EDIP_A");
	writer->PutDouble(EDIP_B, "EDIP_B");
	writer->PutDouble(EDIP_Q0, "EDIP_Q0");
	writer->PutDouble(EDIP_a, "EDIP_a");
	writer->PutDouble(EDIP_c, "EDIP_c");
	writer->PutDouble(EDIP_alpha, "EDIP_alpha");
	writer->PutDouble(EDIP_beta, "EDIP_beta");
	writer->PutDouble(EDIP_gamma, "EDIP_gamma");
	writer->PutDouble(EDIP_eta, "EDIP_eta");
	writer->PutDouble(EDIP_lambda, "EDIP_lambda");
	writer->PutDouble(EDIP_mu, "EDIP_mu");
	writer->PutDouble(EDIP_rho, "EDIP_rho");
	writer->PutDouble(EDIP_sigma, "EDIP_sigma");
}

void SiliconEDIPPotential::Load(DMReader *reader)
{
	Potential::Load(reader);

	CheckPotentialParameter(reader, EDIP_A, "EDIP_A");
	CheckPotentialParameter(reader, EDIP_B, "EDIP_B");
	CheckPotentialParameter(reader, EDIP_Q0, "EDIP_Q0");
	CheckPotentialParameter(reader, EDIP_a, "EDIP_a");
	CheckPotentialParameter(reader, EDIP_c, "EDIP_c");
	CheckPotentialParameter(reader, EDIP_alpha, "EDIP_alpha");
	CheckPotentialParameter(reader, EDIP_beta, "EDIP_beta");
	CheckPotentialParameter(reader, EDIP_gamma, "EDIP_gamma");
	CheckPotentialParameter(reader, EDIP_eta, "EDIP_eta");
	CheckPotentialParameter(reader, EDIP_lambda, "EDIP_lambda");
	CheckPotentialParameter(reader, EDIP_mu, "EDIP_mu");
	CheckPotentialParameter(reader, EDIP_rho, "EDIP_rho");
	CheckPotentialParameter(reader, EDIP_sigma, "EDIP_sigma");
}



