// SiliconEDIPPotential.h: interface for the SiliconEDIPPotential class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef SILICONEDIPPOTENTIAL_H
#define SILICONEDIPPOTENTIAL_H

#include "DMFirstHeader.h"

#include "DigitalMaterial.h"
#include "Potential.h"

//! Silicon potential with environmentally dependent 2 and 3 body terms  developed by Justo, Kaxiras et al.

/*!
  The environmental dependence means that there are effectively higher order
  than 3 body terms present. The potential is described in detail in 
  Justo et al, Phys Rev B 58, 2539 (1998)
  


  \date August 8, 2001
  \author Nicholas Bailey

 */

class SiliconEDIPPotential : public Potential
{
public:
	SiliconEDIPPotential();
	virtual ~SiliconEDIPPotential();
	void Allocate(int nAtoms);
	void CalculateForces(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces);
	//! forces due to the term which gives the atomic energy on atom i
	double AtomicForcesEnergy(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces, int i);
	double CalculateEnergy(ListOfAtoms *atoms);

	double CalculateForcesAndEnergy(ListOfAtoms *atoms, DMArray<double, DIMENSION> *forces);

	double CalculateAtomicEnergy(ListOfAtoms *atoms, int i);
	
	void CalculateStressTensor(ListOfAtoms *atoms, int i, double stress[DIMENSION*(DIMENSION+1)/2]);
	void CalculateStressTensor(ListOfAtoms *atoms, DMArray<double, DIMENSION*(DIMENSION+1)/2> *stress);
	

	double GetLengthScale() const
	{
		return 2.35147433;
	}

	double GetEnergyCurvatureScale() const
	{
		return 0.; // needs to be determined.
	}

	void Save(DMWriter *writer) const;
	void Load(DMReader *reader);
	//! Get the string name of the class for saving or loading.
	const std::string &GetType() const
	{
		return className;
	}

private:
#ifndef SWIG	
	void StressTensorKernel(ListOfAtoms *atoms, int i, double stress[(DIMENSION+1)*DIMENSION/2]);
	int         nAtoms;
	double      *displacementsSq;
	double      (*displacements)[DIMENSION];
	int         *neighbors;


	// parameters for the EDIP potential
	const double EDIP_A;
	const double EDIP_B;
	const double EDIP_Q0; 
	const double EDIP_a;
	const double EDIP_c;
	const double EDIP_alpha;
	const double EDIP_beta;
	const double EDIP_gamma;
	const double EDIP_eta;
	const double EDIP_lambda;
	const double EDIP_mu;
	const double EDIP_rho;
	const double EDIP_sigma;

	const double u1;
	const double u2;
	const double u3;
	const double u4;
private:
	static const std::string className;

#endif // SWIG 
};

#endif // !defined(SILICONEDIPPOTENTIAL_H)
