// SortAtoms.h: interface for the SortAtoms class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef SORTATOMS_H
#define SORTATOMS_H

#include "DMFirstHeader.h"

// #include "ListOfAtoms.h"
#include "DMArray.h"
#include <vector>
#include <algorithm>
#include <assert.h>
#include <float.h>
#include "IEEEFloatingPoint.h"

#include "DMLastHeader.h"

// This class was written to sort the atoms from back to front for plotting
// It uses the vector standard-template library sort function, and passes back
// the array of sorted integers pointing to the atoms in increasing order of height z.

class SortAtoms  
{
public:
	SortAtoms(int coordToSort = DIMENSION  -1) {this->coordToSort = coordToSort;}

	virtual ~SortAtoms() {}

	// Matt Kuntz suggested this method.

	// The vector class needs to know when one atom is higher than another. 
	// We don't want to redefine ">" and "<" for integers, so we make a function
	// The vector class needs a class with () overloaded to use for sorting
	// We use the SortAtoms class as the sorting function too!
	bool operator () (int atomA, int atomB)
	{
		return ((*positions)[atomA][coordToSort]  < (*positions)[atomB][coordToSort]);
	}

	// We use the standard-template library sort function to sort the atoms

	const int *Sort(const DMArray<double, DIMENSION> *positions)
	{
		this->positions = positions;
		int atom, nAtoms = positions->Size();

		// this->atoms = atoms;
		// positions = atoms->GetCartesianPositions() ;

		

		sortedAtomNumbers.resize(0);	// Empty vector; fill with atoms that aren't at infinity or bad
		badAtomNumbers.resize(0);		// Store atoms at infinity: Microsoft sort dies with -1#IND000
		for( atom = 0; atom < nAtoms; ++atom )
		{
			if(finite((*positions)[atom][coordToSort]))
			{
				sortedAtomNumbers.push_back(atom);
			}
			else
			{
				assert(false);	// In debug mode, die here!
				badAtomNumbers.push_back(atom);		// Draws bad atoms on top of others ...
			}
		}
		std::sort(sortedAtomNumbers.begin(), sortedAtomNumbers.end(), *this);

		for (atom=0; atom<badAtomNumbers.size(); atom++)
		{
			sortedAtomNumbers.push_back(badAtomNumbers[atom]);
		}
		// We'd like to return an array of integers, rather than a vector, since 
		// naive users, other languages, and wrapping programs like the current version of SWIG
		// know about vectors.
		// We're not sure that vectors are guaranteed to be stored in contiguous blocks of memory
		// Of course, it might be contiguous in debug mode and not in release mode, but we hope not.

		assert(sortedAtomNumbers.end() - sortedAtomNumbers.begin() == nAtoms);

		return (const int *) &(sortedAtomNumbers[0]);
	}

protected:
	
	// const ListOfAtoms *atoms;

	const DMArray<double,DIMENSION> *positions;

	int coordToSort; // sort by which coord (set in constructor, default DIMENSION -1)

	// We use a vector of ints labeling the atoms. This could have been a vector
	// of pointers to the atom coordinates just as well.
	std::vector<int> sortedAtomNumbers;
	std::vector<int> badAtomNumbers;
};

#endif // !defined(SORTATOMS_H)
