// AnisotropicDislocationMaker.cpp: implementation of the AnisotropicDislocationMaker class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "DMFirstHeader.h"
#include <fstream>
#include "DigitalMaterial.h"

#include "AnisotropicDislocationMaker.h"
#include "DMArray.h"

#include "DMLastHeader.h"

using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

AnisotropicDislocationMaker::AnisotropicDislocationMaker(const double burgers[DIMENSION],const double direction[DIMENSION],double center[DIMENSION],double *elasticConstants, double thetaLowerCutoff):DislocationMaker(burgers,direction,center,0.0,thetaLowerCutoff)
{  
	for(int a=0;a<3;++a)
		this->elasticConstants[a] = elasticConstants[a]; // Elastic constants

	// nu has no role in the anisotropic case, so is set to zero (see last argument to DislocationMaker constructor)


	// should the following be put into UpdateMemberVariables?

	double angle = 0.5*acos((2.*elasticConstants[1]*elasticConstants[2] + 
							 elasticConstants[1]*elasticConstants[1] - 
							 elasticConstants[0]*elasticConstants[0]) /
							(2.*elasticConstants[0]*elasticConstants[2]));	

	anisotropyPhaseFactor = complex<double>(cos(angle), sin(angle));

	yCompFactor = (elasticConstants[2]*anisotropyPhaseFactor + 
				   elasticConstants[0]*conj(anisotropyPhaseFactor))/ 
		(elasticConstants[1]+elasticConstants[2]);

	complex<double>I(0,1);
	complex<double>phaseFactorSq(cos(2.*angle),sin(2.*angle));

	logCoefficient = -0.25*edge*(phaseFactorSq*elasticConstants[0] - 
								 elasticConstants[1])/
		elasticConstants[0]/sin(2.*angle)/PI ;

}

AnisotropicDislocationMaker::~AnisotropicDislocationMaker()
{

}


void AnisotropicDislocationMaker::ElasticDisplacement(const double position[DIMENSION], double displacement[DIMENSION])
{
	int coord,coord2;
	double displacementLocalBasis[2];


	CalculateGeometry(position);

	complex<double> z1(orient1Comp,0);
	complex<double> z2(z1);
	z1 += orient2Comp*anisotropyPhaseFactor;
	z2 -= orient2Comp*anisotropyPhaseFactor;

	complex<double> log1 = log(z1);
	complex<double> log2 = log(z2);
	

	if(theta >= PI) // ie the physical angle is in the range which corresponds to having had 2 PI added onto the angle (so it would have been initially in the range -PI to -0.5*PI)
	{
		if(imag(log1) < 0.) // if theta is Pi imag(log1) could be Pi, resulting in it becoming 3 Pi
			log1 += complex<double>(0,2*PI);
		if(imag(log2) > 0.) // not sure about this
			log2 -= complex<double>(0,2*PI);
	}


	if(rPerp > cutoff)
	{
		// calculate the screw displacement which is as in isotropic case
		// since we are considering the dislocation line to be parallel
		// to a cubic axis
	    for(coord=0;coord<DIMENSION;coord++)
			displacement[coord] += screw*direction[coord]*theta/2./PI;

		// then calculate the edge displacement
		displacementLocalBasis[0] = real(logCoefficient * (log1 - log2)); 
		// minus sign for second term because for that the D coefficient
		// should be divided by - 2 PI i

		displacementLocalBasis[1] = real(-yCompFactor * 
										 logCoefficient * (log1 + log2));
		// plus sign because yCompFactor for the second term is minus that for the first (A_2 versus A_1 in Eshelby paper)

		for(coord=0;coord<DIMENSION;++coord)
			for(coord2=0;coord2<2;++coord2)// there are only two orientingVectors
				displacement[coord] += orientingVector[coord2][coord]*displacementLocalBasis[coord2];
		
	}
}


void AnisotropicDislocationMaker::GradientWRTPosition(const double position[DIMENSION],double gradientWRTPosition[DIMENSION][DIMENSION])
{
    
	double gradientLocalBasis[2][2];
	
	CalculateGeometry(position);

	complex<double> z1(orient1Comp,0);
	complex<double> z2(z1);
	z1 += orient2Comp*anisotropyPhaseFactor;
	z2 -= orient2Comp*anisotropyPhaseFactor;

	complex<double> z1inv(1./z1);
	complex<double> z2inv(1./z2);
	
    
	if (rPerp > cutoff) 
	{ 
		// contribution from screw part (same as isotropic--see above)
		for(int coord1=0;coord1<DIMENSION;++coord1)
			for(int coord2=0;coord2<DIMENSION;++coord2)
				gradientWRTPosition[coord1][coord2] += screw*direction[coord1]*rPerpInverse*(-sinTheta*orientingVector[0][coord2] + cosTheta*orientingVector[1][coord2])/(2.*PI);
		
		// contribution from edge part

		gradientLocalBasis[0][0] = real(logCoefficient * (z1inv - z2inv));
		gradientLocalBasis[0][1] = real(logCoefficient * anisotropyPhaseFactor * (z1inv + z2inv));
		gradientLocalBasis[1][0] = real(-yCompFactor * logCoefficient * (z1inv + z2inv));
		gradientLocalBasis[1][1] = real(-yCompFactor * logCoefficient * anisotropyPhaseFactor * (z1inv - z2inv));


		for(int a=0;a<DIMENSION;++a)
			for(int b=0;b<DIMENSION;++b)
				for(int c=0;c<2;++c)
					for(int d=0;d<2;++d)
					{
						gradientWRTPosition[a][b] += gradientLocalBasis[c][d]*orientingVector[c][a]*orientingVector[d][b];
					}

		
	}       

  
}
