// AnisotropicMultipoleField.cpp: implementation of the AnisotropicMultipoleField class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "DMFirstHeader.h"
#include <fstream>
#include "DigitalMaterial.h"

#include "AnisotropicMultipoleField.h"
#include "DMArray.h"

#include "DMLastHeader.h"

using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

AnisotropicMultipoleField::AnisotropicMultipoleField(int maximumOrder,const double direction[DIMENSION], double center[DIMENSION],double *elasticConstants):MultipoleField(maximumOrder,direction,center)
{
	for(int a=0;a<3;++a)
		this->elasticConstants[a] = elasticConstants[a]; // Elastic constants


	double angle = 0.5*acos((2. * elasticConstants[1]*elasticConstants[2] + 
							 elasticConstants[1]*elasticConstants[1] - 
							 elasticConstants[0]*elasticConstants[0]) /
							(2.*elasticConstants[0]*elasticConstants[2]));	

	anisotropyPhaseFactor = complex<double>(cos(angle), sin(angle));

	yCompFactor = (elasticConstants[2]*anisotropyPhaseFactor + 
				   elasticConstants[0]*conj(anisotropyPhaseFactor)) /
		(elasticConstants[1]+elasticConstants[2]);

	
	cutoff = 0.;
	
}

AnisotropicMultipoleField::~AnisotropicMultipoleField()
{


}



void AnisotropicMultipoleField::ElasticDisplacement(const double position[DIMENSION], double displacement[DIMENSION])
{
	int coord,coord2;
	double displacementLocalBasis[2] = {0.,0.};
	
	CalculateGeometry(position);

	complex<double> z1(orient1Comp,0);
	complex<double> z2(z1);
	z1 += orient2Comp*anisotropyPhaseFactor;
	z2 -= orient2Comp*anisotropyPhaseFactor;
	complex<double> z1inv(1./z1);
	complex<double> z2inv(1./z2);
	
   	complex<double> z1inv_n(z1inv),z2inv_n(z2inv);


	if(rPerp > cutoff)
		for(int orderIdx=1;orderIdx<=maximumOrder;++orderIdx)
		{
			complex<double> complexCoeff1(coefficients[orderIdx-1][0],coefficients[orderIdx-1][1]);
			complex<double> complexCoeff2(coefficients[orderIdx-1][2],coefficients[orderIdx-1][3]);
			displacementLocalBasis[0] += real(complexCoeff1*z1inv_n + complexCoeff2*z2inv_n);
			displacementLocalBasis[1] += real(-yCompFactor*(complexCoeff1*z1inv_n - complexCoeff2*z2inv_n));
			z1inv_n *= z1inv;
			z2inv_n *= z2inv;
		}
	
	for(coord=0;coord<DIMENSION;++coord)
		for(coord2=0;coord2<2;++coord2)// there are only two orientingVectors
			displacement[coord] += orientingVector[coord2][coord]*displacementLocalBasis[coord2];

}


void AnisotropicMultipoleField::GradientWRTParameters(const double position[DIMENSION], double (*displacementGradient)[DIMENSION])
{
	int orderIdx,coord,coord2;
	double (*gradientLocalBasis)[2] = new double[4*maximumOrder][2];


	for(orderIdx=0;orderIdx<4*maximumOrder;++orderIdx)
	{
		for(coord=0;coord<DIMENSION;++coord)
			displacementGradient[orderIdx][coord] = 0.;    
		for(coord=0;coord<2;++coord) // in a quasi 2D clas, the local basis is 2-dimensional,
			// even when DIMENSION is 3
			gradientLocalBasis[orderIdx][coord] = 0;
	}


	
	CalculateGeometry(position);

	complex<double> z1(orient1Comp,0);
	complex<double> z2(z1);
	z1 += orient2Comp*anisotropyPhaseFactor;
	z2 -= orient2Comp*anisotropyPhaseFactor;
	complex<double> z1inv(1./z1);
	complex<double> z2inv(1./z2);


 	complex<double> z1inv_n(z1inv),z2inv_n(z2inv);
	complex<double> I(0.,1.);
	
	if(rPerp > cutoff)
		for(orderIdx=1;orderIdx<=maximumOrder;++orderIdx)
		{
			gradientLocalBasis[4*(orderIdx-1)][0] += real(z1inv_n);
			gradientLocalBasis[4*(orderIdx-1)+1][0] += real(I*z1inv_n);
			gradientLocalBasis[4*(orderIdx-1)+2][0] += real(z2inv_n);
			gradientLocalBasis[4*(orderIdx-1)+3][0] += real(I*z2inv_n);
			
			gradientLocalBasis[4*(orderIdx-1)][1] += real(-yCompFactor*z1inv_n);
			gradientLocalBasis[4*(orderIdx-1)+1][1] += real(-yCompFactor*I*z1inv_n);
			gradientLocalBasis[4*(orderIdx-1)+2][1] += real(yCompFactor*z2inv_n);
			gradientLocalBasis[4*(orderIdx-1)+3][1] += real(yCompFactor*I*z2inv_n);
			z1inv_n *= z1inv;
			z2inv_n *= z2inv;
		}
	
	
	for(int paramIdx=0;paramIdx<4*maximumOrder;++paramIdx)
		for(coord=0;coord<DIMENSION;++coord)
			for(coord2=0;coord2<2;++coord2)// there are only two orientingVectors
				displacementGradient[paramIdx][coord] += orientingVector[coord2][coord]*gradientLocalBasis[paramIdx][coord2];	

	delete [] gradientLocalBasis;
}

void AnisotropicMultipoleField::GradientWRTPosition(const double position[DIMENSION],double gradientWRTPosition[DIMENSION][DIMENSION])
{
	int coord,coord2;
	double gradientLocalBasis[2][2];
	for(coord=0;coord<2;++coord)
		for(coord2=0;coord2<2;++coord2)
			gradientLocalBasis[coord][coord2] = 0.;
	
	CalculateGeometry(position);
	complex<double> z1(orient1Comp,0);
	complex<double> z2(z1);
	z1 += orient2Comp*anisotropyPhaseFactor;
	z2 -= orient2Comp*anisotropyPhaseFactor;
	complex<double> z1inv(1./z1);
	complex<double> z2inv(1./z2);

	complex<double> z1inv_nP1(z1inv*z1inv),z2inv_nP1(z2inv*z2inv);
 
    
	if (rPerp > cutoff) 
		for(int orderIdx=1;orderIdx<=maximumOrder;++orderIdx)
		{ 
			complex<double> complexCoeff1(coefficients[orderIdx-1][0],coefficients[orderIdx-1][1]);
			complex<double> complexCoeff2(coefficients[orderIdx-1][2],coefficients[orderIdx-1][3]);
			
			gradientLocalBasis[0][0] += -orderIdx*real(complexCoeff1*z1inv_nP1 + complexCoeff2*z2inv_nP1);
			gradientLocalBasis[0][1] += -orderIdx*real(anisotropyPhaseFactor*(complexCoeff1*z1inv_nP1 - complexCoeff2*z2inv_nP1));
			gradientLocalBasis[1][0] += orderIdx*real(yCompFactor*(complexCoeff1*z1inv_nP1 - complexCoeff1*z2inv_nP1));
			gradientLocalBasis[1][1] += orderIdx*real(yCompFactor*anisotropyPhaseFactor*(complexCoeff1*z1inv_nP1 + complexCoeff1*z2inv_nP1));
			z1inv_nP1 *= z1inv;
			z2inv_nP1 *= z2inv;		
		}
	
	for(int a=0;a<DIMENSION;++a)
		for(int b=0;b<DIMENSION;++b)
			for(int c=0;c<2;++c)
				for(int d=0;d<2;++d)
				{
					gradientWRTPosition[a][b] += gradientLocalBasis[c][d]*orientingVector[c][a]*orientingVector[d][b];
				}
	
}
