// ConfigurationShifter.h: interface for the ConfigurationShifter class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef CONFIGURATIONSHIFTER_H
#define CONFIGURATIONSHIFTER_H

#include "DigitalMaterial.h"

#include "Transformer.h"

class CellNeighborLocator;
class SimpleNeighborLocator;
#include "DMArray.h"

//! This transformer shifts the configuration by a certain displacement vector up to possible permutations of the atoms, to minimise movement of individual atoms

/**
 * The transform function loops through the old positions and adds the
 * displacement vector to each one, giving a position in the new configuration.
 * Then it uses a NeighborLocator to find the closest atom in the old
 * configuration to this new position. It stores the new position in a separate
 * positions array (newPositions) at the location corresponding to this nearest
 * atom. Having done this for all atoms it calls SetCartesianPositions passing  * newPositions as the argument. Along the way the routine checks that there is
 * no double assignment of atoms.
 *
 * \ingroup moleculardynamics
 * \author nbailey
 * \date May 15, 2000
 */


class ConfigurationShifter : public Transformer
{
public:
	ConfigurationShifter(double shift[DIMENSION],double transWeightFactor = 1.0);
	virtual ~ConfigurationShifter();

	void SetShift(const double shift[DIMENSION]);
	const double *GetShift() const {return shift;}

	const double *GetCOMShift() const{return COMShift;}

	virtual void Transform(ListOfAtoms *atoms);
	class Map_error
	  {
	  public:
	    Map_error(int nn,int aa,int ii): nearest(nn),a(aa),i(ii){}
	    int nearest,a,i;
	    double nearest_pos[DIMENSION];
	    double a_pos[DIMENSION];
	    double i_pos[DIMENSION];
	  };
	void SetTransWeightFactor(double twf) {transWeightFactor = twf;}
	double GetTransWeightFactor() const {return transWeightFactor;}
 private:
	void Allocate(int nAtoms, const DMArray<double, DIMENSION> *array);
	double shift[DIMENSION];
	double COMShift[DIMENSION];
	int nAtoms;
	double shiftMag;
	double transWeightFactor;
	double *displacementsSq;
	double (*displacements)[DIMENSION];
	int *neighbors;
	CellNeighborLocator *cellNL;
	SimpleNeighborLocator *simpleNL;
};

#endif // !defined(CONFIGURATIONSHIFTER_H)
