// DislocationMaker.cpp: implementation of the DislocationMaker class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "DMFirstHeader.h"
#include <fstream>
#include "DigitalMaterial.h"

#include "DislocationMaker.h"
#include "DMArray.h"

#include "DMLastHeader.h"


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

DislocationMaker::DislocationMaker(const double burgers[DIMENSION],const double direction[DIMENSION],double center[DIMENSION],const double nu,double thetaLowerCutoff):Quasi2DElasticField(direction,center,thetaLowerCutoff)
{  
	// nu is (E - 2mu)/2mu, where E is Young's modulus and mu is the shear modulus.
	assert((DIMENSION==2) || (DIMENSION==3));
	for(int coord=0; coord<DIMENSION; coord++)
		this->burgers[coord] = burgers[coord];

	this->nu = nu;
	cutoff = -1.0; // default is no cutoff
	UpdateMemberVariables();
}

DislocationMaker::~DislocationMaker()
{

}


void DislocationMaker::UpdateMemberVariables()
{
	int coord;
	double norm = 0., dotProduct = 0.;
	for(coord=0;coord<DIMENSION;coord++) 
	{
		norm += direction[coord]*direction[coord];
		dotProduct += burgers[coord]*direction[coord];
	}
	norm = sqrt(norm);
	if(DIMENSION == 2)
	{
		norm = 1.;
		screw = 0.;
	}
	else
	{
		assert(norm > 0.);
		screw = dotProduct/norm;
	}
	edge = 0.;
	for(coord=0;coord<DIMENSION;coord++)
	{
		if(DIMENSION==3)
		{	
			direction[coord] = direction[coord]/norm;
		}
		else
		{
			direction[coord] = 0.;
		}
		edgeComponent[coord] = burgers[coord] - screw*direction[coord];
		edge += edgeComponent[coord]*edgeComponent[coord];
	}
	edge = sqrt(edge);
	if(edge != 0.) // maybe should test for edge > some tolerance instead
	{
	    for(coord=0;coord<DIMENSION;++coord)
			orientingVector[0][coord] = edgeComponent[coord]/edge;
	}
	else // ie pure screw
	{ 
	    // choose two perpendicular vectors
	    for(coord=0;coord<DIMENSION;++coord)
		{
			orientingVector[0][coord] = orientingVector[1][coord] = 0.;
			orientingVector[coord][coord] = 1.;
		}
		
	    double dotProd1=0.,dotProd2=0.;
	    for(coord=0;coord<DIMENSION;++coord)
		{
			dotProd1 += direction[coord]*orientingVector[0][coord];
			dotProd2 += direction[coord]*orientingVector[1][coord];
		}
	    // we want the one which is most orthogonal to the direction (the case we're trying to avoid is choosing a vector parallel to direction because then we can't get an orthonormal triad)
	    if(fabs(dotProd2) < fabs(dotProd1))
		{
			for(coord=0;coord<DIMENSION;++coord)
				orientingVector[0][coord] = orientingVector[1][coord];
			dotProd1 = dotProd2;
			// now we deal only with orientingVector[0], so don't need to swap completely
		}
	    // Get the perpendicular component and normalize
	    double ov1Norm = 0.;
	    for(coord=0;coord<DIMENSION;++coord)
		{
			orientingVector[0][coord] -= dotProd1*direction[coord];
			ov1Norm += orientingVector[0][coord]*orientingVector[0][coord];
		}
	    for(coord=0;coord<DIMENSION;++coord)
			orientingVector[0][coord] /= sqrt(ov1Norm);
		
	}
	
	
	if(DIMENSION==3)
	{
	    
	    orientingVector[1][0] = direction[1]*orientingVector[0][2] - direction[2]*orientingVector[0][1]; 
	    orientingVector[1][1] = direction[2]*orientingVector[0][0] - direction[0]*orientingVector[0][2]; 
	    orientingVector[1][2] = direction[0]*orientingVector[0][1] - direction[1]*orientingVector[0][0];
	    
	}  
	else
	{
		orientingVector[1][0] = -orientingVector[0][1];
		orientingVector[1][1] =  orientingVector[0][0];
	}

	edge_over_2PI = edge/ (2.*PI);
	nu_factor = 1./(2*(1.-nu));
}


void DislocationMaker::ElasticDisplacement(const double position[DIMENSION], double displacement[DIMENSION])
{
    int coord;
    
    CalculateGeometry(position);

    
    if(rPerp > cutoff)
    {
		// calculate the screw displacement
		for(coord=0;coord<DIMENSION;coord++)
			displacement[coord] += screw*direction[coord]*theta/2./PI;
		
		// then calculate the edge displacement
		double thetaCorr = -thetaLowerCutoff - PI; // chosen so that the average displacement around a circle is zero
		for(coord=0;coord<DIMENSION;coord++)
			displacement[coord] += orientingVector[0][coord] *
				(theta + thetaCorr + cosTheta * sinTheta * nu_factor)
				* edge_over_2PI -
				orientingVector[1][coord] * edge_over_2PI *
				( (1 - 2. * nu) * log(rPerp * rPerp) +
				 cosTheta * cosTheta - sinTheta * sinTheta) * 0.5 * nu_factor;
    }
}


void DislocationMaker::GradientWRTPosition(const double position[DIMENSION],double gradientWRTPosition[DIMENSION][DIMENSION])
{
    
	double gradientLocalBasis[2][2];
	
	CalculateGeometry(position);
	
	if (rPerp > cutoff) 
	{ 
		//  Contributions from screw part
		for(int coord1=0;coord1<DIMENSION;++coord1)
			for(int coord2=0;coord2<DIMENSION;++coord2)
				gradientWRTPosition[coord1][coord2] += screw*direction[coord1]*rPerpInverse*(-sinTheta*orientingVector[0][coord2] + cosTheta*orientingVector[1][coord2])/(2.*PI);
		
		
		// Contributions from the edge part
		gradientLocalBasis[0][0] = -edge_over_2PI*(nu_factor*sinTheta*cos(2*theta) + sinTheta)*rPerpInverse;
		gradientLocalBasis[0][1] = -edge_over_2PI*(nu_factor*(sinTheta*sin(2*theta) - cosTheta) - cosTheta)*rPerpInverse;
		gradientLocalBasis[1][0] = -edge_over_2PI*nu_factor*(sinTheta*sin(2*theta) + (1.-2.*nu)*cosTheta)*rPerpInverse;
		gradientLocalBasis[1][1] = -edge_over_2PI*nu_factor*(-sinTheta*cos(2.*theta)-2.*nu*sinTheta)*rPerpInverse;
		
		for(int a=0;a<DIMENSION;++a)
			for(int b=0;b<DIMENSION;++b)
				for(int c=0;c<2;++c)
					for(int d=0;d<2;++d)
					{
						gradientWRTPosition[a][b] += gradientLocalBasis[c][d]*orientingVector[c][a]*orientingVector[d][b];
					}
		
		
	}         
}


void DislocationMaker::SetBurgers(const double burgers[DIMENSION])
{
    for(int coord=0; coord<DIMENSION;coord++) this->burgers[coord] = burgers[coord];
    UpdateMemberVariables();
}

void DislocationMaker::SetDirection(const double direction[DIMENSION])
{
	Quasi2DElasticField::SetDirection(direction);
	UpdateMemberVariables();
}

void DislocationMaker::SetCenter(const double center[DIMENSION])
{
	Quasi2DElasticField::SetCenter(center);
    UpdateMemberVariables();
}
