// EulerCoordinateTransformer.h: interface for the EulerCoordinateTransformer class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef EULERCOORDINATETRANSFORMER_H
#define EULERCOORDINATETRANSFORMER_H



#include "DMFirstHeader.h"
#include <math.h>
#include "DigitalMaterial.h"
#include <assert.h>

#include <vector>
#include <math.h>
#include <assert.h>

class BravaisLatticeWithBasis;
class BoundaryConditions;

#include "Transformer.h"
#include "ListOfAtoms.h"
#include "ElasticFieldTransformer.h"



//! The Transform function of this class self-consistently calculates the Eulerian displacements due to the ElasticDisplacement functions of a list of ElasticFieldTransformers

/*!
  
  The field are specified by a list of pointers to other 
  ElasticFieldTransformers. The user can use the InitialDisplace function to 
  make a Lagrangian displacement equal to the sum of displacmeents from the 
  individual fields, or use the Transform function which iterates until the
  total displacement field evaluated at the current position equals the actual
  displacement field from a lattice point (the nearest lattice point to the 
  current position minus the current displacement field!)

  Hence this can be used in two ways: one just to sum other elastic fields; one
  to apply the self-consistent Eulerian displacement

  
  \author nbailey
  \date 2001
  \ingroup moleculardynamics

 */


class EulerCoordinateTransformer : public ElasticFieldTransformer
{
public:
	EulerCoordinateTransformer(const double tolerance, BoundaryConditions *bc = 0);
	virtual ~EulerCoordinateTransformer();

	void Transform(ListOfAtoms *atoms);

	void InitialDisplace(ListOfAtoms *atoms);
	
	void SetUndeformedLattice(const BravaisLatticeWithBasis *undeformedLattice) {this->undeformedLattice = undeformedLattice;}
	
	void AddElasticField(ElasticFieldTransformer* strainField){strainTerms.push_back(strainField);}
	
	void ElasticDisplacement(const double position[DIMENSION], double displacement[DIMENSION]);
	void GradientWRTPosition(const double position[DIMENSION],  double gradientWRTPosition[DIMENSION][DIMENSION]);
	
	void CalculateDerivativeMatrix(const double position[DIMENSION],double derivativeMatrix[DIMENSION][DIMENSION]);
	
	void CurrentDisplacementField(const double position[DIMENSION], double displacement[DIMENSION]);


	double GetTolerance() const {return tolerance;}
	void SetTolerance(double t) {tolerance = t;}
	int GetMaxIterations() const {return maxIterations;}
	void SetMaxIterations(int n) {maxIterations = n;}
	
	// Accessing "return" information
	double GetFinalError() const {return error;}
	int GetNIterations() const {return nIterations;}

	void SetUndeformedPositions(const DMArray<double, DIMENSION> *currentPositions);
	const DMArray<double, DIMENSION> *GetUndeformedPositions() const { return undeformedPositions; }

private:
#ifndef SWIG
	std::vector<ElasticFieldTransformer *> strainTerms;
	const BravaisLatticeWithBasis *undeformedLattice;
	const DMArray<double, DIMENSION> *undeformedPositions;
	double tolerance;
	double error;
	int maxIterations;
	int nIterations;
	BoundaryConditions *bc;
#endif // SWIG
};

#endif // !defined(EULERCOORDINATETRANSFORMER_H)
