// NotchFieldDisplacer.cpp: implementation of the NotchFieldDisplacer class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "DMFirstHeader.h"
#include <fstream>
#include "DigitalMaterial.h"

#include "NotchFieldDisplacer.h"
#include "DMArray.h"

#include "DMLastHeader.h"


const double NotchFieldDisplacer::TINY = 1.0e-13;

using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

NotchFieldDisplacer::NotchFieldDisplacer(const double direction[DIMENSION],double center[DIMENSION], DMArray<double, DIMENSION> *strohVectorsReal, DMArray<double, DIMENSION> *strohVectorsImag, double strohEigenvaluesReal[DIMENSION], double strohEigenvaluesImag[DIMENSION], double notchAlphaAngle, double lambdaExponent[2], double stressIntensity[2]):Quasi2DElasticField(direction,center)
{  
	ASSERT (DIMENSION == 3);
	if(DIMENSION != 3)
		throw DMException("NotchFieldDisplacer","NotchFieldDisplacer","Attempt to create a NotchFieldDisplacer with DIMENSION not equal to 3");
	int a, coord;

	if(strohVectorsReal->Size() != 2*DIMENSION || strohVectorsImag->Size() != 2*DIMENSION)
		throw DMException("NotchFieldDisplacer","NotchFieldDisplacer","Wrong size strohVectors array passed");
	
	this->strohVectors = new complex<double>*[2*DIMENSION];
	
	for(a=0;a<2*DIMENSION;++a)
	{
		strohVectors[a] = new complex<double>[DIMENSION];
		
		for(coord=0;coord<DIMENSION;++coord)
			strohVectors[a][coord] = complex<double>((*strohVectorsReal)[a][coord],(*strohVectorsImag)[a][coord]);
	}
	for(coord=0;coord<DIMENSION;++coord)
		this->strohEigenvalues[coord] = complex<double>(strohEigenvaluesReal[coord], strohEigenvaluesImag[coord]);
	
	this->notchAlphaAngle = notchAlphaAngle;
	for(a = 0; a < 2; ++a )
	{
		this->stressIntensity[a] = stressIntensity[a];
		this->lambdaExponent[a] = lambdaExponent[a];
	}	

	cutoff = -1.0; // default is no cutoff
	UpdateMemberVariables();


}

NotchFieldDisplacer::~NotchFieldDisplacer()
{
	delete strohVectors;
}


void NotchFieldDisplacer::UpdateMemberVariables()
{
	// should set up the orienting vectors here....

	int alpha, mode, coord1, coord2;
	for(alpha=0;alpha<DIMENSION;++alpha)
		XiMinusAlpha[alpha] = cos(-notchAlphaAngle) + strohEigenvalues[alpha] * sin(-notchAlphaAngle);

	// next, construct the matrix K, which we need for the vector q
	double Kmatrix[DIMENSION][DIMENSION];
	for(coord1=0;coord1<DIMENSION;++coord1)
		for(coord2=0;coord2<DIMENSION;++coord2)
			Kmatrix[coord1][coord2] = 0.;

	for(alpha=0;alpha<DIMENSION;++alpha)
	{
		complex<double> XiAlpha = cos(notchAlphaAngle) + strohEigenvalues[alpha] * sin(notchAlphaAngle);
		complex<double> kernel = pow(XiAlpha,lambdaExponent[0])/pow(XiMinusAlpha[alpha],lambdaExponent[0]);
		for(coord1=0;coord1<DIMENSION;++coord1)
			for(coord2=0;coord2<DIMENSION;++coord2)
				Kmatrix[coord1][coord2] += 2 * real(kernel * strohVectors[alpha+DIMENSION][coord1] * strohVectors[alpha+DIMENSION][coord2]);// add DIMENSION to get the b Stroh vectors
	}
	// can check that the determinant of K is (close to) zero, or at 
	// least the determinant of the relevant submatrix
	// cout << "Det K is " << (Kmatrix[0][0] * Kmatrix[1][1] - Kmatrix[0][1] * Kmatrix[1][0]) << endl;
	
	// All we really need to compute individual displacements are the dot 
	// products of the b vectors with q. 

	complex<double> bTq[DIMENSION];

	if (0. <= notchAlphaAngle && notchAlphaAngle < PI - TINY)
	{
		// In this case the eigenvalues are distinct and the elements of K are
		// non-zero. Assuming lambda corresponding to mode 1 has been passed
		// we relate write q2 in terms of q1 and normalized q1 as in the paper
		// of Labossiere and Dunn 
		for( alpha = 0; alpha < DIMENSION; ++alpha )
			bTq[alpha] = strohVectors[alpha+DIMENSION][0]
				- (Kmatrix[0][0] / Kmatrix[0][1])
				* strohVectors[alpha + DIMENSION][1];
		
		double q1[2] = {0.,0.}; // 1 refers to the first cpt of the vector q
		
		for(alpha=0;alpha<DIMENSION;++alpha)
		{
			// need separate lines because second index of strohVectors
			// is different (I guess it could be written 1-mode)
			q1[0] += 2. * real(strohVectors[alpha+DIMENSION][1] * pow(XiMinusAlpha[alpha],-lambdaExponent[0]) * bTq[alpha]); 
			q1[1] += 2. * real(strohVectors[alpha+DIMENSION][0] * pow(XiMinusAlpha[alpha],-lambdaExponent[1]) * bTq[alpha]); 
			
		}
		
		for( mode = 0; mode < 2; ++mode )		
			q1[mode] = 1./q1[mode];
		// finally
		for( mode = 0; mode < 2; ++mode )
			for( alpha = 0; alpha < DIMENSION; ++alpha )
				bKqFactor[mode][alpha] = bTq[alpha] * q1[mode];
	}
	else if(fabs(notchAlphaAngle - PI) < TINY)
	{
		// In the case of a crack the matrix K is zero at the common eigenvalue
		// 0.5 for both mode I and II. There are no restrictions on q
		// so how do we distinguish mode I from mode II?
		// taking the q1 and q2 contributions separately gives almost
		// antisymmetric and symmetric fields respectively when the crack 
		// plane is 111 (elastic constants not invariant under reflection 
		// through this plane in general cubic crystal). If you orient the
		// crystal so that the crack plane is a mirror plane
		// then the q1 and q2 fields are the mode II and mode I fields
		// respectively

		double q[DIMENSION], alf[2]; 
		// here the array index refers to the component fo the vector q
		// there's a one to one correspondance with the mode index in this case
		// alf is indexed by mode. For mode I it's the ratio of
		// q1 to q2; for mode II it's the ratio of q2 to q1
		
		// first compute the f12, q1 and q2, field factors evaluated at theta = 0
		double f12fieldFactor[2] = {0.,0.};
		double f22fieldFactor[2] = {0.,0.};
		for( coord1 = 0; coord1 < 2; ++coord1 )
			for( alpha = 0; alpha < DIMENSION; ++alpha )
			{
				// the factor of I is from dividing by minus I, minus I being
				// xi(-pi)^0.5
				f12fieldFactor[coord1] += 2. * real(strohVectors[alpha+DIMENSION][0] * strohVectors[alpha+DIMENSION][coord1] * complex<double>(0.,1.)); 
				f22fieldFactor[coord1] += 2. * real(strohVectors[alpha+DIMENSION][1] * strohVectors[alpha+DIMENSION][coord1] * complex<double>(0.,1.)); 
			}

		alf[0] = - f12fieldFactor[1] / f12fieldFactor[0];
		alf[1] = - f22fieldFactor[0] / f22fieldFactor[1];

		// use q[0] for mode II
		q[0] = 1. / (f12fieldFactor[0] + alf[1] * f12fieldFactor[1]);
		// use q[1] for mode I
		q[1] = 1. / ( alf[0] * f22fieldFactor[0] + f22fieldFactor[1]);
		q[2]= 0.;

		cout << "q[0] = " << q[0] << " q[1] = " << q[1] << endl;
		// finally
		for( alpha = 0; alpha < DIMENSION; ++alpha )
		{		
			// mode I (q2, ie q[1] plus a bit (alf[0]) of q[0])
			bKqFactor[0][alpha] = (alf[0] * strohVectors[alpha+DIMENSION][0]
								   + strohVectors[alpha+DIMENSION][1]) * q[1];
			// mode II (q1, ie q[0] plus a bit (alf[1]) of q[1])
			bKqFactor[1][alpha] = (strohVectors[alpha+DIMENSION][0] + alf[1] * strohVectors[alpha+DIMENSION][1]) * q[0];
		}
	}
	else throw DMException("NotchFieldDisplacer","UpdateMemberVariables","notchAlphaAngle too big");
}


void NotchFieldDisplacer::ElasticDisplacement(const double position[DIMENSION], double displacement[DIMENSION])
{
    int coord, alpha;

    CalculateGeometry(position);

	// need to put in the orienting vectors--right now it only works if the 
	// notch is parallel to the z-axis
    if(rPerp > cutoff)
    {
		double preFactor[2];
		for( int mode = 0; mode < 2; ++mode )
		{
			preFactor[mode] = stressIntensity[mode] * pow(rPerp,lambdaExponent[mode]) / lambdaExponent[mode];
		}
		
		for( alpha = 0; alpha < DIMENSION; ++alpha )
		{
			complex<double> XiTheta = cosTheta + strohEigenvalues[alpha] *  sinTheta;		
			complex<double> kernel[2];

			for( int mode = 0; mode < 2; ++mode )
			{
				kernel[mode] = preFactor[mode] *  (pow(XiTheta,lambdaExponent[mode])/pow(XiMinusAlpha[alpha],lambdaExponent[mode])) * bKqFactor[mode][alpha] ;			
				for( coord = 0; coord < DIMENSION; ++coord )
					displacement[coord] += 2 * real(kernel[mode] * strohVectors[alpha][coord]);
			}
			
		}
	}

}


void NotchFieldDisplacer::ElasticStress(const double position[DIMENSION], double stress[(DIMENSION+1)*DIMENSION/2])
{
    int alpha;

	if(stressIntensity[1] != 0.)
		throw DMException("NotchFieldDisplacer","ElasticStress","Not updated to handle mode II yet");

    CalculateGeometry(position);

	// need to put in the orienting vectors--right now it only works if the 
	// notch is parallel to the z-axis
    if(rPerp > cutoff)
    {
		for( int cp = 0; cp < (DIMENSION+1)*DIMENSION/2; ++ cp )
			stress[cp] = 0.;
		double preFactor = stressIntensity[0] * pow(rPerp,lambdaExponent[0]-1);
		for( alpha = 0; alpha < DIMENSION; ++alpha )
		{

			complex<double> XiTheta = cosTheta + strohEigenvalues[alpha] *  sinTheta;		
			complex<double> kernel = preFactor *  (pow(XiTheta,lambdaExponent[0])/pow(XiMinusAlpha[alpha],lambdaExponent[0])) * bKqFactor[0][alpha] / XiTheta;			
			
			stress[0] += - 2 * real( strohEigenvalues[alpha] * strohVectors[alpha + DIMENSION][0] * kernel); // 11 component
			stress[1] += 2 * real( strohVectors[alpha + DIMENSION][1] * kernel); // 22 component
			stress[2] += 0.; // ?? 33 component
			stress[3] += 2 * real( strohVectors[alpha + DIMENSION][2] * kernel); // 23/32 component
			stress[4] += - 2 * real( strohEigenvalues[alpha] * strohVectors[alpha + DIMENSION][2] * kernel); // 13/31 component
			stress[5] += - 2 * real( strohEigenvalues[alpha] * strohVectors[alpha + DIMENSION][1] * kernel); // 12/21 component

		}
	}

}


void NotchFieldDisplacer::GradientWRTPosition(const double position[DIMENSION],double gradientWRTPosition[DIMENSION][DIMENSION])
{
	throw DMException("NotchFieldDisplacer","GradientWRTPosition","Not implemented yet");

    // NOT IMPLEMENTED!
	double gradientLocalBasis[2][2];
	for(int a=0;a<2;++a)
		for(int b=0;a<b;++b)
			gradientLocalBasis[a][b] = 0.;

	CalculateGeometry(position);
	
	if (rPerp > cutoff) 
	{ 

		
		for(int a=0;a<DIMENSION;++a)
			for(int b=0;b<DIMENSION;++b)
				for(int c=0;c<2;++c)
					for(int d=0;d<2;++d)
					{
						gradientWRTPosition[a][b] += gradientLocalBasis[c][d]*orientingVector[c][a]*orientingVector[d][b];
					}
		
		
	}         
}


void NotchFieldDisplacer::SetStressIntensity(double stressIntensity[2])
{
	for( int a = 0; a < 2; ++a )
		this->stressIntensity[a] = stressIntensity[a];
}

void NotchFieldDisplacer::SetDirection(const double direction[DIMENSION])
{
	Quasi2DElasticField::SetDirection(direction);
	UpdateMemberVariables();
}

void NotchFieldDisplacer::SetCenter(const double center[DIMENSION])
{
	Quasi2DElasticField::SetCenter(center);
    UpdateMemberVariables();
}
