// NotchFieldDisplacer.h: interface for the NotchFieldDisplacer class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef NOTCHFIELDMAKER_H
#define NOTCHFIELDMAKER_H

#include "DMFirstHeader.h"
#include <math.h>
#include "DigitalMaterial.h"
#include <assert.h>
#include <complex>

#include "Quasi2DElasticField.h"
#include "ListOfAtoms.h"

//! This class implements continuum displacements corresponding to a Notch, for given mode I stress intensity factor

/*!
  The continuum solution is from the papers of Suwito, Dunn, Labossiere and
  others (e.g. J. Appl. Phys. 83, No 7, 3574, (1998)) and is based on the
  Stroh formalism for anisotropic elasticity in cases of two dimensional 
  deformations (see T. C. T. Ting, Anisotropic Elasticity: Theory and 
  Applications, OUP, NY, 1996)

  The class is derived from Quasi2DElasticField, and thus one passes a
  direction and center. Orienting vectors (basis vectors within the plane
  normal to direction) are chosen or may be set (via Quasi2DElasticField
  functions). Also passed in is the stress intensity factor and the notch angle
  knowns as alpha (sometimes beta) in the papers. This is not the same angle 
  is the notch angle in NotchMaker (though it is simply related to it). Finally
  one must pass in the Stroh eigenvectors and eigenvalues. Perhaps one could 
  pass in the elastic constants and then calculate the eigenvectors and 
  eigenvalues but this seems unnecessary.

  For now this assumes that DIMENSION is 3. The formalism assumes so and would
  have to be adapted for purely 2D problems.

  need to put in the orienting vectors--right now it only works if the 
  notch is parallel to the z-axis


  \author nbailey
  \date 7/3/2001
  \ingroup moleculardynamics
*/


class NotchFieldDisplacer : public Quasi2DElasticField
{
public:
	NotchFieldDisplacer(const double direction[DIMENSION], double center[DIMENSION], DMArray<double, DIMENSION> *strohVectorsReal, DMArray<double, DIMENSION> *strohVectorsImag, double strohEigenValuesReal[DIMENSION], double strohEigenValuesImag[DIMENSION],double notchAlphaAngle, double lambdaExponent[2], double stressIntensity[2]);
	virtual ~NotchFieldDisplacer();


	

	// Gets and Sets for member variables
	void SetStressIntensity( double stressIntensity[2] );
	double GetStressIntensity(int a) { return stressIntensity[a]; }

	void SetDirection(const double direction[DIMENSION]); 
	const double *GetDirection() const {return (const double *)direction;}
	void SetCenter(const double center[DIMENSION]); 
	const double *GetCenter() const {return (const double *)center;}
	void ElasticDisplacement(const double position[DIMENSION], double displacement[DIMENSION]);
	void ElasticStress(const double position[DIMENSION], double stress[(DIMENSION+1)*DIMENSION/2]);

#ifndef SWIG

	void GradientWRTPosition(const double position[DIMENSION],  double gradientWRTPosition[DIMENSION][DIMENSION]);


protected:
	virtual void UpdateMemberVariables();
	std::complex<double> **strohVectors;
	std::complex<double> strohEigenvalues[DIMENSION];
	double notchAlphaAngle;
	double lambdaExponent[2];
	double stressIntensity[2];

	std::complex<double> XiMinusAlpha[DIMENSION];
	std::complex<double> bKqFactor[2][DIMENSION];
	static const double TINY;

#endif //SWIG
};

#endif // !defined(NOTCHFIELDMAKER_H)
