// OverlapPruner.h: interface for the OverlapPruner class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#ifndef OVERLAPPRUNER_H
#define OVERLAPPRUNER_H

#include "DMFirstHeader.h"
#include <vector>

#include "Transformer.h"

/**
The duty of this class is to remove overlapping atoms when Transform(atoms) is called.
There are two possibilities:
1. If no fixed atoms (from a different list of atoms) are specified then we take all pairs 
of atoms closer than minSeparation and replaces the atoms by one at the mean position.
2. If we specify some fixed atoms (again a different list of atoms) then we remove all atoms that are
closer than minSeparation to any fixed position.

Note that in order to determine the pairs too close, we use a CellNeighborLocator. If the cutoff distance 
is very small, then the cells will we small too and we'll have problems of allocating too many cells.
This is why we can set cellSize by hand. 
\author Thierry Cretegny and Nick Bailey
\date May 2000
\ingroup moleculardynamics
**/

class ListOfAtoms;
#ifndef SWIG
template<class Type, int N> class DMArray ;
#endif

class OverlapPruner : public Transformer  
{
public:
	OverlapPruner(double minSeparation, const DMArray<double, DIMENSION> *fixed=NULL, double *metric = 0);
	virtual ~OverlapPruner();
	void SetMinSeparation(double minSeparation);
	double GetMinSeparation() {return minSeparation;}
	void SetCellSize(double cellSize);
	double GetCellSize() {return cellSize;}
	void Transform(ListOfAtoms *atoms);
#ifndef SWIG
	const std::set<int> &GetAtomsRemoved() {return atomsRemoved;}
private:
/**The principle of the SelfPruner is to treat the pairs by increasing separation. 
All pairs are treated once before an averaged atome is treated again.
Atoms are really removed only at the end, such that the number of atoms in the list doesn't change 
during the calculation of the list of neighbors.
**/
	void SelfPrune(ListOfAtoms *atoms);
	void RemoveToCloseAtoms(ListOfAtoms *atoms);
	double minSeparation;
	double cellSize;
	const DMArray<double, DIMENSION> *fixed;
	std::set<int> atomsRemoved;
	double *metric;
#endif //SWIG
};

#endif // !defined(OVERLAPPRUNER_H)
