// Quasi2DElasticField.cpp: implementation of the Quasi2DElasticField class.
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "DMFirstHeader.h"
#include <fstream>
#include "DigitalMaterial.h"

#include "Quasi2DElasticField.h"
#include "DMArray.h"

#include "DMLastHeader.h"


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////


Quasi2DElasticField::Quasi2DElasticField(const double direction[DIMENSION],
					 double center[DIMENSION],double thetaLowerCutoff)
{
    int coord;
	
	this->thetaLowerCutoff = thetaLowerCutoff;// default is zero (so that theta goes from 0 to 2 PI)
    
    for(coord=0; coord<DIMENSION; coord++)
    {
		this->center[coord] = center[coord];
		this->direction[coord] = direction[coord];
    }
    
    
    // choose two perpendicular vectors
    for(coord=0;coord<DIMENSION;++coord)
   		orientingVector[0][coord] = orientingVector[1][coord] = 0.;

	for(coord=0;coord<2;++coord)
		orientingVector[coord][coord] = 1.;
    
	

    if(DIMENSION == 3)
    {
		
		double dotProd1=0.,dotProd2=0.;
		for(coord=0;coord<DIMENSION;++coord)
		{
			dotProd1 += direction[coord]*orientingVector[0][coord];
			dotProd2 += direction[coord]*orientingVector[1][coord];
		}
		// we want the one which is most orthogonal to the direction
		// (the case we're trying to avoid is choosing a vector parallel
		// to direction because then we can't get an orthonormal triad)
		if(fabs(dotProd2) < fabs(dotProd1))
		{
			for(coord=0;coord<DIMENSION;++coord)
				orientingVector[0][coord] = orientingVector[1][coord];
			dotProd1 = dotProd2;
			// now we deal only with orientingVector[0], so don't need to
			// swap completely
		}
		// Get the perpendicular component and normalize
		double ov1Norm = 0.;
		for(coord=0;coord<DIMENSION;++coord)
		{
			orientingVector[0][coord] -= dotProd1*direction[coord];
			ov1Norm += orientingVector[0][coord]*orientingVector[0][coord];
		}
		for(coord=0;coord<DIMENSION;++coord)
			orientingVector[0][coord] /= sqrt(ov1Norm);
		
		orientingVector[1][0] = direction[1]*orientingVector[0][2] -
			direction[2]*orientingVector[0][1]; 
		orientingVector[1][1] = direction[2]*orientingVector[0][0] -
			direction[0]*orientingVector[0][2]; 
		orientingVector[1][2] = direction[0]*orientingVector[0][1] -
			direction[1]*orientingVector[0][0];
		
    }
}

void Quasi2DElasticField::CalculateGeometry(const double position[DIMENSION])
{
    int coord;
    double dotProduct = 0., vect[DIMENSION];
    orient1Comp=0.;orient2Comp=0.; rPerp=0.;
    
    
    for(coord=0;coord<DIMENSION;coord++) 
    {
		vect[coord] = position[coord] - center[coord];
		dotProduct += vect[coord]*direction[coord];
    }
    
    // calculate the coordinates of the atom in the system formed by the
    // edge and screw comp and orentingvect.	
    for(coord=0;coord<DIMENSION;coord++)
    {
		parallelComp[coord] = dotProduct*direction[coord];
		perpComp[coord] = vect[coord] - parallelComp[coord];
		orient1Comp += perpComp[coord] * orientingVector[0][coord];
		orient2Comp += perpComp[coord] * orientingVector[1][coord];
		rPerp += perpComp[coord]*perpComp[coord];
    }
    cosTheta = 0.; sinTheta = 0.;
    rPerp = sqrt(rPerp); 
    
    
    if (rPerp != 0.)
    { 
		rPerpInverse = 1./rPerp;
		for(coord=0;coord<DIMENSION;coord++)
			cosTheta += orientingVector[0][coord]*perpComp[coord]*rPerpInverse;
		for(coord=0;coord<DIMENSION;coord++)
			sinTheta += orientingVector[1][coord]*perpComp[coord]*rPerpInverse;
		theta = acos(cosTheta) * (sinTheta > 0 ? 1.:-1);
		if(theta < thetaLowerCutoff) theta += 2*PI;
    }
    else 
    {
		theta = 0.;
		rPerpInverse = 0.;
    }
}

void Quasi2DElasticField::GradientWRTCenter(const double position[DIMENSION],
					    double gradientWRTCenter[DIMENSION][DIMENSION])
{
    // doesn't work when you're adding contribution s from different classes:
    // the different contributions will get multiplied by minus 1 different
    // numbers of times!
    int coord1,coord2;
    GradientWRTPosition(position,gradientWRTCenter);
    
    for(coord1=0;coord1<DIMENSION;++coord1)
	for(coord2=0;coord2<DIMENSION;++coord2)
	    gradientWRTCenter[coord1][coord2] *= -1.;
}


void Quasi2DElasticField::SetDirection(const double direction[DIMENSION])
{
    for(int coord=0; coord<DIMENSION;coord++)
	this->direction[coord] = direction[coord];
}

void Quasi2DElasticField::SetCenter(const double center[DIMENSION])
{
    for(int coord=0; coord<DIMENSION;coord++)
	this->center[coord] = center[coord];
}
