// MDView.cpp : implementation of the CMDView class
//
/*********************************************************************************

    Copyright (C) 2005 Nicholas P. Bailey, Thierry Cretegny, Andrew J. Dolgert,
    Christopher R. Myers, Jakob Schiotz, James P. Sethna, Valerie R. Coffman

    This library is free software; you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as published by the
    Free Software Foundation; either version 2.1 of the License, or (at your
    option) any later version.

    This library is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
    for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA or visit
    http://www.gnu.org/licenses/lgpl.html

    Author Contact Information:
    Jim Sethna      
    sethna@lassp.cornell.edu  
    117 Clark Hall
    Cornell University
    Ithaca, NY 14850

**********************************************************************************/

#include "stdafx.h"
#include "MD.h"

#include "MDDoc.h"
#include "MDView.h"
#include <math.h>
#include "Thread.h"
#include "SortAtoms.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

const long WINDOW_BUFFER_RESOLUTION=10000;	// Matt Kuntz's way of drawing atoms with more resolution than screen

/////////////////////////////////////////////////////////////////////////////
// CMDView

IMPLEMENT_DYNCREATE(CMDView, CView)

BEGIN_MESSAGE_MAP(CMDView, CView)
	//{{AFX_MSG_MAP(CMDView)
	ON_COMMAND(ID_CONTROL_RESTART, OnControlRestart)
	ON_COMMAND(ID_CONTROL_RESTARTPAUSED, OnControlRestartPaused)
	//}}AFX_MSG_MAP
	// Standard printing commands
	ON_COMMAND(ID_FILE_PRINT, CView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_DIRECT, CView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_PREVIEW, CView::OnFilePrintPreview)
	ON_MESSAGE(WM_USER_UPDATE_SIM, OnUpdateSim)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CMDView construction/destruction

CMDView::CMDView()
{
	buffer = NULL;	// Set up in OnInitialUpdate
	stored_pDC = NULL;	// Set up first time used in Update, then store
	for (int coord=0; coord<DIMENSION; coord++) center[coord]=0;
	diameter = 1;
	plottedRegionSize=30;
	phWndStore = NULL;
	atoms = 0;
}

CMDView::~CMDView()
{
	// Get us off the listOfAtoms' list of Observers.
	CMDDoc *pcmdDoc = this->GetDocument();
	pcmdDoc->sim.GetListOfAtoms()->Detach(this);

	delete buffer;
	if (stored_pDC != NULL && ::IsWindow(m_hWnd)) ReleaseDC(stored_pDC);	// from Update
	if (phWndStore) delete phWndStore;
}

BOOL CMDView::PreCreateWindow(CREATESTRUCT& cs)
{
	// TODO: Modify the Window class or styles here by modifying
	//  the CREATESTRUCT cs

	return CView::PreCreateWindow(cs);
}

/////////////////////////////////////////////////////////////////////////////
// CMDView drawing

void CMDView::OnDraw(CDC* pDC)
{
	CMDDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);

	// From Matt Kuntz's program, Spring 1998

	pDC->SetMapMode(MM_ISOTROPIC);
	pDC->SetWindowExt(WINDOW_BUFFER_RESOLUTION,WINDOW_BUFFER_RESOLUTION);	// Resolution of window
	pDC->SetViewportExt(window_width,-window_height);	// Reverse direction of y
	pDC->SetViewportOrg(0,window_height);	// Set 0 to LLHS

	// Put buffer to window

	pDC->BitBlt(0,0,WINDOW_BUFFER_RESOLUTION,WINDOW_BUFFER_RESOLUTION,&bufferDC,0,0,SRCCOPY);
}

void CMDView::OnUpdate(CView *pSender, LPARAM lHint, CObject *pHint)
{
	if ((lHint==1) && (atoms!=0)) {
		if(!::IsWindow(m_hWnd)) return;
		if (stored_pDC == NULL)
			stored_pDC = GetDC();	// Make new if first time
		
		// clear screen
		
		bufferDC.PatBlt(0,0,WINDOW_BUFFER_RESOLUTION,WINDOW_BUFFER_RESOLUTION,WHITENESS);
		
		double radius = GetDiameter()/2.0;
		atoms->ReadLock();
		if (DIMENSION < 3)
		{
			for(long i=0; i<atoms->GetNumber(); i++)
			{
				double x = atoms->GetCartesianPosition(i,0)-center[0]+0.5*plottedRegionSize;
				double y = atoms->GetCartesianPosition(i,1)-center[1]+0.5*plottedRegionSize;
				bufferDC.Ellipse((int)((x-radius)*dX), (int)((y-radius)*dY),
					(int)((x+radius)*dX), (int)((y+radius)*dY));
			}
		}
		else
		{	
			static SortAtoms sort3D;
			
			const int* sortedAtomNumbers = sort3D.Sort(atoms);
			
			for(long i=0; i<atoms->GetNumber(); i++)
			{
				double x = atoms->GetCartesianPosition(sortedAtomNumbers[i],0)-center[0]+0.5*plottedRegionSize;
				double y = atoms->GetCartesianPosition(sortedAtomNumbers[i],1)-center[1]+0.5*plottedRegionSize;
				bufferDC.Ellipse((int)((x-radius)*dX), (int)((y-radius)*dY),
					(int)((x+radius)*dX), (int)((y+radius)*dY));
			}
		}
		atoms->ReadRelease();
		Invalidate(false);
	} else {
		this->CView::OnUpdate(pSender,lHint,pHint);
	}
}


/////////////////////////////////////////////////////////////////////////////
// CMDView printing

BOOL CMDView::OnPreparePrinting(CPrintInfo* pInfo)
{
	// default preparation
	return DoPreparePrinting(pInfo);
}

void CMDView::OnBeginPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
	// TODO: add extra initialization before printing
}

void CMDView::OnEndPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
	// TODO: add cleanup after printing
}

/////////////////////////////////////////////////////////////////////////////
// CMDView diagnostics

#ifdef _DEBUG
void CMDView::AssertValid() const
{
	CView::AssertValid();
}

void CMDView::Dump(CDumpContext& dc) const
{
	CView::Dump(dc);
}

CMDDoc* CMDView::GetDocument() // non-debug version is inline
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CMDDoc)));
	return (CMDDoc*)m_pDocument;
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CMDView message handlers

void CMDView::SetPlottedRegionSize(double maxSize)
{
	plottedRegionSize = maxSize;
	dX = WINDOW_BUFFER_RESOLUTION/plottedRegionSize;
	dY = WINDOW_BUFFER_RESOLUTION/plottedRegionSize;
}


void CMDView::Update(Subject *theSubject)
{
	// Rather than
	// ListOfAtoms *atoms = (ListOfAtoms *) theSubject;
	// use this: Safer, better.
	// For Windows (MFC), Needs Project Settings / C++ / Enable Run Time Type Information 
	atoms = dynamic_cast<ListOfAtoms *>(theSubject);

	if(!::IsWindow(m_hWnd)) return;
	// We use messages here because this function is actually called
	// by the simulation thread so it can't call UpdateAllViews and such.
	// The reason is that each thread has its own set of internal handle tables
	// for things like Windows and MFC likes to check those handles to see if
	// they are correct.  Here, they would not be, so we pass the information
	// to the main thread to process (or handle).
	// This sends a Windows message for immediate execution so the
	// display of atoms is synchronous with the simulation.  It is a real windows
	// message so it runs in the Windows MFC thread, but it is of a kind that
	// automatically goes to the front of the queue.
	// SendNotifyMessage posts the message and returns while
	// SendMessage would block this thread until the message was taken from the queue.
	// By block I mean create a deadlock because the sim thread would be waiting for the
	// message to post and the GUI thread could be waiting for the simthread
	// to notify the GUI it paused.
	//::SendNotifyMessage(this->m_hWnd,WM_USER_UPDATE_SIM,NULL,1);
	// Posting the message instead would permit the system to handle it in batches.
	// The message goes to the back of the message queue with normal priority.
	// This might be faster for the sim but more jerky for the viewer.
	::PostMessage(this->m_hWnd,WM_USER_UPDATE_SIM,NULL,1);
}

// This function is called when we receive a WM_USER_UPDATE_SIM message
// which we sent ourself.  It should just tell the document it needs to redraw.
LRESULT CMDView::OnUpdateSim(WPARAM wParam, LPARAM lParam)
{
	// Get old updates out of the queue in case the drawing is too slow.
	if(!::IsWindow(m_hWnd)) return 0;
	MSG msg;
	while (::PeekMessage(&msg,m_hWnd,WM_USER_UPDATE_SIM,WM_USER_UPDATE_SIM,
		PM_REMOVE)) ;
	// When we call update, we give a hint=1, a note to ourselves that we
	// are the ones requesting the update so we know not to clear the window
	// before we draw.
	this->GetDocument()->UpdateAllViews(NULL,1);
	return 0; // Is this the right thing to return?
}

void CMDView::OnInitialUpdate() 
{
	CView::OnInitialUpdate();


	CMDDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);

	double width = plottedRegionSize;
	double height = plottedRegionSize;

	CClientDC dc(this);

	CRect window_rect;		// Find size of window to draw to
	GetClientRect(&window_rect);


	window_height = window_rect.Height();
	window_width = window_rect.Width();

	bufferDC.CreateCompatibleDC(&dc);

	buffer = new CBitmap;

	buffer->CreateCompatibleBitmap(&dc,window_width,window_height);
	oldBuffer = bufferDC.SelectObject(buffer);

	bufferDC.SetMapMode(MM_ISOTROPIC);
	bufferDC.SetWindowExt(WINDOW_BUFFER_RESOLUTION,WINDOW_BUFFER_RESOLUTION);       // Resolution of window
	bufferDC.SetViewportExt(window_width,-window_height);   // Reverse direction of y
	bufferDC.SetViewportOrg(0,window_height);       // Set 0 to LLHS

	bufferDC.PatBlt(0,0,WINDOW_BUFFER_RESOLUTION,WINDOW_BUFFER_RESOLUTION,WHITENESS);

	dX = WINDOW_BUFFER_RESOLUTION/width;    // Coordinate conversion factor
	dY = WINDOW_BUFFER_RESOLUTION/height;

	bufferDC.SelectStockObject(BLACK_PEN);
	bufferDC.SelectStockObject(GRAY_BRUSH);
	
	// Start thread
	// Can't pass pointers to windows across threads: must pass handles to the window
	// The handle points to the CWnd unless one attaches it to your specific subclass

	// Should just call OnRestart here, but resets twice

	CMDDoc *pcmdDoc = this->GetDocument();
	Thread *simThread = pcmdDoc->GetSimThread();
	MDSimulation *sim = &(pcmdDoc->sim);
//	this->Detach();	// Detach former view
//	this->Attach(this->GetSafeHwnd());

	sim->GetListOfAtoms()->Attach(this);
	this->SetDiameter(sim->GetDiameter());    // returns the size of an atom
	this->SetPlottedRegionSize(sim->GetMaxSize());
	this->SetCenter(sim->GetAtomsInitializer()->GetCenter());

	//	OnRestart();
}


void CMDView::OnControlRestart() 
{
	this->OnControlRestartPaused();
	CMDDoc *pcmdDoc = this->GetDocument();
	pcmdDoc->GetSimThread()->Resume();
}

void CMDView::OnControlRestartPaused() 
{
	CMDDoc *pcmdDoc = this->GetDocument();
	MDSimulation *pSim = &(pcmdDoc->sim);

	Thread *simThread = pcmdDoc->GetSimThread();
	simThread->Pause();

	pSim->GetListOfAtoms()->Attach(this);
	this->SetDiameter(pSim->GetDiameter());    // returns the size of an atom
	this->SetPlottedRegionSize(pSim->GetMaxSize());
	this->SetCenter(pSim->GetAtomsInitializer()->GetCenter());
	pSim->Reset();
}


